Package: jmvReadWrite
Title: Read and Write 'jamovi' Files ('.omv')
Version: 0.4.13
Author: Sebastian Jentschke [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0003-2576-5432>)
Authors@R: 
    person(given = "Sebastian", family = "Jentschke", role = c("aut", "cre", "cph"), email = "sebastian.jentschke@uib.no", comment = c(ORCID = "0000-0003-2576-5432"))
Maintainer: Sebastian Jentschke <sebastian.jentschke@uib.no>
Description: The free and open a statistical spreadsheet 'jamovi'
    (<https://www.jamovi.org>) aims to make statistical analyses easy and
    intuitive. 'jamovi' produces syntax that can directly be used in R (in
    connection with the R-package 'jmv'). Having import / export routines for
    the data files 'jamovi' produces ('.omv') permits an easy transfer of
    data and analyses between 'jamovi' and R.
License: AGPL-3
LazyData: true
Language: en-GB
Encoding: UTF-8
RoxygenNote: 7.3.3
VignetteBuilder: knitr
URL: https://sjentsch.github.io/jmvReadWrite/
BugReports: https://github.com/sjentsch/jmvReadWrite/issues
Depends: R (>= 3.5.0)
Imports: jsonlite, methods, zip
Suggests: jmv, jmvcore (>= 2.4.7), foreign, haven, knitr, rmarkdown,
        RProtoBuf, testthat (>= 3.0.0)
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-04-27 05:41:27 UTC; sje025
Repository: CRAN
Date/Publication: 2026-04-27 12:50:02 UTC
Built: R 4.6.0; ; 2026-04-27 21:36:13 UTC; unix

Package: jmvReadWrite
Title: Read and Write 'jamovi' Files ('.omv')
Version: 0.4.14
Author: Sebastian Jentschke [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0003-2576-5432>)
Authors@R: 
    person(given = "Sebastian",
           family = "Jentschke",
           role = c("aut", "cre", "cph"),
           email = "sebastian.jentschke@uib.no",
           comment = c(ORCID = "0000-0003-2576-5432"))
Maintainer: Sebastian Jentschke <sebastian.jentschke@uib.no>
Description: The free and open a statistical spreadsheet 'jamovi'
    (<https://www.jamovi.org>) aims to make statistical analyses easy and
    intuitive. 'jamovi' produces syntax that can directly be used in R (in
    connection with the R-package 'jmv'). Having import / export routines for
    the data files 'jamovi' produces ('.omv') permits an easy transfer of
    data and analyses between 'jamovi' and R.
License: AGPL-3
LazyData: true
Language: en-GB
Encoding: UTF-8
VignetteBuilder: knitr
URL: https://sjentsch.github.io/jmvReadWrite/
BugReports: https://github.com/sjentsch/jmvReadWrite/issues
Depends: R (>= 3.5.0)
Imports: jsonlite, methods, zip
Suggests: jmv, jmvcore (>= 2.4.7), foreign, haven, knitr, rmarkdown,
        RProtoBuf, testthat (>= 3.0.0)
Config/testthat/edition: 3
NeedsCompilation: no
Config/roxygen2/version: 8.0.0
Packaged: 2026-07-13 14:48:26 UTC; sje025
Repository: CRAN
Date/Publication: 2026-07-13 15:10:02 UTC
Built: R 4.6.1; ; 2026-07-13 19:00:59 UTC; unix

# Generated by roxygen2: do not edit by hand

export(aggregate_omv)
export(arrange_cols_omv)
export(combine_cols_omv)
export(convert_to_omv)
export(describe_omv)
export(distances_omv)
export(label_vars_omv)
export(long2wide_omv)
export(merge_cols_omv)
export(merge_rows_omv)
export(read_omv)
export(replace_omv)
export(search_omv)
export(sort_omv)
export(transform_vars_omv)
export(transpose_omv)
export(wide2long_omv)
export(write_omv)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# jmvReadWrite <a href="https://sjentsch.github.io/jmvReadWrite/"><img src="man/figures/logo.svg" align="right" width="20%" /></a>

<!---
<br clear="all">
--->
<!-- badges: start -->

[![CRAN](https://www.r-pkg.org/badges/version/jmvReadWrite)](https://cran.r-project.org/package=jmvReadWrite)
[![Project Status: Active – The project has reached a stable, usable
state and is being actively
developed.](https://www.repostatus.org/badges/latest/active.svg)](https://www.repostatus.org/#active)
[![License](https://img.shields.io/badge/License-AGPL%20v3-green.svg)](https://www.gnu.org/licenses/agpl-3.0.html)
[![Downloads](https://cranlogs.r-pkg.org/badges/jmvReadWrite?color=brightgreen)](https://cran.r-project.org/package=jmvReadWrite)
[![Last
commit](https://img.shields.io/github/last-commit/sjentsch/jmvReadWrite?logo=GitHub)](https://github.com/sjentsch/jmvReadWrite)
[![Register an
issue](https://img.shields.io/github/issues/sjentsch/jmvReadWrite?color=%23fa251e&logo=GitHub)](https://github.com/sjentsch/jmvReadWrite/issues)
[![R-hub](https://github.com/sjentsch/jmvReadWrite/actions/workflows/rhub.yaml/badge.svg)](https://github.com/sjentsch/jmvReadWrite/actions/workflows/rhub.yaml)
[![R-CMD-check](https://github.com/sjentsch/jmvReadWrite/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/sjentsch/jmvReadWrite/actions/workflows/R-CMD-check.yaml)
[![CI](https://github.com/sjentsch/jmvReadWrite/actions/workflows/CI.yml/badge.svg)](https://github.com/sjentsch/jmvReadWrite/actions/workflows/CI.yml)
[![pkgcheck](https://github.com/sjentsch/jmvReadWrite/workflows/pkgcheck/badge.svg)](https://github.com/sjentsch/jmvReadWrite/actions?query=workflow%3Apkgcheck)
[![Codecov
coverage](https://codecov.io/gh/sjentsch/jmvReadWrite/branch/main/graph/badge.svg)](https://app.codecov.io/gh/sjentsch/jmvReadWrite?branch=main)
[![Documentation](https://img.shields.io/badge/documentation-is_here-blue)](https://sjentsch.github.io/jmvReadWrite/)
<!-- badges: end -->

<!---
[//]: [![Build Status](https://app.travis-ci.com/sjentsch/jmvReadWrite.svg?branch=main)](https://app.travis-ci.com/sjentsch/jmvReadWrite)
--->

The R-package `jmvReadWrite` reads and writes the .omv-files that are
used by the statistical spreadsheet `jamovi` (<https://www.jamovi.org>).
It is supposed to ease using jamovi and R together, provide helper
functions for some often required data management tasks, and to adjust
and use syntax for statistical analyses that were created using the GUI
in `jamovi` in R (in connection with the R-library `jmv`). More
recently, `jmvReadWrite` became easily available from within `jamovi` by
becoming part of the `Rj` module (where you can use it by writing R
commands, documented below), and via the `jTransform` module that
provides a graphical user interface for most helper functions.

## Installation

In R, you can either install a stable version of `jmvReadWrite` which is
available on [CRAN](https://cran.r-project.org/package=jmvReadWrite)
using the following command:

``` r
install.packages("jmvReadWrite")
```

or you can install the development version of the `jmvReadWrite` package
from [GitHub](https://github.com/sjentsch/jmvReadWrite/):

``` r
if (!require(devtools)) install.packages("devtools")
devtools::install_github("sjentsch/jmvReadWrite")
```

## How to use the package?

[**read_omv**](https://sjentsch.github.io/jmvReadWrite/reference/read_omv.html)

The following code uses the ToothGrowth-data set that is part of the
data sets included in R (the current file contains some modifications
though for testing the reading and writing routines: `read_omv` and
`write_omv`). With this data set, a syntax to conduct an ANOVA is run.

The results should be similar to those obtained when running the same
analysis in jamovi (using the GUI). To do so, open the file menu (☰)
choose `Open`, `Data Library` and `ToothGrowth`. Afterwards, click on
the `ANOVA`-button in the `Analyses`-tab and choose `ANOVA`. There, you
assign the variable `len` to `Dependent Variable` and `supp` and `dose`
to `Fixed Factors`. Afterwards, you choose / tick `Overall Model Test`
and `ω²`. Open the drop-down menu `Assumption Checks` and tick
`Homogeneity test` and `Normality test`. The results should be identical
apart from that the table output looks nicer in `jamovi` (not only text,
as below), numbers are rounded and maybe one or two other cosmetic
differences.

If you want to copy the syntax generated in jamovi, you have to switch
on the
[`Syntax Mode`](https://jamovi.readthedocs.io/en/latest/usermanual/um_6_jamovi_and_R/#syntax-mode).
Afterwards, the syntax is shown at the top of the analysis and can be
copied from there.

``` r
fleOMV <- system.file("extdata", "ToothGrowth.omv", package = "jmvReadWrite")
data <- jmvReadWrite::read_omv(fleOMV)
# if the "jmv"-package is installed, we can run a test analysis with the data
if ("jmv" %in% rownames(installed.packages())) {
    jmv::ANOVA(
        formula = len ~ supp + dose + supp:dose,
        data = data,
        effectSize = c("omega"),
        modelTest = TRUE,
        homo = TRUE,
        norm = TRUE)
    }
#> 
#>  ANOVA
#> 
#>  ANOVA - len                                                                                      
#>  ──────────────────────────────────────────────────────────────────────────────────────────────── 
#>                     Sum of Squares    df    Mean Square    F            p             ω²          
#>  ──────────────────────────────────────────────────────────────────────────────────────────────── 
#>    Overall model         2740.1033     5      548.02067    41.557178    < .0000001                
#>    supp                   205.3500     1      205.35000    15.571979     0.0002312    0.0554519   
#>    dose                  2426.4343     2     1213.21717    91.999965    < .0000001    0.6925788   
#>    supp:dose              108.3190     2       54.15950     4.106991     0.0218603    0.0236466   
#>    Residuals              712.1060    54       13.18715                                           
#>  ──────────────────────────────────────────────────────────────────────────────────────────────── 
#> 
#> 
#>  ASSUMPTION CHECKS
#> 
#>  Homogeneity of Variances Test (Levene's) 
#>  ──────────────────────────────────────── 
#>    F           df1    df2    p           
#>  ──────────────────────────────────────── 
#>    1.940130      5     54    0.1027298   
#>  ──────────────────────────────────────── 
#> 
#> 
#>  Normality Test (Shapiro-Wilk) 
#>  ───────────────────────────── 
#>    Statistic    p           
#>  ───────────────────────────── 
#>    0.9849884    0.6694242   
#>  ─────────────────────────────
```

Since version 0.2.0,
[`read_omv`](https://sjentsch.github.io/jmvReadWrite/reference/read_omv.html)
also extracts the syntax from analyses that you may have conducted in
the jamovi-GUI and that are stored in the .omv-file. To extract them,
you have to set the parameter `getSyn = TRUE` when calling
[`read_omv`](https://sjentsch.github.io/jmvReadWrite/reference/read_omv.html)
(default is `FALSE`). When the parameter is set, the analyses are stored
in the attribute `syntax`. They can be used as shown in the following
examples:

``` r
fleOMV <- system.file("extdata", "ToothGrowth.omv", package = "jmvReadWrite")
data <- jmvReadWrite::read_omv(fleOMV, getSyn = TRUE)
# shows the syntax of the analyses from the .omv-file
# please note that syntax extraction may not work on all systems
# if the syntax couldn't be extracted, an empty list (length = 0) is returned,
# otherwise, the syntax of the analyses from the .omv-file is shown and
# the commands of the first and the second analysis are run, with the
# output of the second analysis assigned to the variable result
if (length(attr(data, "syntax")) >= 2) {
    attr(data, "syntax")
    # if the "jmv"-package is installed, we can run the analyses in "syntax"
    if ("jmv" %in% rownames(installed.packages())) {
        eval(parse(text = attr(data, "syntax")[[1]]))
        eval(parse(text = paste0("result = ", attr(data, "syntax")[[2]])))
        names(result)
        # -> "main"      "assump"    "contrasts" "postHoc"   "emm"
        # (the names of the five output tables)
    }
}
#> [1] "main"      "assump"    "contrasts" "postHoc"   "emm"       "residsOV"
```

<br/>

[**write_omv**](https://sjentsch.github.io/jmvReadWrite/reference/write_omv.html)

The `jmvReadWrite`-package also enables you to write `.omv`-files in
order to use them in `jamovi`. Let’s assume that you have a large
collection of log-files (e.g., from an experiment) that you compile and
process (summarize, filter, etc.) in R in order to later analyse them in
`jamovi`. You will have those processed log-files stored in a data frame
(called, e.g., `data`) which you then write to a file that you can open
in jamovi afterwards.

``` r
# use the data set "ToothGrowth" and, if it exists, write it as jamovi-file
# using write_omv()
data("ToothGrowth", package = "jmvReadWrite")
# "retDbg" has to be set in order to return debug information to wrtDta
wrtDta <- jmvReadWrite::write_omv(ToothGrowth, "Trial.omv", retDbg = TRUE)
names(wrtDta)
#> [1] "mtaDta" "xtdDta" "dtaFrm"
# -> "mtaDta" "xtdDta" "dtaFrm"
# this debug information contains a list with the metadata ("mtaDta", e.g.,
# column and data type), the extended data ("xtdDta", e.g., variable lables),
# and the data frame (dtaFrm) for checking (understanding the file format) and
# debugging

# check whether the file was written to the disk, get the file information (size, etc.)
# and delete the file afterwards
list.files(".", "Trial.omv")
#> [1] "Trial.omv"
file.info("Trial.omv")[, c("size", "isdir", "mode")]
#>           size isdir mode
#> Trial.omv 2618 FALSE  644
unlink("Trial.omv")
```

Although jamovi reads R-data files (.RData, .rda, .rds)
[`write_omv`](https://sjentsch.github.io/jmvReadWrite/reference/write_omv.html)
permits to store `jamovi`-specific attributes (such as variable labels)
in addition. Please note that if you are reading from an `.omv`-file in
order to write back to an `.omv`-file (perhaps after some
modifications), it is recommended to leave the `sveAtt`-attribute set to
`TRUE` (which is the default).

``` r
# reading and writing a file with the "sveAtt"-parameter permits you to keep
# essential meta-data to ensure that the written file looks and works like the
# original file (plus you modifications)
fleOMV <- system.file("extdata", "ToothGrowth.omv", package = "jmvReadWrite")
data <- jmvReadWrite::read_omv(fleOMV, sveAtt = TRUE)
# shows the names of the attributes for the whole data set (e.g., number of
# rows and columns) and the names of the attributes of the first column
names(attributes(data))
#> [1] "names"       "row.names"   "class"       "fltLst"      "removedRows"
#> [6] "addedRows"   "transforms"
names(attributes(data[[1]]))
#>  [1] "name"           "id"             "columnType"     "dataType"      
#>  [5] "measureType"    "formula"        "formulaMessage" "parentId"      
#>  [9] "width"          "type"           "importName"     "description"   
#> [13] "transform"      "edits"          "missingValues"  "filterNo"      
#> [17] "active"
#
# perhaps do some modifications to the file here and write it back afterwards
jmvReadWrite::write_omv(data, "Trial.omv")
unlink("Trial.omv")
```

If `Trial.omv` in the example above would have been kept, it should look
like the original file (plus your possible modifications). If you, e.g.,
added a new column, you could adjust some attributes (e.g., to enforce a
specific `columnType` or `measurementType`): just look at how attributes
are stored for other columns. <br /><br />

**Helper functions**

`jmvReadWrite` contains a number of helper functions that assist you
with data management tasks that are frequently required:

- [`aggregate_omv`](https://sjentsch.github.io/jmvReadWrite/reference/aggregate_omv.html):
  Aggregates data (calculating, e.g., mean, median, etc.) from data sets
  in long format.

- [`arrange_cols_omv`](https://sjentsch.github.io/jmvReadWrite/reference/arrange_cols_omv.html):
  Re-arranges the columns of your data file in a requested order.

- [`combine_cols_omv`](https://sjentsch.github.io/jmvReadWrite/reference/combine_cols_omv.html):
  Combines the data from two columns into one. The function assumes that
  the data contained in the two columns are the same. It checks for (and
  thereby ensures) equality and replaces missing values in one column by
  replacing those with values from the second column (if those are not
  missing as well).

- [`convert_to_omv`](https://sjentsch.github.io/jmvReadWrite/reference/convert_to_omv.html):
  Converts data sets from other file formats into jamovi-format. This
  function may be helpful if you have to convert a larger amount of
  files.

- [`describe_omv`](https://sjentsch.github.io/jmvReadWrite/reference/describe_omv.html):
  Adds a title and a description to a data set. This function may be
  helpful for documenting what is contained in a data set, e.g. for
  publishing them in a repository such as OSF, or for generated data
  sets, e.g. those used in teaching.

- [`distances_omv`](https://sjentsch.github.io/jmvReadWrite/reference/distances_omv.html):
  Calculates a wide range of distances measures (for continuous,
  frequency or binary data). If can be determined, whether the
  calculation of the distances should be carried out between columns /
  variables or between rows / units of observation. The original data
  can be standardized before the distances are calculated.

- [`label_vars_omv`](https://sjentsch.github.io/jmvReadWrite/reference/label_vars_omv.html):
  Labels columns / variables in jamovi data files, using either a CSV
  file, a data frame (each containing two columns, one with the variable
  names, the other with the variable labels), or a character vector with
  the variable labels.

- [`long2wide_omv`](https://sjentsch.github.io/jmvReadWrite/reference/long2wide_omv.html):
  Converts a data set from long to wide format: Time points for repeated
  measurements are arranged as rows in the original and converted into
  columns.

- [`wide2long_omv`](https://sjentsch.github.io/jmvReadWrite/reference/wide2long_omv.html):
  Converts a data set from wide to long format: Time points for repeated
  measurements are arranged as columns in the original and converted
  into rows.

- [`merge_cols_omv`](https://sjentsch.github.io/jmvReadWrite/reference/merge_cols_omv.html):
  Add variables from several data sets, i.e. the variables / columns in
  the second, etc. input data set are added as columns to the first data
  set.

- [`merge_rows_omv`](https://sjentsch.github.io/jmvReadWrite/reference/merge_rows_omv.html):
  Add cases from several data sets, i.e. the cases / rows in the second,
  etc. data set are added as rows to the first data set.

- [`sort_omv`](https://sjentsch.github.io/jmvReadWrite/reference/sort_omv.html):
  Sort a data set according to one or more variable(s).

- [`transform_vars_omv`](https://sjentsch.github.io/jmvReadWrite/reference/transform_vars_omv.html):
  Transform skewed variables (aiming at they better conform to a normal
  distribution).

- [`transpose_omv`](https://sjentsch.github.io/jmvReadWrite/reference/transpose_omv.html):
  Transpose a data set: Make rows into columns and vice versa.

------------------------------------------------------------------------

[Changelog](https://github.com/sjentsch/jmvReadWrite/blob/main/NEWS.md)

## Authors

[Sebastian Jentschke](https://www.uib.no/en/persons/Sebastian.Jentschke)

## License

[AGPL 3](https://github.com/sjentsch/jmvReadWrite/blob/main/LICENSE)

## Giving back

If you find this package helpful, please consider donating to the jamovi
project (via the Patreon-link on the left side). If you can’t give
money, but would like to support us in another way, you may contribute
to translating [jamovi](https://hosted.weblate.org/engage/jamovi/), the
[jamovi documentation](https://hosted.weblate.org/engage/jamovidocs/),
or the textbook [”learning statistics with
jamovi“](https://hosted.weblate.org/engage/jamovi/) into your language.

Thank you for your support!

<!-- README.md is generated from README.Rmd. Please edit that file -->

# jmvReadWrite <a href="https://sjentsch.github.io/jmvReadWrite/"><img src="man/figures/logo.svg" align="right" width="20%" /></a>

<!---
<br clear="all">
--->
<!-- badges: start -->

[![CRAN](https://www.r-pkg.org/badges/version/jmvReadWrite)](https://cran.r-project.org/package=jmvReadWrite)
[![Project Status: Active – The project has reached a stable, usable
state and is being actively
developed.](https://www.repostatus.org/badges/latest/active.svg)](https://www.repostatus.org/#active)
[![License](https://img.shields.io/badge/License-AGPL%20v3-green.svg)](https://www.gnu.org/licenses/agpl-3.0.html)
[![Downloads](https://cranlogs.r-pkg.org/badges/jmvReadWrite?color=brightgreen)](https://cran.r-project.org/package=jmvReadWrite)
[![Last
commit](https://img.shields.io/github/last-commit/sjentsch/jmvReadWrite?logo=GitHub)](https://github.com/sjentsch/jmvReadWrite)
[![Register an
issue](https://img.shields.io/github/issues/sjentsch/jmvReadWrite?color=%23fa251e&logo=GitHub)](https://github.com/sjentsch/jmvReadWrite/issues)
[![R-hub](https://github.com/sjentsch/jmvReadWrite/actions/workflows/rhub.yaml/badge.svg)](https://github.com/sjentsch/jmvReadWrite/actions/workflows/rhub.yaml)
[![R-CMD-check](https://github.com/sjentsch/jmvReadWrite/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/sjentsch/jmvReadWrite/actions/workflows/R-CMD-check.yaml)
[![CI](https://github.com/sjentsch/jmvReadWrite/actions/workflows/CI.yml/badge.svg)](https://github.com/sjentsch/jmvReadWrite/actions/workflows/CI.yml)
[![pkgcheck](https://github.com/sjentsch/jmvReadWrite/workflows/pkgcheck/badge.svg)](https://github.com/sjentsch/jmvReadWrite/actions?query=workflow%3Apkgcheck)
[![Codecov
coverage](https://codecov.io/gh/sjentsch/jmvReadWrite/branch/main/graph/badge.svg)](https://app.codecov.io/gh/sjentsch/jmvReadWrite?branch=main)
[![Documentation](https://img.shields.io/badge/documentation-is_here-blue)](https://sjentsch.github.io/jmvReadWrite/)
<!-- badges: end -->

<!---
[//]: [![Build Status](https://app.travis-ci.com/sjentsch/jmvReadWrite.svg?branch=main)](https://app.travis-ci.com/sjentsch/jmvReadWrite)
--->

The R-package `jmvReadWrite` reads and writes the .omv-files that are
used by the statistical spreadsheet `jamovi` (<https://www.jamovi.org>).
It is supposed to ease using jamovi and R together, provide helper
functions for some often required data management tasks, and to adjust
and use syntax for statistical analyses that were created using the GUI
in `jamovi` in R (in connection with the R-library `jmv`). More
recently, `jmvReadWrite` became easily available from within `jamovi` by
becoming part of the `Rj` module (where you can use it by writing R
commands, documented below), and via the `jTransform` module that
provides a graphical user interface for most helper functions.

## Installation

In R, you can either install a stable version of `jmvReadWrite` which is
available on [CRAN](https://cran.r-project.org/package=jmvReadWrite)
using the following command:

``` r
install.packages("jmvReadWrite")
```

or you can install the development version of the `jmvReadWrite` package
from [GitHub](https://github.com/sjentsch/jmvReadWrite/):

``` r
if (!require(devtools)) install.packages("devtools")
devtools::install_github("sjentsch/jmvReadWrite")
```

## How to use the package?

[**read_omv**](https://sjentsch.github.io/jmvReadWrite/reference/read_omv.html)

The following code uses the ToothGrowth-data set that is part of the
data sets included in R (the current file contains some modifications
though for testing the reading and writing routines: `read_omv` and
`write_omv`). With this data set, a syntax to conduct an ANOVA is run.

The results should be similar to those obtained when running the same
analysis in jamovi (using the GUI). To do so, open the file menu (☰)
choose `Open`, `Data Library` and `ToothGrowth`. Afterwards, click on
the `ANOVA`-button in the `Analyses`-tab and choose `ANOVA`. There, you
assign the variable `len` to `Dependent Variable` and `supp` and `dose`
to `Fixed Factors`. Afterwards, you choose / tick `Overall Model Test`
and `ω²`. Open the drop-down menu `Assumption Checks` and tick
`Homogeneity test` and `Normality test`. The results should be identical
apart from that the table output looks nicer in `jamovi` (not only text,
as below), numbers are rounded and maybe one or two other cosmetic
differences.

If you want to copy the syntax generated in jamovi, you have to switch
on the
[`Syntax Mode`](https://jamovi.readthedocs.io/en/latest/usermanual/um_6_jamovi_and_R/#syntax-mode).
Afterwards, the syntax is shown at the top of the analysis and can be
copied from there.

``` r
fleOMV <- system.file("extdata", "ToothGrowth.omv", package = "jmvReadWrite")
data <- jmvReadWrite::read_omv(fleOMV)
# if the "jmv"-package is installed, we can run a test analysis with the data
if (nzchar(system.file(package = "jmv"))) {
    jmv::ANOVA(
        formula = len ~ supp + dose + supp:dose,
        data = data,
        effectSize = "omega",
        modelTest = TRUE,
        homo = TRUE,
        norm = TRUE)
    }
#> 
#>  ANOVA
#> 
#>  ANOVA - len                                                                                      
#>  ──────────────────────────────────────────────────────────────────────────────────────────────── 
#>                     Sum of Squares    df    Mean Square    F            p             ω²          
#>  ──────────────────────────────────────────────────────────────────────────────────────────────── 
#>    Overall model         2740.1033     5      548.02067    41.557178    < .0000001                
#>    supp                   205.3500     1      205.35000    15.571979     0.0002312    0.0554519   
#>    dose                  2426.4343     2     1213.21717    91.999965    < .0000001    0.6925788   
#>    supp:dose              108.3190     2       54.15950     4.106991     0.0218603    0.0236466   
#>    Residuals              712.1060    54       13.18715                                           
#>  ──────────────────────────────────────────────────────────────────────────────────────────────── 
#> 
#> 
#>  ASSUMPTION CHECKS
#> 
#>  Homogeneity of Variances Test (Levene's) 
#>  ──────────────────────────────────────── 
#>    F           df1    df2    p           
#>  ──────────────────────────────────────── 
#>    1.940130      5     54    0.1027298   
#>  ──────────────────────────────────────── 
#> 
#> 
#>  Normality Test (Shapiro-Wilk) 
#>  ───────────────────────────── 
#>    Statistic    p           
#>  ───────────────────────────── 
#>    0.9849884    0.6694242   
#>  ─────────────────────────────
```

Since version 0.2.0,
[`read_omv`](https://sjentsch.github.io/jmvReadWrite/reference/read_omv.html)
also extracts the syntax from analyses that you may have conducted in
the jamovi-GUI and that are stored in the .omv-file. To extract them,
you have to set the parameter `getSyn = TRUE` when calling
[`read_omv`](https://sjentsch.github.io/jmvReadWrite/reference/read_omv.html)
(default is `FALSE`). When the parameter is set, the analyses are stored
in the attribute `syntax`. They can be used as shown in the following
examples:

``` r
fleOMV <- system.file("extdata", "ToothGrowth.omv", package = "jmvReadWrite")
data <- jmvReadWrite::read_omv(fleOMV, getSyn = TRUE)
# shows the syntax of the analyses from the .omv-file
# please note that syntax extraction may not work on all systems
# if the syntax couldn't be extracted, an empty list (length = 0) is returned,
# otherwise, the syntax of the analyses from the .omv-file is shown and
# the commands of the first and the second analysis are run, with the
# output of the second analysis assigned to the variable result
if (length(attr(data, "syntax")) >= 2) {
    attr(data, "syntax")
    # if the "jmv"-package is installed, we can run the analyses in "syntax"
    if (nzchar(system.file(package = "jmv"))) {
        eval(parse(text = attr(data, "syntax")[[1]]))
        eval(parse(text = paste0("result = ", attr(data, "syntax")[[2]])))
        names(result)
        # -> "main"      "assump"    "contrasts" "postHoc"   "emm"
        # (the names of the five output tables)
    }
}
#> [1] "main"      "assump"    "contrasts" "postHoc"   "emm"       "residsOV"
```

<br/>

[**write_omv**](https://sjentsch.github.io/jmvReadWrite/reference/write_omv.html)

The `jmvReadWrite`-package also enables you to write `.omv`-files in
order to use them in `jamovi`. Let’s assume that you have a large
collection of log-files (e.g., from an experiment) that you compile and
process (summarize, filter, etc.) in R in order to later analyse them in
`jamovi`. You will have those processed log-files stored in a data frame
(called, e.g., `data`) which you then write to a file that you can open
in jamovi afterwards.

``` r
# use the data set "ToothGrowth" and, if it exists, write it as jamovi-file
# using write_omv()
data("ToothGrowth", package = "jmvReadWrite")
# "retDbg" has to be set in order to return debug information to wrtDta
wrtDta <- jmvReadWrite::write_omv(ToothGrowth, "Trial.omv", retDbg = TRUE)
names(wrtDta)
#> [1] "mtaDta" "xtdDta" "dtaFrm"
# -> "mtaDta" "xtdDta" "dtaFrm"
# this debug information contains a list with the metadata ("mtaDta", e.g.,
# column and data type), the extended data ("xtdDta", e.g., variable lables),
# and the data frame (dtaFrm) for checking (understanding the file format) and
# debugging

# check whether the file was written to the disk, get the file information (size, etc.)
# and delete the file afterwards
list.files(".", "Trial.omv")
#> [1] "Trial.omv"
file.info("Trial.omv")[, c("size", "isdir", "mode")]
#>           size isdir mode
#> Trial.omv 2521 FALSE  644
unlink("Trial.omv")
```

Although jamovi reads R-data files (.RData, .rda, .rds)
[`write_omv`](https://sjentsch.github.io/jmvReadWrite/reference/write_omv.html)
permits to store `jamovi`-specific attributes (such as variable labels)
in addition. Please note that if you are reading from an `.omv`-file in
order to write back to an `.omv`-file (perhaps after some
modifications), it is recommended to leave the `sveAtt`-attribute set to
`TRUE` (which is the default).

``` r
# reading and writing a file with the "sveAtt"-parameter permits you to keep
# essential meta-data to ensure that the written file looks and works like the
# original file (plus you modifications)
fleOMV <- system.file("extdata", "ToothGrowth.omv", package = "jmvReadWrite")
data <- jmvReadWrite::read_omv(fleOMV, sveAtt = TRUE)
# shows the names of the attributes for the whole data set (e.g., number of
# rows and columns) and the names of the attributes of the first column
names(attributes(data))
#> [1] "names"       "row.names"   "class"       "fltLst"      "removedRows"
#> [6] "addedRows"   "transforms"
names(attributes(data[[1]]))
#>  [1] "name"           "id"             "columnType"     "dataType"      
#>  [5] "measureType"    "formula"        "formulaMessage" "parentId"      
#>  [9] "width"          "type"           "importName"     "description"   
#> [13] "transform"      "edits"          "missingValues"  "filterNo"      
#> [17] "active"
#
# perhaps do some modifications to the file here and write it back afterwards
jmvReadWrite::write_omv(data, "Trial.omv")
unlink("Trial.omv")
```

If `Trial.omv` in the example above would have been kept, it should look
like the original file (plus your possible modifications). If you, e.g.,
added a new column, you could adjust some attributes (e.g., to enforce a
specific `columnType` or `measurementType`): just look at how attributes
are stored for other columns. <br /><br />

**Helper functions**

`jmvReadWrite` contains a number of helper functions that assist you
with data management tasks that are frequently required:

- [`aggregate_omv`](https://sjentsch.github.io/jmvReadWrite/reference/aggregate_omv.html):
  Aggregates data (calculating, e.g., mean, median, etc.) from data sets
  in long format.

- [`arrange_cols_omv`](https://sjentsch.github.io/jmvReadWrite/reference/arrange_cols_omv.html):
  Re-arranges the columns of your data file in a requested order.

- [`combine_cols_omv`](https://sjentsch.github.io/jmvReadWrite/reference/combine_cols_omv.html):
  Combines the data from two columns into one. The function assumes that
  the data contained in the two columns are the same. It checks for (and
  thereby ensures) equality and replaces missing values in one column by
  replacing those with values from the second column (if those are not
  missing as well).

- [`convert_to_omv`](https://sjentsch.github.io/jmvReadWrite/reference/convert_to_omv.html):
  Converts data sets from other file formats into jamovi-format. This
  function may be helpful if you have to convert a larger amount of
  files.

- [`describe_omv`](https://sjentsch.github.io/jmvReadWrite/reference/describe_omv.html):
  Adds a title and a description to a data set. This function may be
  helpful for documenting what is contained in a data set, e.g. for
  publishing them in a repository such as OSF, or for generated data
  sets, e.g. those used in teaching.

- [`distances_omv`](https://sjentsch.github.io/jmvReadWrite/reference/distances_omv.html):
  Calculates a wide range of distances measures (for continuous,
  frequency or binary data). If can be determined, whether the
  calculation of the distances should be carried out between columns /
  variables or between rows / units of observation. The original data
  can be standardized before the distances are calculated.

- [`label_vars_omv`](https://sjentsch.github.io/jmvReadWrite/reference/label_vars_omv.html):
  Labels columns / variables in jamovi data files, using either a CSV
  file, a data frame (each containing two columns, one with the variable
  names, the other with the variable labels), or a character vector with
  the variable labels.

- [`long2wide_omv`](https://sjentsch.github.io/jmvReadWrite/reference/long2wide_omv.html):
  Converts a data set from long to wide format: Time points for repeated
  measurements are arranged as rows in the original and converted into
  columns.

- [`wide2long_omv`](https://sjentsch.github.io/jmvReadWrite/reference/wide2long_omv.html):
  Converts a data set from wide to long format: Time points for repeated
  measurements are arranged as columns in the original and converted
  into rows.

- [`merge_cols_omv`](https://sjentsch.github.io/jmvReadWrite/reference/merge_cols_omv.html):
  Add variables from several data sets, i.e. the variables / columns in
  the second, etc. input data set are added as columns to the first data
  set.

- [`merge_rows_omv`](https://sjentsch.github.io/jmvReadWrite/reference/merge_rows_omv.html):
  Add cases from several data sets, i.e. the cases / rows in the second,
  etc. data set are added as rows to the first data set.

- [`sort_omv`](https://sjentsch.github.io/jmvReadWrite/reference/sort_omv.html):
  Sort a data set according to one or more variable(s).

- [`transform_vars_omv`](https://sjentsch.github.io/jmvReadWrite/reference/transform_vars_omv.html):
  Transform skewed variables (aiming at they better conform to a normal
  distribution).

- [`transpose_omv`](https://sjentsch.github.io/jmvReadWrite/reference/transpose_omv.html):
  Transpose a data set: Make rows into columns and vice versa.

------------------------------------------------------------------------

[Changelog](https://github.com/sjentsch/jmvReadWrite/blob/main/NEWS.md)

## Authors

[Sebastian Jentschke](https://www.uib.no/en/persons/Sebastian.Jentschke)

## License

[AGPL 3](https://github.com/sjentsch/jmvReadWrite/blob/main/LICENSE)

## Giving back

If you find this package helpful, please consider donating to the jamovi
project (via the Patreon-link on the left side). If you can’t give
money, but would like to support us in another way, you may contribute
to translating [jamovi](https://hosted.weblate.org/engage/jamovi/), the
[jamovi documentation](https://hosted.weblate.org/engage/jamovidocs/),
or the textbook [”learning statistics with
jamovi“](https://hosted.weblate.org/engage/jamovi/) into your language.

Thank you for your support!

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(echo = TRUE,
                      comment = "#>",
                      collapse = TRUE,
                      message = FALSE,
                      warning = FALSE,
                      fig.height = 5,
                      fig.width = 7,
                      fig.align = "center",
                      out.width = "100%")

## ----eval=FALSE---------------------------------------------------------------
# install.packages("jmvReadWrite")

## ----eval=FALSE---------------------------------------------------------------
# if (!require(devtools)) install.packages("devtools")
# devtools::install_github("sjentsch/jmvReadWrite")

## ----echo=TRUE----------------------------------------------------------------
fleOMV <- system.file("extdata", "ToothGrowth.omv", package = "jmvReadWrite")
data <- jmvReadWrite::read_omv(fleOMV)
# if the "jmv"-package is installed, we can run a test analysis with the data
if ("jmv" %in% rownames(installed.packages())) {
    jmv::ANOVA(
        formula = len ~ supp + dose + supp:dose,
        data = data,
        effectSize = c("omega"),
        modelTest = TRUE,
        homo = TRUE,
        norm = TRUE)
    }

## ----echo=TRUE----------------------------------------------------------------
fleOMV <- system.file("extdata", "ToothGrowth.omv", package = "jmvReadWrite")
data <- jmvReadWrite::read_omv(fleOMV, getSyn = TRUE)
# shows the syntax of the analyses from the .omv-file
# please note that syntax extraction may not work on all systems
# if the syntax couldn't be extracted, an empty list (length = 0) is returned,
# otherwise, the syntax of the analyses from the .omv-file is shown and
# the commands of the first and the second analysis are run, with the
# output of the second analysis assigned to the variable result
if (length(attr(data, "syntax")) >= 2) {
    attr(data, "syntax")
    # if the "jmv"-package is installed, we can run the analyses in "syntax"
    if ("jmv" %in% rownames(installed.packages())) {
        eval(parse(text = attr(data, "syntax")[[1]]))
        eval(parse(text = paste0("result = ", attr(data, "syntax")[[2]])))
        names(result)
        # -> "main"      "assump"    "contrasts" "postHoc"   "emm"
        # (the names of the five output tables)
    }
}

## ----echo=TRUE----------------------------------------------------------------
# use the data set "ToothGrowth" and, if it exists, write it as jamovi-file
# using write_omv()
data("ToothGrowth", package = "jmvReadWrite")
# "retDbg" has to be set in order to return debug information to wrtDta
wrtDta <- jmvReadWrite::write_omv(ToothGrowth, "Trial.omv", retDbg = TRUE)
names(wrtDta)
# -> "mtaDta" "xtdDta" "dtaFrm"
# this debug information contains a list with the metadata ("mtaDta", e.g.,
# column and data type), the extended data ("xtdDta", e.g., variable lables),
# and the data frame (dtaFrm) for checking (understanding the file format) and
# debugging

# check whether the file was written to the disk, get the file information (size, etc.)
# and delete the file afterwards
list.files(".", "Trial.omv")
file.info("Trial.omv")[, c("size", "isdir", "mode")]
unlink("Trial.omv")

## ----echo=TRUE----------------------------------------------------------------
# reading and writing a file with the "sveAtt"-parameter permits you to keep
# essential meta-data to ensure that the written file looks and works like the
# original file (plus you modifications)
fleOMV <- system.file("extdata", "ToothGrowth.omv", package = "jmvReadWrite")
data <- jmvReadWrite::read_omv(fleOMV, sveAtt = TRUE)
# shows the names of the attributes for the whole data set (e.g., number of
# rows and columns) and the names of the attributes of the first column
names(attributes(data))
names(attributes(data[[1]]))
#
# perhaps do some modifications to the file here and write it back afterwards
jmvReadWrite::write_omv(data, "Trial.omv")
unlink("Trial.omv")

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(echo = TRUE,
                      comment = "#>",
                      collapse = TRUE,
                      message = FALSE,
                      warning = FALSE,
                      fig.height = 5,
                      fig.width = 7,
                      fig.align = "center",
                      out.width = "100%")

## ----eval=FALSE---------------------------------------------------------------
# install.packages("jmvReadWrite")

## ----eval=FALSE---------------------------------------------------------------
# if (!require(devtools)) install.packages("devtools")
# devtools::install_github("sjentsch/jmvReadWrite")

## ----echo=TRUE----------------------------------------------------------------
fleOMV <- system.file("extdata", "ToothGrowth.omv", package = "jmvReadWrite")
data <- jmvReadWrite::read_omv(fleOMV)
# if the "jmv"-package is installed, we can run a test analysis with the data
if (nzchar(system.file(package = "jmv"))) {
    jmv::ANOVA(
        formula = len ~ supp + dose + supp:dose,
        data = data,
        effectSize = "omega",
        modelTest = TRUE,
        homo = TRUE,
        norm = TRUE)
    }

## ----echo=TRUE----------------------------------------------------------------
fleOMV <- system.file("extdata", "ToothGrowth.omv", package = "jmvReadWrite")
data <- jmvReadWrite::read_omv(fleOMV, getSyn = TRUE)
# shows the syntax of the analyses from the .omv-file
# please note that syntax extraction may not work on all systems
# if the syntax couldn't be extracted, an empty list (length = 0) is returned,
# otherwise, the syntax of the analyses from the .omv-file is shown and
# the commands of the first and the second analysis are run, with the
# output of the second analysis assigned to the variable result
if (length(attr(data, "syntax")) >= 2) {
    attr(data, "syntax")
    # if the "jmv"-package is installed, we can run the analyses in "syntax"
    if (nzchar(system.file(package = "jmv"))) {
        eval(parse(text = attr(data, "syntax")[[1]]))
        eval(parse(text = paste0("result = ", attr(data, "syntax")[[2]])))
        names(result)
        # -> "main"      "assump"    "contrasts" "postHoc"   "emm"
        # (the names of the five output tables)
    }
}

## ----echo=TRUE----------------------------------------------------------------
# use the data set "ToothGrowth" and, if it exists, write it as jamovi-file
# using write_omv()
data("ToothGrowth", package = "jmvReadWrite")
# "retDbg" has to be set in order to return debug information to wrtDta
wrtDta <- jmvReadWrite::write_omv(ToothGrowth, "Trial.omv", retDbg = TRUE)
names(wrtDta)
# -> "mtaDta" "xtdDta" "dtaFrm"
# this debug information contains a list with the metadata ("mtaDta", e.g.,
# column and data type), the extended data ("xtdDta", e.g., variable lables),
# and the data frame (dtaFrm) for checking (understanding the file format) and
# debugging

# check whether the file was written to the disk, get the file information (size, etc.)
# and delete the file afterwards
list.files(".", "Trial.omv")
file.info("Trial.omv")[, c("size", "isdir", "mode")]
unlink("Trial.omv")

## ----echo=TRUE----------------------------------------------------------------
# reading and writing a file with the "sveAtt"-parameter permits you to keep
# essential meta-data to ensure that the written file looks and works like the
# original file (plus you modifications)
fleOMV <- system.file("extdata", "ToothGrowth.omv", package = "jmvReadWrite")
data <- jmvReadWrite::read_omv(fleOMV, sveAtt = TRUE)
# shows the names of the attributes for the whole data set (e.g., number of
# rows and columns) and the names of the attributes of the first column
names(attributes(data))
names(attributes(data[[1]]))
#
# perhaps do some modifications to the file here and write it back afterwards
jmvReadWrite::write_omv(data, "Trial.omv")
unlink("Trial.omv")

#' Search values in .omv-files for the statistical spreadsheet 'jamovi' (<https://www.jamovi.org>)
#'
#' @param dtaInp Either a data frame or the name of a jamovi data file to be read (including the path, if required; "FILENAME.omv"; default: NULL)
#' @param fleOut Name of the data file to be written (including the path, if required; "FILE_OUT.omv"; default: ""); if empty, the resulting data frame is
#'               returned instead
#' @param rplLst A list where each entry is a vector (with length 2) containing the original value and the to-replace-value (default: list())
#' @param whlTrm Whether the search term (first entry in the vectors) must be found exactly (TRUE) or whether a partial match is sufficient (FALSE; default:
#'               TRUE)
#' @param varInc Names of variables (character vector) to be included in the replacement (default: c())
#' @param varExc Names of variables (character vector) to be excluded from the replacement (default: c())
#' @param incNum Whether to include continuous variables in the replacement (default: TRUE)
#' @param incOrd Whether to include ordinal variables in the replacement (default: TRUE)
#' @param incNom Whether to include nominal variables in the replacement (default: TRUE)
#' @param incID  Whether to include ID variables in the replacement (default: TRUE)
#' @param incCmp Whether to include Computed variables in the replacement (default: TRUE)
#' @param incRcd Whether to include Recoded variables in the replacement (default: TRUE)
#' @param psvAnl Whether analyses that are contained in the input file shall be transferred to the output file (default: FALSE)
#' @param ...    Additional arguments passed on to methods; see Details below
#'
#' @return a data frame (only returned if `fleOut` is empty) with the replaced values
#'
#' @details
#' * `rplLst` is a list. Each list entry contains a vector (with length 2), where the first entry is the original value, and the second entry is the value the
#'   original value is to be replaced with.
#' * `whlTrm` indicates whether partial matches of the original value(s) shall replaced (e.g., for original: 24 and replacement: 34, 241 will be changed into
#'   341).
#' * `varInc` and `varExc` determine which variables are included or excluded from the replacement. If both are given, a warning is issued and `varInc` takes
#'   precedence. `varInc` makes that only in these variables, the replacement requested by `rplLst` is carried out, if `varExc` is given, for all variables of
#'   the input data set, except those defined in `varExc`, the replacement is carried out.
#' * The ellipsis-parameter (`...`) can be used to submit arguments / parameters to the function that is used for reading and writing the data. Clicking on the
#'   respective function under “See also”, you can get a more detailed overview over which parameters each of those functions take. The functions are:
#'   `read_omv` and `write_omv` (for jamovi-files).
#'
#' @seealso `replace_omv` uses [jmvReadWrite::read_omv()] and [jmvReadWrite::write_omv()] for reading and writing jamovi-files.
#'
#' @examples
#' \dontrun{
#' bfi_sample <- jmvReadWrite::bfi_sample
#' # the gender in the original data file is plural...
#' table(bfi_sample$gender)
#' # and shall be converted to singular
#' rplDF <- jmvReadWrite::replace_omv(dtaInp = bfi_sample,
#'            rplLst = list(c("Females", "Female"), c("Males", "Male")))
#' table(rplDF$gender)
#' # with giving an output file name, the data set is written
#' nmeOut <- tempfile(fileext = ".omv")
#' jmvReadWrite::replace_omv(bfi_sample, fleOut = nmeOut,
#'   rplLst = list(c("Females", "Female"), c("Males", "Male")))
#' file.exists(nmeOut)
#' rplDF <- jmvReadWrite::read_omv(nmeOut)
#' table(rplDF$gender)
#' unlink(nmeOut)
#' # it is sensible to check / search for the original values before running replace_omv
#' jmvReadWrite::search_omv(bfi_sample, 24, whlTrm = TRUE)
#' rplDF <- jmvReadWrite::replace_omv(bfi_sample, rplLst = list(c(24, NA)))
#' table(rplDF$age)
#' }
#'
#' @export replace_omv
#'
replace_omv <- function(dtaInp = NULL, fleOut = "", rplLst = list(), whlTrm = TRUE, varInc = c(), varExc = c(), incNum = TRUE, incOrd = TRUE, incNom = TRUE,
                        incID = TRUE, incCmp = TRUE, incRcd = TRUE, psvAnl = FALSE, ...) {

    # check the input parameter:
    if (length(rplLst) < 1 || !is.list(rplLst) || !all(vapply(rplLst, length, integer(1)) == 2)) {
        stop("Calling replace_omv requires the parameter rplLst (a list where each entry is a vector with length 2; see Details in help).")
    }

    # check and import input data set (either as data frame or from a file)
    if (!is.null(list(...)[["fleInp"]])) stop("Please use the argument dtaInp instead of fleInp.")
    dtaFrm <- inp2DF(dtaInp = dtaInp, ...)
    dtaFrm <- jmvAtt(dtaFrm)

    srcNme <- chkInE(dtaFrm, varInc, varExc, incNum, incOrd, incNom, incID, incCmp, incRcd)
    for (i in seq_along(srcNme)) {
        for (j in seq_along(rplLst)) {
            dtaFrm <- rplVal(dtaFrm, srcNme[i], rplLst[[j]], whlTrm)
        }
    }

    # rtnDta in globals.R (unified function to either write the data frame, open it in a new jamovi session or return it)
    rtnDta(dtaFrm = dtaFrm, fleOut = fleOut, dtaTtl = jmvTtl("_rplc"), psvAnl = psvAnl, dtaInp = dtaInp, ...)
}

chkInE <- function(dtaFrm = NULL, varInc = c(), varExc = c(), incNum = TRUE, incOrd = TRUE, incNom = TRUE, incID = TRUE, incCmp = TRUE, incRcd = TRUE) {
    if (!is.null(varInc) && length(varInc) > 0 && !is.null(varExc) && length(varExc) > 0) warning("Both varInc and varExc are given, varInc takes precedence.")
    if        (!is.null(varInc) && length(varInc) > 0 && all(nzchar(varInc))) {
        if (!all(varInc %in% names(dtaFrm))) {
            stop(sprintf("All variables in varInc must be contained in the original data set (%s are not).", paste(varInc[!(varInc %in% names(dtaFrm))], collapse = ", ")))
        }
        srcNme <- varInc
    } else if (!is.null(varExc) && length(varExc) > 0 && all(nzchar(varExc))) {
        if (!all(varExc %in% names(dtaFrm))) {
            stop(sprintf("All variables in varExc must be contained in the original data set (%s are not).", paste(varExc[!(varExc %in% names(dtaFrm))], collapse = ", ")))
        }
        srcNme <- setdiff(names(dtaFrm), varExc)
    } else {
        srcNme <- names(dtaFrm)
    }

    incClT <- c("Data", rep("Computed", incCmp), rep("Recoded", incRcd))
    incMsT <- c(rep("ID", incID), rep("Nominal", incNom), rep("Ordinal", incOrd), rep("Continuous", incNum))
    srcNme <- srcNme[vapply(dtaFrm[srcNme], function(x) is.null(attr(x,  "columnType")) || any(attr(x,  "columnType") == incClT), logical(1))]
    srcNme <- srcNme[vapply(dtaFrm[srcNme], function(x) is.null(attr(x, "measureType")) || any(attr(x, "measureType") == incMsT), logical(1))]

    srcNme
}

rplVal <- function(dtaFrm = NULL, crrCll = "", crrRpl = NULL, whlTrm = TRUE) {
    if (is.factor(dtaFrm[[crrCll]])) {
        if (crrRpl[1] %in% levels(dtaFrm[[crrCll]])) levels(dtaFrm[[crrCll]]) <- gsub(crrRpl[1], crrRpl[2], levels(dtaFrm[[crrCll]]))
    } else if (is.numeric(dtaFrm[[crrCll]])) {
        srcSel <- srcClm(dtaFrm[[crrCll]], gsub("\\.", "\\\\.", as.character(crrRpl[1])), whlTrm)
        if (any(srcSel)) dtaFrm[srcSel, crrCll] <- dtaFrm[srcSel, crrCll] + diff(as.numeric(crrRpl))
    } else if (is.character(dtaFrm[[crrCll]])) {
        srcSel <- srcClm(dtaFrm[[crrCll]], gsub("\\.", "\\\\.", as.character(crrRpl[1])), whlTrm)
        if (any(srcSel)) dtaFrm[srcSel, crrCll] <- crrRpl[2]
    }
    # other variable types are already caught by jmvAtt() above

    dtaFrm
}

test_that("replace_omv works", {
    tmpDF <- jmvReadWrite::bfi_sample
    expect_equal(as.list(table(tmpDF$gender)),                                                                                  list(Females = 174L, Males = 80L))
    expect_equal(as.list(table(replace_omv(dtaInp = tmpDF, rplLst = list(c("Females", "Female"), c("Males", "Male")))$gender)), list(Female  = 174L, Male  = 80L))
    expect_equal(as.list(table(replace_omv(dtaInp = tmpDF, rplLst = list(c("females", "Female"), c("males", "Male")))$gender)), list(Females = 174L, Males = 80L))
    expect_equal(which(is.na(replace_omv(dtaInp = tmpDF, rplLst = list(c(24, NA)))$age)), c(2, 6, 11, 12, 61, 75, 77, 99, 109, 169, 197, 200))
    expect_equal(which(is.na(replace_omv(dtaInp = tmpDF, rplLst = list(c(24, NA)))$age)), which(tmpDF$age == 24))
    expect_false(any(replace_omv(dtaInp = tmpDF, rplLst = list(c(24, NA)))$age == 24, na.rm = TRUE))

    tmpDF <- jmvReadWrite::bfi_sample2
    tmpDF <- cbind(tmpDF, data.frame(age_log = log(tmpDF$age)))
    expect_equal(vapply(replace_omv(dtaInp = tmpDF, rplLst = list(c(1, 6)), varExc = c("A1", "A2", "A3", "A4", "A5"))[2:26], function(x) c(sum(x == 1), sum(x == 6)), integer(2), USE.NAMES = FALSE),
      t(matrix(c(78, 5, 8, 9, 6, rep(0, 20), 9, 82, 69, 106, 65, 49, 55, 47, 78, 75, 73, 76, 54, 68, 68, 81, 65, 72, 74, 94, 83, 95, 54, 110, 66), ncol = 2)))
    expect_equal(vapply(replace_omv(dtaInp = tmpDF, rplLst = list(c(1, 6)), varInc = c("A1", "A2", "A3", "A4", "A5"))[2:26], function(x) c(sum(x == 1), sum(x == 6)), integer(2), USE.NAMES = FALSE),
      t(matrix(c(rep(0, 5), 2, 7, 7, 71, 45, 52, 51, 15, 14, 4, 59, 30, 42, 41, 64, 3, 75, 4, 4, 60,
                 87, 87, 77, 115, 71, 47, 48, 40, 7, 30, 21, 25, 39, 54, 64, 22, 35, 30, 33, 30, 80, 20, 50, 106, 6), ncol = 2)))
    expect_equal(jmvReadWrite::replace_omv(dtaInp = tmpDF, rplLst = list(c("66574", "Trial")))[244, "ID"],  "Trial")
    expect_equal(which(jmvReadWrite::replace_omv(dtaInp = tmpDF, rplLst = list(c("66574", "Trial"))) != tmpDF),  244)
    expect_equal(jmvReadWrite::replace_omv(dtaInp = tmpDF, rplLst = list(c("A7209", "Trial")))[244, "ID2"], "Trial")
    expect_equal(which(jmvReadWrite::replace_omv(dtaInp = tmpDF, rplLst = list(c("A7209", "Trial"))) != tmpDF), 7244)
    expect_equal(unname(colSums(jmvReadWrite::replace_omv(dtaInp = tmpDF, rplLst = list(c("4.", "0.")), whlTrm = FALSE) != tmpDF)), c(rep(0, 29), 5))
    expect_equal(mean(jmvReadWrite::replace_omv(dtaInp = tmpDF, rplLst = list(c("4.", "0.")), whlTrm = FALSE)[, "age_log"]), 3.1970427)

    expect_warning(expect_equal(vapply(replace_omv(dtaInp = tmpDF, rplLst = list(c(1, 6)), varInc = c("A1", "A2", "A3", "A4", "A5"), varExc = c("ID"))[2:26],
      function(x) c(sum(x == 1), sum(x == 6)), integer(2), USE.NAMES = FALSE),
      t(matrix(c(rep(0, 5), 2, 7, 7, 71, 45, 52, 51, 15, 14, 4, 59, 30, 42, 41, 64, 3, 75, 4, 4, 60,
                 87, 87, 77, 115, 71, 47, 48, 40, 7, 30, 21, 25, 39, 54, 64, 22, 35, 30, 33, 30, 80, 20, 50, 106, 6), ncol = 2))),
      regexp = "Both varInc and varExc are given, varInc takes precedence\\.")

    expect_error(replace_omv(dtaInp = tmpDF),
      regexp = "^Calling replace_omv requires the parameter rplLst \\(a list where each entry is a vector with length 2; see Details in help\\)\\.")
    expect_error(replace_omv(dtaInp = tmpDF, rplLst = c()),
      regexp = "^Calling replace_omv requires the parameter rplLst \\(a list where each entry is a vector with length 2; see Details in help\\)\\.")
    expect_error(replace_omv(dtaInp = tmpDF, rplLst = list(c(1, 2), c(3, 4, 5), c(6, 7))),
      regexp = "^Calling replace_omv requires the parameter rplLst \\(a list where each entry is a vector with length 2; see Details in help\\)\\.")
    expect_error(replace_omv(fleInp = tmpDF, rplLst = list(c(1, 2))),    regexp = "^Please use the argument dtaInp instead of fleInp\\.")
    expect_error(replace_omv(dtaInp = cbind(tmpDF, data.frame(AD2 = as.complex(tmpDF$age))), rplLst = list(c(24, NA))), regexp = "Variable type \\w+ not implemented.")
    expect_error(replace_omv(dtaInp = tmpDF, rplLst = list(c(1, 6)), varInc = c("not_a_var")), regexp = "^All variables in varInc must be contained in the original data set \\(\\w+ are not\\)\\.")
    expect_error(replace_omv(dtaInp = tmpDF, rplLst = list(c(1, 6)), varExc = c("not_a_var")), regexp = "^All variables in varExc must be contained in the original data set \\(\\w+ are not\\)\\.")

    rm(tmpDF)
})

#' Read files created of the statistical spreadsheet 'jamovi' (<https://www.jamovi.org>)
#'
#' @param fleInp Name (including the path, if required) of the 'jamovi'-file to be read ("FILENAME.omv"; default: "")
#' @param useFlt Apply filters (remove the lines where the filter is set to 0; default: FALSE)?
#' @param rmMsVl Remove values defined as missing values (replace them with NA; default: FALSE)?
#' @param sveAtt Store attributes that are not required in the data set (if you want to write the same data set using write_omv; default: FALSE)?
#' @param getSyn Extract syntax from the analyses in the 'jamovi'-file and store it in the attribute "syntax" (default: FALSE)?
#' @param getHTM Store index.html in the attribute "HTML" (default: FALSE)?
#'
#' @return data frame (can be directly used with functions included in the R-package `jmv` and syntax from 'jamovi'; also compatible with the format of the
#'         R-package `foreign`)
#'
#' @examples
#' nmeInp <- system.file("extdata", "ToothGrowth.omv", package = "jmvReadWrite")
#' data <- jmvReadWrite::read_omv(fleInp = nmeInp)
#' str(data)
#' # shows the data frame including all jamovi attributes:
#' # 'data.frame':	60 obs. of  16 variables:
#' #  $ Filter 1          : logi  TRUE TRUE TRUE TRUE TRUE TRUE ...
#' #   ..- attr(*, "name")= chr "Filter 1"
#' #   ..- attr(*, "id")= int 18
#' #   ..- attr(*, "columnType")= chr "Filter"
#' #   ..- attr(*, "dataType")= chr "Integer"
#' #   ..- attr(*, "measureType")= chr "Nominal"
#' #   ..- attr(*, "formula")= chr "logLen < 1.5"
#' #   ..- attr(*, "formulaMessage")= chr ""
#' #   ..- attr(*, "parentId")= int 0
#' #   ..- attr(*, "width")= int 78
#' #   ..- attr(*, "type")= chr "integer"
#' #   ..- attr(*, "importName")= chr ""
#' #   ..- attr(*, "description")= chr ""
#' #   ..- attr(*, "transform")= int 0
#' #   ..- attr(*, "edits")= list()
#' #   ..- attr(*, "missingValues")= list()
#' #   ..- attr(*, "filterNo")= int 0
#' #   ..- attr(*, "active")= logi FALSE
#' # ... (continued)
#'
#' # getSyn requires jmvcore and RProtoBuf, and is thus not run here
#' \dontrun{
#' data <- jmvReadWrite::read_omv(fleInp = nmeInp, getSyn = TRUE)
#' }
#' # if the syntax couldn't be extracted, an empty list - length = 0 - is returned,
#' # otherwise, the commands are shown and the first analysis is run, with the output
#' # from the second analysis being assigned to the variable result
#' if (length(attr(data, "syntax")) >= 1) {
#'     print(attr(data, "syntax"))
#'     if (nzchar(system.file(package = "jmv"))) {
#'         # the print-function is only used to force devtools::run_examples() to show output
#'         eval(parse(text = paste0("result = ", attr(data, "syntax")[1])))
#'         # without assigning the output to a variable, the command would be:
#'         # eval(parse(text = attr(data, "syntax")[1]))
#'         print(names(result))
#'         print(result$main)
#'         # -> "main"      "assump"    "contrasts" "postHoc"   "emm"       "residsOV"
#'         # (the names of the six output tables)
#'     }
#' }
#'
#' @export read_omv
#'
read_omv <- function(fleInp = "", useFlt = FALSE, rmMsVl = FALSE, sveAtt = TRUE, getSyn = FALSE, getHTM = FALSE) {
    if (nchar(fleInp) == 0) stop("File name to the input data file needs to be given as parameter (fleInp = ...).")

    # check and format input file names - do not use chkExt here
    if (!hasExt(fleInp, c("omv", "omt"))) {
        stop("read_omv only reads jamovi files (.omv / .omt), use convert_to_omv first, if you want to read other files types.")
    }
    fleInp <- fmtFlI(fleInp, maxLng = 1)
    fleLst <- zip::zip_list(fleInp)$filename
    # check whether the file list contains either the file "meta" (newer jamovi file format) or MANIFEST.MF (older format)
    chkFle(fleInp, isZIP = TRUE)
    chkMnf(fleInp, fleLst[grepl("^meta$|MANIFEST.MF$", fleLst)])

    # https://github.com/jamovi/jamovi/blob/current-dev/server/jamovi/server/formatio/omv.py describes
    # how to handle the different jamovi-archive-versions
    # read and decode files: Manifest, metadata (metadata.json), metadata about value labels (xdata.json), binary numeric data (data.bin)
    # and binary string data (strings.bin; if present: it only exists if there are columns that contain text variables)
    strBin <- any(grepl("strings.bin", fleLst))

    # load the meta-data (global and data column attributes of the data set) and the extended
    # data (value labels)
    mtaDta <- getTxt(fleInp, "metadata.json")[["dataSet"]]
    xtdDta <- getTxt(fleInp, "xdata.json")
    binHdl <- getHdl(fleInp, "data.bin", "rb")
    if (strBin) {
        strHdl <- getHdl(fleInp, "strings.bin", "rb")
        strPos <- 0
    }

    # check meta-data, change weights if NULL
    if (!all(grepl(grpMta, names(mtaDta)))) stop("Unimplemeted field in the meta data (data frame).")

    # determine rows and columns and create data frame
    rowNum <- mtaDta$rowCount
    colNum <- mtaDta$columnCount
    if (length(mtaDta$fields) != colNum) stop("Number of fields in the metadata is not matching up the number of columns.")
    dtaFrm <- stats::setNames(data.frame(matrix(NA, nrow = rowNum, ncol = colNum)), vapply(mtaDta$fields, "[[", character(1), "name"))
    fltLst <- c()

    # iterate through fields
    for (i in seq_len(colNum)) {
        # assign metadate for the current column, and check whether there are any unimplemented entries
        mtaCol <- mtaDta$fields[[i]]
        if (!all(grepl(grpMta, names(mtaCol)))) stop("Unimplemeted field in the meta data (column).")

        # type: determines the format in the binary file
        if        (chkFld(mtaCol, "type", "integer")) {
            crrCol <- readBin(binHdl, integer(), n = rowNum)
        } else if (chkFld(mtaCol, "type", "number"))  {
            crrCol <- readBin(binHdl,  double(), n = rowNum)
        } else if (chkFld(mtaCol, "type", "string"))  {
            crrCol <- vector("character", length = rowNum)
            crrIdx <- readBin(binHdl, integer(), n = rowNum)
            if (rowNum > 0) {
                for (j in seq(rowNum)) {
                    if (is.na(crrIdx[j])) next
                    if (crrIdx[j] == strPos) {
                        crrCol[j] <- readBin(strHdl, "character", n = 1)
                        strPos <- strPos + length(charToRaw(crrCol[j])) + 1
                    } else {
                        stop("Mismatch between the assumed and the actual position in the file when reading string variables.")
                    }
                }
            }
            rm(crrIdx)
        } else {
            stop(sprintf("Variable type \"%s\" not implemented.", mtaCol$type))
        }

        # assign name of the current column, add value labels, assign crrCol to dtaFrm, and add attributes (if present)
        crrNme <- mtaCol$name
        dtaFrm[[crrNme]] <- valLbl(crrCol, mtaCol, xtdDta)
        if (chkFld(mtaCol, "measureType", "ID")) attr(dtaFrm[[crrNme]], "jmv-id")   <- TRUE
        if (chkFld(mtaCol, "description", ".+")) attr(dtaFrm[[crrNme]], "jmv-desc") <- mtaCol[["description"]]
        if (chkFld(mtaCol, "columnType", "Filter") || chkFld(mtaCol, "name", "^Filter [0-9]+$")) fltLst <- c(fltLst, i)

        if (sveAtt) {
            dtaFrm[crrNme] <- setAtt(names(mtaFld),   mtaCol, dtaFrm[crrNme])
        } else {
            dtaFrm[crrNme] <- setAtt("missingValues", mtaCol, dtaFrm[crrNme])
        }

        if (rmMsVl) {
            mssLst <- attr(dtaFrm[[crrNme]], "missingValues")
            if (length(mssLst) > 0) {
               crrAtt <- attributes(dtaFrm[crrNme])
               rmvLvl <- rep(FALSE, length(levels(dtaFrm[[crrNme]])))
               for (j in seq_along(mssLst)) {
                   dtaFrm[[crrNme]][eval(parse(text = paste0("dtaFrm[[\"", crrNme, "\"]]", mssLst[[j]])))] <- NA
                   rmvLvl <- rmvLvl | eval(parse(text = paste0("levels(dtaFrm[[\"", crrNme, "\"]]) ", mssLst[j])))
               }
               dtaFrm[[crrNme]] <- dtaFrm[[crrNme]][, drop = TRUE]
               crrAtt$missingValues <- list()
               crrAtt$values <- crrAtt$values[!rmvLvl]
               dtaFrm[crrNme] <- setAtt(setdiff(names(crrAtt), names(attributes(dtaFrm[crrNme]))), crrAtt, dtaFrm[crrNme])
               rm(crrAtt, rmvLvl)
            }
        }
        rm(crrCol)
    }

    # close and remove the binary file(s)
    clsHdl(binHdl)
    rm(binHdl)
    if (strBin) {
        clsHdl(strHdl)
        rm(strHdl, strPos)
    }

    # handle filters
    if (useFlt) {
        fltInc <- rep(TRUE, dim(dtaFrm)[1])
        for (i in fltLst) fltInc <- fltInc & dtaFrm[[i]]
        dtaFrm <- dtaFrm[fltInc, ]
        dtaFrm[fltLst] <- NULL
    } else if (length(fltLst) > 0) {
        attr(dtaFrm, "fltLst") <- names(dtaFrm)[fltLst]
    }

    # store data frame attributes except rowCount, columnCount and fields (the first two available
    # as data frame dimensions, the latter stored in the variables / columns)
    if (sveAtt) {
        dtaFrm <- setAtt(setdiff(names(mtaGlb), c("rowCount", "columnCount", "fields")), mtaDta, dtaFrm)
    }

    # reset / empty some attributes ("addedRows" [for the data.frame] and "edits" for each column) for jamovi templates (.omt)
    if (hasExt(fleInp, "omt")) {
        dtaFrm <- nllAtt(dtaFrm, "addedRows")
        for (crrNme in names(dtaFrm)) dtaFrm[[crrNme]] <- nllAtt(dtaFrm[[crrNme]], "edits")
    }

    # handle weights
    if (is.character(mtaDta$weights) && nzchar(mtaDta$weights)) {
        # TO-DO: check whether the protobuffers (.. weights) contain
        # anything useful beyond that information
        attr(dtaFrm, "jmv-weights-name") <- mtaDta$weights
        attr(dtaFrm, "jmv-weights")      <- as.vector(dtaFrm[, mtaDta$weights])
        # it would be possible to keep both jmv-weights-name and weights, but
        # this might be rather confusing, so the version jamovi uses internally
        # (jmv-weights-name) is chosen and the other (weights) disregarded
        attr(dtaFrm, "weights")          <- NULL
    }

    # import and extract syntax from the analyses
    if (getSyn) {
        anlLst <- fleLst[grepl("[0-9]+\\s\\w+/analysis", fleLst)]
        savSyn <- c()
        savPtB <- list()
        if (length(anlLst) > 0 && jmvPtB()) {
            for (anlNme in anlLst) {
                anlPtB <- RProtoBuf::read(jamovi.coms.AnalysisResponse, anlHdl <- getHdl(fleInp, anlNme, "rb"))
                clsHdl(anlHdl)
                rm(anlHdl)
                if (!grepl("empty$", dirname(anlNme))) {
                    savSyn <- c(savSyn, gsub("\\( ", "\\(", gsub("\\n\\s+", " ", fndSyn(anlPtB$results))))
                }
                anlPtB$results <- NULL
                savPtB[[anlNme]] <- anlPtB
            }
        }
        attr(dtaFrm, "syntax")   <- savSyn
        attr(dtaFrm, "protobuf") <- savPtB
    }

    # import the HTML output
    if (getHTM) {
        attr(dtaFrm, "HTML") <- getTxt(fleInp, "index.html")
    }

    # return the resulting data frame
    dtaFrm
}

valLbl <- function(crrCol = NULL, mtaCol = NULL, xtdDta = NULL) {
    crrNme <- mtaCol$name
    if (any(crrNme == names(xtdDta))) {
        if        (chkFld(mtaCol, "columnType", "Filter") || chkFld(mtaCol, "name", "^Filter [0-9]+$")) {
            crrCol <- as.logical(crrCol)
        } else if (chkFld(mtaCol, "columnType", "Data|Recoded")) {
            crrCol <- factor(crrCol, levels = unlist(lapply(xtdDta[[crrNme]]$labels, function(m) m[1])),
                                     labels = unlist(lapply(xtdDta[[crrNme]]$labels, function(m) m[2])),
                                     ordered = chkFld(mtaCol, "measureType", "Ordinal"))
            if (identical(sort(levels(crrCol)), c("0", "1")))        crrCol <- as.logical(gsub("^1$", "TRUE", gsub("^0$", "FALSE", crrCol)))
            if (identical(sort(levels(crrCol)), c("FALSE", "TRUE"))) crrCol <- as.logical(crrCol)
            if (!is.logical(crrCol) && chkFld(mtaCol, "dataType", "Integer") &&
              all(vapply(xtdDta[[crrNme]]$labels, function(m) grepl("^\\d+$", m[2]) && m[1] == as.integer(m[2]), logical(1)))) {
                attr(crrCol, "values") <- unlist(lapply(xtdDta[[crrNme]]$labels, function(m) m[1]))
            }
        } else {
            stop(sprintf("Error when reading value label - likely the column type is not implemented (yet): %s - %s - %s",
                         crrNme, mtaCol$dataType, mtaCol$columnType))
        }
    }

    crrCol
}

# =================================================================================================
# read_all: for reading data files from various formats (incl. functions that are called)
read_all <- function(fleInp = "", usePkg = c("foreign", "haven"), selSet = "", ...) {
    if (nchar(fleInp) == 0) stop("File name to the input data file needs to be given as parameter (fleInp = ...).")

    # check whether the file exists
    fleInp <- fmtFlI(fleInp, maxLng = 1)
    varArg <- list(...)
    usePkg <- match.arg(usePkg)
    dtaFrm <- NULL

    # OMV / OMT
    if        (hasExt(fleInp, c("omv", "omt"))) {
        dtaFrm <- tryCatch(do.call(read_omv, adjArg("read_omv", list(fleInp = fleInp), varArg, "fleInp")),
                           error   = function(errMsg) tryErr(fleInp, errMsg),
                           warning = function(wrnMsg) tryWrn(fleInp, wrnMsg))
    # CSV
    } else if (hasExt(fleInp, c("csv"))) {
        dtaFrm <- tryCatch(rmvQtn(do.call(utils::read.table, adjArg("read.table", list(file = fleInp, sep = ",",  quote = "\"", header = TRUE, fill = TRUE), varArg, "file"))),
                           error   = function(errMsg) tryErr(fleInp, errMsg),
                           warning = function(wrnMsg) tryWrn(fleInp, wrnMsg))
    # TSV
    } else if (hasExt(fleInp, c("tsv"))) {
        dtaFrm <- tryCatch(rmvQtn(do.call(utils::read.table, adjArg("read.table", list(file = fleInp, sep = "\t", quote = "\"", header = TRUE, fill = TRUE), varArg, "file"))),
                           error   = function(errMsg) tryErr(fleInp, errMsg),
                           warning = function(wrnMsg) tryWrn(fleInp, wrnMsg))
    # Rdata
    } else if (hasExt(fleInp, c("rdata", "rda"))) {
        dtaFrm <- tryCatch({
                             load(fleInp, rdaTmp <- new.env())
                             if (length(rdaTmp) != 1 && selSet == "") stop("The Rdata-file must include only one object.")
                             rdaTmp[[ifelse(selSet == "", names(rdaTmp)[1], selSet)]]
                           },
                           error   = function(errMsg) tryErr(fleInp, errMsg),
                           warning = function(wrnMsg) tryWrn(fleInp, wrnMsg))
        if (methods::is(dtaFrm, "tbl_df") || any(vapply(dtaFrm, function(C) methods::is(C, "haven_labelled"), logical(1))))
            dtaFrm <- clnTbb(dtaFrm, c("format.sas", "format.spss", "format.stata", "display_width"), jmvLbl = TRUE)
    # RDS
    } else if (hasExt(fleInp, c("rds"))) {
        dtaFrm <- tryCatch(do.call(readRDS, adjArg("readRDS", list(file = fleInp), varArg, "file")),
                           error   = function(errMsg) tryErr(fleInp, errMsg),
                           warning = function(wrnMsg) tryWrn(fleInp, wrnMsg))
        if (methods::is(dtaFrm, "tbl_df") || any(vapply(dtaFrm, function(C) methods::is(C, "haven_labelled"), logical(1))))
            dtaFrm <- clnTbb(dtaFrm, c("format.sas", "format.spss", "format.stata", "display_width"), jmvLbl = TRUE)
    # SPSS (haven / foreign)
    } else if (hasExt(fleInp, c("sav", "zsav"))) {
        dtaFrm <-  getSPSS(fleInp, usePkg, varArg)
    # Stata (haven / foreign)
    } else if (hasExt(fleInp, c("dta"))) {
        dtaFrm <- getStata(fleInp, usePkg, varArg)
    # SAS (haven / foreign [not all formats])
    } else if (hasExt(fleInp, c("sas7bdat", "sd2", "sd7", "xpt", "stx", "stc"))) {
        dtaFrm <-   getSAS(fleInp, usePkg, varArg)
    }

    # check whether the input data are a data frame with the correct dimensions
    if (!is.null(dtaFrm)) chkDtF(dtaFrm)

    dtaFrm
}


# =================================================================================================
# helper functions for read_omv and read_all (in alphabetical order)

chkMnf <- function(fleOMV = "", fleMnf = c("")) {
    if (length(fleMnf) < 1) {
        stop(sprintf("File \"%s\" has not the correct file format (is missing the jamovi-file-manifest).", basename(fleOMV)))
    }

    # check the version information in the manifest and whether they are currently supported
    # [[1]] points to the first manifest file, in case both (MANIFEST.MF and meta) exist
    crrTxt <- getTxt(fleOMV, fleMnf[[1]])
    if (length(crrTxt) != length(lstMnf) || !all(grepl(paste(vapply(lstMnf, "[[", character(1), 1), collapse = "|"), crrTxt))) {
        stop(sprintf(paste("The file you are trying to read (%s) has an improper manifest file (meta) and is likely corrupted.",
                           "If the error persists, send the file to sebastian.jentschke@uib.no!"), basename(fleOMV)))
    }
    for (i in seq_along(lstMnf)) {
        # if no version number is given, skip this entry
        if (length(lstMnf[[i]]) == 1) break
        # compare versions, and if the current version is higher then the version defined in lstMnf, issue a warning
        crrVer <- trimws(strsplit(crrTxt[grepl(paste0(lstMnf[[i]][1], ":"), crrTxt)], ":")[[1]])[-1]
        if (utils::compareVersion(crrVer, lstMnf[[i]][-1]) > 0) {
             warning(sprintf(paste("The file that you are trying to read (%s) was written with a version of jamovi that currently is not implemented",
                                   "(%s: %s vs. %s) and therefore may be read incorrectly. Please send the file to sebastian.jentschke@uib.no!"),
                                   basename(fleOMV), lstMnf[[i]][1], crrVer, lstMnf[[i]][-1]))
        }
    }

    TRUE
}

chkSyn <- function(resElm = NULL) {
   # checks whether the results element is a syntax entry
   # it has to have the name syntax and the preformatted attribute must not be empty
   utils::hasName(resElm, "name")         && resElm[["name"]]         == "syntax" &&
   utils::hasName(resElm, "preformatted") && resElm[["preformatted"]] != ""
}

clsHdl <- function(crrHdl = NULL) {
    crrFle <- summary(crrHdl)$description
    close(crrHdl)
    unlink(crrFle)
    rm(crrFle, crrHdl)
}

clnFgn <- function(dtaFrm = NULL) {
    if (is.null(dtaFrm)) return(dtaFrm)
    chkDtF(dtaFrm)

    if (chkAtt(dtaFrm, "variable.labels")) {
        varLbl <- attr(dtaFrm, "variable.labels")
        Encoding(varLbl) <- "latin1"
        for (crrNme in names(dtaFrm)) {
            if (crrNme %in% names(varLbl) && varLbl[[crrNme]] != "") {
                attr(dtaFrm[[crrNme]], "jmv-desc") <- cnvUTF(varLbl[[crrNme]])
            }
        }
        attr(dtaFrm, "variable.labels") <- NULL
    }

    dtaFrm
}

clnTbb <- function(dtaFrm = NULL, rmvAtt = c(), jmvLbl = FALSE) {
    if (is.null(dtaFrm)) return(dtaFrm)

    # convert tibble to data.frame (remove tbl_df, tbl from class)
    class(dtaFrm) <- setdiff(class(dtaFrm), c("tbl_df", "tbl", rmvAtt))
    chkDtF(dtaFrm)

    for (crrNme in names(dtaFrm)) {
        # convert haven_labelled to factors
        if (methods::is(dtaFrm[[crrNme]], "haven_labelled")) {
            tmpLbl <- gsub("value label", "", names(attr(dtaFrm[[crrNme]], "labels")))
            if (all(nzchar(tmpLbl))) {
                dtaFrm[[crrNme]] <- cnvCol(dtaFrm[[crrNme]], "factor")
            } else if (!any(nzchar(tmpLbl))) {
                class(dtaFrm[[crrNme]]) <- setdiff(class(dtaFrm[[crrNme]]), c("haven_labelled", "vctrs_vctr"))
                attr(dtaFrm[[crrNme]], "labels") <- NULL
                if (is.numeric(dtaFrm[[crrNme]]) && all(dtaFrm[[crrNme]] %% 1 == 0, na.rm = TRUE))
                    class(dtaFrm[[crrNme]]) <- gsub("numeric", "integer", class(dtaFrm[[crrNme]]))
            } else {
                tmpLbl[!nzchar(tmpLbl)] <- sprintf("%d", which(!nzchar(tmpLbl)))
                names(attr(dtaFrm[[crrNme]], "labels")) <- tmpLbl
                dtaFrm[[crrNme]] <- cnvCol(dtaFrm[[crrNme]], "factor")
            }
        }
        # remove attributes given in rmvAtt
        for (crrAtt in rmvAtt) {
            if (! is.null(attr(dtaFrm[[crrNme]], crrAtt))) {
                attr(dtaFrm[[crrNme]], crrAtt) <- NULL
            }
        }
        # convert label to jamovi label
        if (jmvLbl && !is.null(attr(dtaFrm[[crrNme]], "label"))) {
            if (nzchar(gsub("variable label", "", attr(dtaFrm[[crrNme]], "label"))))
                attr(dtaFrm[[crrNme]], "jmv-desc") <- attr(dtaFrm[[crrNme]], "label")
            attr(dtaFrm[[crrNme]], "label") <- NULL
        }
    }

    dtaFrm
}

fndSyn <- function(resElm = NULL) {
    if (chkSyn(resElm)) {
        resElm[["preformatted"]]
    } else if (utils::hasName(resElm, "group") && length(resElm[["group"]]) > 0) {
        for (obj in resElm[["group"]][["elements"]]) {
            ret <- Recall(obj)
            if (!is.null(ret)) return(ret)
        }
    }
}

getHdl <- function(fleOMV = "", crrFle = "", crrMde = "r") {
    tryCatch({
                 zip::unzip(fleOMV, crrFle, exdir = tempdir(), junkpaths = TRUE)
                 file(file.path(tempdir(), list.files(path = tempdir(), pattern = basename(crrFle))), crrMde)
             },
             error = function(errMsg) {
                 message(sprintf("The file \"%s\" could not be extracted from \"%s\".\nPlease send the file to sebastian.jentschke@uib.no!\nError message: %s\n",
                           crrFle, fleOMV, errMsg))
                 invisible(NULL)
             }
        )
}

getSAS   <- function(fleInp = "", usePkg = "", varArg = list()) {
    # SAS data (haven)
    if (hasExt(fleInp, c("sas7bdat", "sd2", "sd7"))) {
        if        (usePkg == "haven"   && hasPkg("haven"))   {
            hvnTmp <- tryCatch(do.call(haven::read_sas, adjArg("haven::read_sas", list(data_file = fleInp), varArg, "data_file")),
                               error   = function(errMsg) tryErr(fleInp, errMsg),
                               warning = function(wrnMsg) tryErr(fleInp, wrnMsg))
            if (is.null(hvnTmp)) return(hvnTmp)
            clnTbb(hvnTmp, c("format.sas", "display_width"), jmvLbl = TRUE)
        } else {
            stop(sprintf("In order to read the SAS-file \"%s\" the R-packages \"haven\" needs to be installed.", basename(fleInp)))
        }
    # SAS-transport-files (haven / foreign)
    } else if (hasExt(fleInp, c("xpt", "stx", "stc"))) {
        if        (usePkg == "haven"   && hasPkg("haven"))   {
            hvnTmp <- tryCatch(do.call(haven::read_xpt, adjArg("haven::read_xpt", list(file = fleInp), varArg, "file")),
                               error   = function(errMsg) tryErr(fleInp, errMsg),
                               warning = function(wrnMsg) tryErr(fleInp, wrnMsg))
            clnTbb(hvnTmp, c("format.sas", "display_width"), jmvLbl = TRUE)
        } else if (usePkg == "foreign" && hasPkg("foreign")) {
            fgnTmp <- tryCatch(do.call(foreign::read.xport, adjArg("foreign::read.xport", list(file = fleInp), varArg, c("file"))),
                               error   = function(errMsg) tryErr(fleInp, errMsg))
            clnFgn(fgnTmp)
        } else {
            stop(sprintf("In order to read the SAS-transport-file \"%s\" either of the R-packages \"haven\" or \"foreign\" needs to be installed.", basename(fleInp)))
        }
    }
}

getSPSS  <- function(fleInp = "", usePkg = "", varArg = list()) {
    if        (usePkg == "haven"   && hasPkg("haven"))   {
        hvnTmp <- tryCatch(do.call(haven::read_sav, adjArg("haven::read_sav", list(file = fleInp), varArg, "file")),
                           error   = function(errMsg) tryErr(fleInp, errMsg),
                           warning = function(wrnMsg) tryErr(fleInp, wrnMsg))
        clnTbb(hvnTmp, c("format.spss", "display_width"), jmvLbl = TRUE)
    } else if (usePkg == "foreign" && hasPkg("foreign")) {
        fgnTmp <- tryCatch(suppressWarnings(do.call(foreign::read.spss,
                             adjArg("foreign::read.spss", list(file = fleInp, to.data.frame = TRUE, trim_values = TRUE, trim.factor.names = TRUE),
                               varArg, c("file", "to.data.frame", "trim_values", "trim.factor.names")))),
                           error   = function(errMsg) tryErr(fleInp, errMsg))
        clnFgn(fgnTmp)
    } else {
        stop(sprintf("In order to read the SPSS-file \"%s\" either of the R-packages \"haven\" or \"foreign\" needs to be installed.", basename(fleInp)))
    }
}

getStata  <- function(fleInp = "", usePkg = "", varArg = list()) {
    # NB: more recent versions of the Stata-format require "haven" and can't be read with foreign
    usePkg <- ifelse(grepl("^<stata_dta><header>", readBin(fleInp, character(), n = 1)), "haven", usePkg)
    if        (usePkg == "haven"   && hasPkg("haven"))   {
        hvnTmp <- tryCatch(do.call(haven::read_dta, adjArg("haven::read_dta", list(file = fleInp), varArg, "file")),
                           error   = function(errMsg) tryErr(fleInp, errMsg),
                           warning = function(wrnMsg) tryErr(fleInp, wrnMsg))
        clnTbb(hvnTmp, c("format.stata", "display_width"), jmvLbl = TRUE)
    } else if (usePkg == "foreign" && hasPkg("foreign")) {
        tmpFgn <- tryCatch(do.call(foreign::read.dta, adjArg("foreign::read.dta", list(file = fleInp), varArg, c("file"))),
                           error = function(errMsg) tryErr(fleInp, errMsg))
        clnFgn(tmpFgn)
    } else {
        stop(sprintf("In order to read the Stata-file \"%s\" either of the R-packages \"haven\" or \"foreign\" needs to be installed.", basename(fleInp)))
    }
}

getTxt <- function(fleOMV = "", crrFle = "") {
    crrTxt <- readLines(crrHdl <- getHdl(fleOMV, crrFle), warn = FALSE)
    clsHdl(crrHdl)
    rm(crrHdl)

    # depending on whether the original was a JSON file or not, return the appropriate result
    if (hasExt(crrFle, "json")) {
        crrTxt <- jsonlite::fromJSON(crrTxt, simplifyVector = FALSE)
    }

    crrTxt
}

rmvQtn <- function(dtaFrm = NULL) {
    for (crrNme in names(which(vapply(dtaFrm, is.character, logical(1))))) {
        dtaFrm[[crrNme]] <- trimws(gsub("\"", "", dtaFrm[[crrNme]]))
    }

    dtaFrm
}

tryErr <- function(fleInp = "", errMsg = NULL) {
    message(sprintf("File \"%s\" couldn\'t be read.\nThe error message was: %s\n", basename(fleInp), conditionMessage(errMsg)))
    return(invisible(NULL))
}

tryWrn <- function(fleInp = "", wrnMsg = NULL) {
    message(sprintf("Warnings were issued when reading the file \"%s\".\nThe warning was: %s\n", basename(fleInp), conditionMessage(wrnMsg)))
    return(invisible(NULL))
}

test_that("aggregate_omv works", {
    nmeInp <- tempfile(fileext = ".rds")
    nmeOut <- tempfile(fileext = "_A.omv")
    clcStr <- c("N",  "Mss",   "Mn",   "Mdn",    "Mde",  "Sum",  "SD",  "Var",      "Rng",   "Min",  "Max",  "IQR")
    clcDsc <- c("N",  "Miss.", "Mean", "Median", "Mode", "Sum",  "SD",  "Variance", "Range", "Min.", "Max.", "IQR")

    # simple data set with a factorial and a continuous variable ======================================================
    set.seed(1234)
    dtaTmp <- data.frame(ID = rep(as.character(seq(1, 100)), each = 10), Measure = rep(seq(10), times = 100),
                         V1 = runif(n = 100 * 10, 0, 100), V2 = as.factor(round(rnorm(n = 100 * 10, 3, 2 / 3))))
    attr(dtaTmp[, "V1"], "jmv-desc") <- "Variable V1"
    attr(dtaTmp[, "V2"], "jmv-desc") <- "Variable V2"
    saveRDS(dtaTmp, nmeInp)

    # using a grouping variable (ID)
    expect_null(aggregate_omv(dtaInp = nmeInp, fleOut = nmeOut, varAgg = c("V1", "V2"), grpAgg = "ID",
                              clcN = TRUE, clcMss = TRUE, clcMn = TRUE, clcMdn = TRUE, clcMde = TRUE, clcSum = TRUE,
                              clcSD = TRUE, clcVar = TRUE, clcRng = TRUE, clcMin = TRUE, clcMax = TRUE, clcIQR = TRUE))
    df4Chk <- read_omv(nmeOut)
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(100, 25))
    expect_equal(vapply(df4Chk, typeof, character(1), USE.NAMES = FALSE),
                 c("character", rep("integer", 2), rep("double", 10), rep("integer", 2), rep("double", 2), rep("integer", 2), rep("double", 2),
                   rep("integer", 3), "double"))
    expect_equal(names(df4Chk), c("ID", paste0(rep(c("V1_", "V2_"), each = 12), rep(clcStr, 2))))
    expect_equal(unname(colMeans(df4Chk[-1])),
                 c(10, 0, 50.7273, 50.3691, 8.9539, 507.2735, 28.7392, 848.0571, 82.6450, 8.9539, 91.5989, 41.2718,
                   10, 0,  2.9940,  2.9800, 2.9700,  29.9400,  0.6790,   0.4862,  2.0500, 1.9700,  4.0200,  0.5900),
                 tolerance = 1e-4)
    expect_equal(vapply(df4Chk[-1], attr, character(1), "jmv-desc", USE.NAMES = FALSE),
                 sprintf("Variable %s (%s)", rep(c("V1", "V2"), each = 12), rep(clcDsc, 2)))
    unlink(nmeOut)

    # not using a grouping variable
    expect_null(aggregate_omv(dtaInp = nmeInp, fleOut = nmeOut, varAgg = c("V1", "V2"),
                              clcN = TRUE, clcMss = TRUE, clcMn = TRUE, clcMdn = TRUE, clcMde = TRUE, clcSum = TRUE,
                              clcSD = TRUE, clcVar = TRUE, clcRng = TRUE, clcMin = TRUE, clcMax = TRUE, clcIQR = TRUE))
    df4Chk <- read_omv(nmeOut)
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(1, 24))
    expect_equal(vapply(df4Chk, typeof, character(1), USE.NAMES = FALSE),
                 c(rep("integer", 2), rep("double", 10), rep("integer", 2), "double", rep("integer", 3), rep("double", 2),
                   rep("integer", 4)))
    expect_equal(names(df4Chk), paste0(rep(c("V1_", "V2_"), each = 12), rep(clcStr, 2)))
    expect_equal(as.numeric(unname(df4Chk)),
                 c(1000, 0, 50.7273, 51.0192, 0.0342, 50727.3463, 29.1208, 848.0223, 99.8961, 0.0342, 99.9303, 50.0280,
                   1000, 0,  2.9940,  3.0000, 3.0000,  2994.0000,  0.6946,   0.4824,  4.0000, 1.0000,  4.0200,  0.5900),
                 tolerance = 1e-4)
    expect_equal(vapply(df4Chk, attr, character(1), "jmv-desc", USE.NAMES = FALSE),
                 sprintf("Variable %s (%s)", rep(c("V1", "V2"), each = 12), rep(clcDsc, 2)))
    unlink(nmeOut)


    # simple data set with two continuous variables ===================================================================
    set.seed(1)
    dtaTmp <- data.frame(ID = rep(as.character(seq(1, 100)), each = 10), Measure = rep(seq(10), times = 100),
                         V1 = runif(n = 100 * 10, 0, 100), V2 = round(rnorm(n = 100 * 10, 100, 15)))
    attr(dtaTmp[, "V1"], "jmv-desc") <- "Variable V1"
    attr(dtaTmp[, "V2"], "jmv-desc") <- "Variable V2"
    saveRDS(dtaTmp, nmeInp)

    expect_null(aggregate_omv(dtaInp = nmeInp, fleOut = nmeOut, varAgg = c("V1", "V2"), grpAgg = "ID",
                              clcN = TRUE, clcMss = TRUE, clcMn = TRUE, clcMdn = TRUE, clcMde = TRUE, clcSum = TRUE,
                              clcSD = TRUE, clcVar = TRUE, clcRng = TRUE, clcMin = TRUE, clcMax = TRUE, clcIQR = TRUE))
    df4Chk <- read_omv(nmeOut)
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(100, 25))
    expect_equal(vapply(df4Chk, typeof, character(1), USE.NAMES = FALSE),
                 c("character", rep("integer", 2), rep("double", 10), rep("integer", 2), rep("double", 2), rep("integer", 2), rep("double", 2),
                   rep("integer", 3), "double"))
    expect_equal(names(df4Chk), c("ID", paste0(rep(c("V1_", "V2_"), each = 12), rep(clcStr, 2))))
    expect_equal(unname(colMeans(df4Chk[-1])),
                 c(10, 0, 49.9692, 49.8635,  9.3976, 499.6917, 28.2575, 818.0221, 81.4983,  9.3976,  90.8959, 39.0899,
                   10, 0, 99.6310, 99.3150, 88.6200, 996.3100, 15.1117, 239.1143, 47.6100, 76.1400, 123.7500, 18.2550),
                 tolerance = 1e-4)
    expect_equal(vapply(df4Chk[-1], attr, character(1), "jmv-desc", USE.NAMES = FALSE),
                 sprintf("Variable %s (%s)", rep(c("V1", "V2"), each = 12), rep(clcDsc, 2)))
    unlink(nmeOut)

    expect_null(aggregate_omv(dtaInp = nmeInp, fleOut = nmeOut, varAgg = c("V1", "V2"),
                              clcN = TRUE, clcMss = TRUE, clcMn = TRUE, clcMdn = TRUE, clcMde = TRUE, clcSum = TRUE,
                              clcSD = TRUE, clcVar = TRUE, clcRng = TRUE, clcMin = TRUE, clcMax = TRUE, clcIQR = TRUE))
    df4Chk <- read_omv(nmeOut)
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(1, 24))
    expect_equal(vapply(df4Chk, typeof, character(1), USE.NAMES = FALSE),
                 c(rep("integer", 2), rep("double", 10), rep("integer", 2), "double", rep("integer", 3), rep("double", 2),
                   rep("integer", 4)))
    expect_equal(names(df4Chk), paste0(rep(c("V1_", "V2_"), each = 12), rep(clcStr, 2)))
    expect_equal(as.numeric(unname(df4Chk)),
                 c(1000, 0, 49.9692, 48.3260,  0.1315, 49969.1673, 28.8387, 831.6708,  99.8616,  0.1315,  99.9931, 48.8804,
                   1000, 0, 99.6310, 99.0000, 96.0000, 99631.0000, 15.5205, 240.8857, 104.0000, 51.0000, 155.0000, 22.0000),
                 tolerance = 1e-4)
    expect_equal(vapply(df4Chk, attr, character(1), "jmv-desc", USE.NAMES = FALSE),
                 sprintf("Variable %s (%s)", rep(c("V1", "V2"), each = 12), rep(clcDsc, 2)))
    unlink(nmeOut)

    # simple data set with NAs ========================================================================================
    set.seed(1)
    dtaTmp[sample(nrow(dtaTmp), 10), "V1"] <- NA
    dtaTmp[sample(nrow(dtaTmp), 10), "V2"] <- NA

    expect_null(aggregate_omv(dtaInp = dtaTmp, fleOut = nmeOut, varAgg = c("V1", "V2"), grpAgg = "ID",
                              clcN = TRUE, clcMss = TRUE, clcMn = TRUE, clcMdn = TRUE, clcMde = TRUE, clcSum = TRUE,
                              clcSD = TRUE, clcVar = TRUE, clcRng = TRUE, clcMin = TRUE, clcMax = TRUE, clcIQR = TRUE))
    df4Chk <- read_omv(nmeOut)
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(100, 25))
    expect_equal(sort(dtaTmp[is.na(dtaTmp$V1), "ID"]), df4Chk[df4Chk$V1_N < 10, "ID"])
    expect_equal(sort(dtaTmp[is.na(dtaTmp$V2), "ID"]), df4Chk[df4Chk$V2_N < 10, "ID"])
    expect_equal(unname(colMeans(df4Chk[-1])),
                 c(9.9, 0.1, 50.0834, 50.1054,  9.4463, 495.3516, 28.3369, 822.7721, 81.4495,  9.4463,  90.8959, 39.0855,
                   9.9, 0.1, 99.5748, 99.2950, 88.7600, 985.7900, 15.0837, 238.6681, 47.3300, 76.2800, 123.6100, 18.2550),
                 tolerance = 1e-4)
    unlink(nmeOut)

    expect_null(aggregate_omv(dtaInp = dtaTmp, fleOut = nmeOut, varAgg = c("V1", "V2"), grpAgg = "ID", drpNA = FALSE,
                              clcN = TRUE, clcMss = TRUE, clcMn = TRUE, clcMdn = TRUE, clcMde = TRUE, clcSum = TRUE,
                              clcSD = TRUE, clcVar = TRUE, clcRng = TRUE, clcMin = TRUE, clcMax = TRUE, clcIQR = TRUE))
    df4Chk <- read_omv(nmeOut)
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(100, 25))
    expect_true(all(is.na(df4Chk[df4Chk$V1_N < 10, sprintf("V1_%s", setdiff(clcStr, c("N", "Mss")))])))
    expect_true(all(is.na(df4Chk[df4Chk$V2_N < 10, sprintf("V2_%s", setdiff(clcStr, c("N", "Mss")))])))
    expect_equal(unname(colMeans(df4Chk[-1])), c(9.9, 0.1, rep(NA, 10), 9.9, 0.1, rep(NA, 10)))
    unlink(nmeOut)

    # simple data set with NAs in the grouping variable ===============================================================
    set.seed(1)
    dtaTmp <- data.frame(ID = rep(as.character(seq(1, 100)), each = 10), Measure = rep(seq(10), times = 100),
                         V1 = runif(n = 100 * 10, 0, 100), V2 = round(rnorm(n = 100 * 10, 100, 15)))
    dtaTmp[sample(nrow(dtaTmp), 10), "ID"] <- NA

    expect_null(aggregate_omv(dtaInp = dtaTmp, fleOut = nmeOut, varAgg = c("V1", "V2"), grpAgg = "ID", clcMn = TRUE))
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(100, 25))
    expect_true(all(is.na(df4Chk[df4Chk$V1_N < 10, sprintf("V1_%s", setdiff(clcStr, c("N", "Mss")))])))
    expect_true(all(is.na(df4Chk[df4Chk$V2_N < 10, sprintf("V2_%s", setdiff(clcStr, c("N", "Mss")))])))
    expect_equal(unname(colMeans(df4Chk[-1])), c(9.9, 0.1, rep(NA, 10), 9.9, 0.1, rep(NA, 10)))
    unlink(nmeOut)

    expect_error(aggregate_omv(dtaInp = dtaTmp, fleOut = nmeOut, varAgg = c("V1", "V2"), grpAgg = "ID", drpNA = FALSE, clcMn = TRUE),
                 regexp = "The grouping variables must not contain empty or NA values \\(if drpNA is set to FALSE\\)\\.")

    # code coverage ===================================================================================================
    expect_error(aggregate_omv(fleInp = nmeInp, varAgg = c("V1", "V2"), grpAgg = "ID"),
                 regexp = "Please use the argument dtaInp instead of fleInp\\.")

    expect_error(aggregate_omv(dtaInp = nmeInp, varAgg = c("X", "Y"), grpAgg = "ID", clcMn = TRUE),
                 regexp = paste("Calling aggregate_omv requires giving at least one \\(valid\\) variable to aggregate,",
                                "and the grouping variable\\(s\\) needs to be empty or valid\\."))
    expect_error(aggregate_omv(dtaInp = nmeInp, varAgg = c("V1", "V2"), grpAgg = "ID"),
                 regexp = "At least one aggregation calculation \\(clc\\.\\.\\.\\) needs to be set to TRUE\\.")

    unlink(nmeInp)
})

test_that("read_omv works", {
    nmeInp <- file.path("..", "ToothGrowth.omv")

    # read the data set (without default arguments) and test its properties: data frame, size, correct column type, attributes (data frame and fourth column),
    # whether the "missingValues"-attribute is an empty list, and whether the factor levels at columns 4, 6, and 8 have the correct length and content
    df4Chk <- read_omv(fleInp = nmeInp, useFlt = FALSE, rmMsVl = FALSE, sveAtt = FALSE, getSyn = FALSE, getHTM = FALSE)
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(60, 16))
    expect_equal(vapply(df4Chk, typeof, character(1), USE.NAMES = FALSE),
      c("logical", "character", "double", "integer", "double", "integer", "double", "integer", "integer", "integer", "double", "double", "double", "integer", "logical", "integer"))
    expect_equal(names(attributes(df4Chk)), c("names", "row.names", "class", "fltLst"))
    expect_equal(names(attributes(df4Chk[[2]])),  c("jmv-id", "missingValues"))
    expect_equal(names(attributes(df4Chk[[4]])),  c("levels", "class", "values", "missingValues"))
    expect_equal(names(attributes(df4Chk[[9]])),  c("levels", "class", "values", "missingValues"))
    expect_equal(names(attributes(df4Chk[[10]])), c("levels", "class", "jmv-desc", "missingValues"))
    expect_vector(attr(df4Chk[[4]],  "missingValues"), list(), 0)
    expect_vector(attr(df4Chk[[14]], "missingValues"), list(), 2)
    expect_vector(attr(df4Chk[[4]], "levels"), c(), 2)
    expect_vector(attr(df4Chk[[6]], "levels"), c(), 2)
    expect_vector(attr(df4Chk[[8]], "levels"), c(), 3)
    expect_equal(attr(df4Chk[[4]], "levels"), c("1", "2"))
    expect_equal(attr(df4Chk[[6]], "levels"), c("OJ", "VC"))
    expect_equal(attr(df4Chk[[8]], "levels"), c("0.5", "1.0", "2.0"))
    expect_equal(attr(df4Chk[[14]], "missingValues"), list("== 99", "== 88"))

    # read the data set (with the useFlt-argument set TRUE) and test its properties: data frame, size, correct column type (NB: two rows and the first column - "Filter 1" are removed)
    df4Chk <- read_omv(fleInp = nmeInp, useFlt = TRUE, rmMsVl = FALSE, sveAtt = FALSE, getSyn = FALSE, getHTM = FALSE)
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(58, 15))
    expect_equal(vapply(df4Chk, typeof, character(1), USE.NAMES = FALSE),
      c("character", "double", "integer", "double", "integer", "double", "integer", "integer", "integer", "double", "double", "double", "integer", "logical", "integer"))

    # read the data set (with the rmMsVl-argument set TRUE) and test its properties: data frame, size, correct column type
    df4Chk <- read_omv(fleInp = nmeInp, useFlt = FALSE, rmMsVl = TRUE, sveAtt = FALSE, getSyn = FALSE, getHTM = FALSE)
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(60, 16))
    expect_equal(vapply(df4Chk, typeof, character(1), USE.NAMES = FALSE),
      c("logical", "character", "double", "integer", "double", "integer", "double", "integer", "integer", "integer", "double", "double", "double", "integer", "logical", "integer"))

    # read the data set (with the sveAtt-argument set TRUE) and test its properties: data frame, size, correct column type, and several attributes of the whole data frame and columns within it
    df4Chk <- read_omv(fleInp = nmeInp, useFlt = FALSE, rmMsVl = FALSE, sveAtt = TRUE, getSyn = FALSE, getHTM = FALSE)
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(60, 16))
    expect_equal(vapply(df4Chk, typeof, character(1), USE.NAMES = FALSE),
      c("logical", "character", "double", "integer", "double", "integer", "double", "integer", "integer", "integer", "double", "double", "double", "integer", "logical", "integer"))
    expect_equal(names(attributes(df4Chk)), c("names", "row.names", "class", "fltLst", "removedRows", "addedRows", "transforms"))
    expect_equal(names(attributes(df4Chk[[4]])),  c("levels", "class", "values", "name", "id", "columnType", "dataType", "measureType", "formula", "formulaMessage",
                                                    "parentId", "width", "type", "importName", "description", "transform", "edits", "missingValues", "trimLevels"))
    expect_equal(names(attributes(df4Chk[[9]])),  c("levels", "class", "values", "name", "id", "columnType", "dataType", "measureType", "formula", "formulaMessage",
                                                    "parentId", "width", "type", "importName", "description", "transform", "edits", "missingValues", "trimLevels"))
    expect_equal(names(attributes(df4Chk[[11]])), c("jmv-desc", "name", "id", "columnType", "dataType", "measureType", "formula", "formulaMessage", "parentId", "width",
                                                    "type", "outputAnalysisId", "outputOptionName", "outputName", "outputDesiredColumnName", "outputAssignedColumnName", "importName",
                                                    "description", "transform", "edits", "missingValues"))

    # read the data set (with the getSyn-argument set TRUE) and test its properties: correct attributes (should contain "syntax" and "protobuf"), as well as the type, size and content of "syntax"
    # (should start with "jmv::")
    df4Chk <- read_omv(fleInp = nmeInp, useFlt = FALSE, rmMsVl = FALSE, sveAtt = FALSE, getSyn = TRUE, getHTM = FALSE)
    expect_equal(names(attributes(df4Chk)), c("names", "row.names", "class", "fltLst", "syntax", "protobuf"))
    # the next two command actually work in both cases: when a list with "syntax" is filled with command and if it's empty
    expect_vector(attr(df4Chk, "syntax"), c(), 2)
    expect_true(all(grepl("^jmv::", attr(df4Chk, "syntax"))))
    expect_equal(attr(df4Chk, "syntax"),
      c(paste("jmv::ANOVA(formula = len ~ supp + dose2 + supp:dose2, data = data, effectSize = \"partEta\", modelTest = TRUE, qq = TRUE,",
              "contrasts = list(list(var=\"supp\", type=\"none\"), list(var=\"dose2\", type=\"polynomial\")), postHoc = ~ supp + dose2, emMeans = ~ dose2:supp)"),
           "jmv::ancova(formula = len ~ supp + dose, data = data, effectSize = \"partEta\", modelTest = TRUE)"))

    # read the data set (with the getSyn-argument set TRUE) and test its properties: correct attributes (should contain "syntax" and "protobuf"), as well as the type, size and content of "syntax"
    df4Chk <- read_omv(fleInp = nmeInp, useFlt = FALSE, rmMsVl = FALSE, sveAtt = FALSE, getSyn = FALSE, getHTM = TRUE)
    expect_equal(names(attributes(df4Chk)), c("names", "row.names", "class", "fltLst", "HTML"))
    expect_vector(attr(df4Chk, "HTML"), character(), 257)
    expect_equal(attr(df4Chk, "HTML")[1],   "<!DOCTYPE html>")
    expect_equal(attr(df4Chk, "HTML")[2],   "<html>")
    expect_equal(attr(df4Chk, "HTML")[257], "</html>")
    expect_equal(length(grep("<.*?>", attr(df4Chk, "HTML"))), 72)

    # test cases for code coverage ============================================================================================================================
    # fleInp is not given or empty
    expect_error(read_omv(),   regexp = "^File name to the input data file needs to be given as parameter \\(fleInp = \\.\\.\\.\\)\\.")
    expect_error(read_omv(""), regexp = "^File name to the input data file needs to be given as parameter \\(fleInp = \\.\\.\\.\\)\\.")
    # fleInp is not a jamovi-file (.omv)
    expect_error(read_omv("Trial.rds"),
      regexp = "^read_omv only reads jamovi files \\(\\.omv\\ / \\.omt\\), use convert_to_omv first, if you want to read other files types\\.")
    # the manifest must have a file name as second parameter and
    # the file has to be a valid manifest file (which is not the
    # case for "index.html" [exists, but isn't a manifest])
    expect_error(chkMnf(nmeInp, c()), regexp = "^File \".*?\" has not the correct file format \\(is missing the jamovi-file-manifest\\)\\.")
    expect_error(chkMnf(nmeInp, "index.html"),
      regexp = "^The file you are trying to read \\(ToothGrowth\\.omv\\) has an improper manifest file \\(meta\\) and is likely corrupted\\.")

    # .omv-file isn't a ZIP
    nmeTmp <- tempfile(fileext = ".omv")
    writeBin("", con = nmeTmp)
    expect_error(chkFle(nmeTmp, isZIP = TRUE), regexp = "^chkFle: File \".*\" has not the correct file format \\(is not a ZIP archive\\)\\.")
    unlink(nmeTmp)

    # invalid manifest (wrong version number)
    nmeTmp <- tempfile(fileext = ".omv")
    add2ZIP(nmeTmp, crrFle = c("meta", "wb"), txtOut = gsub("jamovi-Archive-Version: 11.0", "jamovi-Archive-Version: 99.0", mnfTxt()))
    suppressMessages(expect_warning(expect_error(read_omv(nmeTmp), regexp = "^'con' is not a connection"),
      regexp = "^The file that you are trying to read \\(.*\\) was written with a version of jamovi that currently is not implemented"))
    suppressMessages(expect_null(getHdl(fleOMV = nmeTmp, crrFle = "MANIFEST.MF")))
    unlink(nmeTmp)

    # convert vector of 0 and 1 into logical
    tmpCol <- sample(c(0, 1), 100, replace = TRUE)
    cl4Chk <- valLbl(crrCol = tmpCol, mtaCol = list(name = "A", columnType = "Data", dataType = "Integer"),
                     xtdDta = list(A = list(labels = list(list(0, "0", "0", FALSE), list(1, "1", "1", FALSE)))))
    expect_true(is.logical(cl4Chk))
    expect_null(attributes(cl4Chk))
    expect_equal(c(mean(cl4Chk), sd(cl4Chk)), c(mean(tmpCol), sd(tmpCol)))

    # unimplemented columnType when assigning value labels
    expect_error(valLbl(mtaCol = list(name = "Trial", columnType = "Trial", dataType = "Trial"), xtdDta = list(Trial = NULL)),
      regexp = "Error when reading value label - likely the column type is not implemented \\(yet\\): Trial - Trial - Trial")
})

test_that("read_all works", {
    # read_all should work as read_omv (with the sveAtt-argument set TRUE) and have the same properties: data frame, size, correct column type, and several attributes of the data frame and columns
    nmeInp <- file.path("..", "ToothGrowth.omv")
    df4Chk <- read_all(nmeInp)
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(60, 16))
    expect_equal(vapply(df4Chk, typeof, character(1), USE.NAMES = FALSE),
      c("logical", "character", "double", "integer", "double", "integer", "double", "integer", "integer", "integer", "double", "double", "double", "integer", "logical", "integer"))
    expect_equal(names(attributes(df4Chk)), c("names", "row.names", "class", "fltLst", "removedRows", "addedRows", "transforms"))
    expect_equal(names(attributes(df4Chk[[4]])),  c("levels", "class", "values", "name", "id", "columnType", "dataType", "measureType", "formula", "formulaMessage",
                                                    "parentId", "width", "type", "importName", "description", "transform", "edits", "missingValues", "trimLevels"))
    expect_equal(names(attributes(df4Chk[[9]])),  c("levels", "class", "values", "name", "id", "columnType", "dataType", "measureType", "formula", "formulaMessage",
                                                    "parentId", "width", "type", "importName", "description", "transform", "edits", "missingValues", "trimLevels"))

    # read_all for ToothGrowth as Rdata-set has fewer columns and attributes, check data frame, size, correct column type, and several attributes of the data frame and columns
    nmeTmp <- tempfile(fileext = ".rds")
    saveRDS(jmvReadWrite::ToothGrowth, nmeTmp)
    df4Chk <- read_all(nmeTmp)
    expect_equal(dim(df4Chk), c(60, 7))
    expect_equal(vapply(df4Chk, typeof,                  character(1), USE.NAMES = FALSE), c("character", "integer", "integer", "double", "integer", "double", "double"))
    expect_equal(vapply(df4Chk, function(x) class(x)[1], character(1), USE.NAMES = FALSE), c("character", "factor", "factor", "numeric", "ordered", "numeric", "numeric"))
    expect_equal(names(attributes(df4Chk)), c("names", "row.names", "class"))
    expect_equal(names(attributes(df4Chk[[3]])),  c("levels", "class", "description"))
    expect_equal(names(attributes(df4Chk[[5]])),  c("levels", "class"))
    unlink(nmeTmp)

    # test cases for code coverage ============================================================================================================================
    # empty file name
    expect_error(read_all(),   regexp = "^File name to the input data file needs to be given as parameter \\(fleInp = \\.\\.\\.\\)\\.")
    expect_error(read_all(""), regexp = "^File name to the input data file needs to be given as parameter \\(fleInp = \\.\\.\\.\\)\\.")

    # more than one object when using Rdata
    D1 <- data.frame(A = runif(n = 100))
    D2 <- data.frame(B = runif(n = 100))
    nmeInp <- tempfile(fileext = ".rda")
    save(list = ls()[grepl("D[1-2]", ls())], file = nmeInp)
    # throw error when selSet is not sepcified
    expect_message(expect_null(read_all(nmeInp)), regexp = "^File \".*\" couldn't be read\\.\nThe error message was: The Rdata-file must include only one object\\.")
    # check whether reading works correct with using selSet
    df4Chk <- read_all(nmeInp, selSet = "D1")
    expect_s3_class(df4Chk, class = "data.frame")
    expect_equal(dim(df4Chk), c(100, 1))
    unlink(nmeInp)

    # test cases for clnTbl in RData and RDS
    fleInp <- tempfile()
    if (requireNamespace("haven", quietly = TRUE)) {
        haven::write_sav(jmvReadWrite::ToothGrowth, paste0(fleInp, ".sav"))
        tmpDta <- haven::read_sav(paste0(fleInp, ".sav"))
        unlink(paste0(fleInp, ".sav"))
        save(tmpDta, file = paste0(fleInp, ".RData"))
        df4Chk <- read_all(paste0(fleInp, ".RData"))
        expect_equal(attributes(df4Chk)[c("names", "row.names", "class")],
                     list(names = c("ID", "supp", "supp2", "dose", "dose2", "len", "logLen"), row.names = seq(60), class = "data.frame"))
        expect_equal(lapply(df4Chk, attributes),
                     list(ID = NULL, supp = list(levels = c("OJ", "VC"), class = "factor"), supp2 = list(levels = c("1", "2"), class = "factor"),
                          dose = NULL, dose2 = list(levels = c("0.5", "1.0", "2.0"), class = "factor"), len = NULL, logLen = NULL))
        unlink(paste0(fleInp, ".RData"))
        saveRDS(tmpDta, file = paste0(fleInp, ".rds"))
        df4Chk <- read_all(paste0(fleInp, ".rds"))
        expect_equal(attributes(df4Chk)[c("names", "row.names", "class")],
                     list(names = c("ID", "supp", "supp2", "dose", "dose2", "len", "logLen"), row.names = seq(60), class = "data.frame"))
        expect_equal(lapply(df4Chk, attributes),
                     list(ID = NULL, supp = list(levels = c("OJ", "VC"), class = "factor"), supp2 = list(levels = c("1", "2"), class = "factor"),
                          dose = NULL, dose2 = list(levels = c("0.5", "1.0", "2.0"), class = "factor"), len = NULL, logLen = NULL))
        unlink(paste0(fleInp, ".rds"))
    }

    # test cases for CSV / TSV
    nmeInp <- tempfile(fileext = ".tsv")
    write.table(D1, file = nmeInp, sep = "\t")
    df4Chk <- read_all(nmeInp)
    expect_s3_class(df4Chk, class = "data.frame")
    expect_equal(dim(df4Chk), c(100, 1))
    unlink(nmeInp)

    nmeInp <- tempfile(fileext = ".csv")
    write.table(D1, file = nmeInp, sep = ",")
    df4Chk <- read_all(nmeInp)
    expect_s3_class(df4Chk, class = "data.frame")
    expect_equal(dim(df4Chk), c(100, 1))
    unlink(nmeInp)

    nmeInp <- tempfile(fileext = ".csv")
    write.table(D1, file = nmeInp, sep = ";")
    # if the separator is not given, the input column can't be converted into a number
    expect_type(read_all(nmeInp)$A, "character")
    df4Chk <- read_all(nmeInp, sep = ";")
    expect_s3_class(df4Chk, class = "data.frame")
    expect_equal(dim(df4Chk), c(100, 1))
    expect_type(df4Chk$A, "double")
    unlink(nmeInp)

    nmeInp <- tempfile(fileext = ".csv")
    writeBin("X1,X2\n1.2,2.2\n7.1,3.2", nmeInp)
    expect_message(expect_null(read_all(nmeInp)),
      regexp = "^Warnings were issued when reading the file \".*\"\\.\nThe warning was: line 3 appears to contain embedded nulls")
    unlink(nmeInp)

    fleInp <- tempfile(fileext = ".sav")
    writeBin("$FL2@(#) IBM SPSS STATISTICS 64-bit Linux 25.0.0.0              \002", con = fleInp)
    expect_message(expect_null(read_all(fleInp, usePkg = "haven")),
      regexp = "^File \".*\" couldn't be read\\.\nThe error message was: Failed to parse .*: Unable to read from file\\.")
    expect_message(expect_null(read_all(fleInp, usePkg = "foreign")), regexp = "^File \".*\" couldn't be read\\.")
    unlink(fleInp)
    if (requireNamespace("haven", quietly = TRUE)) {
        haven::write_sav(jmvReadWrite::ToothGrowth, fleInp)
        df4Chk <- read_all(fleInp, usePkg = "haven")
        expect_equal(attributes(df4Chk)[c("names", "row.names", "class")],
                     list(names = c("ID", "supp", "supp2", "dose", "dose2", "len", "logLen"), row.names = seq(60), class = "data.frame"))
        expect_equal(lapply(df4Chk, attributes),
                     list(ID = NULL, supp = list(levels = c("OJ", "VC"), class = "factor"), supp2 = list(levels = c("1", "2"), class = "factor"),
                          dose = NULL, dose2 = list(levels = c("0.5", "1.0", "2.0"), class = "factor"), len = NULL, logLen = NULL))
        unlink(fleInp)
    }

    fleInp <- tempfile(fileext = ".dta")
    writeBin("<stata_dta><header><release>117</release><byteorder>LSF</byteorder><K>\x8f", con = fleInp)
    suppressWarnings(expect_message(expect_null(read_all(fleInp, usePkg = "haven")),
      regexp = "File \".*\" couldn't be read\\.\nThe error message was: Failed to parse .*: Unable to read from file\\."))
    suppressWarnings(expect_message(expect_null(read_all(fleInp, usePkg = "foreign")),
      regexp = "File \".*\" couldn't be read\\.\nThe error message was: not a Stata version 5-12 .dta file"))
    unlink(fleInp)
    if (requireNamespace("haven", quietly = TRUE)) {
        haven::write_dta(jmvReadWrite::ToothGrowth, fleInp)
        df4Chk <- read_all(fleInp, usePkg = "haven")
        expect_equal(attributes(df4Chk)[c("names", "row.names", "class")],
                     list(names = c("ID", "supp", "supp2", "dose", "dose2", "len", "logLen"), row.names = seq(60), class = "data.frame"))
        expect_equal(lapply(df4Chk, attributes),
                     list(ID = NULL, supp = list(levels = c("OJ", "VC"), class = "factor"), supp2 = list(levels = c("1", "2"), class = "factor"),
                          dose = NULL, dose2 = list(levels = c("0.5", "1.0", "2.0"), class = "factor"), len = NULL, logLen = NULL))
        unlink(fleInp)
    }

    fleInp <- tempfile(fileext = ".sas7bdat")
    writeBin("", con = fleInp)
    invisible(capture.output(expect_message(expect_null(read_all(fleInp, usePkg = "haven")), "File \".*\" couldn't be read")))
    expect_error(read_all(fleInp, usePkg = "foreign"), "In order to read the SAS-file \".*\" the R-packages \"haven\" needs to be installed\\.")
    unlink(fleInp)
    if (requireNamespace("haven", quietly = TRUE)) {
        suppressWarnings(haven::write_sas(jmvReadWrite::ToothGrowth, fleInp))
        df4Chk <- read_all(fleInp, usePkg = "haven")
        expect_equal(attributes(df4Chk)[c("names", "row.names", "class")],
                     list(names = c("ID", "supp", "supp2", "dose", "dose2", "len", "logLen"), row.names = seq(60), class = "data.frame"))
        expect_equal(lapply(df4Chk, attributes),
                     list(ID = NULL, supp = NULL, supp2 = NULL, dose = NULL, dose2 = NULL, len = NULL, logLen = NULL))
        unlink(fleInp)
    }

    fleInp <- tempfile(fileext = ".xpt")
    writeBin("HEADER RECORD*******LIBRARY HEADER RECORD!!!!!!!", con = fleInp)
    expect_message(expect_null(read_all(fleInp, usePkg = "haven")),
      regexp = "^File \".*\" couldn't be read\\.\nThe error message was: Failed to parse .*: Unable to read from file\\.")
    expect_message(expect_null(read_all(fleInp, usePkg = "foreign")),
      regexp = "^File \".*\" couldn't be read\\.\nThe error message was: file not in SAS transfer format")
    unlink(fleInp)
    if (requireNamespace("haven", quietly = TRUE)) {
        haven::write_xpt(jmvReadWrite::ToothGrowth, fleInp)
        df4Chk <- read_all(fleInp, usePkg = "haven")
        expect_equal(attributes(df4Chk)[c("names", "row.names", "class")],
                     list(names = c("ID", "supp", "supp2", "dose", "dose2", "len", "logLen"), row.names = seq(60), class = "data.frame"))
        expect_equal(lapply(df4Chk, attributes),
                     list(ID = NULL, supp = NULL, supp2 = NULL, dose = NULL, dose2 = NULL, len = NULL, logLen = NULL))
        unlink(fleInp)
    }

    dtaTmp <- jmvReadWrite::ToothGrowth
    expect_null(attributes(clnTbb(dtaTmp, c("jmv-id", "jmv-desc"))[[1]]))
    expect_null(attributes(clnTbb(dtaTmp, c("jmv-id", "jmv-desc"))[[7]]))
    attr(dtaTmp[[6]], "label") <- "Trial for label conversion"
    expect_equal(attributes(clnTbb(dtaTmp, jmvLbl = TRUE)[[6]]), list(`jmv-desc` = "Trial for label conversion"))

    set.seed(2)
    dtaTmp <- structure(list(value = structure(sample(c(1, 2, 4), 200, replace = TRUE), format.sas = "LEVELS",
                             class = c("haven_labelled", "vctrs_vctr", "double"),
                             labels = c(`Level 1` = 1, `Level 2` = 2, `Level 3` = 4))),
                        class = c("tbl_df", "tbl", "data.frame"), row.names = c(NA, -200L))
    expect_s3_class(clnTbb(dtaTmp), "data.frame")
    expect_s3_class(clnTbb(dtaTmp)[, "value"], "factor")
    expect_equal(attributes(clnTbb(dtaTmp)), list(names = "value", row.names = seq(200), class = "data.frame"))
    expect_equal(attributes(clnTbb(dtaTmp)[, "value"]), list(levels = sprintf("Level %d", seq(3)), class = "factor", format.sas = "LEVELS"))
    expect_equal(attributes(clnTbb(dtaTmp, rmvAtt = "format.sas")[, "value"]), list(levels = sprintf("Level %d", seq(3)), class = "factor"))
    expect_equal(as.integer(table(clnTbb(dtaTmp)[, "value"])), c(71, 69, 60))
    expect_equal(names(table(clnTbb(dtaTmp)[, "value"])), sprintf("Level %d", seq(3)))

    dtaTmp <- structure(list(value = structure(sample(c(1, 2, 4), 200, replace = TRUE), format.sas = "LEVELS",
                             class = c("haven_labelled", "vctrs_vctr", "double"), labels = setNames(c(1, 2, 4), rep("", 3)))),
                        class = c("tbl_df", "tbl", "data.frame"), row.names = c(NA, -200L))
    expect_s3_class(clnTbb(dtaTmp), "data.frame")
    expect_type(clnTbb(dtaTmp)[, "value"], "integer")
    expect_equal(attributes(clnTbb(dtaTmp)), list(names = "value", row.names = seq(200), class = "data.frame"))
    expect_equal(attributes(clnTbb(dtaTmp)[, "value"]), list(format.sas = "LEVELS"))
    expect_null(attributes(clnTbb(dtaTmp, rmvAtt = "format.sas")[, "value"]))
    expect_equal(as.integer(clnTbb(dtaTmp)[, "value"])[1:10], c(4, 4, 2, 2, 4, 2, 4, 1, 2, 4))
    expect_equal(as.integer(table(clnTbb(dtaTmp)[, "value"])), c(66, 67, 67))
    expect_equal(names(table(clnTbb(dtaTmp)[, "value"])), sprintf("%d", c(1, 2, 4)))

    dtaTmp <- structure(list(value = structure(sample(c(1, 2, 4), 200, replace = TRUE), format.sas = "LEVELS",
                             class = c("haven_labelled", "vctrs_vctr", "double"), labels = setNames(c(1, 2, 4), rep("value label", 3)))),
                        class = c("tbl_df", "tbl", "data.frame"), row.names = c(NA, -200L))
    expect_s3_class(clnTbb(dtaTmp), "data.frame")
    expect_type(clnTbb(dtaTmp)[, "value"], "integer")
    expect_equal(attributes(clnTbb(dtaTmp)), list(names = "value", row.names = seq(200), class = "data.frame"))
    expect_equal(attributes(clnTbb(dtaTmp)[, "value"]), list(format.sas = "LEVELS"))
    expect_null(attributes(clnTbb(dtaTmp, rmvAtt = "format.sas")[, "value"]))
    expect_equal(as.integer(clnTbb(dtaTmp)[, "value"])[1:10], c(1, 1, 2, 1, 1, 4, 1, 2, 1, 1))
    expect_equal(as.integer(table(clnTbb(dtaTmp)[, "value"])), c(70, 72, 58))
    expect_equal(names(table(clnTbb(dtaTmp)[, "value"])), sprintf("%d", c(1, 2, 4)))

    dtaTmp <- structure(list(value = structure(sample(seq(4), 200, replace = TRUE), format.sas = "LEVELS",
                             class = c("haven_labelled", "vctrs_vctr", "double"), labels = setNames(seq(4), c("Smallest", "", "", "Largest")))),
                        class = c("tbl_df", "tbl", "data.frame"), row.names = c(NA, -200L))
    expect_s3_class(clnTbb(dtaTmp), "data.frame")
    expect_s3_class(clnTbb(dtaTmp)[, "value"], "factor")
    expect_equal(attributes(clnTbb(dtaTmp)), list(names = "value", row.names = seq(200), class = "data.frame"))
    expect_equal(attributes(clnTbb(dtaTmp)[, "value"]), list(levels = c("Smallest", "2", "3", "Largest"), class = "factor", format.sas = "LEVELS"))
    expect_equal(attributes(clnTbb(dtaTmp, rmvAtt = "format.sas")[, "value"]), list(levels = c("Smallest", "2", "3", "Largest"), class = "factor"))
    expect_equal(as.integer(clnTbb(dtaTmp)[, "value"])[1:10], c(4, 2, 1, 3, 2, 1, 3, 3, 2, 2))
    expect_equal(as.integer(table(clnTbb(dtaTmp)[, "value"])), c(52, 47, 46, 55))
    expect_equal(names(table(clnTbb(dtaTmp)[, "value"])), c("Smallest", "2", "3", "Largest"))

    dtaTmp <- jmvReadWrite::AlbumSales[-1]
    attr(dtaTmp, "variable.labels") <- c(Adverts = "Advertsing budget (thousands)",
                                         Airplay = "No. of plays on radio",
                                         Image = "Band image rating (0-10)",
                                         Sales = "Album sales (thousands)")
    expect_identical(vapply(clnFgn(dtaTmp), attr, character(1), "jmv-desc"), attr(dtaTmp, "variable.labels"))
})

test_that("arrange_cols_omv works", {
    nmeInp <- tempfile(fileext = ".rds")
    nmeOut <- tempfile(fileext = "_A.omv")
    saveRDS(jmvReadWrite::AlbumSales, nmeInp)

    expect_null(arrange_cols_omv(dtaInp = nmeInp, fleOut = nmeOut, varOrd = c("selSbj", "Sales", "Adverts", "Airplay", "Image")))
    expect_true(chkFle(nmeOut))
    expect_gt(file.info(nmeOut)$size, 1)
    expect_true(chkFle(nmeOut, isZIP = TRUE))
    expect_true(chkFle(nmeOut, fleCnt = "meta"))
    expect_true(chkFle(nmeOut, fleCnt = "metadata.json"))
    expect_true(chkFle(nmeOut, fleCnt = "data.bin"))
    df4Chk <- read_omv(nmeOut, sveAtt = FALSE)
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(200, 5))
    expect_equal(names(df4Chk), c("selSbj", "Sales", "Adverts", "Airplay", "Image"))
    expect_equal(vapply(df4Chk, typeof, character(1), USE.NAMES = FALSE), c("integer", "integer", "double", "integer", "integer"))
    unlink(nmeOut)

    expect_null(arrange_cols_omv(dtaInp = nmeInp, fleOut = nmeOut, varMve = list(Sales = -3, Adverts = 2)))
    expect_true(chkFle(nmeOut))
    expect_gt(file.info(nmeOut)$size, 1)
    expect_true(chkFle(nmeOut, isZIP = TRUE))
    expect_true(chkFle(nmeOut, fleCnt = "meta"))
    expect_true(chkFle(nmeOut, fleCnt = "metadata.json"))
    expect_true(chkFle(nmeOut, fleCnt = "data.bin"))
    df4Chk <- read_omv(nmeOut, sveAtt = FALSE)
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(200, 5))
    expect_equal(names(df4Chk), c("selSbj", "Sales", "Airplay", "Image", "Adverts"))
    expect_equal(vapply(df4Chk, typeof, character(1), USE.NAMES = FALSE), c("integer", "integer", "integer", "integer", "double"))
    unlink(nmeOut)

    df4Chk <- arrange_cols_omv(dtaInp = nmeInp, varMve = list(Sales = -3, Adverts = 2))
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(200, 5))
    expect_equal(names(df4Chk), c("selSbj", "Sales", "Airplay", "Image", "Adverts"))
    expect_equal(vapply(df4Chk, typeof, character(1), USE.NAMES = FALSE), c("integer", "integer", "integer", "integer", "double"))

    # test cases for code coverage ============================================================================================================================
    expect_error(arrange_cols_omv(fleInp = nmeInp, varMve = list(len = -2, supp = -1)), regexp = "Please use the argument dtaInp instead of fleInp\\.")
    expect_error(arrange_cols_omv(dtaInp = nmeInp, fleOut = nmeOut),
      regexp = "^Calling arrange_cols_omv requires either the parameter varOrd \\(a character vector\\) or the parameter varMve \\(a named list\\), using the correct format")
    expect_false(file.exists(nmeOut))
    expect_error(arrange_cols_omv(dtaInp = nmeInp, fleOut = nmeOut, varOrd = list()),
      regexp = "^Calling arrange_cols_omv requires either the parameter varOrd \\(a character vector\\) or the parameter varMve \\(a named list\\), using the correct format")
    expect_false(file.exists(nmeOut))
    expect_error(arrange_cols_omv(dtaInp = nmeInp, fleOut = nmeOut, varOrd = c()),
      regexp = "^Calling arrange_cols_omv requires either the parameter varOrd \\(a character vector\\) or the parameter varMve \\(a named list\\), using the correct format")
    expect_false(file.exists(nmeOut))
    expect_error(arrange_cols_omv(dtaInp = nmeInp, fleOut = nmeOut, varOrd = c(1, 2)),
      regexp = "^Calling arrange_cols_omv requires either the parameter varOrd \\(a character vector\\) or the parameter varMve \\(a named list\\), using the correct format")
    expect_false(file.exists(nmeOut))
    expect_error(arrange_cols_omv(dtaInp = nmeInp, fleOut = nmeOut, varOrd = c("", "")),
      regexp = "^Calling arrange_cols_omv requires either the parameter varOrd \\(a character vector\\) or the parameter varMve \\(a named list\\), using the correct format")
    expect_false(file.exists(nmeOut))
    expect_error(arrange_cols_omv(dtaInp = nmeInp, fleOut = nmeOut, varMve = c()),
      regexp = "^Calling arrange_cols_omv requires either the parameter varOrd \\(a character vector\\) or the parameter varMve \\(a named list\\), using the correct format")
    expect_false(file.exists(nmeOut))
    expect_error(arrange_cols_omv(dtaInp = nmeInp, fleOut = nmeOut, varMve = list()),
      regexp = "^Calling arrange_cols_omv requires either the parameter varOrd \\(a character vector\\) or the parameter varMve \\(a named list\\), using the correct format")
    expect_false(file.exists(nmeOut))
    expect_error(arrange_cols_omv(dtaInp = nmeInp, fleOut = nmeOut, varMve = list(1, 1)),
      regexp = "^Calling arrange_cols_omv requires either the parameter varOrd \\(a character vector\\) or the parameter varMve \\(a named list\\), using the correct format")
    expect_false(file.exists(nmeOut))
    expect_error(arrange_cols_omv(dtaInp = nmeInp, fleOut = nmeOut, varMve = list(A = "1", B = "1")),
      regexp = "^Calling arrange_cols_omv requires either the parameter varOrd \\(a character vector\\) or the parameter varMve \\(a named list\\), using the correct format")
    expect_false(file.exists(nmeOut))
    expect_error(arrange_cols_omv(dtaInp = nmeInp, fleOut = nmeOut, varMve = list(A = 1, B = 1.1)),
      regexp = "^Calling arrange_cols_omv requires either the parameter varOrd \\(a character vector\\) or the parameter varMve \\(a named list\\), using the correct format")
    expect_false(file.exists(nmeOut))
    expect_error(arrange_cols_omv(dtaInp = nmeInp, fleOut = nmeOut, varMve = list(A = 1, B = 0)),
      regexp = "^Calling arrange_cols_omv requires either the parameter varOrd \\(a character vector\\) or the parameter varMve \\(a named list\\), using the correct format")
    expect_false(file.exists(nmeOut))
    expect_error(arrange_cols_omv(dtaInp = nmeInp, fleOut = nmeOut, varOrd = c("Sale", "Adverts", "Airplay", "Image")),
      regexp = "^The variable\\(s\\) .* are not contained in the current data set\\.")
    expect_false(file.exists(nmeOut))
    expect_error(arrange_cols_omv(dtaInp = nmeInp, fleOut = nmeOut, varMve = list(Sale = -3)),
      regexp = "^The variable\\(s\\) .* are not contained in the current data set\\.")
    expect_false(file.exists(nmeOut))
    expect_error(arrange_cols_omv(dtaInp = nmeInp, fleOut = nmeOut, varMve = list(Sales = -5)),
      regexp = "^The value given in varMve must be chosen so that the element isn't moved before the first or after the last column.")
    expect_false(file.exists(nmeOut))
    expect_error(arrange_cols_omv(dtaInp = nmeInp, fleOut = nmeOut, varMve = list(Sales = -3, Image = 2)),
      regexp = "^The value given in varMve must be chosen so that the element isn't moved before the first or after the last column.")
    expect_false(file.exists(nmeOut))

    expect_warning(arrange_cols_omv(dtaInp = nmeInp, fleOut = nmeOut, varOrd = c("Sales", "Adverts", "Airplay", "Image")),
      regexp = "^The following variable\\(s\\) from the original data set are not contained in varOrd: .*")
    expect_true(file.exists(nmeOut))
    unlink(nmeOut)
    expect_warning(arrange_cols_omv(dtaInp = nmeInp, fleOut = nmeOut, varOrd = c("selSbj", "Sales", "Adverts", "Airplay", "Image"), varMve = list(Sales = -3)),
      regexp = "^Both, varOrd and varMve given as input parameters. varOrd takes precedence\\.")
    expect_true(file.exists(nmeOut))
    unlink(nmeOut)
    unlink(nmeInp)

    # test cases for the transfer of analyses =================================================================================================================
    nmeInp <- file.path("..", "ToothGrowth.omv")
    arrange_cols_omv(dtaInp = nmeInp, fleOut = nmeOut, varMve = list(len = -2, supp = -1), psvAnl = TRUE)
    expect_true(chkFle(nmeOut))
    expect_gt(file.info(nmeOut)$size, 1)
    expect_true(chkFle(nmeOut, isZIP = TRUE))
    expect_true(chkFle(nmeOut, fleCnt = "meta"))
    expect_true(chkFle(nmeOut, fleCnt = "metadata.json"))
    expect_true(chkFle(nmeOut, fleCnt = "data.bin"))
    df4Chk <- read_omv(nmeOut, getSyn = TRUE)
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(60, 16))
    expect_equal(names(df4Chk), c("Filter 1", "ID", "len", "logLen", "supp", "supp - Transform 1", "dose", "dose2", "dose3", "Trial", "Residuals", "J", "K", "L", "M", "weights"))
    expect_equal(as.vector(vapply(df4Chk, typeof, character(1))),
      c("logical", "character", "double", "double", "integer", "integer", "double", "integer", "integer", "integer", "double", "double", "double", "integer", "logical", "integer"))
    expect_equal(sort(zip::zip_list(nmeOut)$filename),
      c("01 empty/analysis", "02 anova/analysis", "02 anova/resources/65167cb3bdaf8761.png", "02 anova/resources/99f9b5d34a92049b.png", "03 empty/analysis",
        "04 ancova/analysis", "05 empty/analysis", "data.bin", "index.html", "meta", "metadata.json", "strings.bin", "xdata.json"))
    expect_equal(attr(df4Chk, "syntax"),
      c(paste("jmv::ANOVA(formula = len ~ supp + dose2 + supp:dose2, data = data, effectSize = \"partEta\", modelTest = TRUE, qq = TRUE,",
              "contrasts = list(list(var=\"supp\", type=\"none\"), list(var=\"dose2\", type=\"polynomial\")), postHoc = ~ supp + dose2, emMeans = ~ dose2:supp)"),
           "jmv::ancova(formula = len ~ supp + dose, data = data, effectSize = \"partEta\", modelTest = TRUE)"))
    unlink(nmeOut)
    expect_warning(arrange_cols_omv(dtaInp = jmvReadWrite::AlbumSales, fleOut = nmeOut, varOrd = c("selSbj", "Sales", "Adverts", "Airplay", "Image"), psvAnl = TRUE),
      regexp = "^psvAnl is only possible if dtaInp is a file name \\(analyses are not stored in data frames, only in the jamovi files\\)\\.")
    expect_warning(arrange_cols_omv(dtaInp = jmvReadWrite::AlbumSales, varOrd = c("selSbj", "Sales", "Adverts", "Airplay", "Image"), psvAnl = TRUE),
      regexp = "^psvAnl is only possible if fleOut is a file name \\(analyses are not stored in data frames, only in the jamovi files\\)\\.")
    unlink(nmeOut)
    # do not unlink nmeInp, this isn't a generated file, but a link
})

test_that("transpose_omv works", {
    set.seed(1)
    tmpDF <- stats::setNames(as.data.frame(matrix(sample(6, 1200, replace = TRUE), nrow = 16)), sprintf("sbj_%03d", seq(75)))
    nmeInp <- tempfile(fileext = ".rds")
    nmeOut <- tempfile(fileext = "_T.omv")
    saveRDS(tmpDF, nmeInp)

    expect_null(transpose_omv(dtaInp = tmpDF, fleOut = nmeOut))
    expect_true(chkFle(nmeOut))
    expect_gt(file.info(nmeOut)$size, 1)
    expect_true(chkFle(nmeOut, isZIP = TRUE))
    expect_true(chkFle(nmeOut, fleCnt = "meta"))
    expect_true(chkFle(nmeOut, fleCnt = "metadata.json"))
    expect_true(chkFle(nmeOut, fleCnt = "data.bin"))
    df4Chk <- read_omv(nmeOut, sveAtt = FALSE)
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(75, 17))
    expect_equal(names(df4Chk), c("ID", sprintf("V_%d", seq(16))))
    expect_equal(vapply(df4Chk, typeof, character(1), USE.NAMES = FALSE), c("character", rep("integer", 16)))
    unlink(nmeOut)

    expect_null(transpose_omv(dtaInp = cbind(list(qstItm = sprintf("Qst_%02d", seq(16))), tmpDF), fleOut = nmeOut, varNme = "qstItm"))
    expect_true(chkFle(nmeOut))
    expect_gt(file.info(nmeOut)$size, 1)
    expect_true(chkFle(nmeOut, isZIP = TRUE))
    expect_true(chkFle(nmeOut, fleCnt = "meta"))
    expect_true(chkFle(nmeOut, fleCnt = "metadata.json"))
    expect_true(chkFle(nmeOut, fleCnt = "data.bin"))
    df4Chk <- read_omv(nmeOut, sveAtt = FALSE)
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(75, 17))
    expect_equal(names(df4Chk), c("ID", sprintf("Qst_%02d", seq(16))))
    expect_equal(vapply(df4Chk, typeof, character(1), USE.NAMES = FALSE), c("character", rep("integer", 16)))
    unlink(nmeOut)

    expect_null(transpose_omv(dtaInp = tmpDF, fleOut = nmeOut, varNme = sprintf("Trl_%02d", seq(16))))
    expect_true(chkFle(nmeOut))
    expect_gt(file.info(nmeOut)$size, 1)
    expect_true(chkFle(nmeOut, isZIP = TRUE))
    expect_true(chkFle(nmeOut, fleCnt = "meta"))
    expect_true(chkFle(nmeOut, fleCnt = "metadata.json"))
    expect_true(chkFle(nmeOut, fleCnt = "data.bin"))
    df4Chk <- read_omv(nmeOut, sveAtt = FALSE)
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(75, 17))
    expect_equal(names(df4Chk), c("ID", sprintf("Trl_%02d", seq(16))))
    expect_equal(vapply(df4Chk, typeof, character(1), USE.NAMES = FALSE), c("character", rep("integer", 16)))
    unlink(nmeOut)

    expect_null(transpose_omv(dtaInp = nmeInp, fleOut = nmeOut))
    expect_true(chkFle(nmeOut))
    expect_gt(file.info(nmeOut)$size, 1)
    expect_true(chkFle(nmeOut, isZIP = TRUE))
    expect_true(chkFle(nmeOut, fleCnt = "meta"))
    expect_true(chkFle(nmeOut, fleCnt = "metadata.json"))
    expect_true(chkFle(nmeOut, fleCnt = "data.bin"))
    df4Chk <- read_omv(nmeOut, sveAtt = FALSE)
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(75, 17))
    expect_equal(names(df4Chk), c("ID", sprintf("V_%d", seq(16))))
    expect_equal(vapply(df4Chk, typeof, character(1), USE.NAMES = FALSE), c("character", rep("integer", 16)))
    unlink(nmeOut)

    df4Chk <- transpose_omv(dtaInp = tmpDF)
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(75, 17))
    expect_equal(names(df4Chk), c("ID", sprintf("V_%d", seq(16))))
    expect_equal(vapply(df4Chk, typeof, character(1), USE.NAMES = FALSE), c("character", rep("integer", 16)))

    expect_error(transpose_omv(fleInp = tmpDF, fleOut = nmeOut), regexp = "Please use the argument dtaInp instead of fleInp\\.")
    unlink(nmeInp)


    expect_error(transpose_omv(dtaInp = tmpDF, fleOut = nmeOut, varNme = "notExist"),
      regexp = "^.*\\(varNme\\) not contained in the input data frame\\.")
    expect_error(transpose_omv(dtaInp = tmpDF, fleOut = nmeOut, varNme = 1), regexp = "^varNme must be a character variable\\.")
    expect_error(transpose_omv(dtaInp = tmpDF, fleOut = nmeOut, varNme = c("Qst_01", "Qst_02")),
      regexp = "^varNme must either be empty, have one element or as many elements as there are rows in the input data frame\\.")
})

test_that("sort_omv works", {
    nmeInp <- tempfile(fileext = ".rds")
    nmeOut <- tempfile(fileext = "_S.omv")
    saveRDS(jmvReadWrite::AlbumSales, nmeInp)

    expect_null(sort_omv(dtaInp = nmeInp, fleOut = nmeOut, varSrt = "Image"))
    expect_true(file.exists(nmeOut))
    expect_gt(file.info(nmeOut)$size, 1)
    expect_true(chkFle(nmeOut, isZIP = TRUE))
    expect_true(chkFle(nmeOut, fleCnt = "meta"))
    expect_true(chkFle(nmeOut, fleCnt = "metadata.json"))
    expect_true(chkFle(nmeOut, fleCnt = "data.bin"))
    df4Chk <- read_omv(nmeOut, sveAtt = FALSE)
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(200, 5))
    expect_equal(vapply(df4Chk, typeof, character(1), USE.NAMES = FALSE), c("integer", "double", "integer", "integer", "integer"))
    expect_equal(as.integer(table(df4Chk[["Image"]])), c(3, 1, 1, 4, 17, 44, 73, 44, 12, 1))
    expect_equal(which(diff(as.integer(df4Chk[["Image"]])) == 1), c(3, 4, 5, 9, 26, 70, 143, 187, 199))
    expect_false(is.unsorted(df4Chk[["Image"]]))
    unlink(nmeInp)
    unlink(nmeOut)

    expect_null(sort_omv(dtaInp = jmvReadWrite::AlbumSales, fleOut = nmeOut, varSrt = "Image"))
    expect_true(file.exists(nmeOut))
    expect_gt(file.info(nmeOut)$size, 1)
    expect_true(chkFle(nmeOut, isZIP = TRUE))
    expect_true(chkFle(nmeOut, fleCnt = "meta"))
    expect_true(chkFle(nmeOut, fleCnt = "metadata.json"))
    expect_true(chkFle(nmeOut, fleCnt = "data.bin"))
    df4Chk <- read_omv(nmeOut, sveAtt = FALSE)
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(200, 5))
    expect_equal(vapply(df4Chk, typeof, character(1), USE.NAMES = FALSE), c("integer", "double", "integer", "integer", "integer"))
    expect_false(is.unsorted(df4Chk[["Image"]]))
    unlink(nmeOut)

    df4Chk <- sort_omv(dtaInp = jmvReadWrite::AlbumSales, varSrt = "Image")
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(200, 5))
    expect_equal(vapply(df4Chk, typeof, character(1), USE.NAMES = FALSE), c("integer", "double", "integer", "integer", "integer"))
    expect_false(is.unsorted(df4Chk[["Image"]]))

    # test cases for code coverage and for the transfer of analyses ===========================================================================================
    expect_error(sort_omv(fleInp = jmvReadWrite::AlbumSales, varSrt = "Image"), regexp = "Please use the argument dtaInp instead of fleInp\\.")

    expect_error(sort_omv(dtaInp = nmeInp, fleOut = nmeOut, varSrt = c()), regexp = "^Calling sort_omv requires giving at least one variable to sort after\\.")
    sort_omv(dtaInp = file.path("..", "ToothGrowth.omv"), fleOut = nmeOut, varSrt = "len", psvAnl = TRUE)
    expect_true(chkFle(nmeOut))
    expect_gt(file.info(nmeOut)$size, 1)
    expect_true(chkFle(nmeOut, isZIP = TRUE))
    expect_true(chkFle(nmeOut, fleCnt = "meta"))
    expect_true(chkFle(nmeOut, fleCnt = "metadata.json"))
    expect_true(chkFle(nmeOut, fleCnt = "data.bin"))
    df4Chk <- read_omv(nmeOut, sveAtt = FALSE, getSyn = TRUE)
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(60, 16))
    expect_equal(names(df4Chk),
      c("Filter 1", "ID", "logLen", "supp - Transform 1", "len", "supp", "dose", "dose2", "dose3", "Trial", "Residuals", "J", "K", "L", "M", "weights"))
    expect_equal(vapply(df4Chk, typeof, character(1), USE.NAMES = FALSE),
      c("logical", "character", "double", "integer", "double", "integer", "double", "integer", "integer", "integer", "double", "double", "double", "integer", "logical", "integer"))
    expect_equal(sort(zip::zip_list(nmeOut)$filename),
      c("01 empty/analysis", "02 anova/analysis", "02 anova/resources/65167cb3bdaf8761.png", "02 anova/resources/99f9b5d34a92049b.png", "03 empty/analysis",
        "04 ancova/analysis", "05 empty/analysis", "data.bin", "index.html", "meta", "metadata.json", "strings.bin", "xdata.json"))
    expect_equal(attr(df4Chk, "syntax"),
      c(paste("jmv::ANOVA(formula = len ~ supp + dose2 + supp:dose2, data = data, effectSize = \"partEta\", modelTest = TRUE, qq = TRUE,",
              "contrasts = list(list(var=\"supp\", type=\"none\"), list(var=\"dose2\", type=\"polynomial\")), postHoc = ~ supp + dose2, emMeans = ~ dose2:supp)"),
           "jmv::ancova(formula = len ~ supp + dose, data = data, effectSize = \"partEta\", modelTest = TRUE)"))
    unlink(nmeOut)

    expect_warning(sort_omv(dtaInp = jmvReadWrite::AlbumSales, fleOut = nmeOut, varSrt = "Sales", psvAnl = TRUE),
      regexp = "^psvAnl is only possible if dtaInp is a file name \\(analyses are not stored in data frames, only in the jamovi files\\)\\.")
    unlink(nmeOut)
    expect_warning(sort_omv(dtaInp = jmvReadWrite::AlbumSales, varSrt = "Sales", psvAnl = TRUE),
      regexp = "^psvAnl is only possible if fleOut is a file name \\(analyses are not stored in data frames, only in the jamovi files\\)\\.")
    unlink(nmeOut)
})

test_that("long2wide_omv works", {
    # rather simple data set ==================================================================================================================================
    set.seed(1)
    dtaTmp <- data.frame(Year = sort(rep(1900:2020, 12)), Month = rep(month.abb[1:12], 121), X = runif(n = 121 * 12, 0, 100), Y = runif(n = 121 * 12, 0, 100))
    attr(dtaTmp[["X"]], "jmv-desc") <- "Variable X"
    attr(dtaTmp[["Y"]], "jmv-desc") <- "Variable Y"
    nmeInp <- tempfile(fileext = ".rds")
    nmeOut <- tempfile(fileext = "_W.omv")
    saveRDS(dtaTmp, nmeInp)

    expect_null(long2wide_omv(dtaInp = nmeInp, fleOut = nmeOut, varID = "Year", varTme = "Month"))
    df4Chk <- read_omv(nmeOut)
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(121, 25))
    expect_equal(vapply(df4Chk, typeof, character(1), USE.NAMES = FALSE), c("character", rep("double", 24)))
    expect_equal(names(df4Chk), c("Year", paste0(rep(c("X_", "Y_"), each = 12), month.abb[rep(1:12, times = 2)])))
    expect_equal(unname(colMeans(df4Chk[-1])), c(51.05398, 51.52200, 50.90146, 47.98040, 46.28997, 53.70601, 49.47946, 49.24704, 49.92602, 44.93970, 49.37357, 47.55488,
                                                 48.56846, 48.96117, 47.64545, 46.51572, 50.94652, 47.33624, 47.53437, 55.55701, 51.50431, 50.19580, 52.81145, 43.68338), tolerance = 1e-4)
    expect_equal(vapply(df4Chk[-1], attr, character(1), "jmv-desc", USE.NAMES = FALSE),
      sprintf("%s (Month: %s)", rep(c("Variable X", "Variable Y"), each = 12), month.abb[rep(1:12, times = 2)]))
    unlink(nmeOut)

    expect_null(long2wide_omv(dtaInp = dtaTmp, fleOut = nmeOut, varID = "Year", varTme = "Month", varOrd = "vars"))
    df4Chk <- read_omv(nmeOut)
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(121, 25))
    expect_equal(vapply(df4Chk, typeof, character(1), USE.NAMES = FALSE), c("character", rep("double", 24)))
    expect_equal(names(df4Chk), c("Year", paste0(rep(c("X_", "Y_"), times = 12), month.abb[rep(1:12, each = 2)])))
    expect_equal(unname(colMeans(df4Chk[-1])), c(51.05398, 48.56846, 51.52200, 48.96117, 50.90146, 47.64545, 47.98040, 46.51572, 46.28997, 50.94652, 53.70601, 47.33624,
                                                 49.47946, 47.53437, 49.24704, 55.55701, 49.92602, 51.50431, 44.93970, 50.19580, 49.37357, 52.81145, 47.55488, 43.68338), tolerance = 1e-4)
    expect_equal(vapply(df4Chk[-1], attr, character(1), "jmv-desc", USE.NAMES = FALSE),
      sprintf("%s (Month: %s)", rep(c("Variable X", "Variable Y"), times = 12), month.abb[rep(1:12, each = 2)]))
    unlink(nmeOut)

    expect_null(long2wide_omv(dtaInp = nmeInp, fleOut = nmeOut, varID = "Year", varTme = "Month", varTgt = c("X")))
    df4Chk <- read_omv(nmeOut)
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(121, 13))
    expect_equal(vapply(df4Chk, typeof, character(1), USE.NAMES = FALSE), c("character", rep("double", 12)))
    expect_equal(names(df4Chk), c("Year", paste0("X_", month.abb[1:12])))
    expect_equal(unname(colMeans(df4Chk[-1])), c(51.05398, 51.52200, 50.90146, 47.98040, 46.28997, 53.70601, 49.47946, 49.24704, 49.92602, 44.93970, 49.37357, 47.55488), tolerance = 1e-4)
    expect_equal(vapply(df4Chk[-1], attr, character(1), "jmv-desc", USE.NAMES = FALSE),
      sprintf("Variable X (Month: %s)", month.abb[1:12]))
    unlink(nmeOut)

    expect_null(long2wide_omv(dtaInp = nmeInp, fleOut = nmeOut, varID = "Year", varTme = "Month", varTgt = c("Y")))
    df4Chk <- read_omv(nmeOut)
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(121, 13))
    expect_equal(vapply(df4Chk, typeof, character(1), USE.NAMES = FALSE), c("character", rep("double", 12)))
    expect_equal(names(df4Chk), c("Year", paste0("Y_", month.abb[1:12])))
    expect_equal(unname(colMeans(df4Chk[-1])), c(48.56846, 48.96117, 47.64545, 46.51572, 50.94652, 47.33624, 47.53437, 55.55701, 51.50431, 50.19580, 52.81145, 43.68338), tolerance = 1e-4)
    expect_equal(vapply(df4Chk[-1], attr, character(1), "jmv-desc", USE.NAMES = FALSE),
      sprintf("Variable Y (Month: %s)", month.abb[1:12]))
    unlink(nmeOut)

    df4Chk <- long2wide_omv(dtaInp = dtaTmp, varID = "Year", varTme = "Month")
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(121, 25))
    expect_equal(vapply(df4Chk, typeof, character(1), USE.NAMES = FALSE), c("integer", rep("double", 24)))
    expect_equal(names(df4Chk), c("Year", paste0(rep(c("X_", "Y_"), each = 12), month.abb[rep(1:12, 2)])))
    expect_equal(unname(colMeans(df4Chk[-1])), c(51.05398, 51.52200, 50.90146, 47.98040, 46.28997, 53.70601, 49.47946, 49.24704, 49.92602, 44.93970, 49.37357, 47.55488,
                                                 48.56846, 48.96117, 47.64545, 46.51572, 50.94652, 47.33624, 47.53437, 55.55701, 51.50431, 50.19580, 52.81145, 43.68338), tolerance = 1e-4)
    expect_equal(vapply(df4Chk[-1], attr, character(1), "jmv-desc", USE.NAMES = FALSE),
      sprintf("%s (Month: %s)", rep(c("Variable X", "Variable Y"), each = 12), month.abb[rep(1:12, times = 2)]))

    expect_null(long2wide_omv(dtaInp = nmeInp, fleOut = nmeOut, varID = "Year", varTme = "Month", varSep = "."))
    df4Chk <- read_omv(nmeOut)
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(121, 25))
    expect_equal(vapply(df4Chk, typeof, character(1), USE.NAMES = FALSE), c("character", rep("double", 24)))
    expect_equal(names(df4Chk), c("Year", paste0(rep(c("X.", "Y."), each = 12), month.abb[rep(1:12, times = 2)])))
    expect_equal(unname(colMeans(df4Chk[-1])), c(51.05398, 51.52200, 50.90146, 47.98040, 46.28997, 53.70601, 49.47946, 49.24704, 49.92602, 44.93970, 49.37357, 47.55488,
                                                 48.56846, 48.96117, 47.64545, 46.51572, 50.94652, 47.33624, 47.53437, 55.55701, 51.50431, 50.19580, 52.81145, 43.68338), tolerance = 1e-4)
    expect_equal(vapply(df4Chk[-1], attr, character(1), "jmv-desc", USE.NAMES = FALSE),
      sprintf("%s (Month: %s)", rep(c("Variable X", "Variable Y"), each = 12), month.abb[rep(1:12, times = 2)]))
    unlink(nmeOut)

    # more complex data set ===================================================================================================================================
    set.seed(1)
    dtaTmp <- data.frame(ID = rep(sprintf("sbj_%03d", seq(1, 100)), each = 24), sex = rep(sample(rep(c("male", "female"), each = 50)), each = 24),
                         cond = rep(rep(c("cong", "incong", "neutral"), each = 8), times = 100), colour = rep(c("RED", "GREEN", "BLUE", "YELLOW"), times = 600),
                         rep = rep(rep(c(1, 2), each = 4), times = 300), rspCrr = as.integer(runif(2400, 0, 1) > rep(rep(c(0.002, 0.03, 0.02), each = 8), times = 100)),
                         rspTme = rep(rep(c(0.450, 0.530, 0.510), each = 8), times = 100) + rnorm(2400) * 0.05)[sample(2400), ]
    dtaTmp <- dtaTmp[order(dtaTmp$ID, dtaTmp$rep), ]
    lblTmp <- list(ID = "Participant ID", sex = "Sex", cond = "Exp. cond. – congruency", colour = "Colour to be named", rep = "Repetition", rspCrr = "Correct response", rspTme = "Reaction time")
    for (i in seq_along(lblTmp)) attr(dtaTmp[[names(lblTmp[i])]], "jmv-desc") <- lblTmp[[i]]

    df4Chk <- long2wide_omv(dtaInp = dtaTmp, varID = "ID", varTme = c("cond", "colour", "rep"), varTgt = c("rspCrr", "rspTme"), varExc = "sex")
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(100, 50))
    expect_equal(names(df4Chk), c("ID", "sex", sprintf("%s_%s_%s_%d", rep(c("rspCrr", "rspTme"), each = 24), rep(rep(c("neutral", "cong", "incong"), each = 8), times = 2),
                                                       rep(c("RED", "BLUE", "GREEN", "YELLOW"), each = 2), rep(1:2, times = 24))))
    avgTmp <- aggregate(x = dtaTmp[, c("rspCrr", "rspTme")], by = dtaTmp[, c("cond", "colour", "rep")], FUN = mean)
    expect_equal(unname(colMeans(df4Chk[, sprintf("rspCrr_%s_%s_%s", avgTmp$cond, avgTmp$colour, avgTmp$rep)])), avgTmp$rspCrr)
    expect_equal(unname(colMeans(df4Chk[, sprintf("rspTme_%s_%s_%s", avgTmp$cond, avgTmp$colour, avgTmp$rep)])), avgTmp$rspTme)
    expect_equal(vapply(df4Chk, attr, character(1), "jmv-desc", USE.NAMES = FALSE),
      c(unname(unlist(lblTmp[1:2])), sprintf("%s (cond: %s, colour: %s, rep: %d)",
      rep(unname(unlist(lblTmp[6:7])), each = 24), rep(rep(c("neutral", "cong", "incong"), each = 8), times = 2), rep(c("RED", "BLUE", "GREEN", "YELLOW"), each = 2), rep(1:2, times = 24))))

    df4Chk <- long2wide_omv(dtaInp = dtaTmp, varID = "ID", varTme = "cond", varTgt = c("rspCrr", "rspTme"), varExc = "sex")
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(100, 8))
    expect_equal(names(df4Chk), c("ID", "sex", sprintf("%s_%s", rep(c("rspCrr", "rspTme"), each = 3), rep(rep(c("cong", "incong", "neutral"), times = 2)))))
    avgTmp <- aggregate(x = dtaTmp[, c("rspCrr", "rspTme")], by = dtaTmp[, c("sex", "cond"), drop = FALSE], FUN = mean)
    expect_equal(as.vector(unlist(avgTmp[c(-1, -2)])), unname(unlist(aggregate(x = df4Chk[3:8], by = df4Chk[, "sex", drop = FALSE], FUN = mean)[-1])))
    expect_equal(vapply(df4Chk, attr, character(1), "jmv-desc", USE.NAMES = FALSE),
      c(unname(unlist(lblTmp[1:2])), sprintf("%s (cond: %s)",
      rep(unname(unlist(lblTmp[6:7])), each = 3), rep(c("cong", "incong", "neutral"), times = 2))))

    df4Chk <- long2wide_omv(dtaInp = dtaTmp, varID = "ID", varTme = "cond", varTgt = c("rspCrr", "rspTme"))
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(100, 7))
    expect_equal(names(df4Chk), c("ID", sprintf("%s_%s", rep(c("rspCrr", "rspTme"), each = 3), rep(rep(c("cong", "incong", "neutral"), times = 2)))))
    avgTmp <- aggregate(x = dtaTmp[, c("rspCrr", "rspTme")], by = dtaTmp[, c("cond"), drop = FALSE], FUN = mean)
    expect_equal(unname(colMeans(df4Chk[2:7])), as.vector(unlist(avgTmp[-1])))
    expect_equal(vapply(df4Chk, attr, character(1), "jmv-desc", USE.NAMES = FALSE), c(unname(unlist(lblTmp[1])), sprintf("%s (cond: %s)",
      rep(unname(unlist(lblTmp[6:7])), each = 3), rep(c("cong", "incong", "neutral"), times = 2))))

    df4Chk <- long2wide_omv(dtaInp = dtaTmp, varID = "ID", varTme = "cond", varTgt = c("rspCrr", "rspTme"), varExc = "sex", varOrd = "vars")
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(100, 8))
    expect_equal(names(df4Chk), c("ID", "sex", sprintf("%s_%s", rep(c("rspCrr", "rspTme"), times = 3), rep(rep(c("cong", "incong", "neutral"), each = 2)))))
    avgTmp <- aggregate(x = dtaTmp[, c("rspCrr", "rspTme")], by = dtaTmp[, c("cond"), drop = FALSE], FUN = mean)
    expect_equal(unname(colMeans(df4Chk[3:8])), as.vector(t(avgTmp[-1])))
    expect_equal(vapply(df4Chk, attr, character(1), "jmv-desc", USE.NAMES = FALSE), c(unname(unlist(lblTmp[1:2])), sprintf("%s (cond: %s)",
      rep(unname(unlist(lblTmp[6:7])), times = 3), rep(c("cong", "incong", "neutral"), each = 2))))

    df4Chk <- long2wide_omv(dtaInp = dtaTmp, varID = "ID", varTme = "cond", varTgt = c("rspCrr", "rspTme"), varExc = "sex", varAgg = "first")
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(100, 8))
    expect_equal(names(df4Chk), c("ID", "sex", sprintf("%s_%s", rep(c("rspCrr", "rspTme"), each = 3), rep(rep(c("cong", "incong", "neutral"), times = 2)))))
    avgTmp <- aggregate(x = dtaTmp[, c("rspCrr", "rspTme")], by = dtaTmp[, c("ID", "cond")], FUN = "[[", 1)
    expect_equal(unname(as.matrix(df4Chk[3:8])), cbind(matrix(avgTmp[, "rspCrr"], ncol = 3), matrix(avgTmp[, "rspTme"], ncol = 3)))
    expect_equal(vapply(df4Chk, attr, character(1), "jmv-desc", USE.NAMES = FALSE), c(unname(unlist(lblTmp[1:2])), sprintf("%s (cond: %s)",
      rep(unname(unlist(lblTmp[6:7])), each = 3), rep(c("cong", "incong", "neutral"), times = 2))))

    df4Chk <- long2wide_omv(dtaInp = dtaTmp, varID = "ID", varTme = "cond", varTgt = c("rspCrr", "rspTme"), varExc = "sex", varAgg = "mean")
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(100, 8))
    expect_equal(names(df4Chk), c("ID", "sex", sprintf("%s_%s", rep(c("rspCrr", "rspTme"), each = 3), rep(rep(c("cong", "incong", "neutral"), times = 2)))))
    avgTmp <- aggregate(x = dtaTmp[, c("rspCrr", "rspTme")], by = dtaTmp[, c("ID", "cond")], FUN = mean)
    expect_equal(unname(as.matrix(df4Chk[3:8])), cbind(matrix(avgTmp[, "rspCrr"], ncol = 3), matrix(avgTmp[, "rspTme"], ncol = 3)))
    expect_equal(vapply(df4Chk, attr, character(1), "jmv-desc", USE.NAMES = FALSE), c(unname(unlist(lblTmp[1:2])), sprintf("%s (cond: %s)",
      rep(unname(unlist(lblTmp[6:7])), each = 3), rep(c("cong", "incong", "neutral"), times = 2))))

    # ---------------------------------------------------------------------------------------------------------------------------------------------------------
    dtaTmp[seq(1, 2400, 25), c("rspCrr", "rspTme")] <- NA

    df4Chk <- long2wide_omv(dtaInp = dtaTmp, varID = "ID", varTme = "cond", varTgt = c("rspCrr", "rspTme"), varExc = "sex", varAgg = "first")
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(100, 8))
    expect_equal(names(df4Chk), c("ID", "sex", sprintf("%s_%s", rep(c("rspCrr", "rspTme"), each = 3), rep(rep(c("cong", "incong", "neutral"), times = 2)))))
    avgTmp <- aggregate(x = dtaTmp[, c("rspCrr", "rspTme")], by = dtaTmp[, c("ID", "cond")], FUN = function(x) x[!is.na(x)][1])
    expect_equal(unname(as.matrix(df4Chk[3:8])), cbind(matrix(avgTmp[, "rspCrr"], ncol = 3), matrix(avgTmp[, "rspTme"], ncol = 3)))
    expect_equal(vapply(df4Chk, attr, character(1), "jmv-desc", USE.NAMES = FALSE), c(unname(unlist(lblTmp[1:2])), sprintf("%s (cond: %s)",
      rep(unname(unlist(lblTmp[6:7])), each = 3), rep(c("cong", "incong", "neutral"), times = 2))))

    df4Chk <- long2wide_omv(dtaInp = dtaTmp, varID = "ID", varTme = "cond", varTgt = c("rspCrr", "rspTme"), varExc = "sex", varAgg = "mean")
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(100, 8))
    expect_equal(names(df4Chk), c("ID", "sex", sprintf("%s_%s", rep(c("rspCrr", "rspTme"), each = 3), rep(rep(c("cong", "incong", "neutral"), times = 2)))))
    avgTmp <- aggregate(x = dtaTmp[, c("rspCrr", "rspTme")], by = dtaTmp[, c("ID", "cond")], FUN = mean, na.rm = TRUE)
    expect_equal(unname(as.matrix(df4Chk[3:8])), cbind(matrix(avgTmp[, "rspCrr"], ncol = 3), matrix(avgTmp[, "rspTme"], ncol = 3)))
    expect_equal(vapply(df4Chk, attr, character(1), "jmv-desc", USE.NAMES = FALSE), c(unname(unlist(lblTmp[1:2])), sprintf("%s (cond: %s)",
      rep(unname(unlist(lblTmp[6:7])), each = 3), rep(c("cong", "incong", "neutral"), times = 2))))

    # ---------------------------------------------------------------------------------------------------------------------------------------------------------

    dtaTmp$rspCrr <- as.factor(dtaTmp$rspCrr)
    expect_error(long2wide_omv(dtaInp = dtaTmp, varID = "ID", varTme = "cond", varTgt = c("rspCrr", "rspTme"), varExc = "sex"),
      regexp = "^In order to calculate the mean when aggregating the data, all target variables \\(varTgt\\) need to be numeric\\.")

    # test cases for code coverage ============================================================================================================================
    expect_error(long2wide_omv(fleInp = nmeInp, varID = "Year", varTme = "Month"), regexp = "Please use the argument dtaInp instead of fleInp\\.")

    expect_error(long2wide_omv(dtaInp = nmeInp, fleOut = nmeOut, varID = "Year", varTme = ""),
      regexp = "^Using the arguments varID and varTme is mandatory \\(i\\.e\\., they can't be empty\\)\\.")
    expect_error(long2wide_omv(dtaInp = nmeInp, fleOut = nmeOut, varID = "", varTme = "Month"),
      regexp = "^Using the arguments varID and varTme is mandatory \\(i\\.e\\., they can't be empty\\)\\.")
    expect_error(long2wide_omv(dtaInp = nmeInp, fleOut = nmeOut, varID = "Year", varTme = c("Month", "Month2")),
      regexp = "^The variable\\(s\\) .* are not contained in the current data set\\.")
    expect_error(long2wide_omv(dtaInp = dtaTmp, fleOut = nmeOut, varID = "Year", varTme = ""),
      regexp = "^Using the arguments varID and varTme is mandatory \\(i\\.e\\., they can't be empty\\)\\.")
    expect_error(long2wide_omv(dtaInp = dtaTmp, fleOut = nmeOut, varID = "", varTme = "Month"),
      regexp = "^Using the arguments varID and varTme is mandatory \\(i\\.e\\., they can't be empty\\)\\.")
    expect_error(long2wide_omv(dtaInp = dtaTmp, fleOut = nmeOut, varID = "Year", varTme = c("Month", "Month2")),
      regexp = "^The variable\\(s\\) .* are not contained in the current data set\\.")

    unlink(nmeInp)
})

test_that("label_vars_omv works", {

    lblFle <- file.path("..", "label_example.csv")
    lblDtF <- read.csv(lblFle, header = FALSE)
    nmeInp <- system.file("extdata", "bfi_sample.omv", package = "jmvReadWrite")
    dtaInp <- jmvReadWrite::read_omv(fleInp = nmeInp)[, lblDtF[, 1]]
    nmeOut <- tempfile(fileext = "_L.omv")

    # check that the labels in the original file are not set (NULL)
    expect_true(all(vapply(dtaInp[, lblDtF[, 1]], function(c) is.null(attr(c, "jmv-desc")), logical(1))))

    # varLbl: (1) use file name as parameter
    expect_null(label_vars_omv(dtaInp = nmeInp, fleOut = nmeOut, varLbl = lblFle))
    expect_true(chkFle(nmeOut))
    expect_gt(file.info(nmeOut)$size, 1)
    expect_true(chkFle(nmeOut, isZIP = TRUE))
    expect_true(chkFle(nmeOut, fleCnt = "meta"))
    expect_true(chkFle(nmeOut, fleCnt = "metadata.json"))
    expect_true(chkFle(nmeOut, fleCnt = "data.bin"))
    df4Chk <- read_omv(nmeOut, sveAtt = FALSE)
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(250, 33))
    expect_equal(as.character(lapply(df4Chk[, lblDtF[, 1]], attr, "jmv-desc")), lblDtF[, 2])
    unlink(nmeOut)

    # varLbl: (2) use data frame as parameter
    expect_null(label_vars_omv(dtaInp = nmeInp, fleOut = nmeOut, varLbl = lblDtF))
    expect_true(chkFle(nmeOut))
    expect_gt(file.info(nmeOut)$size, 1)
    expect_true(chkFle(nmeOut, isZIP = TRUE))
    expect_true(chkFle(nmeOut, fleCnt = "meta"))
    expect_true(chkFle(nmeOut, fleCnt = "metadata.json"))
    expect_true(chkFle(nmeOut, fleCnt = "data.bin"))
    df4Chk <- read_omv(nmeOut, sveAtt = FALSE)
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(250, 33))
    expect_equal(as.character(lapply(df4Chk[, lblDtF[, 1]], attr, "jmv-desc")), lblDtF[, 2])
    unlink(nmeOut)

    # varLbl: (3) use character vector (with labels) as parameter, here we need to
    # give a data frame with the same number of columns as input
    expect_null(label_vars_omv(dtaInp = dtaInp[, lblDtF[, 1]], fleOut = nmeOut, varLbl = lblDtF))
    expect_true(chkFle(nmeOut))
    expect_gt(file.info(nmeOut)$size, 1)
    expect_true(chkFle(nmeOut, isZIP = TRUE))
    expect_true(chkFle(nmeOut, fleCnt = "meta"))
    expect_true(chkFle(nmeOut, fleCnt = "metadata.json"))
    expect_true(chkFle(nmeOut, fleCnt = "data.bin"))
    df4Chk <- read_omv(nmeOut, sveAtt = FALSE)
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(250, 28))
    expect_equal(as.character(lapply(df4Chk, attr, "jmv-desc")), lblDtF[, 2])
    unlink(nmeOut)

    # test cases for code coverage ============================================================================================================================
    expect_error(label_vars_omv(fleInp = nmeInp, varLbl = lblDtF), regexp = "Please use the argument dtaInp instead of fleInp\\.")
    expect_error(label_vars_omv(dtaInp = nmeInp, fleOut = nmeOut),
      regexp = "^Calling label_vars_omv requires the parameter varLbl, using the correct format \\(see Details in help\\)\\.")
    expect_false(file.exists(nmeOut))
    expect_error(label_vars_omv(dtaInp = nmeInp, fleOut = nmeOut, varLbl = list()),
      regexp = "^Calling label_vars_omv requires the parameter varLbl, using the correct format \\(see Details in help\\)\\.")
    expect_false(file.exists(nmeOut))
    expect_error(label_vars_omv(dtaInp = nmeInp, fleOut = nmeOut, varLbl = lblDtF[, 2]),
      regexp = paste0("^If the parameter varLbl is a character, it eiter needs to be the name of an \\(exisiting\\) ",
                      "file or a vector with labels that has the same length as the number of variables in the input ",
                      "data set \\(\\d+\\).$"))
    expect_false(file.exists(nmeOut))
    expect_error(label_vars_omv(dtaInp = nmeInp, fleOut = nmeOut, varLbl = data.frame(varNme = c("Dummy", lblDtF[-1, 1]), varLbl = lblDtF[2])),
      regexp = "^There must be exactly one column with the variable names \\(currently: 0\\)\\.\\nAll variable names in the label definition must be contained in the input data set\\.")
    expect_false(file.exists(nmeOut))
    expect_error(label_vars_omv(dtaInp = nmeInp, fleOut = nmeOut, varLbl = numeric(1)),
      regexp = "^varLbl was either not a data frame or could not be converted into one.")
    expect_false(file.exists(nmeOut))
    unlink(nmeOut)

    # test cases for the transfer of analyses =================================================================================================================
    nmeInp <- file.path("..", "ToothGrowth.omv")
    lblDtF <- data.frame(varNme = c("ID", "len", "supp", "dose", "dose2", "dose3"), varLbl = c("Participant", "Tooth length", "Type of Nutrition Supplement",
                "Dosage of the Nutrition Supplement (num.)", "Dosage of the Nutrition Supplement (ordered)", "Dosage of the Nutrition Supplement (factor)"))
    # check that the labels in the original file are not set (NULL)
    expect_true(all(vapply(read_omv(nmeInp)[, lblDtF[, 1]], function(c) is.null(attr(c, "jmv-desc")), logical(1))))

    label_vars_omv(dtaInp = nmeInp, fleOut = nmeOut, varLbl = lblDtF, psvAnl = TRUE)
    expect_true(chkFle(nmeOut))
    expect_gt(file.info(nmeOut)$size, 1)
    expect_true(chkFle(nmeOut, isZIP = TRUE))
    expect_true(chkFle(nmeOut, fleCnt = "meta"))
    expect_true(chkFle(nmeOut, fleCnt = "metadata.json"))
    expect_true(chkFle(nmeOut, fleCnt = "data.bin"))
    df4Chk <- read_omv(nmeOut, getSyn = TRUE)
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(60, 16))
    expect_equal(names(df4Chk), c("Filter 1", "ID", "logLen", "supp - Transform 1", "len", "supp", "dose", "dose2", "dose3", "Trial", "Residuals", "J", "K", "L", "M", "weights"))
    expect_equal(as.vector(vapply(df4Chk, typeof, character(1))),
      c("logical", "character", "double", "integer", "double", "integer", "double", "integer", "integer", "integer", "double", "double", "double", "integer", "logical", "integer"))
    expect_equal(as.character(lapply(df4Chk[, lblDtF[, 1]], attr, "jmv-desc")), lblDtF[, 2])
    expect_equal(sort(zip::zip_list(nmeOut)$filename),
      c("01 empty/analysis", "02 anova/analysis", "02 anova/resources/65167cb3bdaf8761.png", "02 anova/resources/99f9b5d34a92049b.png", "03 empty/analysis",
        "04 ancova/analysis", "05 empty/analysis", "data.bin", "index.html", "meta", "metadata.json", "strings.bin", "xdata.json"))
    expect_equal(attr(df4Chk, "syntax"),
      c(paste("jmv::ANOVA(formula = len ~ supp + dose2 + supp:dose2, data = data, effectSize = \"partEta\", modelTest = TRUE, qq = TRUE,",
              "contrasts = list(list(var=\"supp\", type=\"none\"), list(var=\"dose2\", type=\"polynomial\")), postHoc = ~ supp + dose2, emMeans = ~ dose2:supp)"),
           "jmv::ancova(formula = len ~ supp + dose, data = data, effectSize = \"partEta\", modelTest = TRUE)"))
    unlink(nmeOut)
    expect_warning(label_vars_omv(dtaInp = jmvReadWrite::AlbumSales, fleOut = nmeOut, varLbl = c("Include participant?", "Sales", "Adverts", "Airplay", "Image"), psvAnl = TRUE),
      regexp = "^psvAnl is only possible if dtaInp is a file name \\(analyses are not stored in data frames, only in the jamovi files\\)\\.")
    expect_warning(label_vars_omv(dtaInp = jmvReadWrite::AlbumSales,                  varLbl = c("Include participant?", "Sales", "Adverts", "Airplay", "Image"), psvAnl = TRUE),
      regexp = "^psvAnl is only possible if fleOut is a file name \\(analyses are not stored in data frames, only in the jamovi files\\)\\.")
    unlink(nmeOut)
    # do not unlink nmeInp, this isn't a generated file, but a link
})

#' Aggregates data from a data set / frame (in long format) and returns them as a .omv-file for the
#' statistical spreadsheet 'jamovi' (<https://www.jamovi.org>)
#'
#' @param dtaInp Either a data frame or the name of a data file to be read (including the path,
#'               if required; "FILENAME.ext"; default: NULL); files can be of any supported file
#'               type, see Details below.
#' @param fleOut Name of the aggregated data set / file to be written (including the path, if
#'               required; "FILE_OUT.omv"; default: ""); if empty, the resulting data frame is
#'               returned instead.
#' @param varAgg A character vector (default: c()) with the names of the variables which shall be
#'               aggregated.
#' @param grpAgg A character vector (default: c()) with the variables to group the aggregation
#'               variable (`varAgg`) by. If no grouping variable is given, the aggregation happens
#'               over the whole data set. If several grouping variables are given, the aggregation
#'               happens for each step / possible combination of these variables. See Details for
#'               more information.
#' @param clcN   If TRUE, counts the number of valid values for each step / combination of values
#'               in the grouping variable. The suffix appended `"_N"´ is appended to each variable
#'               in the resulting data set.
#' @param clcMss If TRUE, counts the number of missing values for each step / combination of values
#'               in the grouping variable. The suffix appended `"_Mss"` is appended to each
#'               variable in the resulting data set.
#' @param clcMn  If TRUE, calculates the mean for each step / combination of values in the grouping
#'               variable. The suffix appended `"_Mn"` is appended to each variable in the
#'               resulting data set.
#' @param clcMdn If TRUE, calculates the median for each step / combination of values in the
#'               grouping variable. The suffix appended `"_Mdn"` is appended to each variable in
#'               the resulting data set.
#' @param clcMde If TRUE, calculates the mode for each step / combination of values in the grouping
#'               variable. The suffix appended `"_Mde"` is appended to each variable in the
#'               resulting data set.
#' @param clcSum If TRUE, calculates the sum for each step / combination of values in the grouping
#'               variable. The suffix appended `"_Sum"` is appended to each variable in the
#'               resulting data set.
#' @param clcSD  If TRUE, calculates the std. deviation for each step / combination of values in
#'               the grouping variable. The suffix appended `"_SD"` is appended to each variable in
#'               the resulting data set.
#' @param clcVar If TRUE, calculates the variance for each step / combination of values in the
#'               grouping variable. The suffix appended `"_Var"` is appended to each variable in
#'               the resulting data set.
#' @param clcRng If TRUE, calculates the range for each step / combination of values in the
#'               grouping variable. The suffix appended `"_Rng"` is appended to each variable in
#'               the resulting data set.
#' @param clcMin If TRUE, determines the minimum at each step / combination of values in the
#'               grouping variable. The suffix appended `"_Min"` is appended to each variable in
#'               the resulting data set.
#' @param clcMax If TRUE, determines the maximum at each step / combination of values in the
#'               grouping variable. The suffix appended `"_Max"` is appended to each variable in
#'               the resulting data set.
#' @param clcIQR If TRUE, calculates the IQR for each step / combination of values in the grouping
#'               variable. The suffix appended `"_IQR"` is appended to each variable in the
#'               resulting data set.
#' @param drpNA  If TRUE (default: TRUE), NA values are removed before the aggregation, and the
#'               aggregation is calculated using the valid values. If FALSE, the result would be NA
#'               for any step / combination of values in the grouping variable that contains a NA
#'               value.
#' @param usePkg Name of the package: "foreign" or "haven" that shall be used to read SPSS, Stata,
#'               and SAS files; "foreign" is the default (it comes with base R), but "haven" is
#'               newer and more comprehensive.
#' @param selSet Name of the data set that is to be selected from the workspace (only applies when
#'               reading .RData-files)
#' @param ...    Additional arguments passed on to methods; see Details below.
#'
#' @return a data frame containing a symmetric matrix (only returned if `fleOut` is empty)
#'         containing the distances between the variables / columns (clmDst == TRUE) or rows
#'         (clmDst == FALSE)
#'
#' @details
#' * `varAgg` must be a character vector containing the variables to be aggregated. From these
#'   variables, a new data set is created where the values in the original data set are aggregated
#'   for each possible combination of steps in the grouping variable(s) given in `grpAgg`. For
#'   example, if one grouping variable were a participant ID, and another grouping variable were
#'   measurement time points, the resulting new dataset would contain as many rows as the number of
#'   participants multiplied by the number of measurement time points. For each variable in
#'   `varAgg`, and each possible aggregation (`clcN`, `clcMn`, etc. if set to TRUE), a new column
#'   would be generated in the resulting new data set.
#' * `drpNA` determines whether NA should be dropped / removed before calculating an aggregation.
#'   If set to FALSE, the result for a combination of steps in the grouping variable(s) where at
#'   least one value is NA would be NA. If set to TRUE, only the values for any given combination
#'   of steps in the grouping variable(s) that are not NA are considered for the calculation. For
#'   `clcN` and `clcMss`, `drpNA` has no effect.
#'   NB: If `drpNA` is set to TRUE, any row where any of the grouping variable(s) has the value NA
#'   is excluded from the aggregation, if `drpNA` is set to FALSE and there is any row where any of
#'   the group variables has the value NA, an error is thrown and no aggregation is carried out.
#' * The ellipsis-parameter (`...`) can be used to submit arguments / parameters to the functions
#'   that are used for reading and writing the data. By clicking on the respective function under
#'   “See also”, you can get a more detailed overview over which parameters each of those functions
#'   take. The functions are: `read_omv` and `write_omv` (for jamovi-files), `read.table` (for
#'   CSV / TSV files; using similar defaults as `read.csv` for CSV and `read.delim` for TSV which
#'   both are based upon `read.table`), `load` (for .RData-files), `readRDS` (for .rds-files),
#'   `read_sav` (needs the R-package `haven`) or `read.spss` (needs the R-package `foreign`) for
#'   SPSS-files, `read_dta` (`haven`) / `read.dta` (`foreign`) for Stata-files, `read_sas`
#'   (`haven`) for SAS-data-files, and `read_xpt` (`haven`) / `read.xport` (`foreign`) for
#'   SAS-transport-files. If you would like to use `haven`, you may need to install it using
#'   `install.packages("haven", dep = TRUE)`.
#'
#' @seealso `aggregate_omv` internally uses [stats::aggregate()] as function for the aggregation,
#'   and [base::is.na()], [base::mean()], [stats::median()], [base::sum()], [stats::sd()],
#'   [stats::var()], [base::range()], [base::min()], [base::max()], and
#'   [stats::quantile()] for calculation. It furthermore uses the following functions for reading
#'   and writing data files in different formats: [jmvReadWrite::read_omv()] and
#'   [jmvReadWrite::write_omv()] for jamovi-files, [utils::read.table()] for CSV / TSV files,
#'   [load()] for reading .RData-files, [readRDS()] for .rds-files, [haven::read_sav()] or
#'   [foreign::read.spss()] for SPSS-files, [haven::read_dta()] or [foreign::read.dta()] for
#'   Stata-files, [haven::read_sas()] for SAS-data-files, and [haven::read_xpt()] or
#'   [foreign::read.xport()] for SAS-transport-files.
#'
#' @examples
#' # generate a test dataframe with 100 (imaginary) participants / units of
#' #  observation (ID), 10 measurement (measure) of two variables (V1, V2)
#' dtaInp <- data.frame(ID = rep(as.character(seq(1, 100)), each = 10),
#'                      Measure = rep(seq(1, 10), times = 100),
#'                      V1 = runif(1000,   0, 100),
#'                      V2 = round(rnorm(1000, 100,  15)))
#' cat(str(dtaInp))
#' # the output should look like this
#' # 'data.frame': 1000 obs. of  4 variables:
#' #  $ ID     : chr  "1" "1" "1" "1" ...
#' #  $ Measure: int  1 2 3 4 5 6 7 8 9 10 ...
#' #  $ V1     : num  ...
#' #  $ V2     : num  ...
#' # this data set is stored as (temporary) RDS-file and later processed by long2wide
#' nmeInp <- tempfile(fileext = ".rds")
#' nmeOut <- tempfile(fileext = ".omv")
#' saveRDS(dtaInp, nmeInp)
#' jmvReadWrite::aggregate_omv(dtaInp = nmeInp, fleOut = nmeOut, varAgg = c("V1", "V2"),
#'                             grpAgg = "ID", clcN = TRUE, clcMn = TRUE, clcSD = TRUE)
#' # it is required to give at least the arguments dtaInp, varAgg and grpAgg, each of
#' # the different switches (clc...) requests a aggregation measure (e.g., mean, median,
#' # SD, IQR, etc.) to be calculated
#' # check whether the file was created and its size
#' cat(list.files(dirname(nmeOut), basename(nmeOut)))
#' # -> "file[...].omv" ([...] contains a random combination of numbers / characters
#' cat(file.info(nmeOut)$size)
#' # -> 4898 (approximate size; size may differ in every run [in dependence of
#' #          how well the generated random data can be compressed])
#' cat(str(jmvReadWrite::read_omv(nmeOut, sveAtt = FALSE)))
#' # the data set contains now the ID variable identifying the different steps of
#' # aggregation and one column for each combination of aggregation variable (V1 / V2)
#' # and which calculation was requested (N, mean and SD)
#' # 'data.frame':	100 obs. of  7 variables:
#' #  $ ID   : chr  "1" "10" "100" "11" ...
#' #   ..- attr(*, "jmv-id")= logi TRUE
#' #   ..- attr(*, "missingValues")= list()
#' #  $ V1_N : int  10 10 10 10 10 10 10 10 10 10 ...
#' #   ..- attr(*, "jmv-desc")= chr "V1 (N)"
#' #   ..- attr(*, "missingValues")= list()
#' #  $ V1_Mn: num  45.4 51.9 49.4 54.8 47.2 ...
#' #   ..- attr(*, "jmv-desc")= chr "V1 (Mean)"
#' #   ..- attr(*, "missingValues")= list()
#' #  $ V1_SD: num  31.7 31.4 26.5 20.2 29.1 ...
#' #   ..- attr(*, "jmv-desc")= chr "V1 (SD)"
#' #   ..- attr(*, "missingValues")= list()
#' #  $ V2_N : int  10 10 10 10 10 10 10 10 10 10 ...
#' #   ..- attr(*, "jmv-desc")= chr "V2 (N)"
#' #   ..- attr(*, "missingValues")= list()
#' #  $ V2_Mn: num  96.4 102.3 101.6 104.6 108.7 ...
#' #   ..- attr(*, "jmv-desc")= chr "V2 (Mean)"
#' #   ..- attr(*, "missingValues")= list()
#' #  $ V2_SD: num  14.8 18.4 11.2 10.1 14.3 ...
#' #   ..- attr(*, "jmv-desc")= chr "V2 (SD)"
#' #   ..- attr(*, "missingValues")= list()
#'
#' unlink(nmeInp)
#' unlink(nmeOut)
#'
#'
#' @export aggregate_omv
#'
aggregate_omv <- function(dtaInp = NULL, fleOut = "", varAgg = c(), grpAgg = c(), clcN = FALSE,
                          clcMss = FALSE, clcMn = FALSE, clcMdn = FALSE, clcMde = FALSE,
                          clcSum = FALSE, clcSD = FALSE, clcVar = FALSE, clcRng = FALSE,
                          clcMin = FALSE, clcMax = FALSE, clcIQR = FALSE, drpNA = TRUE,
                          usePkg = c("foreign", "haven"), selSet = "", ...) {

    # check and import input data set (either as data frame or from a file)
    if (!is.null(list(...)[["fleInp"]])) stop("Please use the argument dtaInp instead of fleInp.")
    frmInp <- inp2DF(dtaInp = dtaInp, usePkg = usePkg, selSet = selSet, ...)

    if (!all(nzchar(varAgg)) || length(intersect(varAgg, names(frmInp))) < 1 ||
        !all(nzchar(grpAgg)) || length(intersect(grpAgg, names(frmInp))) < length(grpAgg)) {
        stop(paste("Calling aggregate_omv requires giving at least one (valid) variable to",
                   "aggregate, and the grouping variable(s) needs to be empty or valid."))
    }

    if (length(grpAgg) >= 1) {
        grpNA <- rowSums(is.na(frmInp[, grpAgg, drop = FALSE])) > 0
        if (!drpNA && any(grpNA))
            stop("The grouping variables must not contain empty or NA values (if drpNA is set to FALSE).")
    } else {
        grpNA <- rep(FALSE, nrow(frmInp))
    }

    clcSel <- c(clcN, clcMss,  clcMn,  clcMdn,   clcMde, clcSum, clcSD, clcVar,     clcRng,  clcMin, clcMax, clcIQR)
    clcStr <- c("N",  "Mss",   "Mn",   "Mdn",    "Mde",  "Sum",  "SD",  "Var",      "Rng",   "Min",  "Max",  "IQR")
    clcDsc <- c("N",  "Miss.", "Mean", "Median", "Mode", "Sum",  "SD",  "Variance", "Range", "Min.", "Max.", "IQR")
    if (!any(clcSel))
        stop("At least one aggregation calculation (clc...) needs to be set to TRUE.")

    frmOut <- NULL
    dscOut <- list()
    for (crrVar in varAgg) {
        crrDsc <- c(attr(frmInp[, crrVar], "description"), attr(frmInp[, crrVar], "jmv-desc"), crrVar)

        if (!is.numeric(frmInp[, crrVar]))
            frmInp[, crrVar] <- cnvCol(frmInp[, crrVar], "numeric")

        for (i in which(clcSel)) {
            crrNme <- sprintf("%s_%s", crrVar, clcStr[i])
            crrRes <- do.call(stats::aggregate,
                              c(list(x = frmInp[!grpNA, crrVar], by = frmInp[!grpNA, grpAgg, drop = FALSE], FUN = getFnc(clcStr[i], drpNA)),
                                rep(list(na.rm = drpNA), !(clcStr[i] %in% c("N", "Mss", "Mde", "Rng", "IQR")))))
            names(crrRes)[ncol(crrRes)] <- crrNme
            dscOut[crrNme] <- sprintf("%s (%s)", crrDsc[nzchar(crrDsc)][1], clcDsc[i])
            if (is.null(frmOut)) {
                frmOut <- crrRes
            } else {
                frmOut <- merge(frmOut, crrRes, by = grpAgg, all = TRUE)
            }
        }
    }

    for (crrCol in names(dscOut)) attr(frmOut[, crrCol], "description") <- dscOut[[crrCol]]

    # rtnDta in globals.R (unified function to either write the data frame, open it in a new jamovi
    # session or return it)
    rtnDta(dtaFrm = frmOut, fleOut = fleOut, dtaTtl = jmvTtl("_agg"), ...)
}

getFnc <- function(crrClc = "", drpNA = TRUE) {
    if        (crrClc == "N")   {
        eval(parse(text = "function(x) sum(!is.na(x))"))
    } else if (crrClc == "Mss") {
        eval(parse(text = "function(x) sum(is.na(x))"))
    } else if (crrClc == "Mn")  {
        mean
    } else if (crrClc == "Mdn") {
        stats::median
    } else if (crrClc == "Mde") {
        eval(parse(text = sprintf("function(x) ifelse(%s && any(is.na(x)), NA, as.numeric(names(sort(-table(x))))[1])", !drpNA)))
    } else if (crrClc == "Sum") {
        sum
    } else if (crrClc == "SD")  {
        stats::sd
    } else if (crrClc == "Var") {
        stats::var
    } else if (crrClc == "Rng") {
        eval(parse(text = sprintf("function(x) diff(range(x, na.rm = %s))", drpNA)))
    } else if (crrClc == "Min") {
        min
    } else if (crrClc == "Max") {
        max
    } else if (crrClc == "IQR") {
        eval(parse(text = sprintf(paste("function(x) ifelse(%s && any(is.na(x)), NA,",
                                        "stats::quantile(x, 0.75, na.rm = TRUE) - stats::quantile(x, 0.25, na.rm = TRUE))"), !drpNA)))
    }
}

#' Write files to be used with the statistical spreadsheet 'jamovi'
#' (<https://www.jamovi.org>)
#'
#' @param dtaFrm Data frame to be exported (default: NULL)
#' @param fleOut Name / position of the output file to be generated
#'               ("FILENAME.omv"; default: "")
#' @param wrtPtB Whether to write protocol buffers (see Details; default: FALSE)
#' @param frcWrt Whether to overwrite existing files with the same name (see
#'               Details; default: FALSE)
#' @param vldExt Whether to validate that the file name has a correct extension
#'               (default: TRUE)
#' @param retDbg Whether to return a list with debugging information (see
#'               Value; default: FALSE)
#'
#' @return a list (if retDbg == TRUE), containing the meta data (mtaDta,
#'         metadata.json in the OMV-file), the extended data (xtdDta,
#'         xdata.json in the OMV-file) and the original data frame (dtaFrm)
#'
#' @details
#' * jamovi has a specific measurement level / type "ID" (in addition to the
#'   "standard" ones "Nominal", "Ordinal", and "Continuous"). "ID" is used for
#'   columns that contain some form of ID (e.g., a participant code). In order
#'   to set a variable of your data frame to "ID", you have to set the
#'   attribute `jmv-id` (e.g., `attr(dtaFrm$column, "jmv-id") = TRUE`).
#' * CAUTION: Setting wrtPtB to TRUE currently overwrites analyses that already
#'   exist in a data file. It is meant to be used for `describe_omv` only. If
#'   you set wrtPtB to TRUE, ensure to use an output file name that isn't would
#'   not overwrite any existing file. Protocol buffers are used to exchange
#'   data between the different parts of jamovi (the server and the client) and
#'   also the format in which analyses are stored in the jamovi data files.
#' * `write_omv` checks whether the output file already exists and throws an
#'   error if this is the case. frcWrt permits you to overwrite the existing
#'   file.
#'
#' @examples
#' # use the data set "ToothGrowth" and, if it exists, write it as
#' # jamovi-file using write_omv()
#' jmvReadWrite::ToothGrowth
#' nmeOut <- tempfile(fileext = ".omv")
#' # typically, one would use a "real" file name instead of tempfile(),
#' # e.g., "Data1.omv"
#' dtaDbg = jmvReadWrite::write_omv(dtaFrm = ToothGrowth, fleOut = nmeOut, retDbg = TRUE)
#' print(names(dtaDbg))
#' # the print-function is only used to force devtools::run_examples()
#' # to show output
#' # -> "mtaDta" "xtdDta" "dtaFrm"
#' # returns a list with the metadata (mtaDta, e.g., column and data type),
#' # the extended data (xtdDta, e.g., variable lables), and the data frame
#' # (dtaFrm) the purpose of these variables is merely for checking (under-
#' # standing the file format) and debugging
#'
#' # check whether the file was written to the disk, get the file informa-
#' # tion (size, etc.) and delete the file afterwards
#' print(list.files(dirname(nmeOut), basename(nmeOut)))
#' # -> "file[...].omv" ([...] is a combination of random numbers / characters
#' print(file.info(nmeOut)$size)
#' # -> approx. 2600 (size may differ on different OSes)
#' unlink(nmeOut)
#'
#'
#' @export write_omv

write_omv <- function(dtaFrm = NULL, fleOut = "", wrtPtB = FALSE, frcWrt = FALSE, retDbg = FALSE, vldExt = TRUE) {
    if (is.null(dtaFrm)) stop("The data frame to be written needs to be given as parameter (dtaFrm = ...).")
    if (!nzchar(fleOut)) stop("Output file name needs to be given as parameter (fleOut = ...).")

    # check that the file name isn't empty, that the destination directory exists and that it ends in .omv
    fleOut <- nrmFle(fleOut)
    chkDir(fleOut)
    !vldExt || chkExt(fleOut, c("omv", "omt"))
    fleExs(fleOut, frcWrt)

    # [1] handle the attributes "variable.labels" and "value.labels" in the format provided by the R-package "foreign"
    #     the attribute "variable.labels" (attached to the data frame) is converted them to the format used by jamovi ("jmv-desc" attached to the data column)
    # [2] clean tibble attributes, particularly convert haven_labelled to either factors or numeric / integer
    if (chkAtt(dtaFrm, "variable.labels"))
        dtaFrm <- clnFgn(dtaFrm)
    if (methods::is(dtaFrm, "tbl_df") || any(vapply(dtaFrm, function(C) methods::is(C, "haven_labelled"), logical(1))))
        dtaFrm <- clnTbb(dtaFrm, c("format.sas", "format.spss", "format.stata", "display_width"), jmvLbl = TRUE)
    if (chkAtt(dtaFrm, "label.table"))
        stop("R-foreign-style value labels need to be implemented. Please send the data file that caused this problem to sebastian.jentschke@uib.no")

    # check whether dtaFrm is a data frame, remove tibble attributes (if necessary) and attach dataType and measureType attributes
    dtaFrm <- jmvAtt(dtaFrm, blnChC = TRUE)

    # initialize metadata.json
    mtaDta <- mtaGlb
    # use the attributes stored in the data frame (for the whole data set) to
    # update the metadata; jmv-weights-name are stored as weights, jmv-weights
    # needs to be dropped
    if (chkAtt(dtaFrm, "jmv-weights-name")) {
        attr(dtaFrm, "weights") <- attr(dtaFrm, "jmv-weights-name")
        attr(dtaFrm, "jmv-weights-name") <- NULL
        attr(dtaFrm, "jmv-weights")      <- NULL
    }
    mtaDta <- setAtt(names(mtaDta), dtaFrm, mtaDta)
    # the number of rows and columns has to be adjusted to the current data set
    mtaDta$rowCount    <- ifelse(hasExt(fleOut, "omt"), 0, nrow(dtaFrm))
    mtaDta$columnCount <- ncol(dtaFrm)
    # create the entries for storing the column specific information
    mtaDta$fields      <- rep(list(mtaFld), mtaDta$columnCount)

    # initialize xdata.json
    xtdDta <- list()

    # create data.bin
    binHdl <- file(description = file.path(tempdir(), "data.bin"),    open = "wb")

    # create strings.bin
    strHdl <- file(description = file.path(tempdir(), "strings.bin"), open = "wb")
    strPos <- 0

    for (i in seq_along(dtaFrm)) {
        # assign the jamovi-specific-attributes that are stored in data frame for this data column (if available)
        mtaDta$fields[[i]] <- setAtt(names(mtaDta$fields[[i]]), dtaFrm[i], mtaDta$fields[[i]])

        # name
        crrNme <- names(dtaFrm[i])
        mtaDta$fields[[i]][["name"]] <- crrNme

        # variable label: if available, choose "jmv-desc", "label", ...
        # the attributes are concatenated if available (otherwise they will be NULL and dropped), if several are available,
        # the first ("jmv-desc") takes precedence, if all are NULL, the content of mtaDta$fields serves as fallback-option
        mtaDta$fields[[i]][["description"]] <- c(attr(dtaFrm[[i]], "jmv-desc"), mtaDta$fields[[i]][["description"]], attr(dtaFrm[[i]], "label"))[1]

        # remove atrributes that are only used with specific columnTypes
        mtaDta$fields[[i]] <- rmvMta(mtaDta$fields[[i]], dtaFrm[[i]])

        # assign column from the original data frame to crrCol (so that modifications don't affect the original)
        crrCol <- dtaFrm[[i]]

        # factors: store labels / values in xdata.json and convert factors to integer for saving
        if (is.factor(crrCol)) {
            xtdDta[[crrNme]] <- xtdCol(crrCol, crrNme)
            crrCol <- cnvCol(crrCol, "integer")
        }

        # keep value from "type" if it exists, otherwise determine and set "type" based upon whether the variable is an ID,
        # and upon "dataType"
        mtaDta$fields[[i]][["type"]] <- ifelse(chkAtt(crrCol, "type"), attr(crrCol, "type"),
                                               gsub("decimal", "number",
                                               gsub("text", ifelse(isID(crrCol), "string", "integer"), tolower(attr(crrCol, "dataType")))))

        # for debugging: check that dataType, and measureType are set accordingly to type (attribute and column in the data frame)
        #cat(do.call(sprintf, c(fmt = "%02d: %s - %s - %s - %s\n", c(i, mtaDta$fields[[i]][c("name", "type", "dataType", "measureType")]))))

        # write to data.bin according to type
        if (!hasExt(fleOut, "omt")) {
            if        (chkFld(mtaDta$fields[[i]], "type", "integer")) {
                wrtCol <- as.integer(crrCol)
            } else if (chkFld(mtaDta$fields[[i]], "type", "number"))  {
                wrtCol <- as.double(crrCol)
            } else if (chkFld(mtaDta$fields[[i]], "type", "string"))  {
                wrtCol <- cnvStr(crrCol, strHdl, strPos)
                strPos <- attr(wrtCol, "strPos")
                wrtCol <- as.integer(wrtCol)
            }
            writeBin(wrtCol, binHdl, endian = "little")
            # remove temporary variables for writing the current column
            rm(wrtCol)
        }

        # remove temporary variables for modifying the current column from the data set
        rm(crrCol)
    }

    # ensure that the column-ID is not empty(NA) and unique
    mtaDta$fields <- unqID(mtaDta$fields)

    # compress data.bin and discard the temporary file
    add2ZIP(fleOut, crrHdl = binHdl)

    # compress strings.bin (only if it contains data) and discard the temporary file
    add2ZIP(fleOut, crrHdl = strHdl, incZIP = (strPos > 0))
    rm(strPos)

    # create meta, write it and add it to ZIP file
    add2ZIP(fleOut, crrFle = c("meta", "wb"), txtOut = mnfTxt())

    # write metadata.json
    add2ZIP(fleOut, crrFle = "metadata.json", txtOut = fmtJSON(list(dataSet = mtaDta)))

    # write xdata.json
    add2ZIP(fleOut, crrFle = "xdata.json",    txtOut = fmtJSON(xtdDta))

    # if "HTML" attribute exists (e.g., because functions such as describe_omv
    # did modify it), this attribute is written; otherwise, htmTxt (a clean
    # "index.html", i.e., without any results) is written
    if (!is.null(attr(dtaFrm, "HTML"))) {
        add2ZIP(fleOut, crrFle = "index.html", txtOut = attr(dtaFrm, "HTML"))
    } else {
        add2ZIP(fleOut, crrFle = "index.html", txtOut = htmTxt())
    }

    # write ProtoBuffers: if the argument wrtPtB is set and the data set contains no
    # protobuffers, an error message is thrown; for templates, protobuffers are written
    # (if they exist), but their non-existence doesn't result in an error message
    if ((wrtPtB || hasExt(fleOut, "omt")) && hasPtB(dtaFrm)) {
        add2ZIP(fleOut, crrFle = c(names(attr(dtaFrm, "protobuf")[1]), "wb"), ptbOut = attr(dtaFrm, "protobuf")[[1]])
    } else if (wrtPtB && !hasPtB(dtaFrm)) {
        unlink(fleOut)
        stop("The data frame (dtaFrm) must contain the attribute \"protobuf\", there has to be at least one of them, and it has to be of the correct type (a RProtoBuf).")
    }

    # handle weights
    if (chkAtt(dtaFrm, "weights", "\\w+")) {
        # TO-DO: this likely requires creating protobuffers
        warning("Handling of weights not yet implemented.")
    }

    if (retDbg) {
        list(mtaDta = mtaDta, xtdDta = xtdDta, dtaFrm = dtaFrm)
    }
}

jmvAtt <- function(dtaFrm = NULL, blnChC = FALSE) {
    chkDtF(dtaFrm)

    for (i in seq_along(dtaFrm)) {
        crrNme <- names(dtaFrm)[i]
        # (a) jmv-id
        # ID variables represent a special case and are therefore treated first
        # if the jmv-id marker is set or if the measureType is set to "ID" in the original data, or if it is the first column that hasn't the attribute
        # measureType attached, has values that are unique and not NA, is either a factor or an integer (rounded equals the original value) and has a
        # name that (as lower-case) matches "id", "name" or "subject"
        if (isID(dtaFrm[[i]], i, crrNme)) {
            attr(dtaFrm[[i]], "jmv-id")      <- TRUE
            attr(dtaFrm[[i]], "measureType") <- "ID"
            attr(dtaFrm[[i]], "dataType")    <- "Text"
            if (blnChC) dtaFrm[[i]] <- cnvCol(dtaFrm[[i]], "character")
        # (b) date and time - jamovi doesn't support it natively, thus the transformation to integer;
        # back-transformation in R: as.Date(..., origin = "1970-01-01") / hms::as_hms(...)
        # NB: must come before the numerical variables since date values are an integer since R 4.5
        } else if (isDnT(dtaFrm[[i]])) {
            attr(dtaFrm[[i]], "measureType") <- "Continuous"
            attr(dtaFrm[[i]], "dataType")    <- "Integer"
            if (blnChC) {
                # assign the description before the conversion (description depends on the original type)
                attr(dtaFrm[[i]], "description") <- dscDnT(dtaFrm[[i]], crrNme)
                dtaFrm[[i]] <- cnvCol(dtaFrm[[i]], "integer")
            }
        # (c) numerical variables, determine first whether the variable can be integer, if not, use / keep it numeric / float
        } else if (is.numeric(dtaFrm[[i]])) {
            attr(dtaFrm[[i]], "measureType") <- "Continuous"
            attr(dtaFrm[[i]], "dataType")    <- ifelse(is.integer(dtaFrm[[i]]) || detInt(dtaFrm[[i]]), "Integer", "Decimal")
        # (d) factors
        } else if (is.factor(dtaFrm[[i]])) {
            attr(dtaFrm[[i]], "measureType") <- ifelse(is.ordered(dtaFrm[[i]]), "Ordinal", "Nominal")
            attr(dtaFrm[[i]], "dataType")    <- ifelse(!is.null(attr(dtaFrm[[i]], "values")) || intFnC(dtaFrm[[i]]), "Integer", "Text")
        # (e) logical and character are converted to factor (if blnChC)
        } else if (is.logical(dtaFrm[[i]]) || is.character(dtaFrm[[i]])) {
            attr(dtaFrm[[i]], "measureType") <- "Nominal"
            attr(dtaFrm[[i]], "dataType")    <- ifelse(is.logical(dtaFrm[[i]]) || intFnC(dtaFrm[[i]]), "Integer", "Text")
            if (blnChC) dtaFrm[[i]] <- cnvCol(dtaFrm[[i]], "factor")
        # variable type is not implemented
        } else {
            clsRmv()
            stop(sprintf("\n\n%s: Variable type %s not implemented. Please send the data file that caused this problem to sebastian.jentschke@uib.no",
              names(dtaFrm)[i], class(dtaFrm[[i]])))
        }

    }

    dtaFrm
}

# determine whether a column is (i.e., can become) integer without loosing data
detInt <- function(crrCol = NULL) {
    is.numeric(crrCol) && !all(is.na(crrCol)) &&
    max(abs(crrCol), na.rm = TRUE) <= .Machine$integer.max &&
    all(abs(crrCol - round(crrCol)) < sqrt(.Machine$double.eps), na.rm = TRUE)
}

# determine whether a column likely should get the measureType "Continuous", requiring
# that there are enough different values and a high value range and variability (sd)
#detCnt <- function(crrCol = NULL) {
#    length(unique(crrCol)) > length(crrCol) / 5 &&
#    length(unique(crrCol)) > diff(range(crrCol, na.rm = TRUE)) / 5 &&
#    stats::sd(crrCol, na.rm = TRUE) > diff(range(crrCol, na.rm = TRUE)) / 10
#}

isDnT  <- function(crrCol) {
    methods::is(crrCol, "Date") || methods::is(crrCol, "POSIXct") || methods::is(crrCol, "POSIXlt") || methods::is(crrCol, "difftime")
}

dscDnT <- function(crrCol = NULL, crrNme = "") {
    paste(ifelse(chkAtt(crrCol, "description"), attr(crrCol, "description"), crrNme),
          ifelse(methods::is(crrCol, "difftime"), "(time converted to integer; sec since 00:00)", "(date converted to integer; days since 1970-01-01)"))
}

xtdCol <- function(crrCol = NULL, crrNme = "") {
    crrLvl <- levels(crrCol)
    # ensure that the "values" attribute is correct
    if (chkAtt(crrCol, "values") && all(grepl("^\\d+$", crrLvl)) &&
      !identical(attr(crrCol, "values"), as.integer(attr(crrCol, "levels")))) {
        clsRmv()
        stop(sprintf(paste("\"values\"-attribute with unexpected values found for column \"%s\".",
                           "Please send the file to sebastian.jentschke@uib.no for debugging."), crrNme))
    }

    # columns that have previously been logical or where all factor levels can be converted to integer
    if        (chkAtt(crrCol, "dataType", "Integer") && identical(crrLvl, c("FALSE", "TRUE"))) {
        return(list(labels = lapply(c(TRUE, FALSE),    function(l) list(as.integer(l), as.character(l), as.character(l), FALSE))))
    } else if (chkAtt(crrCol, "dataType", "Integer") && intFnC(crrLvl)) {
        return(list(labels = lapply(crrLvl,            function(l) list(as.integer(l),              l,               l,  FALSE))))
    # columns that           or where not all factor levels can be converted to integer
    } else if (chkAtt(crrCol, "dataType", "Text")) {
        return(list(labels = lapply(seq_along(crrLvl), function(l) list(l - 1,               crrLvl[l],       crrLvl[l], FALSE))))
    }
}

rmvMta <- function(crrFld = NULL, dtaCol = NULL) {
    if (chkFld(crrFld, "columnType", "Filter")) {
        # if the variable is a filter, trimLevels is removed in any case
        crrFld[["trimLevels"]]               <- NULL
    } else {
        # if the variable isn't a filter, trimLevels is only removed
        # if the original variable was a factor / logical
        if (all(class(dtaCol) != c("logical", "factor"))) {
            crrFld[["trimLevels"]]           <- NULL
        }
        crrFld[["filterNo"]]                 <- NULL
        crrFld[["active"]]                   <- NULL
    }
    if (!chkFld(crrFld, "columnType", "Output")) {
        crrFld[["outputAnalysisId"]]         <- NULL
        crrFld[["outputOptionName"]]         <- NULL
        crrFld[["outputName"]]               <- NULL
        crrFld[["outputDesiredColumnName"]]  <- NULL
        crrFld[["outputAssignedColumnName"]] <- NULL
    }

    crrFld
}

cnvStr <- function(crrCol = NULL, strHdl = NULL, strPos = NA) {
    crrCol[is.na(crrCol)] <- ""
    wrtCol <- rep(NA, length(crrCol))
    for (j in seq_along(crrCol)) {
        if (crrCol[j] == "") next
        writeBin(crrCol[j], strHdl)
        wrtCol[j] <- strPos
        strPos <- strPos + length(charToRaw(crrCol[j])) + 1
    }

    attr(wrtCol, "strPos") <- strPos
    wrtCol
}

unqID <- function(mtaFld = NULL) {
    id_Lst <- unlist(lapply(seq_along(mtaFld), function(i) mtaFld[[i]][["id"]]))
    if (any(is.na(id_Lst)) || any(duplicated(id_Lst))) {
        for (i in seq_along(mtaFld)) mtaFld[[i]][["id"]] <- i
    }

    mtaFld
}

fleExs <- function(fleOut = c(), frcWrt = FALSE) {
    if (file.exists(fleOut)) {
        if (frcWrt) {
            unlink(fleOut)
        } else {
            stop(sprintf("The output file %s already exists. Either remove the file or set the parameter frcWrt to TRUE.", basename(fleOut)))
        }
    }
}

isID <- function(crrCol = NULL, i = NA, crrNme = "") {
    chkAtt(crrCol, "jmv-id", TRUE) || chkAtt(crrCol, "measureType", "ID") ||
           (i == 1 && !chkAtt(crrCol, "jmv-id") && !chkAtt(crrCol, "measureType") && !any(duplicated(crrCol) | is.na(crrCol)) &&
           (is.character(crrCol) || is.factor(crrCol) || (is.numeric(crrCol) && all(crrCol %% 1 == 0))) &&
           (tolower(crrNme) %in% c("id", "name", "subject")))
}

hasPtB <- function(dtaFrm = NULL) {
    chkAtt(dtaFrm, "protobuf") && inherits(attr(dtaFrm, "protobuf")[[1]], "Message")
}

# convert to JSON and do some beatifying (adding spaces for increased legibility)
fmtJSON <- function(txtJSON = "") {
    gsub("\"weights\": \\{\\}", "\"weights\": null", gsub("00: 00", "00:00", gsub("  ", " ", gsub(":", ": ", gsub(",", ", ",
      jsonlite::toJSON(txtJSON, auto_unbox = TRUE))))))
}

# generates an “empty” index.html (i.e., an index.html not containing any results output)
htmTxt <- function() {
    c("<!DOCTYPE html>",
      "<html>",
      "<head>",
      "    <meta charset=\"utf-8\" />",
      "    <title>Results</title>",
      "    <style>\n",
      "    body {",
      "        font-family: \"Segoe UI\",Roboto,Helvetica,Arial,sans-serif,\"Segoe UI Emoji\",\"Segoe UI Symbol\" ;",
      "        color: #333333 ;",
      "        cursor: default ;",
      "        margin: 24px;",
      "        font-size: 12px ;",
      "    }\n",
      "    h1 {",
      "        font-size: 160% ;",
      "        color: #3E6DA9 ;",
      "        margin-bottom: 12px ;",
      "        white-space: nowrap ;",
      "    }\n",
      "    h2 {",
      "        font-size: 130% ;",
      "        margin-bottom: 12px ;",
      "        color: #3E6DA9 ;",
      "    }\n",
      "    h3, h4, h5 {",
      "        font-size: 110% ;",
      "        margin-bottom: 12px ;",
      "    }\n",
      "    table {",
      "        border-spacing: 0 ;",
      "        page-break-inside: avoid;",
      "    }\n",
      "    table tr td, table tr th {",
      "        page-break-inside: avoid;",
      "        font-size: 12px ;",
      "    }\n",
      "    .ql-align-center {\n        text-align: center;\n    }\n",
      "    .ql-align-right {\n        text-align: right;\n    }\n",
      "    .ql-align-justify {\n        text-align: justify;\n    }\n",
      "    .ql-indent-1 {\n        padding-left: 3em;\n    }\n",
      "    .ql-indent-2 {\n        padding-left: 6em;\n    }\n",
      "    .ql-indent-3 {\n        padding-left: 9em;\n    }\n",
      "    .ql-indent-4 {\n        padding-left: 12em;\n    }\n",
      "    .ql-indent-5 {\n        padding-left: 15em;\n    }\n",
      "    .note {\n        margin: 5px 0px;\n    }\n",
      "    </style>",
      "</head>",
      "<body>\n",
      "    <h1 contenteditable=\"\" spellcheck=\"false\">Results</h1>\n",
      "    <p style=\"text-align:left;padding:0px 0px 0px 0px;\"></p>\n",
      "</body>",
      "</html>")
}

# creates the text output for the manifest file
mnfTxt <- function() {
    crrTxt <- paste0(vapply(lstMnf, "[[", character(1), 1), ":")
    for (i in seq_along(lstMnf)) {
        if (crrTxt[i] == "Created-By:") {
            crrTxt[i] <- paste(crrTxt[i], "jmvReadWrite", utils::packageVersion("jmvReadWrite"))
        } else {
            crrTxt[i] <- paste(crrTxt[i], paste(lstMnf[[i]][2:length(lstMnf[[i]])], collapse = "."))
        }
    }
    enc2utf8(crrTxt)
}

# adds an input file to the .omv-file (which is a ZIP-archive)
add2ZIP <- function(fleZIP = "", crrHdl = NULL, crrFle = c(), txtOut = "", ptbOut = NULL, incZIP = TRUE) {
    if ((!is.character(fleZIP) || length(fleZIP) < 1 || !nzchar(fleZIP)) ||
        ((is.null(crrHdl) || length(crrHdl) < 1) && (!is.character(crrFle) || length(crrFle) < 1 || !all(nzchar(crrFle))))) {
        clsRmv()
        stop("fleZIP (a character with a file name), and either crrHdl (with a connection) or crrFle (with a file name and [optionally] a writing mode) need to be given as arguments.")
    }

    # if a file handle is given, determine the file name, close the handle and remove it (from the calling environment)
    if (!is.null(crrHdl)) {
        if (!all(class(crrHdl) == c("file", "connection"))) {
            clsRmv()
            stop(sprintf("Parameter isn\'t a file handle pointing to a file to be zipped:\n%s", trimws(utils::capture.output(utils::str(crrHdl)))))
        }
        crrFle <- rmvTmp(summary(crrHdl)[["description"]])
        close(crrHdl)
        rm(list = deparse(substitute(crrHdl)), envir = sys.frame(-1))
    # if a file name is given, open a connection, write whatever is required
    } else if (is.character(crrFle)) {
        crrFle[1] <- rmvTmp(crrFle[1])
        crtDir(crrFle)
        crrHdl <- file(file.path(tempdir(), crrFle[1]), open = ifelse(length(crrFle) > 1, crrFle[2], "w"))
        if        (all(nzchar(txtOut))) {
            # writing the manifest requires a bit of special handling because of \r\n on Windows vs. \n on Mac / Linux
            # for the other files (metadata.json, xdata.json, index.html), it doesn't matter, therefore sep = "\n" is added for all
            writeLines(txtOut, crrHdl, sep = "\n")
        } else if (!is.null(ptbOut)) {
            ptbOut$serialize(crrHdl)
        }
        close(crrHdl)
    }

    if (incZIP) {
        crrPrm <- list(zipfile = fleZIP, files = crrFle[1], root = tempdir())
        if (file.exists(fleZIP)) do.call(zip::zip_append, crrPrm) else do.call(zip::zip, crrPrm)
    }

    delTmp(crrFle)
    rm(crrFle, crrHdl)

    return(TRUE)
}

crtDir <- function(crrFle = c()) {
    if (dirname(crrFle[1]) != "." && !dir.exists(file.path(tempdir(), dirname(crrFle[1])))) dir.create(file.path(tempdir(), dirname(crrFle[1])))

    return(invisible(NULL))
}

delTmp <- function(crrFle = c()) {
    if (dirname(crrFle[1]) == ".") unlink(file.path(tempdir(), crrFle[1])) else unlink(file.path(tempdir(), dirname(crrFle[1])), recursive = TRUE)

    return(invisible(NULL))
}

rmvTmp <- function(fleNme = "") {
    return(sub("^/|^\\\\", "", sub(tempdir(), "", fleNme, fixed = TRUE)))
}

test_that("write_omv works", {
    # check whether writing the data is working (file existence, size, contents [.omv-files are ZIP archives and must
    # contain files that include meta, metadata.json, data.bin])
    dtaOut <- jmvReadWrite::ToothGrowth
    rowOut <- dim(dtaOut)[1]
    nmeOut <- tempfile(fileext = ".omv")
    dtaDbg <- write_omv(dtaFrm = dtaOut, fleOut = nmeOut, retDbg = TRUE)
    expect_true(file.exists(nmeOut))
    expect_gt(file.info(nmeOut)$size, 1)
    expect_true(chkFle(nmeOut, isZIP = TRUE))
    expect_true(chkFle(nmeOut, fleCnt = "meta"))
    expect_true(chkFle(nmeOut, fleCnt = "metadata.json"))
    expect_true(chkFle(nmeOut, fleCnt = "data.bin"))
    unlink(nmeOut)

    # check the debugging information: name and type of the three parts that are returned, content of the metadata,
    # whether all entries in xtdDta are labels, and whether dtaFrm as a data frame with the correct sizes and attributes
    expect_equal(names(dtaDbg),                       c("mtaDta", "xtdDta", "dtaFrm"))
    expect_equal(vapply(dtaDbg, class, character(1), USE.NAMES = FALSE), c("list",   "list",   "data.frame"))
    expect_equal(names(dtaDbg$mtaDta),
                 c("rowCount", "columnCount", "removedRows", "addedRows", "fields", "transforms", "weights"))
    expect_true(all(grepl("labels", unlist(lapply(dtaDbg$xtdDta, attributes)))))
    expect_s3_class(dtaDbg$dtaFrm, "data.frame")
    expect_equal(dim(dtaDbg$dtaFrm), c(60, 7))
    expect_equal(names(attributes(dtaDbg$dtaFrm)), c("names", "row.names", "class"))
    expect_equal(names(attributes(dtaDbg$dtaFrm[[3]])), c("levels", "class", "description", "measureType", "dataType"))
    expect_equal(names(attributes(dtaDbg$dtaFrm[[7]])), c("jmv-desc", "measureType", "dataType"))
    expect_equal(names(attributes(dtaDbg$dtaFrm[[4]])), c("measureType", "dataType"))
    expect_equal(lapply(jmvReadWrite::ToothGrowth, class), lapply(dtaDbg$dtaFrm, class))

    # add columns with only NAs to the dataset and attach jamovi-attributes to them, write the resulting data frame
    dtaDbg <- write_omv(dtaFrm = cbind(dtaOut, jmvAtt(data.frame(T1 = rep(as.integer(NA), nrow(dtaOut)), T2 = as.numeric(NA), T3 = factor(NA),
                                                                 T4 = factor(NA, ordered = TRUE), T5 = as.character(NA), T6 = NA))), fleOut = nmeOut, retDbg = TRUE)
    expect_true(file.exists(nmeOut))
    expect_gt(file.info(nmeOut)$size, 1)
    expect_true(chkFle(nmeOut, isZIP = TRUE))
    expect_true(chkFle(nmeOut, fleCnt = "meta"))
    expect_true(chkFle(nmeOut, fleCnt = "metadata.json"))
    expect_true(chkFle(nmeOut, fleCnt = "data.bin"))
    unlink(nmeOut)

    expect_equal(names(dtaDbg),                       c("mtaDta", "xtdDta", "dtaFrm"))
    expect_equal(vapply(dtaDbg, class, character(1), USE.NAMES = FALSE), c("list",   "list",   "data.frame"))
    expect_equal(names(dtaDbg$mtaDta), c("rowCount", "columnCount", "removedRows", "addedRows", "fields", "transforms", "weights"))
    expect_true(all(grepl("labels", unlist(lapply(dtaDbg$xtdDta, attributes)))))
    expect_s3_class(dtaDbg$dtaFrm, "data.frame")
    expect_equal(dim(dtaDbg$dtaFrm), c(60, 13))
    expect_equal(names(attributes(dtaDbg$dtaFrm)), c("names", "row.names", "class"))
    expect_equal(names(attributes(dtaDbg$dtaFrm[[3]])),  c("levels", "class", "description", "measureType", "dataType"))
    expect_equal(names(attributes(dtaDbg$dtaFrm[[7]])),  c("jmv-desc", "measureType", "dataType"))
    expect_equal(names(attributes(dtaDbg$dtaFrm[[10]])), c("class", "levels", "measureType", "dataType"))
    expect_equal(names(attributes(dtaDbg$dtaFrm[[11]])), c("class", "levels", "measureType", "dataType"))
    expect_equal(names(attributes(dtaDbg$dtaFrm[[12]])), c("levels", "class", "measureType", "dataType"))
    expect_equal(names(attributes(dtaDbg$dtaFrm[[13]])), c("levels", "class", "measureType", "dataType"))
    expect_equal(attributes(dtaDbg$dtaFrm[[4]]),  list(measureType = "Continuous", dataType = "Decimal"))
    expect_equal(attributes(dtaDbg$dtaFrm[[8]]),  list(measureType = "Continuous", dataType = "Integer"))
    expect_equal(attributes(dtaDbg$dtaFrm[[9]]),  list(measureType = "Continuous", dataType = "Decimal"))
    expect_equal(attributes(dtaDbg$dtaFrm[[10]]), list(class =              "factor",  levels = character(0), measureType = "Nominal",    dataType = "Integer"))
    expect_equal(attributes(dtaDbg$dtaFrm[[11]]), list(class = c("ordered", "factor"), levels = character(0), measureType = "Ordinal",    dataType = "Integer"))
    expect_equal(attributes(dtaDbg$dtaFrm[[12]]), list(levels = character(0), class = "factor",               measureType = "Nominal",    dataType = "Text"))
    expect_equal(attributes(dtaDbg$dtaFrm[[13]]), list(levels = character(0), class = "factor",               measureType = "Nominal",    dataType = "Integer"))
    expect_equal(c(lapply(jmvReadWrite::ToothGrowth, class), list(T1 = "integer", T2 = "numeric", T3 = "factor", T4 = c("ordered", "factor"), T5 = "factor", T6 = "factor")),
                 lapply(dtaDbg$dtaFrm, class))

    nmeTmp <- tempfile(fileext = ".omt")
    write_omv(dtaFrm = dtaOut, fleOut = nmeTmp)
    expect_true(file.exists(nmeTmp))
    expect_gt(file.info(nmeTmp)$size, 1)
    expect_true(chkFle(nmeTmp, isZIP = TRUE))
    expect_true(chkFle(nmeTmp, fleCnt = "meta"))
    expect_true(chkFle(nmeTmp, fleCnt = "metadata.json"))
    expect_true(chkFle(nmeTmp, fleCnt = "data.bin"))
    dtaTmp <- read_omv(nmeTmp)
    expect_s3_class(dtaTmp, "data.frame")
    expect_equal(dim(dtaTmp), c(0, 7))
    expect_equal(names(attributes(dtaTmp)), c("names", "row.names", "removedRows", "addedRows", "transforms", "class"))
    expect_equal(names(attributes(dtaTmp[[3]]))[1:10],
      c("levels", "class", "values", "jmv-desc", "name", "id", "columnType", "dataType", "measureType", "formula"))
    expect_equal(names(attributes(dtaTmp[[7]]))[1:10],
      c("jmv-desc", "name", "id", "columnType", "dataType", "measureType", "formula", "formulaMessage", "parentId", "width"))
    expect_equal(unname(unlist(attributes(dtaTmp[[1]]))),
      c("TRUE", "ID", "1", "Data", "Text", "ID", "", "", "0", "100", "string", "", "", "0"))
    expect_equal(unname(unlist(attributes(dtaTmp[[4]]))),
      c("dose", "4", "Data", "Decimal", "Continuous", "", "", "0", "100", "number", "", "", "0"))
    expect_equal(lapply(jmvReadWrite::ToothGrowth, class), lapply(dtaTmp, class))
    unlink(nmeTmp)

    nmeTmp <- tempfile(fileext = ".omt")
    # NB: edits (in addCol) and addedRows should be reset (empty object of the same type, i.e., list())
    addCol <- setAtt(c("columnType", "dataType", "measureType", "formula", "edits"),
                     list(columnType = "Computed", dataType = "Decimal", measureType = "Continuous", formula = "LOG10(len)", edits = list(start = 0, end = 499)),
                     setNames(dtaOut[, 7, FALSE], "logLenC"))
    write_omv(dtaFrm = setAtt("addedRows", list(addedRows = list(start = 0, end = 499)), cbind(dtaOut, addCol)), fleOut = nmeTmp)
    expect_true(file.exists(nmeTmp))
    expect_gt(file.info(nmeTmp)$size, 1)
    expect_true(chkFle(nmeTmp, isZIP = TRUE))
    expect_true(chkFle(nmeTmp, fleCnt = "meta"))
    expect_true(chkFle(nmeTmp, fleCnt = "metadata.json"))
    expect_true(chkFle(nmeTmp, fleCnt = "data.bin"))
    dtaTmp <- read_omv(nmeTmp)
    expect_s3_class(dtaTmp, "data.frame")
    expect_equal(dim(dtaTmp), c(0, 8))
    expect_equal(attributes(dtaTmp), list(names = c("ID", "supp", "supp2", "dose", "dose2", "len", "logLen", "logLenC"), row.names = integer(0),
                                          removedRows = list(), addedRows = list(), transforms = list(), class = "data.frame"))
    expect_equal(names(attributes(dtaTmp[[3]]))[1:10],
      c("levels", "class", "values", "jmv-desc", "name", "id", "columnType", "dataType", "measureType", "formula"))
    expect_equal(names(attributes(dtaTmp[[7]]))[1:10],
      c("jmv-desc", "name", "id", "columnType", "dataType", "measureType", "formula", "formulaMessage", "parentId", "width"))
    expect_equal(names(attributes(dtaTmp[[8]]))[1:10],
      c("jmv-desc", "name", "id", "columnType", "dataType", "measureType", "formula", "formulaMessage", "parentId", "width"))
    expect_equal(unname(unlist(attributes(dtaTmp[[1]]))),
      c("TRUE", "ID", "1", "Data", "Text", "ID", "", "", "0", "100", "string", "", "", "0"))
    expect_equal(unname(unlist(attributes(dtaTmp[[4]]))),
      c("dose", "4", "Data", "Decimal", "Continuous", "", "", "0", "100", "number", "", "", "0"))
    expect_equal(unname(unlist(attributes(dtaTmp[[7]]))),
      c("Natural logarithm of the tooth length (len)", "logLen", "7", "Data", "Decimal", "Continuous",
        "",           "", "0", "100", "number", "", "Natural logarithm of the tooth length (len)", "0"))
    expect_equal(unname(unlist(attributes(dtaTmp[[8]]))),
      c("Natural logarithm of the tooth length (len)", "logLenC", "8", "Computed", "Decimal", "Continuous",
        "LOG10(len)", "", "0", "100", "number", "", "Natural logarithm of the tooth length (len)", "0"))
    expect_equal(c(lapply(jmvReadWrite::ToothGrowth, class), list(logLenC = "numeric")), lapply(dtaTmp, class))
    unlink(nmeTmp)

    # test cases for code coverage ============================================================================================================================
    expect_error(write_omv(NULL, nmeOut), regexp = "^The data frame to be written needs to be given as parameter \\(dtaFrm = \\.\\.\\.\\)\\.")
    expect_error(write_omv(data.frame(T1 = sample(9999, 100), T2 = as.complex(rnorm(100))), nmeOut),
      regexp = "Variable type \\w+ not implemented\\. Please send the data file that caused this problem to sebastian\\.jentschke@uib\\.no")
    expect_error(write_omv(dtaDbg$dtaFrm), regexp = "^Output file name needs to be given as parameter \\(fleOut = \\.\\.\\.\\)\\.")
    expect_error(add2ZIP(fleZIP = nmeOut, crrHdl = NULL),
      regexp = paste0("fleZIP \\(a character with a file name\\), and either crrHdl \\(with a connection\\) or crrFle \\(with a file name",
                      " and \\[optionally\\] a writing mode\\) need to be given as arguments."))
    expect_error(add2ZIP(fleZIP = nmeOut, crrHdl = 1),
      regexp = "^Parameter isn't a file handle pointing to a file to be zipped\\:")

    attr(dtaDbg$dtaFrm, "label.table") <- c("A", "B", "C")
    expect_error(write_omv(dtaDbg$dtaFrm, nmeOut),
      regexp = "^R-foreign-style value labels need to be implemented\\. Please send the data file that caused this problem to sebastian\\.jentschke@uib\\.no")
    attr(dtaDbg$dtaFrm, "label.table") <- NULL

    attr(dtaDbg$dtaFrm, "variable.labels") <- stats::setNames(c("Label for ID", "Label for supp", "Label for supp2"), c("ID", "supp", "supp2"))
    dtaDbg$dtaFrm$supp <- as.character(dtaDbg$dtaFrm$supp)
    expect_equal(vapply(c(1, 3), function(n) write_omv(dtaDbg$dtaFrm, nmeOut, frcWrt = TRUE, retDbg = TRUE)[["mtaDta"]][["fields"]][[n]][["description"]], character(1)),
      c("Label for ID", "Label for supp2"))
    unlink(nmeOut)
    expect_identical(vapply(c("dataType", "type"), function(f) write_omv(dtaDbg$dtaFrm, nmeOut, frcWrt = TRUE, retDbg = TRUE)[["mtaDta"]][["fields"]][[2]][[f]], character(1), USE.NAMES = FALSE),
      c("Text", "integer"))
    unlink(nmeOut)

    expect_error(write_omv(data.frame(T1 = sample(9999, 100), T2 = as.complex(rnorm(100))), fleOut = tempfile(fileext = ".omv")),
      regexp = "Variable type \\w+ not implemented\\. Please send the data file that caused this problem to sebastian\\.jentschke@uib\\.no")
    expect_equal(write_omv(data.frame(ID = as.factor(seq(100)),  T1 = sample(9999, 100), T2 = rnorm(100)), nmeOut, retDbg = TRUE)$mtaDta$field[[1]]$type, "string")
    unlink(nmeOut)

    tmpDF <- read_omv(file.path("..", "ToothGrowth.omv"))
    attr(tmpDF[[4]], "values") <- as.integer(c(0, 1))
    expect_error(write_omv(dtaFrm = tmpDF, fleOut = nmeOut),
      regexp = "\"values\"-attribute with unexpected values found for column \"supp - Transform 1\"\\. Please send the file to sebastian\\.jentschke@uib\\.no for debugging\\.")
    attr(tmpDF[[4]], "values") <- as.integer(c(1, 2))
    expect_null(write_omv(tmpDF, nmeOut))
    unlink(nmeOut)

    set.seed(1)
    dtaOut <- cbind(dtaOut, data.frame(Bool  =            sample(c(TRUE, FALSE), rowOut, TRUE),
                                       Date1 =            sample(seq(as.Date("1999/01/01"), as.Date("2000/01/01"), by = "day"), rowOut),
                                       Date2 = as.POSIXct(sample(seq(as.Date("1999/01/01"), as.Date("2000/01/01"), by = "day"), rowOut)),
                                       Date3 = as.POSIXlt(sample(seq(as.Date("1999/01/01"), as.Date("2000/01/01"), by = "day"), rowOut)),
                                       Time  =            sample(as.difftime(tim = seq(0, 3600), units = "secs"), rowOut)))
    write_omv(dtaFrm = dtaOut, fleOut = nmeOut)
    dtaInp <- read_omv(fleInp = nmeOut)
    unlink(nmeOut)
    expect_equal(as.integer(table(dtaInp[["Bool"]])), c(29, 31))
    expect_equal(c(mean(dtaInp[["Date1"]]), sd(dtaInp[["Date1"]])), c(10787.3667, 108.7002), tolerance = 1e-4)
    expect_equal(attr(dtaInp[["Date1"]], "jmv-desc"), "Date1 (date converted to integer; days since 1970-01-01)")
    expect_equal(c(mean(dtaInp[["Date2"]]), sd(dtaInp[["Date2"]])), c(10767.2330, 108.2620), tolerance = 1e-4)
    expect_equal(attr(dtaInp[["Date2"]], "jmv-desc"), "Date2 (date converted to integer; days since 1970-01-01)")
    expect_equal(c(mean(dtaInp[["Date3"]]), sd(dtaInp[["Date3"]])), c(10762.8000, 112.2584), tolerance = 1e-4)
    expect_equal(attr(dtaInp[["Date3"]], "jmv-desc"), "Date3 (date converted to integer; days since 1970-01-01)")
    expect_equal(c(mean(dtaInp[["Time"]]),  sd(dtaInp[["Time"]])),   c(2090.8000, 978.9000), tolerance = 1e-4)
    expect_equal(attr(dtaInp[["Time"]], "jmv-desc"), "Time (time converted to integer; sec since 00:00)")
    dtaOut <- read_omv(file.path("..", "ToothGrowth.omv"))
    attr(dtaOut, "jmv-weights-name") <- "weights"
    attr(dtaOut, "jmv-weights") <- as.vector(dtaOut[, "weights"])
    expect_warning(dtaDbg <- write_omv(dtaFrm = dtaOut, fleOut = nmeOut, retDbg = TRUE), regexp = "Handling of weights not yet implemented\\.")
    expect_equal(dtaDbg$mtaDta$weights, "weights")
    # this actually tests read_omv
    dtaInp <- read_omv(nmeOut)
    unlink(nmeOut)
    expect_equal(names(attributes(dtaInp)), c("names", "row.names", "class", "fltLst", "removedRows", "addedRows", "transforms", "jmv-weights-name", "jmv-weights"))
    expect_equal(attr(dtaInp, "jmv-weights-name"), "weights")
    expect_equal(attr(dtaInp, "jmv-weights"), rep(1, 60))

    dtaOut <- jmvReadWrite::ToothGrowth[, 1, drop = FALSE]
    write_omv(dtaFrm = dtaOut, fleOut = nmeOut)
    expect_true(chkFle(nmeOut, isZIP = TRUE))
    expect_true(chkFle(nmeOut, fleCnt = "strings.bin"))
    unlink(nmeOut)
    dtaOut[c(5, 7, 10), 1] <- as.character(NA)
    write_omv(dtaFrm = dtaOut, fleOut = nmeOut)
    expect_true(chkFle(nmeOut, isZIP = TRUE))
    expect_true(chkFle(nmeOut, fleCnt = "strings.bin"))
    unlink(nmeOut)
    dtaOut[, 1] <- as.character(NA)
    attr(dtaOut[, 1], "jmv-id") <- TRUE
    write_omv(dtaFrm = dtaOut, fleOut = nmeOut)
    expect_true(chkFle(nmeOut, isZIP = TRUE))
    expect_error(chkFle(nmeOut, fleCnt = "strings.bin"), regexp = "chkFle: File \".*?\" doesn't contain the file \"strings.bin\".")
    unlink(nmeOut)

    Sys.setenv(JAMOVI_R_VERSION = paste0(R.version$major, ".", R.version$minor))
    df4Chk <- write_omv(dtaFrm = jmvReadWrite::ToothGrowth, fleOut = nmeOut, retDbg = TRUE)$dtaFrm
    Sys.unsetenv("JAMOVI_R_VERSION")
    expect_true(all(vapply(df4Chk, function(x) identical(c("measureType", "dataType") %in% names(attributes(x)), c(TRUE, TRUE)), logical(1))))
    expect_equal(vapply(df4Chk, function(x) attr(x, "dataType"),    character(1), USE.NAMES = FALSE), c("Text", "Text", "Integer", "Decimal", "Text", "Decimal", "Decimal"))
    expect_equal(vapply(df4Chk, function(x) attr(x, "measureType"), character(1), USE.NAMES = FALSE), c("ID", "Nominal", "Nominal", "Continuous", "Ordinal", "Continuous", "Continuous"))
    # do not unlink to provoke the error underneath
    expect_error(write_omv(dtaFrm = jmvReadWrite::ToothGrowth, fleOut = nmeOut),
      regexp = "The output file .* already exists\\. Either remove the file or set the parameter frcWrt to TRUE\\.")
    unlink(nmeOut)

    dtaOut <- jmvReadWrite::ToothGrowth
    expect_error(write_omv(dtaFrm = dtaOut, fleOut = nmeOut, wrtPtB = TRUE),
      regexp = "The data frame \\(dtaFrm\\) must contain the attribute \"protobuf\", there has to be at least one of them, and it has to be of the correct type \\(a RProtoBuf\\)\\.")
    attr(dtaOut, "protobuf")[["01 empty/analysis"]] <- TRUE
    expect_error(write_omv(dtaFrm = dtaOut, fleOut = nmeOut, wrtPtB = TRUE),
      regexp = "The data frame \\(dtaFrm\\) must contain the attribute \"protobuf\", there has to be at least one of them, and it has to be of the correct type \\(a RProtoBuf\\)\\.")
    attr(dtaOut, "protobuf")[["01 empty/analysis"]] <- RProtoBuf::new(jamovi.coms.AnalysisRequest)
    write_omv(dtaFrm = dtaOut, fleOut = nmeOut, wrtPtB = TRUE)
    expect_true(chkFle(nmeOut, isZIP = TRUE))
    expect_true(chkFle(nmeOut, fleCnt = "01 empty/analysis"))
    unlink(nmeOut)

    tmpDF <- data.frame(ID = sprintf("P_%04d", sample(9999, 100)), I = as.integer(sample(1e6, 100)), D = rnorm(100),
                        OT = factor(sample(c("low", "middle", "high"), 100, replace = TRUE), levels = c("low", "middle", "high"), ordered = TRUE),
                        ON = factor(sample(seq(7), 100, replace = TRUE), levels = seq(7), ordered = TRUE),
                        NT = factor(sample(c("low", "middle", "high"), 100, replace = TRUE), levels = c("low", "middle", "high")),
                        NN = factor(sample(seq(7), 100, replace = TRUE), levels = seq(7)),
                        CR = sample(c("low", "middle", "high"), 100, replace = TRUE))
    attr(tmpDF[["ID"]], "jmv-id")   <- TRUE
    attr(tmpDF[["ON"]], "values")   <- seq(7)
    attr(tmpDF[["NN"]], "values")   <- seq(7)
    attr(tmpDF[["CR"]], "jmv-desc") <- "Trial (is description kept?)"
    expect_equal(lapply(lapply(jmvAtt(tmpDF), attributes), names), list(ID = c("jmv-id", "measureType", "dataType"),
        I = c("measureType", "dataType"), D = c("measureType", "dataType"),
        OT = c("levels", "class", "measureType", "dataType"), ON = c("levels", "class", "values", "measureType", "dataType"),
        NT = c("levels", "class", "measureType", "dataType"), NN = c("levels", "class", "values", "measureType", "dataType"),
        CR = c("jmv-desc", "measureType", "dataType")))
    expect_equal(unlist(attributes(jmvAtt(tmpDF)[["ID"]]), use.names = FALSE), c("TRUE", "ID", "Text"))
    expect_equal(unlist(attributes(jmvAtt(tmpDF)[["I"]]),  use.names = FALSE), c("Continuous", "Integer"))
    expect_equal(unlist(attributes(jmvAtt(tmpDF)[["D"]]),  use.names = FALSE), c("Continuous", "Decimal"))
    expect_equal(unlist(attributes(jmvAtt(tmpDF)[["OT"]]), use.names = FALSE), c("low", "middle", "high", "ordered", "factor", "Ordinal", "Text"))
    expect_equal(unlist(attributes(jmvAtt(tmpDF)[["ON"]]), use.names = FALSE), c(sprintf("%d", seq(1:7)), "ordered", "factor", sprintf("%d", seq(1:7)), "Ordinal", "Integer"))
    expect_equal(unlist(attributes(jmvAtt(tmpDF)[["NT"]]), use.names = FALSE), c("low", "middle", "high", "factor", "Nominal", "Text"))
    expect_equal(unlist(attributes(jmvAtt(tmpDF)[["NN"]]), use.names = FALSE), c(sprintf("%d", seq(1:7)), "factor", sprintf("%d", seq(1:7)), "Nominal", "Integer"))
    expect_equal(unlist(attributes(jmvAtt(tmpDF)[["CR"]]), use.names = FALSE), c("Trial (is description kept?)", "Nominal", "Text"))
    expect_error(jmvAtt("Trial"),      regexp = "^Input data are either not a data frame or have incorrect \\(only one or more than two\\) dimensions\\.")
    expect_error(jmvAtt(data.frame()), regexp = "^The second dimension of the input data frame has not the required size \\(0 < 1\\)\\.")
    expect_error(jmvAtt(cbind(tmpDF, data.frame(EC = sample(as.complex(seq(10)), 100, replace = TRUE)))), regexp = "^\\s+\\w+: Variable type \\w+ not implemented.")

    tmpDF  <- structure(list(value = structure(c(1, 2, 4), format.sas = "LEVELS", class = c("haven_labelled", "vctrs_vctr", "double"),
                             labels = c(level1 = 1, level2 = 2, level3 = 4))), class = c("tbl_df", "tbl", "data.frame"), row.names = c(NA, -3L))
    df4Chk <- write_omv(dtaFrm = tmpDF, fleOut = nmeOut, retDbg = TRUE)$dtaFrm
    unlink(nmeOut)
    expect_equal(attributes(df4Chk), list(names = "value", row.names = seq(3), class = "data.frame"))
    expect_equal(attributes(df4Chk[, "value"]),
        list(levels = c("level1", "level2", "level3"), class = "factor", measureType = "Nominal", dataType = "Text"))

    tmpCol <- rnorm(100)
    attr(tmpCol, "Trial") <- "Check whether attribute is preserved"
    expect_equal(cnvCol(tmpCol, tgtTyp = "integer"), round(tmpCol))
    expect_equal(attr(cnvCol(tmpCol, tgtTyp = "integer"), "Trial"), "Check whether attribute is preserved")
})

#' Sort data (using one or more variables) in .omv-files for the statistical spreadsheet 'jamovi' (<https://www.jamovi.org>)
#'
#' @param dtaInp Either a data frame or the name of a data file to be read (including the path, if required; "FILENAME.ext"; default: NULL); files can be of
#'               any supported file type, see Details below
#' @param fleOut Name of the data file to be written (including the path, if required; "FILE_OUT.omv"; default: ""); if empty, the resulting data frame is
#'               returned instead
#' @param varSrt Variable(s) that are used to sort the data frame (see Details; default: c())
#' @param psvAnl Whether analyses that are contained in the input file shall be transferred to the output file (TRUE / FALSE; default: FALSE)
#' @param usePkg Name of the package: "foreign" or "haven" that shall be used to read SPSS, Stata and SAS files; "foreign" is the default (it comes with
#'               base R), but "haven" is newer and more comprehensive
#' @param selSet Name of the data set that is to be selected from the workspace (only applies when reading .RData-files)
#' @param ...    Additional arguments passed on to methods; see Details below
#'
#' @return a data frame (only returned if `fleOut` is empty) where the input data set is sorted (according to the variables in `varSrt`)
#'
#' @details
#' * `varSrt` can be either a character or a character vector (with one or more variables respectively). The sorting order for a particular variable can be
#'   inverted with preceding the variable name with "-". Please note that this doesn't make sense and hence throws a warning for certain variable types (e.g.,
#'   factors).
#' * The ellipsis-parameter (`...`) can be used to submit arguments / parameters to the functions that are used for reading and writing the data. By clicking
#'   on the respective function under “See also”, you can get a more detailed overview over which parameters each of those functions take. The functions are:
#'   `read_omv` and `write_omv` (for jamovi-files), `read.table` (for CSV / TSV files; using similar defaults as `read.csv` for CSV and `read.delim` for TSV
#'   which both are based upon `read.table`), `load` (for .RData-files), `readRDS` (for .rds-files), `read_sav` (needs the R-package `haven`) or `read.spss`
#'   (needs the R-package `foreign`) for SPSS-files, `read_dta` (`haven`) / `read.dta` (`foreign`) for Stata-files, `read_sas` (`haven`) for SAS-data-files,
#'   and `read_xpt` (`haven`) / `read.xport` (`foreign`) for SAS-transport-files. If you would like to use `haven`, you may need to install it using
#'   `install.packages("haven", dep = TRUE)`.
#'
#' @seealso `sort_omv` internally uses the following functions for reading and writing data files in different formats: [jmvReadWrite::read_omv()] and
#'   [jmvReadWrite::write_omv()] for jamovi-files, [utils::read.table()] for CSV / TSV files, [load()] for reading .RData-files, [readRDS()] for .rds-files,
#'   [haven::read_sav()] or [foreign::read.spss()] for SPSS-files, [haven::read_dta()] or [foreign::read.dta()] for Stata-files, [haven::read_sas()] for
#'   SAS-data-files, and [haven::read_xpt()] or [foreign::read.xport()] for SAS-transport-files.
#'
#' @examples
#' nmeInp <- system.file("extdata", "AlbumSales.omv", package = "jmvReadWrite")
#' nmeOut <- tempfile(fileext = ".omv")
#' jmvReadWrite::sort_omv(dtaInp = nmeInp, fleOut = nmeOut, varSrt = "Image")
#' dtaFrm <- jmvReadWrite::read_omv(nmeOut)
#' unlink(nmeOut)
#' cat(dtaFrm$Image)
#' # shows that the variable "Image" is sorted in ascending order
#' cat(is.unsorted(dtaFrm$Image))
#' # is.unsorted (which checks for whether the variable is NOT sorted) returns FALSE
#' jmvReadWrite::sort_omv(dtaInp = nmeInp, fleOut = nmeOut, varSrt = "-Image")
#' # variables can also be sorted in descending order by preceding them with "-"
#' dtaFrm <- jmvReadWrite::read_omv(nmeOut)
#' unlink(nmeOut)
#' cat(dtaFrm$Image)
#' # shows that the variable "Image" is now sorted in descending order
#' cat(is.unsorted(dtaFrm$Image))
#' # this first returns TRUE (the variable is not in ascending order, i.e., unsorted)
#' cat(is.unsorted(-dtaFrm$Image))
#' # if the sign of the variable is changed, it returns FALSE (i.e., the variable is
#' # NOT unsorted)
#'
#'
#' @export sort_omv
#'
sort_omv <- function(dtaInp = NULL, fleOut = "", varSrt = c(), psvAnl = FALSE, usePkg = c("foreign", "haven"), selSet = "", ...) {
    if (length(varSrt) == 0 || !all(nzchar(varSrt))) {
        stop("Calling sort_omv requires giving at least one variable to sort after.")
    }

    # check and import input data set (either as data frame or from a file)
    if (!is.null(list(...)[["fleInp"]])) stop("Please use the argument dtaInp instead of fleInp.")
    dtaFrm <- inp2DF(dtaInp = dtaInp, usePkg = usePkg, selSet = selSet, ...)

    # sort data set
    dtaFrm <- srtFrm(dtaFrm, varSrt)

    # rtnDta in globals.R (unified function to either write the data frame, open it in a new jamovi session or return it)
    rtnDta(dtaFrm = dtaFrm, fleOut = fleOut, dtaTtl = jmvTtl("_sort"), psvAnl = psvAnl, dtaInp = dtaInp, ...)
}

srtFrm <- function(dtaFrm = NULL, varSrt = c()) {
    # if the sorting variable(s) are found, generate an order according to them and afterwards remove / reset the rownames
    if (chkVar(dtaFrm, gsub("^-", "", varSrt))) {
#       srtOrd <- eval(parse(text = paste0("order(", paste0(gsub("dtaFrm[[\"-", "-dtaFrm[[\"", paste0("dtaFrm[[\"", varSrt, "\"]]"), fixed = TRUE), collapse = ", "), ")")))
        srtOrd <- eval(parse(text = paste0("order(", paste0(vapply(varSrt, function(x) {
            s <- ifelse(grepl("^-", x), "-", "")
            ifelse(!any(is.na(suppressWarnings(as.numeric(dtaFrm[[x]])))),
                paste0(s, "as.numeric(dtaFrm[[\"", sub("^-", "", x), "\"]])"),
                paste0(s, "dtaFrm[[\"", sub("^-", "", x), "\"]]"))
            }, character(1)), collapse = ", "), ")")))
        # sorting makes the data.frame lose it's attributes which are therefore stored and later restored
        attLst <- bckAtt(dtaFrm)
        dtaFrm <- dtaFrm[srtOrd, , drop = FALSE]
        rownames(dtaFrm) <- NULL
        dtaFrm <- rstAtt(dtaFrm, attLst)
    }

    dtaFrm
}

#' Merges two or more data files by adding the content of other input files as columns to the first
#' input file and outputs them as files for the statistical spreadsheet 'jamovi'
#' (<https://www.jamovi.org>)
#'
#' @param dtaInp Either a data frame (with the attribute "fleInp" containing the files to merge) or
#'               vector with the names of the input files (including the path, if required;
#'               "FILENAME.ext"; default: NULL); files can be of any supported file type, see
#'               Details below
#' @param fleOut Name of the data file to be written (including the path, if required;
#'               "FILE_OUT.omv"; default: ""); if empty, the resulting data frame is returned
#'               instead
#' @param typMrg Type of merging operation: "outer" (default), "inner", "left" or "right"; see
#'               Details below
#' @param varBy  Name of the variable by which the data sets are matched, can either be a string,
#'               a character or a list (see Details below; default: list())
#' @param varSrt Variable(s) that are used to sort the data frame (see Details; if empty, the order
#'               after merging is kept; default: c())
#' @param psvAnl Whether analyses that are contained in the input file shall be transferred to the
#'               output file (TRUE / FALSE; default: FALSE)
#' @param usePkg Name of the package: "foreign" or "haven" that shall be used to read SPSS, Stata
#'               and SAS files; "foreign" is the default (it comes with base R), but "haven" is
#'               newer and more comprehensive
#' @param selSet Name of the data set that is to be selected from the workspace (only applies when
#'               reading .RData-files)
#' @param ...    Additional arguments passed on to methods; see Details below
#'
#' @return a data frame (only returned if `fleOut` is empty) where the columns of all input data
#'         sets (given in the `dtaInp`-argument) are concatenated
#'
#' @details
#' * Using data frames with the input parameter `dtaInp` is primarily thought to be used when
#'   calling `merge_cols_omv` from the jamovi-modules `jTransform` and `Rj`. For the use in R, it
#'   is strongly recommended to use a character vector with the file names instead.
#' * There are four different types of merging operations (defined via `typMrg`): "outer" keeps all
#'   cases (but columns in the resulting data set may contain empty cells / missing values if same
#'   input data sets did not have a row containing the matching variable (defined in `varBy`).
#'   "inner" keeps only those cases where all datasets contain the same value in the matching
#'   variable, for "left" all cases from the first data set in `dtaInp` are kept (whereas cases
#'   that are only contained in the second or any later input data set are dropped), for "right"
#'   all cases from the second (or any higher) data set in `dtaInp` are kept. The behaviour of
#'   "left" and "right" may be somewhat difficult to predict in case of merging several data sets,
#'   therefore "outer" might be a safer choice if several data sets are merged.
#' * The variable that is used for matching (`varBy`) can either be a string (if all datasets
#'   contain a matching variable with the same name), a character vector (containing more than one
#'   matching variables that are contained in / the same for all data sets) or a list with the same
#'   length as dtaInp. In such list, each cell can again contain either a string (one matching
#'   variable for each data set in dtaInp) or a character vector (several matching variables for
#'   each data set in dtaInp; NB: all character vectors in the cells of the list must have the same
#'   length as it is necessary to always use the same number of matching variables when merging).
#' * `varSrt` can be either a character or a character vector (with one or more variables
#'   respectively). The sorting order for a particular variable can be inverted with preceding the
#'   variable name with "-". Please note that this doesn't make sense and hence throws a warning
#'   for certain variable types (e.g., factors).
#' * The ellipsis-parameter (`...`) can be used to submit arguments / parameters to the functions
#'   that are used for transforming or reading the data. By clicking on the respective function
#'   under “See also”, you can get a more detailed overview over which parameters each of those
#'   functions take.
#' * Adding columns uses `merge`. `typMrg` is implemented by setting `TRUE` or `FALSE` to `all.x`
#'   and `all.y` in `merge`, `varBy` matches `by.x` and `by.y`. The help for `merge` can be
#'   accessed by clicking on the link under “See also”.
#' * The functions for reading and writing the data are: `read_omv` and `write_omv` (for
#'   jamovi-files), `read.table` (for CSV / TSV files; using similar defaults as `read.csv` for CSV
#'   and `read.delim` for TSV which both are based upon `read.table`), `load` (for .RData-files),
#'   `readRDS` (for .rds-files), `read_sav` (needs R-package `haven`) or `read.spss` (needs
#'   R-package `foreign`) for SPSS-files, `read_dta` (`haven`) / `read.dta` (`foreign`) for
#'   Stata-files, `read_sas` (`haven`) for SAS-data-files, and `read_xpt` (`haven`) / `read.xport`
#'   (`foreign`) for SAS-transport-files. If you would like to use `haven`, you may need to install
#'   it using `install.packages("haven", dep = TRUE)`.
#'
#' @seealso `merge_cols_omv` internally uses the following functions: Adding columns uses
#'   [merge()]. For reading and writing data files in different formats: [jmvReadWrite::read_omv()]
#'   and [jmvReadWrite::write_omv()] for jamovi-files, [utils::read.table()] for CSV / TSV files,
#'   [load()] for reading .RData-files, [readRDS()] for .rds-files, [haven::read_sav()] or
#'   [foreign::read.spss()] for SPSS-files, [haven::read_dta()] or [foreign::read.dta()] for
#'   Stata-files, [haven::read_sas()] for SAS-data-files, and [haven::read_xpt()] or
#'   [foreign::read.xport()] for SAS-transport-files.
#'
#' @examples
#' dtaInp <- jmvReadWrite::bfi_sample2
#' nmeInp <- paste0(tempfile(), "_", 1:3, ".rds")
#' nmeOut <- tempfile(fileext = ".omv")
#' for (i in seq_along(nmeInp)) {
#'     saveRDS(stats::setNames(dtaInp, c("ID", paste0(names(dtaInp)[-1], "_", i))), nmeInp[i])
#' }
#' # save dtaInp three times (i.e., the length of nmeInp), adding "_" + 1 ... 3 as index
#' # to the data variables (A1 ... O5, gender, age → A1_1, ...)
#' jmvReadWrite::merge_cols_omv(dtaInp = nmeInp, fleOut = nmeOut, varBy = "ID")
#' cat(file.info(nmeOut)$size)
#' # -> 17731 (size may differ on different OSes)
#' dtaOut <- jmvReadWrite::read_omv(nmeOut, sveAtt = FALSE)
#' # read the data set where the three original datasets were added as columns and show
#' # the variable names
#' cat(names(dtaOut))
#' cat(names(dtaInp))
#' # compared to the input data set, we have the same names (expect for "ID" which was
#' # used for matching and that each variable had added an indicator from which data
#' # set they came)
#' cat(dim(dtaInp), dim(dtaOut))
#' # the first dimension of the data sets (rows) stayed the same (250), whereas the
#' # second dimension is now approx. three times as large (28 -> 82):
#' # 28 - 1 (for "ID") = 27 * 3 + 1 (for "ID") = 82
#' cat(colMeans(dtaInp[2:11]))
#' cat(colMeans(dtaOut[2:11]))
#' # it's therefore not much surprise that the values of the column means for the first
#' # 10 variables of dtaInp and dtaOut are the same too
#'
#' unlink(nmeInp)
#' unlink(nmeOut)
#'
#'
#' @export merge_cols_omv
#'
merge_cols_omv <- function(dtaInp = NULL, fleOut = "", typMrg = c("outer", "inner", "left", "right"), varBy = list(),
                           varSrt = c(), psvAnl = FALSE, usePkg = c("foreign", "haven"), selSet = "", ...) {

    # check and import input data set (either as data frame or from a file)
    if (!is.null(list(...)[["fleInp"]])) stop("Please use the argument dtaInp instead of fleInp.")
    dtaFrm <- inp2DF(dtaInp = dtaInp, minDF = 2, maxDF = Inf, rmvEmp = TRUE, usePkg = usePkg, selSet = selSet, ...)

    # store attributes from the data sets
    attLst <- bckAtt(dtaFrm, bckCol = TRUE)

    # check the matching variable(s)
    varBy <- chkByV(varBy, dtaFrm)

    # merge files ([1] determine arguments, [2] merge using a temporary variable which afterwards is written back)
    typMrg <- match.arg(typMrg)
    crrArg <- list(x = NULL, y = NULL, by.x = "", by.y = "", all.x = ifelse(any(typMrg %in% c("outer", "left")), TRUE, FALSE),
                   all.y = ifelse(any(typMrg %in% c("outer", "right")), TRUE, FALSE))
    crrArg <- adjArg(c("merge", "data.frame"), crrArg, list(...), c("x", "y", "by.x", "by.y", "all.x", "all.y"))
    tmpMrg <- dtaFrm[[1]]
    for (i in setdiff(seq_along(dtaFrm), 1)) {
        tmpMrg <- do.call(merge, c(list(x = tmpMrg, y = dtaFrm[[i]], by.x = varBy[[1]], by.y = varBy[[i]]),
                                   crrArg[!grepl("^x$|^y$|^by.x$|^by.y$", names(crrArg))]))
        # if there are duplicate columns (i.e., columns with the same name in two of the input data sets), unify them
        for (unfClm in setdiff(names(attLst[["dtaCol"]]), names(tmpMrg))) {
            dplClm <- grep(paste0(unfClm, "\\."), names(tmpMrg))
            if (length(dplClm) == 0) next
            names(tmpMrg)[dplClm[1]] <- unfClm
            rmvClm <- NULL
            for (i in seq(2, length(dplClm))) {
                if (identical(tmpMrg[, dplClm[1]], tmpMrg[, dplClm[i]])) {
                    rmvClm <- c(rmvClm, -dplClm[i])
                } else {
                    addClm <- sprintf("%s_%d", unfClm, sum(grepl(paste0(unfClm, "_"), names(tmpMrg))) + 2)
                    names(tmpMrg)[dplClm[i]] <- addClm
                    attLst[["dtaCol"]][[addClm]] <- attLst[["dtaCol"]][[unfClm]]
                }
            }
            if (length(rmvClm) > 0) tmpMrg <- tmpMrg[, rmvClm]
        }
    }
    dtaFrm <- tmpMrg

    # sort data frame (if varSrt not empty)
    dtaFrm <- srtFrm(dtaFrm, varSrt)

    # restore attributes
    dtaFrm <- rstAtt(dtaFrm, attLst)

    # rtnDta in globals.R (unified function to either write the data frame, open it in a new jamovi session or return it)
    rtnDta(dtaFrm = dtaFrm, fleOut = fleOut, dtaTtl = jmvTtl("_mrg_cols"), psvAnl = psvAnl, dtaInp = dtaInp, ...)
}

chkByV <- function(varBy = list(), dtaFrm = NULL) {
    # varBy is empty
    if (empByV(varBy)) {
        varBy <- rep(list(mtcVar(dtaFrm)), length(dtaFrm))
    # varBy is a character vector (without empty elements) or a string
    } else if ((is.vector(varBy) && !is.list(varBy) && length(varBy) >= 1 && all(nzchar(varBy))) || is.character(varBy)) {
        if (all(vapply(dtaFrm, function(x) all(varBy %in% names(x)), logical(length(varBy))))) {
            varBy <- rep(list(varBy), length(dtaFrm))
        } else {
            stop("Not all data sets given in dtaInp contain the variable(s) / column(s) that shall be used for matching.")
        }
    # varBy is a list with the same length as dtaFrm
    } else if (is.list(varBy) && length(varBy) == length(dtaFrm)) {
        if (!all(vapply(seq_along(dtaFrm), function(i) all(varBy[[i]] %in% names(dtaFrm[[i]])), logical(1)))) {
            stop("Not all data sets given in dtaInp contain the variable(s) / column(s) that shall be used for matching.")
        }
        # else: leave varBy unchanged
    } else {
        stop("varBy must be either a list (with the same length as dtaInp), a character vector, or a string.")
    }

    # check whether all entries in the ID variables data sets contain a value
    for (i in seq_along(dtaFrm)) {
        if (any(vapply(varBy[[i]],
                       function(c) {
                           any(is.na(dtaFrm[[i]][, c])) || any(is.null(dtaFrm[[i]][, c])) ||
                           (all(is.character(dtaFrm[[i]][, c])) && !all(nzchar(dtaFrm[[i]][, c])))
                       },
                       logical(length(varBy[[i]]))))) {
            stop(sprintf("Values in the ID variable can't be empty (empty values found in %s).",
                           ifelse(i == 1, "the original data set", sprintf("data set %d to be merged", i - 1))))
        }
    }

    varBy
}

empByV <- function(varBy = NULL) {
    (is.list(varBy)      && length(varBy) == 0) ||
    (is.vector(varBy)    && length(varBy) == 0) ||
    (is.character(varBy) && !nzchar(varBy)) ||
    is.null(varBy)
}

mtcVar <- function(dtaFrm = NULL) {
    varCmm <- names(dtaFrm[[1]])
    for (i in setdiff(seq_along(dtaFrm), 1)) {
        varCmm <- intersect(varCmm, names(dtaFrm[[i]]))
    }
    varCmm
}

test_that("merge_cols_omv works", {
    nmeOut <- tempfile(fileext = "_W.omv")
    nmeInp <- vector(mode = "character", length = 3)
    dtaTmp <- rmvAtt(jmvReadWrite::bfi_sample2)
    varTmp <- names(dtaTmp)[-1]
    for (i in seq_along(nmeInp)) {
        nmeInp[i] <- tempfile(fileext = ".rds")
        names(dtaTmp)[-1] <- paste0(varTmp, "_", i)
        saveRDS(dtaTmp[sample(seq_len(dim(dtaTmp)[1]), size = round(dim(dtaTmp)[1] * (0.97 + 0.01 * i))), ], nmeInp[i])
    }

    expect_null(merge_cols_omv(dtaInp = nmeInp, fleOut = nmeOut, typMrg = "outer", varBy = "ID", varSrt = c("gender_3", "age_3")))
    expect_true(file.exists(nmeOut))
    expect_gt(file.info(nmeOut)$size, 1)
    expect_true(chkFle(nmeOut, isZIP = TRUE))
    expect_true(chkFle(nmeOut, fleCnt = "meta"))
    expect_true(chkFle(nmeOut, fleCnt = "metadata.json"))
    expect_true(chkFle(nmeOut, fleCnt = "data.bin"))
    unlink(nmeOut)

    dtaFrm <- merge_cols_omv(dtaInp = nmeInp, typMrg = "outer", varBy = "ID", varSrt = c("gender_3", "age_3"))
    expect_s3_class(dtaFrm, "data.frame")
    expect_equal(dim(dtaFrm), c(250, 85))
    expect_equal(vapply(dtaFrm, typeof, character(1), USE.NAMES = FALSE), c("character", rep(c(rep("integer", 27), "character"), 3)))
    expect_equal(names(dtaFrm), c("ID", paste0(paste0(varTmp, "_"), sort(rep(1:3, length(varTmp))))))
    expect_equal(unname(colSums(is.na(dtaFrm[, paste0("age_", 1:3)]))), c(5, 2, 0))
    expect_equal(as.integer(table(dtaFrm[["gender_3"]])), c(172, 78))
    expect_equal(which(diff(as.integer(dtaFrm[["gender_3"]])) == 1), 172)
    expect_equal(length(which(diff(as.integer(dtaFrm[["age_3"]])) == 1)), 52)
    expect_equal(max(diff(which(diff(as.integer(dtaFrm[["age_3"]])) == 1))), 12)

    dtaFrm <- merge_cols_omv(dtaInp = nmeInp[-2], typMrg = "inner", varBy = "ID", varSrt = c("gender_3", "age_3"))
    expect_s3_class(dtaFrm, "data.frame")
    expect_equal(dim(dtaFrm), c(245, 57))
    unlink(nmeInp)

    nmeInp <- vector(mode = "character", length = 3)
    dtaTmp <- jmvReadWrite::bfi_sample2
    for (i in seq_along(nmeInp)) {
        nmeInp[i] <- tempfile(fileext = ".rds")
        saveRDS(dtaTmp, nmeInp[i])
    }
    dtaFrm <- merge_cols_omv(dtaInp = nmeInp, typMrg = "outer", varBy = "ID", varSrt = c("ID"))
    expect_s3_class(dtaFrm, "data.frame")
    expect_equal(dim(dtaFrm), c(250, 29))
    expect_true(all(dtaFrm == dtaTmp[order(dtaTmp[, "ID"]), ]))
    unlink(nmeInp)

    nmeInp <- vector(mode = "character", length = 5)
    dtaTmp <- jmvReadWrite::bfi_sample2
    for (i in seq_along(nmeInp)) {
        nmeInp[i] <- tempfile(fileext = ".rds")
        strAtt <- attributes(dtaTmp[, "age"])
        dtaTmp[, "age"] <- dtaTmp[sample(nrow(dtaTmp)), "age"]
        attributes(dtaTmp[, "age"]) <- strAtt
        saveRDS(dtaTmp, nmeInp[i])
    }
    dtaFrm <- merge_cols_omv(dtaInp = nmeInp, typMrg = "outer", varBy = "ID", varSrt = c("ID"))
    dplClm <- gsub("age_1", "age", paste0("age_", seq(5)))
    expect_s3_class(dtaFrm, "data.frame")
    expect_equal(dim(dtaFrm), c(250, 33))
    expect_true(all(dplClm %in% names(dtaFrm)))
    expect_true(all(apply(vapply(dtaFrm[, dplClm], sort, numeric(dim(dtaFrm)[1])), 1, diff) == 0))
    expect_true(all(diff(colMeans(dtaFrm[, dplClm])) == 0))
    expect_true(all(unlist(lapply(dtaFrm[, dplClm], attributes)) == "Age of the respondent (years)"))

    # test cases for code coverage ============================================================================================================================
    expect_error(merge_cols_omv(fleInp = nmeInp, typMrg = "outer", varBy = "ID"), regexp = "Please use the argument dtaInp instead of fleInp\\.")
    expect_warning(merge_cols_omv(dtaInp = nmeInp, typMrg = "outer", varBy = "ID", psvAnl = TRUE),
      regexp = "psvAnl is only possible if fleOut is a file name \\(analyses are not stored in data frames, only in the jamovi files\\)\\.")
    attr(dtaTmp, "fleInp") <- nmeInp
    expect_warning(merge_cols_omv(dtaInp = dtaTmp, fleOut = nmeOut, typMrg = "outer", varBy = "ID", psvAnl = TRUE),
      regexp = "psvAnl is only possible if dtaInp is a file name \\(analyses are not stored in data frames, only in the jamovi files\\)\\.")
    unlink(nmeOut)
    unlink(nmeInp)

    dtaFrm <- list(data.frame(ID = sample(10), A = runif(10)), data.frame(ID = sample(10), B = runif(10)), data.frame(ID = sample(10), C = runif(10)), data.frame(ID = sample(10), D = runif(10)))
    expect_equal(chkByV(list(), dtaFrm), rep(list("ID"), 4))
    expect_equal(chkByV(rep(list("ID"), 4), dtaFrm), rep(list("ID"), 4))
    expect_error(chkByV(rep(list("ID2"), 4), dtaFrm),
      regexp = "^Not all data sets given in dtaInp contain the variable\\(s\\) / column\\(s\\) that shall be used for matching\\.")
    expect_equal(chkByV("ID", dtaFrm), rep(list("ID"), 4))
    expect_error(chkByV("ID2", dtaFrm), regexp = "^Not all data sets given in dtaInp contain the variable\\(s\\) / column\\(s\\) that shall be used for matching\\.")
    expect_error(chkByV(rep(list("ID"), 3), dtaFrm),
      regexp = "^varBy must be either a list \\(with the same length as dtaInp\\), a character vector, or a string\\.")
    expect_error(chkByV(rep(list("ID"), 5), c(dtaFrm, list(data.frame(ID = c(sample(9), NA), E = runif(10))))),
      regexp = "^Values in the ID variable can't be empty \\(empty values found in data set 4 to be merged\\)\\.")
    expect_error(chkByV(rep(list("ID"), 5), c(list(data.frame(ID = c(sample(9), NA), Z = runif(10))), dtaFrm)),
      regexp = "^Values in the ID variable can't be empty \\(empty values found in the original data set\\)\\.")


    nmeInp <- tempfile(fileext = ".rds")
    saveRDS(data.frame(ID = seq(60), A = rnorm(60), B = rnorm(60)), nmeInp)
    expect_null(merge_cols_omv(c(file.path("..", "ToothGrowth.omv"), nmeInp), fleOut = nmeOut, typMrg = "outer", varBy = "ID", psvAnl = TRUE))
    expect_true(chkFle(nmeOut))
    expect_gt(file.info(nmeOut)$size, 1)
    expect_true(chkFle(nmeOut, isZIP = TRUE))
    expect_true(chkFle(nmeOut, fleCnt = "meta"))
    expect_true(chkFle(nmeOut, fleCnt = "metadata.json"))
    expect_true(chkFle(nmeOut, fleCnt = "data.bin"))
    df4Chk <- read_omv(nmeOut, sveAtt = FALSE, getSyn = TRUE)
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(60, 18))
    expect_equal(names(df4Chk),
      c("ID", "Filter 1", "logLen", "supp - Transform 1", "len", "supp", "dose", "dose2", "dose3", "Trial", "Residuals", "J", "K", "L", "M", "weights", "A", "B"))
    expect_equal(vapply(df4Chk, typeof, character(1), USE.NAMES = FALSE),
      c("character", "logical", "double", "integer", "double", "integer", "double", "integer", "integer", "integer",
        "double", "double", "double", "integer", "logical", "integer", "double", "double"))
    expect_equal(sort(zip::zip_list(nmeOut)$filename),
      c("01 empty/analysis", "02 anova/analysis", "02 anova/resources/65167cb3bdaf8761.png", "02 anova/resources/99f9b5d34a92049b.png", "03 empty/analysis",
        "04 ancova/analysis", "05 empty/analysis", "data.bin", "index.html", "meta", "metadata.json", "strings.bin", "xdata.json"))
    expect_equal(attr(df4Chk, "syntax"),
      c(paste("jmv::ANOVA(formula = len ~ supp + dose2 + supp:dose2, data = data, effectSize = \"partEta\", modelTest = TRUE, qq = TRUE,",
              "contrasts = list(list(var=\"supp\", type=\"none\"), list(var=\"dose2\", type=\"polynomial\")), postHoc = ~ supp + dose2, emMeans = ~ dose2:supp)"),
           "jmv::ancova(formula = len ~ supp + dose, data = data, effectSize = \"partEta\", modelTest = TRUE)"))
    unlink(nmeOut)
    unlink(nmeInp)
})

#' Transpose .omv-files for the statistical spreadsheet 'jamovi' (<https://www.jamovi.org>)
#'
#' @param dtaInp Either a data frame or the name of a data file to be read (including the path, if required; "FILENAME.ext"; default: NULL); files can be of
#'               any supported file type, see Details below
#' @param fleOut Name of the data file to be written (including the path, if required; "FILE_OUT.omv"; default: ""); if empty, the resulting data frame is
#'               returned instead
#' @param varNme Name of the variables in the output data frame; see Details below
#' @param usePkg Name of the package: "foreign" or "haven" that shall be used to read SPSS, Stata and SAS files; "foreign" is the default (it comes with
#'               base R), but "haven" is newer and more comprehensive
#' @param selSet Name of the data set that is to be selected from the workspace (only applies when reading .RData-files)
#' @param ...    Additional arguments passed on to methods; see Details below
#'
#' @return a data frame (only returned if `fleOut` is empty) where the input data set is transposed
#'
#' @details
#' * If `varNme` empty, the row names of the input data set are used (preceded by "V_" if all row names are numbers); if `varNme` has length 1, then it is
#'   supposed to point to a variable in the input data frame; if `varNme` has the same length as the number of rows in the input data frame, then the values
#'   in `varNme` are assigned as column names to the output data frame.
#' * The ellipsis-parameter (`...`) can be used to submit arguments / parameters to the functions that are used for reading and writing the data. By clicking
#'   on the respective function under “See also”, you can get a more detailed overview over which parameters each of those functions take. The functions are:
#'   `read_omv` and `write_omv` (for jamovi-files), `read.table` (for CSV / TSV files; using similar defaults as `read.csv` for CSV and `read.delim` for TSV
#'   which both are based upon `read.table`), `load` (for .RData-files), `readRDS` (for .rds-files), `read_sav` (needs the R-package `haven`) or `read.spss`
#'   (needs the R-package `foreign`) for SPSS-files, `read_dta` (`haven`) / `read.dta` (`foreign`) for Stata-files, `read_sas` (`haven`) for SAS-data-files,
#'   and `read_xpt` (`haven`) / `read.xport` (`foreign`) for SAS-transport-files. If you would like to use `haven`, you may need to install it using
#'   `install.packages("haven", dep = TRUE)`.
#'
#' @seealso `transpose_omv` internally uses the following functions for reading and writing data files in different formats: [jmvReadWrite::read_omv()] and
#'   [jmvReadWrite::write_omv()] for jamovi-files, [utils::read.table()] for CSV / TSV files, [load()] for reading .RData-files, [readRDS()] for .rds-files,
#'   [haven::read_sav()] or [foreign::read.spss()] for SPSS-files, [haven::read_dta()] or [foreign::read.dta()] for Stata-files, [haven::read_sas()] for
#'   SAS-data-files, and [haven::read_xpt()] or [foreign::read.xport()] for SAS-transport-files.
#'
#' @examples
#' set.seed(1)
#' tmpDF <- stats::setNames(as.data.frame(matrix(sample(6, 1200, replace = TRUE), nrow = 16)),
#'                          sprintf("sbj_%03d", seq(75)))
#' str(tmpDF)
#' # Data sets that were extracted, e.g., from PsychoPy, may look like this (trials as rows
#' # and participants as columns, one for each participant, manually assembled / copy-and-pasted).
#' # However, for analyses, one wants the data set transposed (units / participants as columns)...
#' nmeOut <- tempfile(fileext = ".omv")
#' jmvReadWrite::transpose_omv(dtaInp = tmpDF, fleOut = nmeOut)
#' dtaFrm <- jmvReadWrite::read_omv(nmeOut)
#' unlink(nmeOut)
#' str(dtaFrm)
#' # if no varNme-parameter is given, generic variable names are created (V_...)
#' jmvReadWrite::transpose_omv(dtaInp = tmpDF, fleOut = nmeOut, varNme = sprintf("Trl_%02d", seq(16)))
#' dtaFrm <- jmvReadWrite::read_omv(nmeOut)
#' unlink(nmeOut)
#' str(dtaFrm)
#' # alternatively, the character vector with the desired variable names (of the same length as
#' # the number of rows in tmpDF) may be given, "Trl" can easily be exchanged by the name of your
#' # questionnaire, experimental conditions, etc.
#'
#'
#' @export transpose_omv
#'
transpose_omv <- function(dtaInp = NULL, fleOut = "", varNme = "", usePkg = c("foreign", "haven"), selSet = "", ...) {

    # check and import input data set (either as data frame or from a file)
    if (!is.null(list(...)[["fleInp"]])) stop("Please use the argument dtaInp instead of fleInp.")
    dtaFrm <- inp2DF(dtaInp = dtaInp, rmvEmp = TRUE, usePkg = usePkg, selSet = selSet, ...)

    # create variable names for the output data frame: if varNme is empty (default), then the row names of the
    # original data frame are used (preceded by "V_" if they contain only numbers); if varNme has the length 1
    # then it is assumed that it points to
    if (is.character(varNme)) {
        # varNme is empty, use row names of the input data frame, if all names are numbers, precede them with "V_"
        if        (length(varNme) == 1 && !nzchar(varNme)) {
            varOut <- paste0(ifelse(all(grepl("^[0-9]*$", row.names(dtaFrm))), "V_", ""), row.names(dtaFrm))
        # varNme has length 1, use the content of respective variable as column names
        } else if (length(varNme) == 1 &&  nzchar(varNme)) {
            selVar <- (names(dtaFrm) %in% varNme)
            if (any(selVar)) {
               varOut <- dtaFrm[,  selVar]
               dtaFrm <- dtaFrm[, !selVar]
            } else {
               stop(sprintf("%s (varNme) not contained in the input data frame.", varNme))
            }
        # varNme has the same length as there are rows in the input data frame
        } else if (length(varNme) == dim(dtaFrm)[1]) {
            varOut <- varNme
        } else {
            stop("varNme must either be empty, have one element or as many elements as there are rows in the input data frame.")
        }
    } else {
        stop("varNme must be a character variable.")
    }

    # transpose data frame
    dtaFrm <- stats::setNames(as.data.frame(t(dtaFrm)), varOut)

    # transfer the former variable / column names into a variable called "ID", and reset
    # the row.names to numbers (those wouldn't be written to the output data set anyway)
    dtaFrm <- cbind(list(ID = as.character(row.names(dtaFrm))), dtaFrm)
    attr(dtaFrm[, "ID"], "jmv-id") <- TRUE
    row.names(dtaFrm) <- NULL

    # rtnDta in globals.R (unified function to either write the data frame, open it in a new jamovi session or return it)
    rtnDta(dtaFrm = dtaFrm, fleOut = fleOut, dtaTtl = jmvTtl("_xpsd"), ...)
}

#' Converts .omv-files for the statistical spreadsheet 'jamovi' (<https://www.jamovi.org>) from long to wide format
#'
#' @param dtaInp Either a data frame or the name of a data file to be read (including the path, if required; "FILENAME.ext"; default: NULL); files can be of
#'               any supported file type, see Details below
#' @param fleOut Name of the data file to be written (including the path, if required; "FILE_OUT.omv"; default: ""); if empty, the resulting data frame is
#'               returned instead
#' @param varTgt Names of one or more variables to be transformed / reshaped (other variables are excluded, if empty(c()) all variables except `varTme`,
#'               `varID` and `varExc` are included; default: c())
#' @param varExc Name of the variable(s) should be excluded from the transformation, typically this will be between-subject-variable(s) (default: c())
#' @param varID  Names of one or more variables that identify the same group / individual (default: c())
#' @param varTme Name of the variable(s) that differentiates multiple records from the same group / individual (default: c())
#' @param varSep Separator character when concatenating the fixed and time-varying part of the variable name ("VAR1_1", "VAR1_2"; default: "_")
#' @param varOrd How variables / columns are organized: for "times" (default) the steps of the time varying variable are adjacent, for "vars" the steps of
#'               the original columns in the long dataset
#' @param varAgg How multiple occurrences of particular combinations of time varying variables are aggregated: either "mean" (calculate the mean over
#'               occurrences), or "first" (take the first occurrence)
#' @param varSrt Variable(s) that are used to sort the data frame (see Details; if empty, the order returned from reshape is kept; default: c())
#' @param usePkg Name of the package: "foreign" or "haven" that shall be used to read SPSS, Stata and SAS files; "foreign" is the default (it comes with
#'               base R), but "haven" is newer and more comprehensive
#' @param selSet Name of the data set that is to be selected from the workspace (only applies when reading .RData-files)
#' @param ...    Additional arguments passed on to methods; see Details below
#'
#' @return a data frame (only returned if `fleOut` is empty) where the input data set is converted from long to wide format
#'
#' @details
#' * If `varTgt` is empty, it is tried to generate it using all variables in the data frame except those defined by `varID`, `varTme` and `varExc`. The
#'   variable(s) in `varID` need to be unique identifiers (in the original dataset), those in `varExc` don't have this requirement. It is generally recommended
#'   that the variable names in `varExc` and `varID` should not contain the variable separator (defined in `varSep`; default: "_").
#' * `varSrt` can be either a character or a character vector (with one or more variables respectively). The sorting order for a particular variable can be
#'   inverted with preceding the variable name with "-". Please note that this doesn't make sense and hence throws a warning for certain variable types (e.g.,
#'   factors).
#' * The ellipsis-parameter (`...`) can be used to submit arguments / parameters to the functions that are used for transforming, reading or writing the data.
#'   By clicking on the respective function under “See also”, you can get a more detailed overview over which parameters each of those functions take.
#' * The transformation from long to wide uses `reshape`. `varTgt` matches (~) `v.names` in `reshape`, `varID` ~ `idvar`, `varTme` ~ `timevar`, and `varSep` ~
#'   `sep`. The help for `reshape` is very explanatory, click on the link under “See also” to access it, particularly what is explained under “Details”.
#' * The functions for reading and writing the data are: `read_omv` and `write_omv` (for jamovi-files), `read.table` (for CSV / TSV files; using similar
#'   defaults as `read.csv` for CSV and `read.delim` for TSV which both are based upon `read.table`), `load` (for .RData-files), `readRDS` (for .rds-files),
#'   `read_sav` (needs R-package `haven`) or `read.spss` (needs R-package `foreign`) for SPSS-files, `read_dta` (`haven`) / `read.dta` (`foreign`) for
#'   Stata-files, `read_sas` (`haven`) for SAS-data-files, and `read_xpt` (`haven`) / `read.xport` (`foreign`) for SAS-transport-files. If you would like to
#'   use `haven`, you may need to install it using `install.packages("haven", dep = TRUE)`.
#'
#' @seealso `long2wide_omv` internally uses the following functions: The transformation from long to wide uses [stats::reshape()]. For reading and writing data
#'   files in different formats: [jmvReadWrite::read_omv()] and [jmvReadWrite::write_omv()] for jamovi-files, [utils::read.table()] for CSV / TSV files,
#'   [load()] for reading .RData-files, [readRDS()] for .rds-files, [haven::read_sav()] or [foreign::read.spss()] for SPSS-files, [haven::read_dta()] or
#'   [foreign::read.dta()] for Stata-files, [haven::read_sas()] for SAS-data-files, and [haven::read_xpt()] or [foreign::read.xport()] for SAS-transport-files.
#'
#' @examples
#' # generate a test dataframe with 100 (imaginary) participants / units of
#' #  observation (ID), 8 measurement (measure) of one variable (X)
#' dtaInp <- data.frame(ID = rep(as.character(seq(1, 100)), each = 8),
#'                      measure = rep(seq(1, 8), times = 100),
#'                      X = runif(800, -10, 10))
#' cat(str(dtaInp))
#' # the output should look like this
#' # 'data.frame': 800 obs. of  3 variables:
#' #  $ ID     : chr  "1" "1" "1" "1" ...
#' #  $ measure: int  1 2 3 4 5 6 7 8 1 2 ...
#' #  $ X      : num  ...
#' # this data set is stored as (temporary) RDS-file and later processed by long2wide
#' nmeInp <- tempfile(fileext = ".rds")
#' nmeOut <- tempfile(fileext = ".omv")
#' saveRDS(dtaInp, nmeInp)
#' jmvReadWrite::long2wide_omv(dtaInp = nmeInp, fleOut = nmeOut, varTgt = "X", varID = "ID",
#'   varTme = "measure")
#' # it is required to give at least the arguments dtaInp, varID and varTme
#' # check whether the file was created and its size
#' cat(list.files(dirname(nmeOut), basename(nmeOut)))
#' # -> "file[...].omv" ([...] contains a random combination of numbers / characters
#' cat(file.info(nmeOut)$size)
#' # -> 6851 (approximate size; size may differ in every run [in dependence of
#' #          how well the generated random data can be compressed])
#' cat(str(jmvReadWrite::read_omv(nmeOut, sveAtt = FALSE)))
#' # the data set is now transformed into wide (and each the measurements is now
#' # indicated as a suffix to X; X_1, X_2, ...)
#' # 'data.frame':	100 obs. of  9 variables:
#' #  $ ID : chr  "1" "2" "3" "4" "5" "6" "7" "8" "9" "10" ...
#' #   ..- attr(*, "jmv-id")= logi TRUE
#' #   ..- attr(*, "missingValues")= list()
#' #  $ X_1: num  ...
#' #   ..- attr(*, "missingValues")= list()
#' #  $ X_2: num  ...
#' #   ..- attr(*, "missingValues")= list()
#' #  $ X_3: num  ...
#' #   ..- attr(*, "missingValues")= list()
#' #  $ X_4: num  ...
#' #   ..- attr(*, "missingValues")= list()
#' #  $ X_5: num  ...
#' #   ..- attr(*, "missingValues")= list()
#' #  $ X_6: num  ...
#' #   ..- attr(*, "missingValues")= list()
#' #  $ X_7: num  ...
#' #   ..- attr(*, "missingValues")= list()
#' #  $ X_8: num  ...
#' #   ..- attr(*, "missingValues")= list()
#'
#' unlink(nmeInp)
#' unlink(nmeOut)
#'
#'
#' @export long2wide_omv
#'
long2wide_omv <- function(dtaInp = NULL, fleOut = "", varTgt = c(), varExc = c(), varID = "ID", varTme = c(), varSep = "_", varOrd = c("times", "vars"),
                          varAgg = c("mean", "first"), varSrt = c(), usePkg = c("foreign", "haven"), selSet = "", ...) {

    # handle / check further input arguments
    # check varID (can be several) and varTme (must be one), neither can be empty
    if (!all(nzchar(c(varID, varTme)))) {
        stop("Using the arguments varID and varTme is mandatory (i.e., they can\'t be empty).")
    }
    varOrd <- match.arg(varOrd)
    varAgg <- match.arg(varAgg)

    # check and import input data set (either as data frame or from a file)
    if (!is.null(list(...)[["fleInp"]])) stop("Please use the argument dtaInp instead of fleInp.")
    dtaFrm <- inp2DF(dtaInp = dtaInp, rmvEmp = TRUE, usePkg = usePkg, selSet = selSet, ...)

    # transform data set
    # [a] check whether varID, varTme and varTgt are not empty and exist in the data set
    chkVar(dtaFrm, varID)
    chkVar(dtaFrm, varTme)
    chkVar(dtaFrm, varExc)
    if (!chkVar(dtaFrm, varTgt)) {
        varTgt <- setdiff(names(dtaFrm), c(varID, varTme, varExc))
    }
    # [b] store the original variable labels, the original time-varying / target variable,
    # and an empty vector for storing labels
    lstLbl <- list(orgLbl = lapply(dtaFrm, attr, "jmv-desc"), orgTgt = varTgt)

    # [c] there might be several occurrences for each combination of varID and varTme; aggregate them
    dtaFrm <- aggDta(dtaFrm = dtaFrm, varAgg = varAgg, varID = varID, varTme = varTme, varExc = varExc, varTgt = varTgt)

    # do the actual work, iterating through the different time-varying variables in varTme
    for (i in seq_along(varTme)) {
        # [d] call "reshape" with having the variable arguments limited to those valid when calling the function
        crrArg <- list(data = dtaFrm, direction = "wide", v.names = varTgt, idvar = c(varID, varTme[seq_along(varTme)[seq_along(varTme) > i]]), timevar = varTme[i], sep = varSep)
        dtaFrm <- do.call(stats::reshape, adjArg("reshape", crrArg, list(), c("data", "direction", "v.names", "idvar", "timevar")))
        # [1] varVry contains the variable names with varTgt / v.names as rows, and the steps of varTme[i] / timevar
        # as columns; [2] when generating an updated varTgt for the next step, this matrix is transformed to a vector
        # either not transposed (in this case, the original variables are adjacent), or transposed (in this case, the
        # different steps of the time-varying variable are adjacent); [3] afterwards the reshapeWide-attribute is removed
        varVry <- attr(dtaFrm, "reshapeWide")$varying
        varTgt <- eval(parse(text = paste0("as.vector(", ifelse(varOrd == "times", "t(varVry)", "varVry"), ")")))
        attr(dtaFrm, "reshapeWide") <- NULL
    }
    if (varOrd == "times") dtaFrm <- dtaFrm[, c(setdiff(names(dtaFrm), as.vector(varVry)), varTgt)]

    # select all variable(s) except those defined by varID and varExc and remove the prefix "measure", if present
#   selVrN <- !grepl(paste0(c(varID, varExc), collapse = "|"), names(dtaFrm))
#   if (all(grepl(paste0("^measure", varSep), names(dtaFrm)[selVrN]))) names(dtaFrm)[selVrN] <- gsub(paste0("^measure", varSep), "", names(dtaFrm)[selVrN])

    # restore the original labels
    dtaFrm <- rstLbl(dtaFrm, lstLbl, varTgt, varTme, varSep)

    # sort data set (if varSrt is not empty)
    dtaFrm <- srtFrm(dtaFrm, varSrt)

    # if varID is unique, set it's measureType to ID
    if (!any(duplicated(dtaFrm[[varID]]))) attr(dtaFrm[[varID]], "jmv-id") <- TRUE

    # rtnDta in globals.R (unified function to either write the data frame, open it in a new jamovi session or return it)
    rtnDta(dtaFrm = dtaFrm, fleOut = fleOut, dtaTtl = jmvTtl("_wide"), ...)
}

aggDta <- function(dtaFrm = NULL, varAgg = "", varID = c(), varTme = c(), varExc = c(), varTgt = c()) {
    # if there exists only one occurence of each possible combination of the variables in varID and
    # varTme, the data don't need to be aggregated, just return the data frame with the relevant
    # columns selected
    if (!any(stats::aggregate(dtaFrm[, varTgt[1]], by = dtaFrm[, c(varID, varTme)], FUN = length)[["x"]] > 1)) {
        dtaFrm[, c(varID, varTme, varExc, varTgt)]
    # otherwise (with more than one occurence), values are aggregate at each possible combination of the
    # variables in varID and varTme
    } else if (varAgg == "first") {
        # [1] if "first" is chosen as aggregation function, the first occurence at each step is returned
        stats::aggregate(x = dtaFrm[, c(varTgt, varExc), drop = FALSE], by = dtaFrm[, c(varID, varTme), drop = FALSE], FUN = function(x) x[!is.na(x)][1])
    } else if (varAgg == "mean")  {
        # [2] if "mean" is chosen as aggregation function, it becomes (a little) more complicated
        # [a] the target variables (for which the mean is calculated) should be numeric
        if (!all(vapply(dtaFrm[, varTgt], is.numeric, logical(1)))) {
            stop(paste("In order to calculate the mean when aggregating the data, all target variables (varTgt) need to be numeric. Use varAgg = \"first\" instead",
                       "(to use the first occuring value) or convert the target variables to numeric."))
        }
        # [b] if there are both target and “excluded” variables, the mean is calculated for the target
        # variables at each possible combination of the variables varID and varTme (first aggregate
        # within merge); afterwards, for the “excluded” variables (i.e., variables not to be
        # transformed to wide, the first occurrence is chosen (second aggregate; variables in
        # varExc, e.g. sex, should be the same for each step of the ID variables, e.g., each
        # participant [ID]); finally the results from the two aggregate-functions are merged again
        # to return the complete data set
        if (length(varExc) > 0) {
            merge(stats::aggregate(x = dtaFrm[, c(varTgt), drop = FALSE], by = dtaFrm[, c(varID, varTme), drop = FALSE], FUN = mean, na.rm = TRUE),
                  stats::aggregate(x = dtaFrm[, c(varExc), drop = FALSE], by = dtaFrm[, c(varID, varTme), drop = FALSE], FUN = function(x) x[!is.na(x)][1]))
        # [c] if there is no “excluded” variable, the mean is calculated for the target variables
        # at each possible combination of the variables varID and varTme
        } else {
            stats::aggregate(x = dtaFrm[, c(varTgt), drop = FALSE], by = dtaFrm[, c(varID, varTme), drop = FALSE], FUN = mean, na.rm = TRUE)
        }
    }
}

rstLbl <- function(dtaFrm = NULL, lstLbl = list(), varTgt = c(), varTme = c(), varSep = "_") {
    for (crrNme in names(lstLbl$orgLbl)) {
        if (crrNme %in% names(dtaFrm)) {
            attr(dtaFrm[[crrNme]], "jmv-desc") <- lstLbl$orgLbl[[crrNme]]
        } else if (crrNme %in% lstLbl$orgTgt) {
            splTgt <- strsplit(varTgt, gsub("\\.", "\\\\.", varSep))
            for (i in seq_along(splTgt)) {
                if (crrNme %in% splTgt[[i]]) {
                    attr(dtaFrm[[varTgt[i]]], "jmv-desc") <-
                      sprintf("%s (%s)", lstLbl$orgLbl[[crrNme]], paste0(apply(rbind(varTme, splTgt[[i]][-1]), 2, paste0, collapse = ": "), collapse = ", "))
                }
            }
        }
    }

    dtaFrm
}

#' Converts .omv-files for the statistical spreadsheet 'jamovi' (<https://www.jamovi.org>) from wide to long format
#'
#' @param dtaInp Either a data frame or the name of a data file to be read (including the path, if required; "FILENAME.ext"; default: NULL); files can be of
#'               any supported file type, see Details below
#' @param fleOut Name of the data file to be written (including the path, if required; "FILE_OUT.omv"; default: ""); if empty, the resulting data frame is
#'               returned instead
#' @param varLst List / set of variables that are to be transformed into single (time-varying) variables in long format (default: c())
#' @param varExc Name of the variable(s) should be excluded from the transformation, typically this will be between-subject-variable(s) (default: c())
#' @param varID  Name(s) of one or more variables that (is created to) identify the same group / individual (if empty, "ID" is added with row numbers
#'               identifying cases; default: NULL)
#' @param varTme Name of the variable (or vector with variable names) that (is created to) differentiate multiple records from the same group / individual #'               (default: "cond"; if required, a counter is added for each time-varying part)
#' @param varSep Character that separates the variables in varLst into a time-varying part and a part that forms the variable name in long format ("_" in
#'               "VAR_1", "VAR_2", default: "_")
#' @param varOrd Whether to arrange the variables before the transformation, so that they are in accordance with the different split levels (default: TRUE)
#' @param varSrt Variable(s) that are used to sort the data frame (see Details; if empty, the order returned from reshape is kept; default: c())
#' @param excLvl Integer (or vector of integers) determining which parts of the variable names in varLst shall not be transformed (default: NULL), see Details
#'               below
#' @param usePkg Name of the package: "foreign" or "haven" that shall be used to read SPSS, Stata and SAS files; "foreign" is the default (it comes with
#'               base R), but "haven" is newer and more comprehensive
#' @param selSet Name of the data set that is to be selected from the workspace (only applies when reading .RData-files)
#' @param ...    Additional arguments passed on to methods; see Details below
#'
#' @return a data frame (only returned if `fleOut` is empty) where the input data set is converted from wide to long format
#'
#' @details
#' * If `varLst` is empty, it is tried to generate it using all variables in the data frame except those defined by `varExc` and `varID`. The variable(s) in
#'   `varID` need to be unique identifiers (in the original dataset), those in `varExc` don't have this requirement. It is recommended that the variable names
#'   in `varExc` and `varID` should not contain the variable separator (defined in `varSep`; default: "_").
#' * `varOrd` determines whether the variables are rearranged to match the order of split levels. Consider the `varLst` X_1, Y_1, X_2, Y_2. If `varOrd` were
#'   set to FALSE, the original order would be preserved and the second part of the variable name (1, 2, ...) would become condition 1, and the first part
#'   condition 2. In most cases, leaving `varOrd` set to TRUE is recommended.
#' * `varSrt` can be either a character or a character vector (with one or more variables respectively). The sorting order for a particular variable can be
#'   inverted with preceding the variable name with "-". Please note that this doesn't make sense and hence throws a warning for certain variable types (e.g.,
#'   factors).
#' * `exclLvl` points to a part of the variable names in `varLst` to be excluded. For example, if the variable name is `PART1_PART2_PART3` (split at _), then
#'   `excLvl` = 1 would exclude PART1 from the transformation. Quite often, one has more that one variable of a particular type (e.g., responses, reaction
#'   times, etc.). Those would typically be the first part of each variable name in `varLst` (the conditions then being PART2, PART3, and so on). `excLvl` = 1
#'   would exclude those variable types / categories from being transformed into long (i.e., they would be kept as separate columns).
#' * The ellipsis-parameter (`...`) can be used to submit arguments / parameters to the functions that are used for transforming or reading the data. By
#'   clicking on the respective function under “See also”, you can get a more detailed overview over which parameters each of those functions take.
#' * The transformation from long to wide uses `reshape`: `varID` matches (~) `idvar` in `reshape`, `varTme` ~ `timevar`, `varLst` ~ `varying`, and `varSep` ~
#'   `sep`. The help for `reshape` is very explanatory, click on the link under “See also” to access it, particularly what is explained under “Details”.
#' * The functions for reading and writing the data are: `read_omv` and `write_omv` (for jamovi-files), `read.table` (for CSV / TSV files; using similar
#'   defaults as `read.csv` for CSV and `read.delim` for TSV which both are based upon `read.table`), `load` (for .RData-files), `readRDS` (for .rds-files),
#'   `read_sav` (needs R-package `haven`) or `read.spss` (needs R-package `foreign`) for SPSS-files, `read_dta` (`haven`) / `read.dta` (`foreign`) for
#'   Stata-files, `read_sas` (`haven`) for SAS-data-files, and `read_xpt` (`haven`) / `read.xport` (`foreign`) for SAS-transport-files. If you would like to
#'   use `haven`, you may need to install it using `install.packages("haven", dep = TRUE)`.
#'
#' @seealso `long2wide_omv` internally uses the following functions: The transformation from long to wide uses [stats::reshape()]. For reading and writing data
#'   files in different formats: [jmvReadWrite::read_omv()] and [jmvReadWrite::write_omv()] for jamovi-files, [utils::read.table()] for CSV / TSV files,
#'   [load()] for reading .RData-files, [readRDS()] for .rds-files, [haven::read_sav()] or [foreign::read.spss()] for SPSS-files, [haven::read_dta()] or
#'   [foreign::read.dta()] for Stata-files, [haven::read_sas()] for SAS-data-files, and [haven::read_xpt()] or [foreign::read.xport()] for SAS-transport-files.
#'
#' @examples
#' \dontrun{
#' # generate a test dataframe with 100 (imaginary) participants / units of
#' # observation (ID), and 8 repeated measurements of variable (X_1, X_2, ...)
#' dtaInp <- cbind(data.frame(ID = as.character(seq(1:100))),
#'                 stats::setNames(
#'                     as.data.frame(matrix(runif(800, -10, 10), nrow = 100)),
#'                     paste0("X_", 1:8)))
#' cat(str(dtaInp))
#' # 'data.frame':	100 obs. of  9 variables:
#' #  $ ID : chr  "1" "2" "3" "4" ...
#' #  $ X_1: num  ...
#' #  $ X_2: num  ...
#' #  $ X_3: num  ...
#' #  $ X_4: num  ...
#' #  $ X_5: num  ...
#' #  $ X_6: num  ...
#' #  $ X_7: num  ...
#' #  $ X_8: num  ...
#' # this data set is stored as (temporary) RDS-file and later processed by wide2long
#' nmeInp <- tempfile(fileext = ".rds")
#' nmeOut <- tempfile(fileext = ".omv")
#' saveRDS(dtaInp, nmeInp)
#' jmvReadWrite::wide2long_omv(dtaInp = nmeInp, fleOut = nmeOut, varID = "ID",
#'     varTme = "measure", varLst = setdiff(names(dtaInp), "ID"),
#'     varSrt = c("ID", "measure"))
#' # it is required to give at least the arguments dtaInp (if dtaInp is a data frame,
#' # fleOut needs to be provided too) and varID
#' # "reshape" then assigns all variables expect the variable defined by varID to
#' # varLst (but throws a warning)
#' # varSrt enforces sorting the data set after the transformation (sorted, the
#' # measurements within one person come after another; unsorted all measurements
#' # for one repetition would come after another)
#'
#' # check whether the file was created and its size
#' cat(list.files(dirname(nmeOut), basename(nmeOut)))
#' # -> "file[...].omv" ([...] contains a random combination of numbers / characters
#' cat(file.info(nmeOut)$size)
#' # -> 6939 (approximate size; size may differ in every run [in dependence of how
#' #          well the generated random data can be compressed])
#' cat(str(jmvReadWrite::read_omv(nmeOut, sveAtt = FALSE)))
#' # the data set is now transformed into long (and each the measurements is now
#' # indicated by the "measure")
#' # 'data.frame':	800 obs. of  3 variables:
#' #  $ ID     : Factor w/ 100 levels "1","2","3","4",..: 1 1 1 1 1 1 1 1 2 2 ...
#' #   ..- attr(*, "missingValues")= list()
#' #  $ measure: Factor w/ 8 levels "1","2","3","4",..: 1 2 3 4 5 6 7 8 1 2 ...
#' #   ..- attr(*, "missingValues")= list()
#' #  $ X      : num  ...
#' #   ..- attr(*, "missingValues")= list()
#'
#' unlink(nmeInp)
#' unlink(nmeOut)
#' }
#'
#' @export wide2long_omv
#'
wide2long_omv <- function(dtaInp = NULL, fleOut = "", varLst = c(), varExc = c(), varID = NULL,
                          varTme = "cond", varSep = "_", varOrd = TRUE, varSrt = c(), excLvl = NULL,
                          usePkg = c("foreign", "haven"), selSet = "", ...) {

    # check and import input data set (either as data frame or from a file)
    if (!is.null(list(...)[["fleInp"]])) stop("Please use the argument dtaInp instead of fleInp.")
    if (is.null(varID)) varID <- "ID"
    dtaFrm <- inp2DF(dtaInp, rmvEmp = TRUE, usePkg = usePkg, selSet = selSet, ...)
    dtaNmV <- names(dtaFrm)
    hasID  <- all(varID %in% dtaNmV)

    # check whether varLst is empty; if so, all variables except the one given as varID are used as varLst
    # otherwise, check whether the variables in varLst are contained in the data frame
    if (is.null(varLst)) {
        cat(paste0("Variable list (varLst) was generated using all variables in the data frame",
                   ifelse(hasID || (!is.null(varExc) && all(nzchar(varExc))),
                   sprintf(" except those defined in varExc or varID (%s).",
                           paste(c(varExc, rep(varID, hasID)), collapse = ", ")), "."), "\n"))
        varLst <- setdiff(dtaNmV, c(varExc, varID))
    } else {
        chkVar(dtaFrm, varLst)
    }
    # check whether they all contain the separator
    if (nzchar(varSep) && !all(grepl(varSep, varLst, fixed = TRUE))) {
        stop(sprintf(paste("\n\nThe variable separator (varSep, \"%s\") must be contained in all variables",
                           "in the variable list (varLst).\nDeviating variables: %s\n"),
                     varSep, paste(varLst[!grepl(varSep, varLst, fixed = TRUE)], collapse = ", ")))
    }
    # re-arrange the variable in the variable list so that they are in accordance with the order
    # of the splits in the variable names
    if (varOrd) varLst <- sort(varLst)
    # use varSep to split the variable names in varLst
    if (nzchar(varSep)) {
        varSpl <- strsplit(varLst, gsub("\\.", "\\\\.", varSep))
        lngSpl <- unique(vapply(varSpl, length, integer(1)))
        if (length(lngSpl) != 1) {
            stop(sprintf(paste("The variable names in varLst need to have the same structure, i.e., the",
                               "same number of separators within all variable names:\n%s\n\n"),
                         paste0(varLst, collapse = ", ")))
        }
    } else {
        varSpl <- as.list(varLst)
        lngSpl <- 1
    }

    # check excLvl: must be numeric and not contain any values that are outside lngSpl
    if (!is.null(excLvl) && length(excLvl) > 0 && (!is.numeric(excLvl) || excLvl < 1 || excLvl > lngSpl)) {
        stop(sprintf("excLvl must be numeric and must not be less (excLvl < 1) or more then the number of available levels (excLvl > %d).", lngSpl))
    }
    # dffSpl determines for each of the splits how soon the first element is differing; it is
    # assumed that the variables in varLst are hierarchically organized, i.e., one condition varies
    # first (here the difference should be smallest), for the next factor in the hierarchy, the
    # difference should be larger and so on; if factors are unique (i.e., if there is no difference
    # for this part of the split), an NA is returned; however: further below (in the while-loop),
    # NAs are used to indicate which positions have already been processed, setting NAs and the
    # levels to be excluded (excLvl) to Inf ensures that they would never be found as crrPos
    # (Inf is larger than any finite number and the while loop stops if there aren't any finite
    # numbers, i.e., values other than Inf or NA) left
    dffSpl <- rep(NA, lngSpl)
    for (i in seq(lngSpl)) dffSpl[i] <- which(c(FALSE, !duplicated(vapply(varSpl, "[[", character(1), i))[-1]))[1]
    dffSpl[is.na(dffSpl)] <- Inf
    dffSpl[excLvl] <- Inf
    numSpl <- sum(is.finite(dffSpl)) > 1
    pfxTme <- ifelse(length(varTme) > 1, "cond", varTme)

    # carry out the transformation
    crrNmV <- varLst
    while (any(is.finite(dffSpl))) {
        # crrPos: the smallest level that is valid (i.e., that is not unique or excluded because of excLvl)
        crrPos <- which(dffSpl == min(dffSpl, na.rm = TRUE))
        # crrTms is used as parameter in crrArg and the if-condition below
        crrTms <- unique(vapply(varSpl, "[[", character(1), crrPos))
        # assemble the list for varying
        crrVry <- crtVry(crrNmV = crrNmV, crrTms = crrTms, varSpl = varSpl, vldPos = which(!is.na(dffSpl)), crrPos = crrPos, varSep = varSep)
        crrArg <- list(data = dtaFrm, direction = "long", idvar = varID, sep = varSep, varying = crrVry,
                       v.names = names(crrVry), timevar = paste0(pfxTme, rep(sum(is.finite(dffSpl)), numSpl)),
                       times = crrTms)
        dtaFrm <- rmvTms(do.call(stats::reshape, adjArg("stats::reshape", crrArg, ...,
                                                        c("data", "direction", "idvar", "sep", "varying", "times", "timevar", "v.names"))), crrTms)
        dtaFrm[[crrArg$timevar]] <- as.factor(dtaFrm[[crrArg$timevar]])
        varID  <- c(varID, crrArg$timevar)
        crrNmV <- names(crrVry)
        dffSpl[crrPos] <- NA
    }
    # remove attributes from reshape attached to the data set, and row.names
    attr(dtaFrm, "reshapeLong") <- NULL
    row.names(dtaFrm) <- NULL

    # correct the column order (ID should come first, otherwise the order is kept with the transformed variables inserted into
    # were the respective original (i.e., before the transformation) variables were
    dtaFrm <- dtaFrm[, ordCol(names(dtaFrm), dtaNmV, varID, varLst)]

    # if varTme was a variable vector, assign the names in it to the columns beginning with the time prefix
    if (length(varTme) > 1) {
        selClm <- grepl(paste0("^", pfxTme), names(dtaFrm))
        if (sum(selClm) == length(varTme)) {
            names(dtaFrm)[selClm] <- varTme
#           varID[grepl(paste0("^", pfxTme), varID)] <- varTme
        } else {
            warning("The number of splits (defined by the variables in varLst and varSep) is not matching the length of varTme.")
        }
    }

    # sort data set (if varSrt is not empty, otherwise sort after the first variable in varID (defining the participant)
    dtaFrm <- srtFrm(dtaFrm, c(rep(varSrt, length(varSrt) > 0), rep(varID[1], length(varSrt) >= 0)))

    # remove the jmv-id, and change the measurement type to "Nominal" (if it was ID; -> the ID variable may be used
    # in analyses, e.g. as random-effects-variable, and this wouldn't be possible if it were still marked as "ID")
    dtaFrm <- rmvID(dtaFrm, varID, hasID)

    # rtnDta in globals.R (unified function to either write the data frame, open it in a new jamovi session or return it)
    rtnDta(dtaFrm = dtaFrm, fleOut = fleOut, dtaTtl = jmvTtl("_long"), ...)
}

crtVry <- function(crrNmV = c(), crrTms = c(), varSpl = c(), vldPos = NA, crrPos = NA, varSep = "_") {
    # assemble the list for varying, if crrTms are the only elements of left in varLst, an output
    # variable “measure” is used as target, otherwise crrVry is assembled as named list with the
    # target as name and all former variables for that step of the hierarchy as entries
    # assemble the arguments to call reshape (limiting the variable arguments - ... - to those permitted)
    # rmvTms also corrects labels (if available) and variable names

    if (all(crrNmV %in% crrTms)) {
        list(measure = crrNmV)
    } else {
        crrVry <- list()
        for (i in seq_along(varSpl)) {
            tmpTgt <- paste0(varSpl[[i]][setdiff(vldPos, crrPos)], collapse = varSep)
            crrVry[[tmpTgt]] <- unique(c(crrVry[[tmpTgt]], paste0(varSpl[[i]][vldPos], collapse = varSep)))
        }
        crrVry
    }
}

ordCol <- function(varNme = c(), dtaNmV = c(), varID = c(), varLst = c()) {
    posVrL <- range(which(dtaNmV %in% varLst))
    varOrd <- c(varID[1], setdiff(dtaNmV[rep(seq(1, posVrL[1] - 1), posVrL[1] > 1)], varID), sort(varID[-1]),
                setdiff(varNme, c(varID, setdiff(dtaNmV, varLst))),
                setdiff(dtaNmV[rep(seq(posVrL[2] + 1, length(dtaNmV)), posVrL[2] < length(dtaNmV))], varID))
    if (length(c(setdiff(varNme, varOrd), setdiff(varOrd, varNme))) != 0) {
        stop(sprintf("Mismatch between old and new variable order - old: (%s); new: (%s).",
                     paste0(varNme, collapse = ", "), paste0(varOrd, collapse = ", ")))
    }

    varOrd
}

rmvID <- function(dtaFrm = NULL, varID = c(), hasID = TRUE) {
    if ("jmv-id" %in% names(attributes(dtaFrm[, varID[1]]))) attr(dtaFrm[, varID[1]], "jmv-id") <- NULL
    if (!hasID || ("measureType" %in% names(attributes(dtaFrm[, varID[1]])) && attr(dtaFrm[, varID[1]], "measureType") == "ID")) {
        dtaFrm[, varID[1]] <- as.factor(dtaFrm[, varID[1]])
        attr(dtaFrm[, varID[1]], "measureType") <- "Nominal"
        attr(dtaFrm[, varID[1]], "dataType")    <- "Text"
    }

    dtaFrm
}

rmvTms <- function(dtaFrm = NULL, crrTms = c()) {
    varNme <- attr(dtaFrm, "reshapeLong")$v.names
    for (crrNme in varNme) {
        attr(dtaFrm[[crrNme]], "name") <- crrNme
        crrDsc <- c(attr(dtaFrm[, crrNme], "jmv-desc"), attr(dtaFrm[, crrNme], "description"))[1]
        if (is.null(crrDsc) || !nzchar(crrDsc)) next
        splDsc <- trimws(strsplit(crrDsc, "\\(|\\)")[[1]])
        if (length(splDsc) == 2) {
            rplDsc <- paste0(trimws(strsplit(splDsc[2], ",")[[1]])[!grepl(paste(paste0("\\w+: ", crrTms, "$"), collapse = "|"),
                             trimws(strsplit(splDsc[2], ",")[[1]]))], collapse = ", ")
            attr(dtaFrm[[crrNme]], "jmv-desc")    <- ifelse(nzchar(rplDsc), paste0(splDsc[1], " (", rplDsc, ")"), splDsc[1])
            attr(dtaFrm[[crrNme]], "description") <- ifelse(nzchar(rplDsc), paste0(splDsc[1], " (", rplDsc, ")"), splDsc[1])
        }
    }

    dtaFrm
}

test_that("wide2long_omv works", {
    # rather simple data set ("." as varSep) ==========================================================================
    set.seed(1)
    nmeTmp <- list(seq(121), paste0("X.", month.abb[seq(12)]))
    dtaTmp <- cbind(data.frame(Year = 1900:2020), as.data.frame(matrix(runif(121 * 12, 0, 100), nrow = 121, dimnames = nmeTmp)))
    for (i in seq(12)) attr(dtaTmp[[i + 1]], "jmv-desc") <- paste0("Test variable (Month: ", month.abb[i], ")")
    nmeInp <- tempfile(fileext = ".rds")
    nmeOut <- tempfile(fileext = "_L.omv")
    saveRDS(dtaTmp, nmeInp)

    expect_null(wide2long_omv(dtaInp = nmeInp, fleOut = nmeOut, varLst = setdiff(names(dtaTmp), "Year"),
                              varID = "Year", varTme = "Month", varSep = "."))
    df4Chk <- read_omv(nmeOut)
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(121 * 12, 3))
    expect_equal(vapply(df4Chk, typeof, character(1), USE.NAMES = FALSE), c("integer", "integer", "double"))
    expect_equal(names(df4Chk), c("Year", "Month", "X"))
    expect_equal(names(attributes(df4Chk)), c("names", "row.names", "class", "removedRows", "addedRows", "transforms"))
    expect_equal(names(attributes(df4Chk[[3]])), c("jmv-desc", "name", "id", "columnType", "dataType", "measureType", "formula", "formulaMessage",
                                                   "parentId", "width", "type", "importName", "description", "transform", "edits", "missingValues"))
    expect_equal(attr(df4Chk[[3]], "jmv-desc"), "Test variable")
    expect_equal(c(mean(df4Chk[[3]]), sd(df4Chk[[3]])), c(49.33121, 28.93480), tolerance = 1e-4)
    unlink(nmeOut)

    expect_null(wide2long_omv(dtaInp = nmeInp, fleOut = nmeOut, varLst = setdiff(names(dtaTmp), "Year"), varID = "Year", varSep = "."))
    df4Chk <- read_omv(nmeOut)
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(121 * 12, 3))
    expect_equal(vapply(df4Chk, typeof, character(1), USE.NAMES = FALSE), c("integer", "integer", "double"))
    expect_equal(names(df4Chk), c("Year", "cond", "X"))
    expect_equal(names(attributes(df4Chk)), c("names", "row.names", "class", "removedRows", "addedRows", "transforms"))
    expect_equal(names(attributes(df4Chk[[3]])), c("jmv-desc", "name", "id", "columnType", "dataType", "measureType", "formula", "formulaMessage",
                                                   "parentId", "width", "type", "importName", "description", "transform", "edits", "missingValues"))
    expect_equal(attr(df4Chk[[3]], "jmv-desc"), "Test variable")
    expect_equal(c(mean(df4Chk[[3]]), sd(df4Chk[[3]])), c(49.33121, 28.93480), tolerance = 1e-4)
    unlink(nmeOut)

    unlink(nmeInp)

    # rather simple data set ("_" as varSep) ==========================================================================
    set.seed(1)
    nmeTmp <- list(seq(121), paste0("X_", month.abb[seq(12)]))
    dtaTmp <- cbind(data.frame(Year = 1900:2020), as.data.frame(matrix(runif(121 * 12, 0, 100), nrow = 121, dimnames = nmeTmp)))
    for (i in seq(12)) attr(dtaTmp[[i + 1]], "jmv-desc") <- paste0("Test variable (Month: ", month.abb[i], ")")
    nmeInp <- tempfile(fileext = ".rds")
    nmeOut <- tempfile(fileext = "_L.omv")
    saveRDS(dtaTmp, nmeInp)

    expect_null(wide2long_omv(dtaInp = nmeInp, fleOut = nmeOut, varLst = setdiff(names(dtaTmp), "Year"),
                              varID = "Year", varTme = "Month", varSep = "_"))
    df4Chk <- read_omv(nmeOut)
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(121 * 12, 3))
    expect_equal(vapply(df4Chk, typeof, character(1), USE.NAMES = FALSE), c("integer", "integer", "double"))
    expect_equal(names(df4Chk), c("Year", "Month", "X"))
    expect_equal(names(attributes(df4Chk)), c("names", "row.names", "class", "removedRows", "addedRows", "transforms"))
    expect_equal(names(attributes(df4Chk[[3]])), c("jmv-desc", "name", "id", "columnType", "dataType", "measureType", "formula", "formulaMessage",
                                                   "parentId", "width", "type", "importName", "description", "transform", "edits", "missingValues"))
    expect_equal(attr(df4Chk[[3]], "jmv-desc"), "Test variable")
    expect_equal(c(mean(df4Chk[[3]]), sd(df4Chk[[3]])), c(49.33121, 28.93480), tolerance = 1e-4)
    unlink(nmeOut)

    expect_null(wide2long_omv(dtaInp = nmeInp, fleOut = nmeOut, varLst = setdiff(names(dtaTmp), "Year"), varID = "Year"))
    df4Chk <- read_omv(nmeOut)
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(121 * 12, 3))
    expect_equal(vapply(df4Chk, typeof, character(1), USE.NAMES = FALSE), c("integer", "integer", "double"))
    expect_equal(names(df4Chk), c("Year", "cond", "X"))
    expect_equal(names(attributes(df4Chk)), c("names", "row.names", "class", "removedRows", "addedRows", "transforms"))
    expect_equal(names(attributes(df4Chk[[3]])), c("jmv-desc", "name", "id", "columnType", "dataType", "measureType", "formula", "formulaMessage",
                                                   "parentId", "width", "type", "importName", "description", "transform", "edits", "missingValues"))
    expect_equal(attr(df4Chk[[3]], "jmv-desc"), "Test variable")
    expect_equal(c(mean(df4Chk[[3]]), sd(df4Chk[[3]])), c(49.33121, 28.93480), tolerance = 1e-4)
    unlink(nmeOut)

    expect_null(wide2long_omv(dtaInp = nmeInp, fleOut = nmeOut, varLst = setdiff(names(dtaTmp), "Year"),
                              varID = "Year", varTme = "Month", varSrt = c("Year", "Month")))
    df4Chk <- read_omv(nmeOut)
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(121 * 12, 3))
    expect_false(is.unsorted(df4Chk[["Year"]]))
    expect_true(all(df4Chk[["Month"]] == rep(sort(month.abb), length(unique(df4Chk[["Year"]])))))
    unlink(nmeOut)

    df4Chk <- wide2long_omv(dtaInp = nmeInp, varLst = setdiff(names(dtaTmp), "Year"), varID = "Year",
                            varTme = "Month", varSrt = c("Year", "Month"))
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(121 * 12, 3))
    expect_false(is.unsorted(df4Chk[["Year"]]))
    expect_true(all(df4Chk[["Month"]] == rep(sort(month.abb), length(unique(df4Chk[["Year"]])))))

    expect_error(wide2long_omv(fleInp = nmeInp, varID = "Year", varTme = "Month", varSep = "_"), regexp = "Please use the argument dtaInp instead of fleInp\\.")
    expect_error(wide2long_omv(dtaInp = dtaTmp, fleOut = nmeOut, varLst = names(dtaTmp), varTme = "Month", varSep = "_", varSrt = c("Year", "Month")),
      regexp = "^\\s+The variable separator \\(.*\\) must be contained in all variables in the variable list \\(varLst\\)\\.")
    expect_error(wide2long_omv(dtaInp = dtaTmp, fleOut = nmeOut, varLst = setdiff(names(dtaTmp), "Year"), varID = "Year", varSep = "x"),
      regexp = "^\\s+The variable separator \\(.*\\) must be contained in all variables in the variable list \\(varLst\\)\\.")
    expect_error(wide2long_omv(dtaInp = dtaTmp, fleOut = nmeOut, varLst = setdiff(names(dtaTmp), "Year"), varID = "Year", varSep = "."),
      regexp = "^\\s+The variable separator \\(.*\\) must be contained in all variables in the variable list \\(varLst\\)\\.")
#   expect_warning(wide2long_omv(dtaInp = nmeInp, nmeOut, varID = "Year", varTme = "Month", varSrt = c("Year", "Month")))

    expect_error(wide2long_omv(dtaInp = data.frame(Year = 1900:2020, X_1 = runif(121), X.2 = runif(121)), fleOut = nmeOut, varLst = c("X_1", "X.2"), varID = "Year", varSep = "_"),
      regexp = "^\\s+The variable separator \\(.*\\) must be contained in all variables in the variable list \\(varLst\\)\\.")
    expect_error(wide2long_omv(dtaInp = data.frame(Year = 1900:2020, X_1 = runif(121), X_2 = runif(121), X_2_A = runif(121)),
      fleOut = nmeOut, varLst = c("X_1", "X_2", "X_2_A"), varID = "Year", varSep = "_"),
      regexp = "^The variable names in varLst need to have the same structure, i\\.e\\., the same number of separators within all variable names\\:")
    expect_error(wide2long_omv(dtaInp = dtaTmp, varLst = setdiff(names(dtaTmp), "Year"), varID = "Year", varTme = "Month", varSep = "_", excLvl = 0),
      regexp = "^excLvl must be numeric and must not be less \\(excLvl < 1\\) or more then the number of available levels \\(excLvl > \\d\\)\\.")
    expect_error(wide2long_omv(dtaInp = dtaTmp, varLst = setdiff(names(dtaTmp), "Year"), varID = "Year", varTme = "Month", varSep = "_", excLvl = Inf),
      regexp = "^excLvl must be numeric and must not be less \\(excLvl < 1\\) or more then the number of available levels \\(excLvl > \\d\\)\\.")

    expect_error(ordCol(varNme = c("Year", "Month", "X"), dtaNmV = c(names(dtaTmp), "Trial"), varID = "Year", varLst = names(dtaTmp)[-1]),
      regexp = "^Mismatch between old and new variable order - old: .*; new: .*\\.")
    expect_error(ordCol(varNme = c("Year", "Month", "X"), dtaNmV = names(dtaTmp), varID = c("Year", "Trial"), varLst = names(dtaTmp)[-1]),
      regexp = "^Mismatch between old and new variable order - old: .*; new: .*\\.")

    dtaTmp <- data.frame(ID = as.character(seq(121)), A = runif(121), B = runif(121), C = runif(121))
    attributes(dtaTmp[[1]]) <- list(`jmv-id` = TRUE, measureType = "ID")
    expect_output(wide2long_omv(dtaInp = dtaTmp, fleOut = nmeOut, varSep = "", varID = "ID"),
                  "Variable list \\(varLst\\) was generated using all variables in the data frame except those defined in varExc or varID \\(ID\\).")
    df4Chk <- read_omv(nmeOut)
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(363, 3))
    expect_equal(names(df4Chk), c("ID", "cond", "measure"))
    expect_equal(vapply(df4Chk, class, character(1), USE.NAMES = FALSE), c("factor", "factor", "numeric"))
    expect_equal(table(df4Chk$cond), table(c(rep("A", 121), rep("B", 121), rep("C", 121))))

    unlink(nmeInp)

    # more complex data set ===========================================================================================
    set.seed(1)
    tgtTmp <- c(0.002, 0.03, 0.02, 0.450, 0.530, 0.510)
    dtaTmp <- cbind(data.frame(ID = as.factor(sprintf("sbj_%03d", seq(100))), sex = as.factor(sample(rep(c("female", "male"), 50)))),
              setNames(as.data.frame(matrix(NA, nrow = 100, ncol = 48)), sprintf("%s_%s_%s_%s", rep(c("rspCrr", "rspTme"), each = 24),
                rep(rep(c("cong", "incong", "neutral"), each = 8), times = 2), rep(rep(c("BLUE", "GREEN", "RED", "YELLOW"), each = 2), times = 6),
                rep(c("1", "2"), times = 24))))
    for (i in seq(1, 48))
        if (i <= 24) dtaTmp[i + 2] <- as.integer(runif(100, 0, 1) > tgtTmp[ceiling((i - 0.1) / 8)]) else dtaTmp[i + 2] <- tgtTmp[ceiling((i - 0.1) / 8)] + rnorm(100) * 0.05

    df4Chk <- wide2long_omv(dtaInp = dtaTmp, fleOut = "", varLst = names(dtaTmp)[c(-1, -2)], varExc = "sex", varID = "ID", varTme = "cond", excLvl = 1)
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(2400, 7))
    expect_equal(vapply(df4Chk, typeof, character(1), USE.NAMES = FALSE), c(rep("integer", 6), "double"))
    expect_equal(names(df4Chk), c("ID", "sex", sprintf("cond%d", 1:3), "rspCrr", "rspTme"))
    expect_equal(df4Chk[, "cond1"], as.factor(rep(rep(c("cong", "incong", "neutral"), each = 8), times = 100)))
    expect_equal(df4Chk[, "cond2"], as.factor(rep(rep(c("BLUE", "GREEN", "RED", "YELLOW"), each = 2), times = 300)))
    expect_equal(df4Chk[, "cond3"], as.factor(rep(c("1", "2"), times = 1200)))
    expect_equal(unname(colMeans(dtaTmp[3:50])), as.vector(unlist(aggregate(x = df4Chk[, c("rspCrr", "rspTme")], by = df4Chk[, c("cond3", "cond2", "cond1")], FUN = mean)[4:5])))

    df4Chk <- wide2long_omv(dtaInp = dtaTmp, fleOut = "", varLst = names(dtaTmp)[c(-1, -2)], varExc = "sex", varID = "ID", varTme = "cond")
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(4800, 7))
    expect_equal(vapply(df4Chk, typeof, character(1), USE.NAMES = FALSE), c(rep("integer", 6), "double"))
    expect_equal(names(df4Chk), c("ID", "sex", sprintf("cond%d", 1:4), "measure"))
    expect_equal(df4Chk[, "cond1"], as.factor(rep(rep(c("rspCrr", "rspTme"), each = 24), times = 100)))
    expect_equal(df4Chk[, "cond2"], as.factor(rep(rep(c("cong", "incong", "neutral"), each = 8), times = 200)))
    expect_equal(df4Chk[, "cond3"], as.factor(rep(rep(c("BLUE", "GREEN", "RED", "YELLOW"), each = 2), times = 600)))
    expect_equal(df4Chk[, "cond4"], as.factor(rep(c("1", "2"), times = 2400)))
    expect_equal(unname(colMeans(dtaTmp[3:50])), aggregate(x = df4Chk[, "measure", drop = FALSE], by = df4Chk[, c("cond4", "cond3", "cond2", "cond1")], FUN = mean)[, 5])
    unlink(nmeOut)

    # more complex data set, varTme as a vector =======================================================================
    set.seed(1)
    nmeTmp <- list(seq(121), sprintf("X_%s_T%d", rep(month.abb[seq(12)], each = 2), rep(seq(2), 12)))
    dtaTmp <- cbind(data.frame(Year = 1900:2020), as.data.frame(matrix(runif(121 * 24, 0, 100), nrow = 121, dimnames = nmeTmp)))
    for (i in seq(12)) for (j in seq(2)) attr(dtaTmp[[i * 2 + j - 1]], "jmv-desc") <- sprintf("Test variable (Month: %s, Test: T%d)", month.abb[i], j)
    nmeInp <- tempfile(fileext = ".rds")
    nmeOut <- tempfile(fileext = "_L.omv")
    saveRDS(dtaTmp, nmeInp)

    expect_null(wide2long_omv(dtaInp = nmeInp, fleOut = nmeOut, varLst = setdiff(names(dtaTmp), "Year"), varID = "Year",
                              varTme = c("Month", "Test"), varSep = "_"))
    df4Chk <- read_omv(nmeOut)
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(121 * 24, 4))
    expect_equal(vapply(df4Chk, typeof, character(1), USE.NAMES = FALSE), c(rep("integer", 3), "double"))
    expect_equal(names(df4Chk), c("Year", "Month", "Test", "X"))
    expect_equal(names(attributes(df4Chk)), c("names", "row.names", "class", "removedRows", "addedRows", "transforms"))
    expect_equal(names(attributes(df4Chk[[4]])), c("jmv-desc", "name", "id", "columnType", "dataType", "measureType", "formula", "formulaMessage",
                                                   "parentId", "width", "type", "importName", "description", "transform", "edits", "missingValues"))
    expect_equal(attr(df4Chk[[4]], "jmv-desc"), "Test variable")
    expect_equal(c(mean(df4Chk[[4]]), sd(df4Chk[[4]])), c(49.30143, 29.22032), tolerance = 1e-4)
    unlink(nmeOut)

    expect_null(wide2long_omv(dtaInp = nmeInp, fleOut = nmeOut, varLst = setdiff(names(dtaTmp), "Year"), varID = "Year", varSep = "_"))
    df4Chk <- read_omv(nmeOut)
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(121 * 24, 4))
    expect_equal(vapply(df4Chk, typeof, character(1), USE.NAMES = FALSE), c(rep("integer", 3), "double"))
    expect_equal(names(df4Chk), c("Year", "cond1", "cond2", "X"))
    expect_equal(names(attributes(df4Chk)), c("names", "row.names", "class", "removedRows", "addedRows", "transforms"))
    expect_equal(names(attributes(df4Chk[[4]])), c("jmv-desc", "name", "id", "columnType", "dataType", "measureType", "formula", "formulaMessage",
                                                   "parentId", "width", "type", "importName", "description", "transform", "edits", "missingValues"))
    expect_equal(attr(df4Chk[[4]], "jmv-desc"), "Test variable")
    expect_equal(c(mean(df4Chk[[4]]), sd(df4Chk[[4]])), c(49.30143, 29.22032), tolerance = 1e-4)
    unlink(nmeOut)

    expect_warning(wide2long_omv(dtaInp = nmeInp, fleOut = nmeOut, varLst = setdiff(names(dtaTmp), "Year"),
                                 varID = "Year", varTme = c("Month", "Test", "AddCol")),
                   regexp = "The number of splits \\(defined by the variables in varLst and varSep\\) is not matching the length of varTme\\.")
    df4Chk <- read_omv(nmeOut)
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(121 * 24, 4))
    expect_equal(vapply(df4Chk, typeof, character(1), USE.NAMES = FALSE), c(rep("integer", 3), "double"))
    expect_equal(names(df4Chk), c("Year", "cond1", "cond2", "X"))
    expect_equal(names(attributes(df4Chk)), c("names", "row.names", "class", "removedRows", "addedRows", "transforms"))
    expect_equal(names(attributes(df4Chk[[4]])), c("jmv-desc", "name", "id", "columnType", "dataType", "measureType", "formula", "formulaMessage",
                                                   "parentId", "width", "type", "importName", "description", "transform", "edits", "missingValues"))
    expect_equal(attr(df4Chk[[4]], "jmv-desc"), "Test variable")
    expect_equal(c(mean(df4Chk[[4]]), sd(df4Chk[[4]])), c(49.30143, 29.22032), tolerance = 1e-4)
    unlink(nmeOut)

    expect_null(wide2long_omv(dtaInp = nmeInp, fleOut = nmeOut, varLst = setdiff(names(dtaTmp), "Year"), varID = "Year",
                              varTme = c("Month", "Test"), varSrt = c("Year", "Month")))
    df4Chk <- read_omv(nmeOut)
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(121 * 24, 4))
    expect_false(is.unsorted(df4Chk[["Year"]]))
    expect_true(all(df4Chk[["Month"]] == rep(rep(sort(month.abb), each = 2), length(unique(df4Chk[["Year"]])))))
    unlink(nmeOut)

    df4Chk <- wide2long_omv(dtaInp = nmeInp, varLst = setdiff(names(dtaTmp), "Year"), varID = "Year",
                            varTme = c("Month", "Test"), varSrt = c("Year", "Month"))
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(121 * 24, 4))
    expect_false(is.unsorted(df4Chk[["Year"]]))
    expect_true(all(df4Chk[["Month"]] == rep(rep(sort(month.abb), each = 2), length(unique(df4Chk[["Year"]])))))

    expect_null(wide2long_omv(dtaInp = nmeInp, fleOut = nmeOut, varLst = setdiff(names(dtaTmp), "Year"), varID = "Year",
                              varTme = c("Month", "Test"), varSrt = c("Year", "Test", "Month")))
    df4Chk <- read_omv(nmeOut)
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(121 * 24, 4))
    expect_false(is.unsorted(df4Chk[["Year"]]))
    expect_true(all(df4Chk[["Month"]] == rep(sort(month.abb), length(unique(df4Chk[["Year"]])) * 2)))
    unlink(nmeOut)

    expect_error(wide2long_omv(dtaInp = nmeInp, varLst = setdiff(names(dtaTmp), "Year"),
                               varID = "Year", varSrt = c("Year", "Month", "AddCol")),
      regexp = "The variable\\(s\\) Month, AddCol are not contained in the current data set\\.")

    expect_error(
      expect_warning(wide2long_omv(dtaInp = nmeInp, varLst = setdiff(names(dtaTmp), "Year"),
                                   varID = "Year", varTme = c("Year", "Month", "AddCol"), varSrt = c("Year", "Month", "AddCol")),
                     regexp = "The number of splits \\(defined by the variables in varLst and varSep\\) is not matching the length of varTme\\."),
      regexp = "The variable\\(s\\) Month, AddCol are not contained in the current data set\\.")

    unlink(nmeInp)
})

#' Adds a title and a description for a data set stored as .omv-file for the statistical spreadsheet 'jamovi' (<https://www.jamovi.org>)
#'
#' @param dtaInp Either a data frame or the name of a data file to be read (including the path, if required; "FILENAME.ext"; default: NULL); files can be of
#'               any supported file type, see Details below
#' @param fleOut Name of the data file to be written (including the path, if required; "FILE_OUT.omv"; default: ""); if empty, the resulting data frame is
#'               returned instead
#' @param dtaTtl Character vector with a title to be added to the data set (see Details; default: "")
#' @param dtaDsc Description of the data set, either as character vector (HTML-formatted) or as named list with the entries "description", "variables",
#'               "references", and "license" (see Details; default: "")
#' @param lngDsc Language of the description (localizes the description components: "Description", "Variables", "References", and "License";
#'               default: "EN")
#' @param usePkg Name of the package: "foreign" or "haven" that shall be used to read SPSS, Stata and SAS files; "foreign" is the default (it comes with
#'               base R), but "haven" is newer and more comprehensive
#' @param selSet Name of the data set that is to be selected from the workspace (only applies when reading .RData-files)
#' @param ...    Additional arguments passed on to methods; see Details below
#'
#' @return a data frame (only returned if `fleOut` is empty) where a description and a title are added to the input data
#'
#' @details
#' * The aim of this function is to add a title and a data set description to jamovi data files. Two typical use cases would be (1) to help creating data sets
#'   to be used in teaching (i.e., either creating or using data sets in R, and afterwards adding a description to those), and (2) to provide "properly
#'   described" data when publishing in a repository such as the OSF).
#' * NB: The data set should not contain any existing analyses. These will be overwritten (a  warning is issued informing you about that).
#' * `dtaTtl` is a title for the dataset (at the top of the results output, i.e., that title which initially is "Results" when you create a new data set in
#'   jamovi).
#' * `dtaDsc` can either be a character vector (with length = 1) containing HTML-formatted text that describes the data set (see `chrDsc` in the examples for
#'   HTML tags that are currently implemented; putting "unformatted" text is not a problem, but then the result is just plain text without formatting).
#'   Alternatively, `dtaDcs` can be a named list with the entries `description`, `variables`, `references`, `license`. All entries except from `variables`
#'   contain character vectors (length = 1); `variables` shall be a named list with the variable name as name and a description what the variable contains as
#'   entry. `description` and `variables` must be given, `references` and `license` can be left blank (""; but the names must be present in the list). An
#'   example for both a named list with a description (`lstDsc`), as well as a character vector with all HTML tags that are implemented (`chrDsc`) can be found
#'   in the examples below.
#' * The ellipsis-parameter (`...`) can be used to submit arguments / parameters to the functions that are used for reading and writing the data. By clicking
#'   on the respective function under "See also", you can get a more detailed overview over which parameters each of those functions take. The functions are:
#'   `read_omv` and `write_omv` (for jamovi-files), `read.table` (for CSV / TSV files; using similar defaults as `read.csv` for CSV and `read.delim` for TSV
#'   which both are based upon `read.table`), `load` (for .RData-files), `readRDS` (for .rds-files), `read_sav` (needs the R-package `haven`) or `read.spss`
#'   (needs the R-package `foreign`) for SPSS-files, `read_dta` (`haven`) / `read.dta` (`foreign`) for Stata-files, `read_sas` (`haven`) for SAS-data-files,
#'   and `read_xpt` (`haven`) / `read.xport` (`foreign`) for SAS-transport-files. If you would like to use `haven`, you may need to install it using
#'   `install.packages("haven", dep = TRUE)`.
#'
#' @seealso `describe_omv` internally uses the following functions for reading and writing data files in different formats: [jmvReadWrite::read_omv()] and
#'   [jmvReadWrite::write_omv()] for jamovi-files, [utils::read.table()] for CSV / TSV files, [load()] for reading .RData-files, [readRDS()] for .rds-files,
#'   [haven::read_sav()] or [foreign::read.spss()] for SPSS-files, [haven::read_dta()] or [foreign::read.dta()] for Stata-files, [haven::read_sas()] for
#'   SAS-data-files, and [haven::read_xpt()] or [foreign::read.xport()] for SAS-transport-files.
#'
#' @examples
#' dtaFrm <- jmvReadWrite::ToothGrowth[, c("len", "supp", "dose")]
#' nmeOut <- tempfile(fileext = ".omv")
#'
#' # the paste's underneath are only for readability (without them, the vignettes are misformatted)
#' lstDsc <- list(description = paste("The response is the length of odontoblasts (cells responsible",
#'                                    "for tooth growth) in 60 guinea pigs. Each animal received one",
#'                                    "of three dose levels of vitamin C (0.5, 1, and 2 mg / day) by",
#'                                    "one of two delivery methods, orange juice or ascorbic acid (a",
#'                                    "form of vitamin C and coded as VC)."),
#'                variables = list(len  = "Tooth length",
#'                                 supp = "Supplement type (VC or OJ)",
#'                                 dose = "Dose (in milligrams / day"),
#'                references = paste("Crampton, E. W. (1947). The growth of the odontoblast of the",
#'                                   "incisor teeth as a criterion of vitamin C intake of the guinea",
#'                                   "pig. <em>The Journal of Nutrition, 33</em>(5), 491-504.",
#'                                   "https://doi.org/10.1093/jn/33.5.491"),
#'                license = "")
#' jmvReadWrite::describe_omv(dtaInp = dtaFrm, fleOut = nmeOut, dtaTtl = "ToothGrowth",
#'   dtaDsc = lstDsc)
#' # don't include the unlink, if you copy the code and want to look at the resulting output file
#' unlink(nmeOut)
#'
#' # the code underneath should cover all formatting options jamovi is able to use (paste0 is only
#' # for readability)
#' chrDsc <- paste0("<p><strong>Trial - all formattings:</strong><br/><strong>bold</strong><br/>",
#'                  "<strong><em>bold, italics</em></strong><br/><em>italics</em><br/><u>underlined",
#'                  "</u><br/><s>strikethrough</s><br/>C<sub>2</sub>H<sub>5</sub>OH<br/>R<sup>2",
#'                  "</sup><br/><span style=\"background-color:#e60000\">background colour: red",
#'                  "</span><br/><span style=\"color:#e60000\">foreground color: red",
#'                  "</span></p><p class=\"ql-align-center\">centered</p><p class=\"ql-align-right\">",
#'                  "right</p><p class=\"ql-align-justify\">justify justify justify justify justify ",
#'                  "justify justify justify justify justify justify justify justify justify justify ",
#'                  "justify justify justify justify justify justify justify justify justify justify",
#'                  "</p><p><br/></p><ol><li>numbered list</li><li>numbered list</li></ol><p><br/>",
#'                  "</p><ul><li>bullet point</li><li>bullet point</li></ul><p class=\"ql-indent-1\">",
#'                  "indented once</p><p class=\"ql-indent-2\">indented twice</p><p ",
#'                  "class=\"ql-indent-1\">indented once</p><p>Formula: <span class=\"ql-formula\">",
#'                  "e=mc^2</span></p><pre>Preformatted</pre><p>normal again</p><h2>Heading</h2>")
#' jmvReadWrite::describe_omv(dtaInp = dtaFrm, fleOut = nmeOut, dtaTtl = "ToothGrowth",
#'   dtaDsc = chrDsc)
#' unlink(nmeOut)
#'
#'
#' @export describe_omv
#'
describe_omv <- function(dtaInp = NULL, fleOut = "", dtaTtl = c(), dtaDsc = c(), lngDsc = "EN", usePkg = c("foreign", "haven"), selSet = "", ...) {

    # check the input parameters: either dtaTtl or dtaDsc need to be given (and in the correct format)
    dscPrm(dtaTtl, dtaDsc)

    # check and import input data set (either as data frame or from a file)
    if (!is.null(list(...)[["fleInp"]])) stop("Please use the argument dtaInp instead of fleInp.")
    dtaFrm <- inp2DF(dtaInp = dtaInp, getSyn = TRUE, getHTM = TRUE, usePkg = usePkg, selSet = selSet, ...)

    if (utils::packageVersion("jmvcore") < "2.4.3") {
        warning("jmvcore version 2.4.3 (or higher) is required for using describe_omv.\n\n")
        return(invisible(NULL))
    }
    if (!jmvPtB())
        stop("The R-packages RProtoBuf and jmvcore must be installed for using describe_omv (see warnings() for further details).")

    # check whether dtaDsc is a list, and if so, convert it to a HTML desription;
    # in addition: add a description if there doesn't exist one already
    if (is.list(dtaDsc)) {
        if (length(setdiff(names(dtaDsc[["variables"]]), names(dtaFrm))) != 0 ||
            length(setdiff(names(dtaFrm), names(dtaDsc[["variables"]]))) != 0)
            stop("The variable description (in dtaDsc) should contain all variables in the data set and no variables not contained in it.")
        for (crrClm in names(dtaDsc[["variables"]])) {
            if (!any(c("jmv-desc", "description") %in% names(attributes(dtaFrm[, crrClm])))) {
                attr(dtaFrm[, crrClm], "jmv-desc")    <- dtaDsc[["variables"]][[crrClm]]
                attr(dtaFrm[, crrClm], "description") <- dtaDsc[["variables"]][[crrClm]]
            }
        }
        dtaDsc <- crtHTM(dtaDsc, defHdr(lngDsc))
    }

    # check whether the data set contains analyses and, if so warn that they will be overwritten
    if (!is.null(attr(dtaFrm, "syntax")) && length(attr(dtaFrm, "syntax")) > 0) {
        warning("The data set contains analyses. Those will be overwritten.")
        if (is.character(dtaInp) && identical(dtaInp, fleOut)) return(invisible(NULL))
        attr(dtaFrm, "syntax")   <- NULL
        attr(dtaFrm, "protobuf") <- NULL
        attr(dtaFrm, "HTML")     <- htmTxt()
    }

    # check whether the data set contains the attribute HTML, if not, add text of an empty index.html
    if (is.null(attr(dtaFrm, "HTML"))) attr(dtaFrm, "HTML") <- htmTxt()

    # add title and description in HMTL
    attr(dtaFrm, "HTML") <- addHTM(attr(dtaFrm, "HTML"), dtaTtl, dtaDsc)

    # add title and description as protobuffer (calls optPtB to create protobuffer options)
    attr(dtaFrm, "protobuf")[["01 empty/analysis"]] <-
      RProtoBuf::new(jamovi.coms.AnalysisResponse, analysisId = 1, name = "empty", ns = "jmv", options = optPtB(dtaTtl, dtaDsc),
        results = RProtoBuf::new(jamovi.coms.ResultsElement, title = "Results", status = 3, group = RProtoBuf::new(jamovi.coms.ResultsGroup)),
        status = 3, index = 1, title = "Results", hasTitle = TRUE)

    # rtnDta in globals.R (unified function to either write the data frame, open it in a new jamovi session or return it)
    rtnDta(dtaFrm = dtaFrm, fleOut = fleOut, dtaTtl = jmvTtl("_desc"), wrtPtB = TRUE, ...)
}

# =================================================================================================
# helper functions (in alphabetical order)
# =================================================================================================

# add / replace HTML for title and description within index.html
addHTM <- function(crrHTM = "", dtaTtl = c(), dtaDsc = c()) {
    vldHTM <- which(!grepl("^\\s*$", crrHTM))
    fndTtl <- grep("<h1.*?>", crrHTM[vldHTM])[1]
    if (length(dtaTtl) > 0 && nzchar(dtaTtl)) {
        splLne <- splHTM(crrHTM[vldHTM[fndTtl + 0]])
        splLne[grep("<h1.*?>", splLne) + 1] <- dtaTtl
        crrHTM[vldHTM[fndTtl + 0]] <- paste0(splLne, collapse = "")
    }
    if (length(dtaDsc) > 0 && nzchar(dtaDsc)) {
        # issue warning if the target line hasn't the expected format
        if (!grepl("^\\s*<p.*?></p>\\s*$", crrHTM[vldHTM[fndTtl + 1]])) {
            warning(sprintf("Unexpected HTML to be replaced with description: %s\n\n", crrHTM[vldHTM[fndTtl + 1]]))
        }
        # assign description embedded in <div> / </div> tags
        crrHTM[vldHTM[fndTtl + 1]] <- paste0("<div class=\"note\">", clnHTM(dtaDsc, toHTM = TRUE), "</div>")
    }

    return(crrHTM)
}

# ensure that the input line is concatened (if required) and remove leading and trailing <p> / </p> tags,
# then split the input (into HTML tags and content) and replace characters that have a HTML-function (e.g., "/")
# in the content
clnHTM <- function(inpLne = c(), toHTM = FALSE) {
    inpLne <- vapply(inpLne, function(x) paste0(ifelse(grepl("^\\s*<p.*?>|^\\s*<li.*?>", x), "", "<p>"), x, ifelse(grepl("</p.*?>\\s*$|</li.*?>\\s*$", x), "", "</p>")),
                character(1), USE.NAMES = FALSE)
    splLne <- splHTM(paste0(inpLne, collapse = ""))
    selCnt <- !grepl("<.*?>", splLne)
    splLne[selCnt] <- rplHTM(splLne[selCnt], toHTM)
    return(paste0(splLne, collapse = ""))
}

# create HTML from the list-version of dtaDsc
crtHTM <- function(inpDsc = NULL, nmeHdr = NULL) {
    outDsc <- c()
    outDsc <- paste0(outDsc,      "<p><strong>", nmeHdr["description"], ":</strong></p>", clnHTM(inpDsc[["description"]]))
    outDsc <- paste0(outDsc, "<p><br/><strong>", nmeHdr["variables"],   ":</strong></p><ul>")
    for (varNme in names(inpDsc[["variables"]])) {
         outDsc <- paste0(outDsc, clnHTM(paste0("<li><strong><em>", varNme, ":</em></strong> ", inpDsc[["variables"]][[varNme]], "</li>")))
    }
    outDsc <- paste0(outDsc, "</ul>")
    if (length(inpDsc[["references"]]) > 0 && all(nzchar(inpDsc[["references"]]))) {
        outDsc <- paste0(outDsc, "<p><br/><strong>", nmeHdr["references"], ":</strong></p>", clnHTM(inpDsc[["references"]]))
    }
    if (length(inpDsc[["license"]]) > 0    && all(nzchar(inpDsc[["license"]]))) {
        outDsc <- paste0(outDsc, "<p><br/>", clnHTM(paste0("<em>", inpDsc[["license"]], "</em>")))
    }

    return(outDsc)
}

# define headers for crtHTM (in different languages)
defHdr <- function(crrLng = "EN") {
    if        (crrLng == "DE") {
        c(description = "Beschreibung", variables = "Variablen", references = "Referenzen")
    } else if (crrLng == "JP") {
#       NB: UTF-characters must be converted using stringi::stri_escape_unicode()
        c(description = "\u8aac\u660e", variables = "\u5909\u6570", references = "\u5f15\u7528\u6587\u732e")
    } else if (crrLng == "NB") {
        c(description = "Beskrivelse", variables = "Variabler", references = "Referanser")
    } else {  # defaults to English
        c(description = "Description", variables = "Variables", references = "References")
    }
}

# define some common licenses and make them translateable
defLic <- function(licNme = "", crrLng = "EN", licHld = "") {
    if        (licNme == "CC0") {
        if        (crrLng == "DE") {
            paste("Dieser Datensatz ist uneingeschr\u00e4nkt zug\u00e4nglich, und der Autor hat auf alle seine Rechte an dem",
                  "Werk verzichtet. Sie k\u00f6nnen den Datensatz daher ohne vorherige Genehmigung kopieren, ver\u00e4ndern,",
                  "und verbreiten, auch f\u00fcr kommerzielle Zwecke. Wenn der Datensatz auf empirischen Daten basiert,",
                  "sind die Autoren im Abschnitt 'Referenzen' des Datensatzes angegeben. Zitieren Sie in diesem Fall",
                  "bitte die entsprechenden Referenzen, wenn Sie den Datensatz verwenden.")
        } else if (crrLng == "EN") {
            paste("This data set is in the public domain, and the author has waived all of his or her rights to the",
                  "work. You can therefore copy, modify and distribute the data set, even for commercial purposes,",
                  "all without asking permission. If the data set is based upon empirical work, the authors are",
                  "given in the 'References'-section of the data set. In such case, please cite the respective",
                  "reference(s) when using the dataset.")
        } else if (crrLng == "JP") {
            stop(sprintf("No translation available (yet) for %s.", crrLng))
        } else if (crrLng == "NB") {
            paste("Datasettet er offentlig tilgjengelig, og forfatteren har fraskrevet seg alle rettigheter til",
                  "datasettet. Du kan derfor kopiere, endre og distribuere datasettet, ogs\u00e5 til kommersielle",
                  "form\u00e5l, uten \u00e5 be om tillatelse. Hvis datasettet er empirisk, er forfatterne oppgitt under",
                  "'Referanser', vennligst siter de aktuelle referansene hvis det er tilfellet.")
        } else {
            stop(sprintf("No translation available (yet) for %s.", crrLng))
        }
    } else if (licNme == "DT_CC4-BY-NC-ND") {
        if        (crrLng == "DE") {
            paste0("Dieser Datensatz enth\u00e4lt Daten einer wissenschaftlichen Studie, deren Urheberrecht bei den ",
                   "Autoren der Studie liegt. Ohne die ausdr\u00fcckliche Zustimmung der Autoren ",
                   rep(paste0("und von ", licHld, " "), nzchar(licHld)), "darf dieser Datensatz nicht zu ",
                   "kommerziellen Zwecken verbreitet, bearbeitet oder verwendet werden ohne die Quelle anzugeben ",
                   "(d. h. der Datensatz unterliegt den Bedingungen der CC BY-NC-ND-Lizenz).")
        } else if (crrLng == "EN") {
            paste0("This data set contains data of a scientific study and the study authors therefore own the ",
                  "copyright. Without the study authors` ", rep(paste0("and ", licHld, "`s "), nzchar(licHld)),
                  "explicit consent, this data set may not be distributed for commercial purposes, not be edited, ",
                  "and not be used without acknowledging its source (i.e., the terms of a CC BY-NC-ND license).")
        } else if (crrLng == "JP") {
            stop(sprintf("No translation available (yet) for %s.", crrLng))
        } else if (crrLng == "NB") {
            paste0("Dette datasettet inneholder data fra en vitenskapelig studie, og studieforfatterne eier derfor ",
                   "opphavsretten. Uten uttrykkelig samtykke fra studieforfatterne ",
                   rep(paste0("og ", licHld, " "), nzchar(licHld)), "kan dette datasettet ikke distribueres for ",
                   "kommersielle form\u00e5l, ikke redigeres og ikke brukes uten \u00e5 oppgi kilden (dvs. ",
                   "vilk\u00e5rene i en CC BY-NC-ND-lisens).")
        } else {
            stop(sprintf("No translation available (yet) for %s.", crrLng))
        }
    } else if (licNme == "FC_CC4-BY-NC-ND") {
        if        (licHld == "")
            stop("When using this license, a license holder needs to be defined.")
        if        (crrLng == "DE") {
            paste0("Dieser Datensatz wurde von ", licHld, " erstellt, der somit das Urheberrecht daran besitzt. ",
                   "Ohne die ausdr\u00fcckliche Zustimmung von ", licHld, " darf dieser Datensatz nicht f\u00fcr ",
                   "kommerzielle Zwecke verbreitet, nicht bearbeitet und nicht ohne Angabe der Quelle verwendet ",
                   "werden (d. h. der Datensatz unterliegt den Bedingungen der CC BY-NC-ND-Lizenz).")
        } else if (crrLng == "EN") {
            paste0("This data set was constructed by ", licHld, ", who therefore owns the copyright. Without the ",
                   "explicit consent of ", licHld, ", this data set may not be distributed for commercial purposes, ",
                   "not be edited, and not be used without acknowledging its source (i.e., the terms of the CC ",
                   "BY-NC-ND license).")
        } else if (crrLng == "JP") {
            stop(sprintf("No translation available (yet) for %s.", crrLng))
        } else if (crrLng == "NB") {
            paste0("Datasettet er utarbeidet av ", licHld, ", som derfor eier opphavsretten. Uten uttrykkelig ",
                   "samtykke fra ", licHld, " kan dette datasettet ikke distribueres for kommersielle form\u00e5l, ",
                   "redigeres eller brukes uten \u00e5 oppgi kilden (dvs. vilk\u00e5rene i CC BY-NC-ND-lisens).")
        } else {
            stop(sprintf("No translation available (yet) for %s.", crrLng))
        }
    } else if (licNme %in% c("RP_GPL2", "RP_GPL3", "RP_AGPL3", "RP_LGPL3")) {
        licFlN <- c(rep("GNU General Public License 2.x",        licNme == "RP_GPL2"),
                    rep("GNU General Public License 3.0",        licNme == "RP_GPL3"),
                    rep("GNU Affero General Public License 3.0", licNme == "RP_AGPL3"),
                    rep("GNU Lesser General Public License 3.0", licNme == "RP_LGPL3"))
        if        (licHld == "")
            stop("When using this license, the R-package where the data are originating from needs to be defined.")
        if        (crrLng == "DE") {
            paste0("Dieser Datensatz ist Teil des R-Pakets '", licHld, ", das unter den Bedingungen der ", licFlN, " ",
                   "ver\u00f6ffentlicht wurde. Sie k\u00f6nnen die Daten sowohl privat als auch kommerziell nutzen, ",
                   "verbreiten oder ver\u00e4ndern. Bei der Verwendung des Datensatzes m\u00fcssen Sie die  Quelle ",
                   "angeben und alle \u00e4nderungen, die Sie vornehmen, unter derselben Lizenz ",
                   "ver\u00f6ffentlichen. Wenn der Datensatz auf empirischer Forschung basiert, sind die Autoren im ",
                   "Abschnitt 'Referenzen' angegeben. Zitieren Sie in diesem Fall die Studie, wenn Sie den Datensatz ",
                   "verwenden.")
        } else if (crrLng == "EN") {
            paste0("This data set was provided as part of the R-package '", licHld, "', which is published under ",
                   "the terms of the ", licFlN, ". You may use the data both privately and commercially, distribute ",
                   "or modify them. When using the data set, you need to disclose the source, and publish any ",
                   "modifications you may make under the same license. If the data set is based on empirical data, ",
                   "the authors are given in the 'References'-section. In such case, please  cite them when using ",
                   "the dataset.")
        } else if (crrLng == "JP") {
            stop(sprintf("No translation available (yet) for %s.", crrLng))
        } else if (crrLng == "NB") {
            paste0("Dette datasettet er en del av R-pakken '", licHld, "', som er publisert under vilk\u00e5rene i ",
                   licFlN, ". Du kan bruke dataene b\u00e5de privat og kommersielt, distribuere eller endre dem. ",
                   "N\u00e5r du bruker datasettet, m\u00e5 du oppgi kilden og publisere eventuelle endringer du ",
                   "gj\u00f8r under samme lisens. Hvis datasettet er basert p\u00e5 empiriske data, er forfatterne ",
                   "oppgitt under 'Referanser'. Vennligst oppgi referansene i slike tilfeller n\u00e5r du bruker ",
                   "datasettet.")
        } else {
            stop(sprintf("No translation available (yet) for %s.", crrLng))
        }
    } else {
        c()
    }
}

# check input parameters
dscPrm <- function(dtaTtl = c(), dtaDsc = NULL) {
    if ((!is.character(dtaTtl) || length(dtaTtl) != 1 || !nzchar(dtaTtl)) &&
        (((!is.character(dtaDsc) || length(dtaDsc) != 1 || !nzchar(dtaDsc)) &&
          (!is.list(dtaDsc) || !all(c("description", "variables", "references", "license") %in% names(dtaDsc)))))) {
        stop("Calling describe_omv requires either the parameter dtaTtl (character vector) or the parameter dtaDsc (character vector or named list).")
    }
}

# transform HTML tags into attributes (bold, italic, heading, etc.)
getAtt <- function(vecChr = c(), vecPos = NA, lstAtt = list()) {
    if (grepl("<p>|</p>", vecChr[vecPos])) return(lstAtt)
    prvAtt <- lstAtt

    # bold
    if (grepl("<strong>|<b>",       vecChr[vecPos])) lstAtt[["bold"]] <- TRUE
    if (grepl("</strong>|</b>",     vecChr[vecPos])) lstAtt[["bold"]] <- NULL
    # italic
    if (grepl("<em>|<i>",           vecChr[vecPos])) lstAtt[["italic"]] <- TRUE
    if (grepl("</em>|</i>",         vecChr[vecPos])) lstAtt[["italic"]] <- NULL
    # underline
    if (grepl("<u>",                vecChr[vecPos])) lstAtt[["underline"]] <- TRUE
    if (grepl("</u>",               vecChr[vecPos])) lstAtt[["underline"]] <- NULL
    # strikethrough
    if (grepl("<s>",                vecChr[vecPos])) lstAtt[["strike"]] <- TRUE
    if (grepl("</s>",               vecChr[vecPos])) lstAtt[["strike"]] <- NULL
    # preformatted / code-block
    if (grepl("<pre.*?>|<code.*?>", vecChr[vecPos])) lstAtt[["code-block"]] <- TRUE
    if (grepl("</pre>|</code>",     vecChr[vecPos])) lstAtt[["code-block"]] <- NULL
    # superscript
    if (grepl("<sup>",              vecChr[vecPos])) lstAtt[["script"]] <- "super"
    if (grepl("</sup>",             vecChr[vecPos])) lstAtt[["script"]] <- NULL
    # subscript
    if (grepl("<sub>",              vecChr[vecPos])) lstAtt[["script"]] <- "sub"
    if (grepl("</sub>",             vecChr[vecPos])) lstAtt[["script"]] <- NULL
    # heading
    if (grepl("<h\\d.*?>",          vecChr[vecPos])) lstAtt[["header"]] <- as.integer(gsub("<h(\\d).*?>", "\\1", vecChr[vecPos]))
    if (grepl("</h\\d>",            vecChr[vecPos])) lstAtt[["header"]] <- NULL
    # links
    if (grepl("<a href=.*?>",       vecChr[vecPos])) lstAtt[["link"]] <- gsub("<a href=\"(.*)\"\\s+.*?>", "\\1", vecChr[vecPos])
    if (grepl("</a>",               vecChr[vecPos])) lstAtt[["link"]] <- NULL
    # ordered list
    if (grepl("<ol>",               vecChr[vecPos])) lstAtt[["list"]] <- "ordered"
    if (grepl("</ol>",              vecChr[vecPos])) lstAtt[["list"]] <- NULL
    # bullet list
    if (grepl("<ul>",               vecChr[vecPos])) lstAtt[["list"]] <- "bullet"
    if (grepl("</ul>",              vecChr[vecPos])) lstAtt[["list"]] <- NULL
    # list entries
    if (grepl("<li>",               vecChr[vecPos])) attr(lstAtt[["list"]], "lstEnt") <- TRUE
    if (grepl("</li>",              vecChr[vecPos])) attr(lstAtt[["list"]], "lstEnt") <- FALSE
    # alignment
    if (grepl("class=\".*?ql-align-\\w+|style=\".*?text-align:\\s*\\w+", vecChr[vecPos])) {
        lstAtt[["align"]] <- gsub(".*?ql-align-(\\w+).*", "\\1", gsub(".*?text-align:\\s*(\\w+).*", "\\1", vecChr[vecPos]))
        attr(lstAtt[["align"]], "endPos") <- nxtAtt(vecChr, vecPos, "class=|style=")
    } else if (chkAtt(lstAtt[["align"]], "endPos") && attr(lstAtt[["align"]], "endPos") <= vecPos) {
        lstAtt[["align"]] <- NULL
    }
    # indent
    if (grepl("class=\".*?ql-indent-\\d", vecChr[vecPos])) {
        lstAtt[["indent"]] <- as.integer(gsub(".*?ql-indent-(\\d+).*", "\\1", vecChr[vecPos]))
        attr(lstAtt[["indent"]], "endPos") <- nxtAtt(vecChr, vecPos, "class=")
    } else if (chkAtt(lstAtt[["indent"]], "endPos") && attr(lstAtt[["indent"]], "endPos") <= vecPos) {
        lstAtt[["indent"]] <- NULL
    }
    # formula
    if (grepl("class=\".*?ql-formula", vecChr[vecPos])) {
        lstAtt[["formula"]] <- TRUE
        attr(lstAtt[["formula"]], "endPos") <- nxtAtt(vecChr, vecPos, "class=")
    } else if (chkAtt(lstAtt[["formula"]], "endPos") && attr(lstAtt[["formula"]], "endPos") <= vecPos) {
        lstAtt[["formula"]] <- NULL
    }
    # colours
    if (grepl("style=\".*?background-color:", vecChr[vecPos])) {
        lstAtt[["background"]] <- gsub(".*?background-color:.*(\\#[0-9,a-f]+).*", "\\1", vecChr[vecPos])
        attr(lstAtt[["background"]], "endPos") <- nxtAtt(vecChr, vecPos, "style=")
    } else if (chkAtt(lstAtt[["background"]], "endPos") && attr(lstAtt[["background"]], "endPos") <= vecPos) {
        lstAtt[["background"]] <- NULL
    }
    if (grepl("style=\".*?color:", vecChr[vecPos]) && !grepl("style=\".*?background-color:", vecChr[vecPos])) {
        lstAtt[["color"]] <- gsub(".*?color:.*(\\#[0-9,a-f]+).*", "\\1", vecChr[vecPos])
        attr(lstAtt[["color"]], "endPos") <- nxtAtt(vecChr, vecPos, "style=")
    } else if (chkAtt(lstAtt[["color"]], "endPos") && attr(lstAtt[["color"]], "endPos") <= vecPos) {
        lstAtt[["color"]] <- NULL
    }

    if (identical(prvAtt, lstAtt)) {
        warning(sprintf("The HTML tag \"%s\" isn't implemented, and hence the jamovi formatting attributes remain unchanged.\n", vecChr[vecPos]))
    }

    return(lstAtt)
}

# find the next HTML tag of a particular type
nxtAtt <- function(vecChr = c(), vecPos = NA, addTrm = c()) {
    grep(paste(c(paste0(c("</", "<"), gsub("<(\\w+)\\s+.*", "\\1", vecChr[vecPos])), addTrm), collapse = "|"),
      vecChr[seq(vecPos + 1, length(vecChr))])[1] + vecPos
}

# store title and description as protobuffer
optPtB <- function(dtaTtl = "", dtaDsc = c()) {
    # create protobuffer for the title
    if (length(dtaTtl) > 0 && nzchar(dtaTtl)) {
        ttlPtB <- var2PB(dtaTtl)
    } else {
        ttlPtB <- NULL
    }
    # create protobuffer for the description
    if (length(dtaDsc) > 0 && nzchar(dtaDsc)) {
        # gsub("^\\s*<p.*?>|</p>\\s*$", "",
        splDsc <- splHTM(gsub("(</h\\d>|</p>|</pre>|</code>)", "\n\\1", gsub("</li>", "</li>\n", gsub("<br\\s*/>|<br>", "\n", rplHTM(dtaDsc)))))
        tgtDsc <- !grepl("<.*?>", splDsc)
        attDsc <- htmPtB <- list()
        for (i in seq_along(splDsc)) {
            if (tgtDsc[i]) {
                if (is.null(attDsc[["formula"]]) || attDsc[["formula"]] == FALSE) crrIns <- splDsc[i] else crrIns <- list(formula = splDsc[i])
                htmPtB[[sum(tgtDsc[seq(i)])]] <- var2PB(c(prpAtt(attDsc), list(insert = crrIns)))
            } else {
                attDsc <- getAtt(splDsc, i, attDsc)
            }
        }

        dscPtB <- RProtoBuf::new(jamovi.coms.AnalysisOption, c = RProtoBuf::new(jamovi.coms.AnalysisOptions, options =
                      RProtoBuf::new(jamovi.coms.AnalysisOption, c = RProtoBuf::new(jamovi.coms.AnalysisOptions, options = htmPtB)),
                    hasNames = TRUE, names = "ops"))
    } else {
        dscPtB <- NULL
    }

    # assemble the two protobuffers (for title, and description) into a AnalysisOptions protobuffer
    RProtoBuf::new(jamovi.coms.AnalysisOptions, options = c(dscPtB, ttlPtB), hasNames = TRUE,
      names = c(rep("results//topText", !is.null(dscPtB)), rep("results//heading", !is.null(ttlPtB))))
}

prpAtt <- function(lstAtt = NULL) {
    # for list entries (<li>), the attribute "list" needs to be removed (it is only attached to the "\n" at
    # the end), and formulas are encoded by wrapping the "insert" (i.e., the attribute itself must not be
    # transformed: [1] assign lstAtt to a temporary variable, [2] remove the "list" entry if inside a list,
    # [3] remove the formula attribute (if present), [4] wrap the arguments in a named list (under the name
    # "attributes"), if crrArg is empty, nothing is added in c(crrAtt, ...)
    crrAtt <- lstAtt
    if (utils::hasName(lstAtt, "list") && identical(attr(lstAtt[["list"]], "lstEnt"), TRUE)) crrAtt[["list"]] <- NULL
    if (utils::hasName(lstAtt, "formula")) crrAtt[["formula"]] <- NULL
    if (length(crrAtt) > 0) return(list(attributes = crrAtt)) else return(invisible(NULL))
}

rplHTM <- function(inpVec = c(), toHTM = FALSE) {
    if (toHTM) {
        return(gsub("\\/", "&#x2F;", inpVec))
    } else {
        return(gsub("&nbsp;", " ", gsub("&#x2F;", "/",   inpVec)))
    }
}

# split HTML into constituents (i.e., tags and content)
splHTM <- function(vecChr = "") {
        splChr <- gsub("^\\s+\\n|\\n\\s+$", "\n", strsplit(vecChr, "(?=[<>])", perl = TRUE)[[1]])
        rplDsc <- grep("<", splChr)
        # has HTML tags
        if (length(rplDsc) > 0) {
            if (!all(splChr[rplDsc + 2] == ">")) {
                stop(paste("Error when decoding HTML. Please check whether the HTML-string the you used for the description is valid (i.e., no typos, etc.).",
                           "If it is valid, send it to sebastian.jentschke@uib.no"))
            }
            splChr[rplDsc + 1] <- paste0("<", splChr[rplDsc + 1], ">")
            return(splChr[-c(rplDsc, rplDsc + 2)])
        } else {
            return(splChr)
        }
}

test_that("describe_omv works", {
    nmeOut <- tempfile(fileext = ".omv")
    lstDsc <- list(description = paste("The response is the length of odontoblasts (cells responsible for tooth growth) in 60 guinea pigs. Each animal received one of three dose",
                                       "levels of vitamin C (0.5, 1, and 2 mg / day) by one of two delivery methods, orange juice or ascorbic acid (a form of vitamin C and coded as VC)."),
                   variables = list(len = "Tooth length",
                                    supp = "Supplement type (VC or OJ)",
                                    dose = "Dose (in milligrams / day"),
                   references = paste("Crampton, E. W. (1947). The growth of the odontoblast of the incisor teeth as a criterion of vitamin C intake of the guinea pig.",
                                      "<em>The Journal of Nutrition, 33</em>(5), 491-504. https://doi.org/10.1093/jn/33.5.491"),
                   license = "The dataset was originally a part of the R-package datasets, and should therefore be licensed under GPL-3.")
    chrDsc <- paste0("<p><strong>Trial – all formattings:</strong><br/><strong>bold</strong><br/><strong><em>bold, italics</em></strong><br/><em>italics</em><br/><u>underlined</u><br/>",
                     "link: <a href=\"https://jamovi.org﻿﻿﻿\" target=\"_blank\">https://jamovi.org﻿﻿﻿</a><br/><s>strikethrough</s><br/>H<sub>2</sub>CO<sub>3</sub><br/>R<sup>2</sup><br/>",
                     "<span style=\"background-color:#e60000\">background colour: red</span><br/><span style=\"color:#e60000\">foreground color: red</span></p><p class=\"ql-align-center\">",
                     "centered</p><p class=\"ql-align-right\">right</p><p class=\"ql-align-justify\">justify justify justify justify justify justify justify justify justify justify ",
                     "justify justify justify justify justify justify justify justify justify justify justify justify justify justify justify</p><p><br/></p><ol><li>numbered list</li>",
                     "<li>numbered list</li></ol><p><br/></p><ul><li>bullet point</li><li>bullet point</li></ul><p class=\"ql-indent-1\">indented once</p><p class=\"ql-indent-2\">indented ",
                     "twice</p><p class=\"ql-indent-1\">indented once</p><p>Formula: <span class=\"ql-formula\">e=mc^2</span></p><pre>Preformatted</pre><p>normal again</p><h2>Heading</h2>")

    if (packageVersion("jmvcore") >= "2.4.3") {
        describe_omv(dtaInp = jmvReadWrite::ToothGrowth[, c("len", "supp", "dose")], fleOut = nmeOut, dtaTtl = "ToothGrowth", dtaDsc = lstDsc)
        expect_true(file.exists(nmeOut))
        expect_gt(file.info(nmeOut)$size, 1)
        expect_true(chkFle(nmeOut, isZIP = TRUE))
        expect_true(chkFle(nmeOut, fleCnt = "01 empty/analysis"))
        df4Chk <- read_omv(fleInp = nmeOut, getSyn = TRUE, getHTM = TRUE)
        expect_s3_class(df4Chk, "data.frame")
        expect_equal(dim(df4Chk), c(60, 3))
        expect_equal(vapply(df4Chk, typeof, character(1), USE.NAMES = FALSE), c("double", "integer", "double"))
        expect_equal(names(attributes(df4Chk)), c("names", "row.names", "class", "removedRows", "addedRows", "transforms", "protobuf", "HTML"))
        expect_type(attr(df4Chk, "protobuf"), "list")
        expect_length(attr(df4Chk, "protobuf"), 1)
        expect_s4_class(attr(df4Chk, "protobuf")[[1]], "Message")
        expect_length(attr(df4Chk, "protobuf")[[1]], 8)
        expect_s4_class(attr(df4Chk, "protobuf")[[1]]$options, "Message")
        expect_length(attr(df4Chk, "protobuf")[[1]]$options, 3)
        expect_equal(vapply(attr(df4Chk, "protobuf")[[1]]$options$options, class, character(1)), rep("Message", 2))
        expect_length(attr(df4Chk, "protobuf")[[1]]$options$options, 2)
        expect_equal(attr(df4Chk, "protobuf")[[1]]$options$names, c("results//topText", "results//heading"))
        expect_length(attr(df4Chk, "protobuf")[[1]]$options$options[[1]]$c$options[[1]]$c$options, 24)
        expect_equal(vapply(attr(df4Chk, "protobuf")[[1]]$options$options[[1]]$c$options[[1]]$c$options, class, character(1)), rep("Message", 24))
        expect_equal(attr(df4Chk, "protobuf")[[1]]$options$options[[1]]$c$options[[1]]$c$options[[1]]$c$names, c("attributes", "insert"))
        expect_equal(attr(df4Chk, "protobuf")[[1]]$options$options[[1]]$c$options[[1]]$c$options[[1]]$c$as.character(),
          paste0("options {\n  c {\n    options {\n      o: TRUE\n    }\n    hasNames: true\n    names: \"bold\"\n  }\n}\n",
                 "options {\n  s: \"Description:\"\n}\nhasNames: true\nnames: \"attributes\"\nnames: \"insert\"\n"))
        expect_equal(attr(df4Chk, "protobuf")[[1]]$options$options[[1]]$c$options[[1]]$c$options[[2]]$c$as.character(),
          "options {\n  s: \"\\n\"\n}\nhasNames: true\nnames: \"insert\"\n")
        expect_equal(attr(df4Chk, "protobuf")[[1]]$options$options[[2]]$s, "ToothGrowth")
        unlink(nmeOut)

        describe_omv(dtaInp = jmvReadWrite::ToothGrowth[, c("len", "supp", "dose")], fleOut = nmeOut, dtaTtl = "ToothGrowth", dtaDsc = chrDsc)
        expect_true(file.exists(nmeOut))
        expect_gt(file.info(nmeOut)$size, 1)
        expect_true(chkFle(nmeOut, isZIP = TRUE))
        expect_true(chkFle(nmeOut, fleCnt = "01 empty/analysis"))
        df4Chk <- read_omv(fleInp = nmeOut, getSyn = TRUE, getHTM = TRUE)
        expect_s3_class(df4Chk, "data.frame")
        expect_equal(dim(df4Chk), c(60, 3))
        expect_equal(vapply(df4Chk, typeof, character(1), USE.NAMES = FALSE), c("double", "integer", "double"))
        expect_equal(names(attributes(df4Chk)), c("names", "row.names", "class", "removedRows", "addedRows", "transforms", "protobuf", "HTML"))
        expect_type(attr(df4Chk, "protobuf"), "list")
        expect_length(attr(df4Chk, "protobuf"), 1)
        expect_s4_class(attr(df4Chk, "protobuf")[[1]], "Message")
        expect_length(attr(df4Chk, "protobuf")[[1]], 8)
        expect_s4_class(attr(df4Chk, "protobuf")[[1]]$options, "Message")
        expect_length(attr(df4Chk, "protobuf")[[1]]$options, 3)
        expect_equal(vapply(attr(df4Chk, "protobuf")[[1]]$options$options, class, character(1)), rep("Message", 2))
        expect_length(attr(df4Chk, "protobuf")[[1]]$options$options, 2)
        expect_equal(attr(df4Chk, "protobuf")[[1]]$options$names, c("results//topText", "results//heading"))
        expect_length(attr(df4Chk, "protobuf")[[1]]$options$options[[1]]$c$options[[1]]$c$options, 46)
        expect_equal(vapply(attr(df4Chk, "protobuf")[[1]]$options$options[[1]]$c$options[[1]]$c$options, class, character(1)), rep("Message", 46))
        expect_equal(attr(df4Chk, "protobuf")[[1]]$options$options[[1]]$c$options[[1]]$c$options[[1]]$c$names, c("attributes", "insert"))
        expect_equal(attr(df4Chk, "protobuf")[[1]]$options$options[[1]]$c$options[[1]]$c$options[[1]]$c$as.character(),
          paste0("options {\n  c {\n    options {\n      o: TRUE\n    }\n    hasNames: true\n    names: \"bold\"\n  }\n}\n",
                 "options {\n  s: \"Trial \\342\\200\\223 all formattings:\"\n}\nhasNames: true\nnames: \"attributes\"\nnames: \"insert\"\n"))
        expect_equal(attr(df4Chk, "protobuf")[[1]]$options$options[[1]]$c$options[[1]]$c$options[[2]]$c$as.character(),
          "options {\n  s: \"\\n\"\n}\nhasNames: true\nnames: \"insert\"\n")
        expect_equal(attr(df4Chk, "protobuf")[[1]]$options$options[[2]]$s, "ToothGrowth")
        unlink(nmeOut)

        describe_omv(dtaInp = jmvReadWrite::ToothGrowth[, c("len", "supp", "dose")], fleOut = nmeOut, dtaTtl = "ToothGrowth")
        expect_true(file.exists(nmeOut))
        expect_gt(file.info(nmeOut)$size, 1)
        expect_true(chkFle(nmeOut, isZIP = TRUE))
        expect_true(chkFle(nmeOut, fleCnt = "01 empty/analysis"))
        df4Chk <- read_omv(fleInp = nmeOut, getSyn = TRUE, getHTM = TRUE)
        expect_s3_class(df4Chk, "data.frame")
        expect_equal(dim(df4Chk), c(60, 3))
        expect_equal(vapply(df4Chk, typeof, character(1), USE.NAMES = FALSE), c("double", "integer", "double"))
        expect_equal(names(attributes(df4Chk)), c("names", "row.names", "class", "removedRows", "addedRows", "transforms", "protobuf", "HTML"))
        expect_type(attr(df4Chk, "protobuf"), "list")
        expect_length(attr(df4Chk, "protobuf"), 1)
        expect_s4_class(attr(df4Chk, "protobuf")[[1]], "Message")
        expect_length(attr(df4Chk, "protobuf")[[1]], 8)
        expect_s4_class(attr(df4Chk, "protobuf")[[1]]$options, "Message")
        expect_length(attr(df4Chk, "protobuf")[[1]]$options, 3)
        expect_equal(vapply(attr(df4Chk, "protobuf")[[1]]$options$options, class, character(1)), rep("Message", 1))
        expect_length(attr(df4Chk, "protobuf")[[1]]$options$options, 1)
        expect_equal(attr(df4Chk, "protobuf")[[1]]$options$names, c("results//heading"))
        expect_equal(attr(df4Chk, "protobuf")[[1]]$options$options[[1]]$s, "ToothGrowth")
        unlink(nmeOut)

        describe_omv(dtaInp = jmvReadWrite::ToothGrowth[, c("len", "supp", "dose")], fleOut = nmeOut, dtaDsc = chrDsc)
        expect_true(file.exists(nmeOut))
        expect_gt(file.info(nmeOut)$size, 1)
        expect_true(chkFle(nmeOut, isZIP = TRUE))
        expect_true(chkFle(nmeOut, fleCnt = "01 empty/analysis"))
        df4Chk <- read_omv(fleInp = nmeOut, getSyn = TRUE, getHTM = TRUE)
        expect_s3_class(df4Chk, "data.frame")
        expect_equal(dim(df4Chk), c(60, 3))
        expect_equal(vapply(df4Chk, typeof, character(1), USE.NAMES = FALSE), c("double", "integer", "double"))
        expect_equal(names(attributes(df4Chk)), c("names", "row.names", "class", "removedRows", "addedRows", "transforms", "protobuf", "HTML"))
        expect_type(attr(df4Chk, "protobuf"), "list")
        expect_length(attr(df4Chk, "protobuf"), 1)
        expect_s4_class(attr(df4Chk, "protobuf")[[1]], "Message")
        expect_length(attr(df4Chk, "protobuf")[[1]], 8)
        expect_s4_class(attr(df4Chk, "protobuf")[[1]]$options, "Message")
        expect_length(attr(df4Chk, "protobuf")[[1]]$options, 3)
        expect_equal(vapply(attr(df4Chk, "protobuf")[[1]]$options$options, class, character(1)), rep("Message", 1))
        expect_length(attr(df4Chk, "protobuf")[[1]]$options$options, 1)
        expect_equal(attr(df4Chk, "protobuf")[[1]]$options$names, c("results//topText"))
        expect_length(attr(df4Chk, "protobuf")[[1]]$options$options[[1]]$c$options[[1]]$c$options, 46)
        expect_equal(vapply(attr(df4Chk, "protobuf")[[1]]$options$options[[1]]$c$options[[1]]$c$options, class, character(1)), rep("Message", 46))
        expect_equal(attr(df4Chk, "protobuf")[[1]]$options$options[[1]]$c$options[[1]]$c$options[[1]]$c$names, c("attributes", "insert"))
        expect_equal(attr(df4Chk, "protobuf")[[1]]$options$options[[1]]$c$options[[1]]$c$options[[1]]$c$as.character(),
          paste0("options {\n  c {\n    options {\n      o: TRUE\n    }\n    hasNames: true\n    names: \"bold\"\n  }\n}\n",
                 "options {\n  s: \"Trial \\342\\200\\223 all formattings:\"\n}\nhasNames: true\nnames: \"attributes\"\nnames: \"insert\"\n"))
        expect_equal(attr(df4Chk, "protobuf")[[1]]$options$options[[1]]$c$options[[1]]$c$options[[2]]$c$as.character(),
          "options {\n  s: \"\\n\"\n}\nhasNames: true\nnames: \"insert\"\n")
        unlink(nmeOut)
    } else {
        expect_warning(expect_null(describe_omv(dtaInp = jmvReadWrite::ToothGrowth[, c("len", "supp", "dose")], fleOut = nmeOut, dtaTtl = "ToothGrowth", dtaDsc = lstDsc)))
        expect_false(file.exists(nmeOut))
        expect_warning(expect_null(describe_omv(dtaInp = jmvReadWrite::ToothGrowth[, c("len", "supp", "dose")], fleOut = nmeOut, dtaTtl = "ToothGrowth", dtaDsc = chrDsc)))
        expect_false(file.exists(nmeOut))
    }

    # code coverage
    expect_error(describe_omv(dtaInp = jmvReadWrite::ToothGrowth[, c("len", "supp", "dose")], fleOut = nmeOut),
      regexp = "^Calling describe_omv requires either the parameter dtaTtl \\(character vector\\) or the parameter dtaDsc \\(character vector or named list\\)\\.")
    expect_error(describe_omv(fleInp = file.path("..", "ToothGrowth.omv"), fleOut = nmeOut, dtaTtl = "ToothGrowth"), regexp = "Please use the argument dtaInp instead of fleInp\\.")
    expect_error(describe_omv(dtaInp = jmvReadWrite::ToothGrowth[, c("len", "supp")], fleOut = nmeOut, dtaTtl = "ToothGrowth", dtaDsc = lstDsc),
      regexp = "^The variable description \\(in dtaDsc\\) should contain all variables in the data set and no variables not contained in it\\.")
    expect_error(describe_omv(dtaInp = jmvReadWrite::ToothGrowth[, c("len", "supp", "supp2", "dose")], fleOut = nmeOut, dtaTtl = "ToothGrowth", dtaDsc = lstDsc),
      regexp = "^The variable description \\(in dtaDsc\\) should contain all variables in the data set and no variables not contained in it\\.")
    expect_equal(defHdr("EN"), c(description = "Description", variables = "Variables", references = "References"))
    expect_equal(defHdr("DE"), c(description = "Beschreibung", variables = "Variablen", references = "Referenzen"))
    expect_equal(defHdr("JP"), c(description = "説明", variables = "変数", references = "引用文献"))
    expect_equal(defHdr("NB"), c(description = "Beskrivelse", variables = "Variabler", references = "Referanser"))
    expect_equal(defLic("CC0"),
                 paste("This data set is in the public domain, and the author has waived all of his or her rights to",
                       "the work. You can therefore copy, modify and distribute the data set, even for commercial",
                       "purposes, all without asking permission. If the data set is based upon empirical work, the",
                       "authors are given in the 'References'-section of the data set. In such case, please cite the",
                       "respective reference(s) when using the dataset."))
    expect_equal(defLic("CC0", crrLng = "DE"),
                 paste("Dieser Datensatz ist uneingeschränkt zugänglich, und der Autor hat auf alle seine Rechte an",
                       "dem Werk verzichtet. Sie können den Datensatz daher ohne vorherige Genehmigung kopieren,",
                       "verändern, und verbreiten, auch für kommerzielle Zwecke. Wenn der Datensatz auf empirischen",
                       "Daten basiert, sind die Autoren im Abschnitt 'Referenzen' des Datensatzes angegeben. Zitieren",
                       "Sie in diesem Fall bitte die entsprechenden Referenzen, wenn Sie den Datensatz verwenden."))
    expect_equal(defLic("CC0", crrLng = "EN"),
                 paste("This data set is in the public domain, and the author has waived all of his or her rights to",
                       "the work. You can therefore copy, modify and distribute the data set, even for commercial",
                       "purposes, all without asking permission. If the data set is based upon empirical work, the",
                       "authors are given in the 'References'-section of the data set. In such case, please cite the",
                       "respective reference(s) when using the dataset."))
    expect_error(defLic("CC0", crrLng = "JP"), "No translation available \\(yet\\) for JP\\.")
    expect_equal(defLic("CC0", crrLng = "NB"),
                 paste("Datasettet er offentlig tilgjengelig, og forfatteren har fraskrevet seg alle rettigheter til",
                       "datasettet. Du kan derfor kopiere, endre og distribuere datasettet, også til kommersielle",
                       "formål, uten å be om tillatelse. Hvis datasettet er empirisk, er forfatterne oppgitt under",
                       "'Referanser', vennligst siter de aktuelle referansene hvis det er tilfellet."))
    expect_error(defLic("CC0", crrLng = "XX"), "No translation available \\(yet\\) for XX\\.")
    expect_equal(defLic("DT_CC4-BY-NC-ND"),
                 paste("This data set contains data of a scientific study and the study authors therefore own the",
                       "copyright. Without the study authors` explicit consent, this data set may not be distributed",
                       "for commercial purposes, not be edited, and not be used without acknowledging its source",
                       "(i.e., the terms of a CC BY-NC-ND license)."))
    expect_equal(defLic("DT_CC4-BY-NC-ND", crrLng = "DE"),
                 paste("Dieser Datensatz enthält Daten einer wissenschaftlichen Studie, deren Urheberrecht bei den",
                       "Autoren der Studie liegt. Ohne die ausdrückliche Zustimmung der Autoren darf dieser Datensatz",
                       "nicht zu kommerziellen Zwecken verbreitet, bearbeitet oder verwendet werden ohne die Quelle",
                       "anzugeben (d. h. der Datensatz unterliegt den Bedingungen der CC BY-NC-ND-Lizenz)."))
    expect_equal(defLic("DT_CC4-BY-NC-ND", crrLng = "EN"),
                 paste("This data set contains data of a scientific study and the study authors therefore own the",
                       "copyright. Without the study authors` explicit consent, this data set may not be distributed",
                       "for commercial purposes, not be edited, and not be used without acknowledging its source",
                       "(i.e., the terms of a CC BY-NC-ND license)."))
    expect_error(defLic("DT_CC4-BY-NC-ND", crrLng = "JP"), "No translation available \\(yet\\) for JP\\.")
    expect_equal(defLic("DT_CC4-BY-NC-ND", crrLng = "NB"),
                 paste("Dette datasettet inneholder data fra en vitenskapelig studie, og studieforfatterne eier",
                       "derfor opphavsretten. Uten uttrykkelig samtykke fra studieforfatterne kan dette datasettet",
                       "ikke distribueres for kommersielle formål, ikke redigeres og ikke brukes uten å oppgi kilden",
                       "(dvs. vilkårene i en CC BY-NC-ND-lisens)."))
    expect_error(defLic("DT_CC4-BY-NC-ND", crrLng = "XX"), "No translation available \\(yet\\) for XX\\.")
    expect_equal(defLic("DT_CC4-BY-NC-ND", licHld = "John Doe"),
                 paste("This data set contains data of a scientific study and the study authors therefore own the",
                       "copyright. Without the study authors` and John Doe`s explicit consent, this data set may not",
                       "be distributed for commercial purposes, not be edited, and not be used without acknowledging",
                       "its source (i.e., the terms of a CC BY-NC-ND license)."))
    expect_equal(defLic("DT_CC4-BY-NC-ND", crrLng = "DE", licHld = "Max Mustermann"),
                 paste("Dieser Datensatz enthält Daten einer wissenschaftlichen Studie, deren Urheberrecht bei den",
                       "Autoren der Studie liegt. Ohne die ausdrückliche Zustimmung der Autoren und von Max",
                       "Mustermann darf dieser Datensatz nicht zu kommerziellen Zwecken verbreitet, bearbeitet oder",
                       "verwendet werden ohne die Quelle anzugeben (d. h. der Datensatz unterliegt den Bedingungen",
                       "der CC BY-NC-ND-Lizenz)."))
    expect_equal(defLic("DT_CC4-BY-NC-ND", crrLng = "EN", licHld = "John Doe"),
                 paste("This data set contains data of a scientific study and the study authors therefore own the",
                       "copyright. Without the study authors` and John Doe`s explicit consent, this data set may not",
                       "be distributed for commercial purposes, not be edited, and not be used without acknowledging",
                       "its source (i.e., the terms of a CC BY-NC-ND license)."))
    expect_error(defLic("DT_CC4-BY-NC-ND", crrLng = "JP", licHld = "\\u5c71\\u7530\\u592a\\u90ce"),
                 "No translation available \\(yet\\) for JP\\.")
    expect_equal(defLic("DT_CC4-BY-NC-ND", crrLng = "NB", licHld = "Kari Normann"),
                 paste("Dette datasettet inneholder data fra en vitenskapelig studie, og studieforfatterne eier",
                       "derfor opphavsretten. Uten uttrykkelig samtykke fra studieforfatterne og Kari Normann kan",
                       "dette datasettet ikke distribueres for kommersielle formål, ikke redigeres og ikke brukes",
                       "uten å oppgi kilden (dvs. vilkårene i en CC BY-NC-ND-lisens)."))
    expect_error(defLic("DT_CC4-BY-NC-ND", crrLng = "XX", licHld = "John Doe"),
                 "No translation available \\(yet\\) for XX\\.")
    expect_error(defLic("FC_CC4-BY-NC-ND"), "When using this license, a license holder needs to be defined\\.")
    expect_error(defLic("FC_CC4-BY-NC-ND", crrLng = "DE"),
                 "When using this license, a license holder needs to be defined\\.")
    expect_equal(defLic("FC_CC4-BY-NC-ND", licHld = "John Doe"),
                 paste("This data set was constructed by John Doe, who therefore owns the copyright. Without the",
                       "explicit consent of John Doe, this data set may not be distributed for commercial purposes,",
                       "not be edited, and not be used without acknowledging its source (i.e., the terms of the CC",
                       "BY-NC-ND license)."))
    expect_equal(defLic("FC_CC4-BY-NC-ND", crrLng = "DE", licHld = "Max Mustermann"),
                 paste("Dieser Datensatz wurde von Max Mustermann erstellt, der somit das Urheberrecht daran besitzt.",
                       "Ohne die ausdrückliche Zustimmung von Max Mustermann darf dieser Datensatz nicht für",
                       "kommerzielle Zwecke verbreitet, nicht bearbeitet und nicht ohne Angabe der Quelle verwendet",
                       "werden (d. h. der Datensatz unterliegt den Bedingungen der CC BY-NC-ND-Lizenz)."))
    expect_equal(defLic("FC_CC4-BY-NC-ND", crrLng = "EN", licHld = "John Doe"),
                 paste("This data set was constructed by John Doe, who therefore owns the copyright. Without the",
                       "explicit consent of John Doe, this data set may not be distributed for commercial purposes,",
                       "not be edited, and not be used without acknowledging its source (i.e., the terms of the CC",
                       "BY-NC-ND license)."))
    expect_error(defLic("FC_CC4-BY-NC-ND", crrLng = "JP", licHld = "\\u5c71\\u7530\\u592a\\u90ce"),
                 "No translation available \\(yet\\) for JP\\.")
    expect_equal(defLic("FC_CC4-BY-NC-ND", crrLng = "NB", licHld = "Kari Normann"),
                 paste("Datasettet er utarbeidet av Kari Normann, som derfor eier opphavsretten. Uten uttrykkelig",
                       "samtykke fra Kari Normann kan dette datasettet ikke distribueres for kommersielle formål,",
                       "redigeres eller brukes uten å oppgi kilden (dvs. vilkårene i CC BY-NC-ND-lisens)."))
    expect_error(defLic("FC_CC4-BY-NC-ND", crrLng = "XX", licHld = "John Doe"),
                 "No translation available \\(yet\\) for XX\\.")
    expect_error(defLic("RP_GPL2"),
                 "When using this license, the R-package where the data are originating from needs to be defined\\.")
    expect_error(defLic("RP_GPL2", crrLng = "DE"),
                 "When using this license, the R-package where the data are originating from needs to be defined\\.")
    expect_equal(defLic("RP_GPL2", licHld = "RPKG"),
                 paste("This data set was provided as part of the R-package 'RPKG', which is published under the",
                       "terms of the GNU General Public License 2.x. You may use the data both privately and",
                       "commercially, distribute or modify them. When using the data set, you need to disclose the",
                       "source, and publish any modifications you may make under the same license. If the data set",
                       "is based on empirical data, the authors are given in the 'References'-section. In such case,",
                       "please  cite them when using the dataset."))
    expect_equal(defLic("RP_GPL2", crrLng = "DE", licHld = "RPKG"),
                 paste("Dieser Datensatz ist Teil des R-Pakets 'RPKG, das unter den Bedingungen der GNU General",
                       "Public License 2.x veröffentlicht wurde. Sie können die Daten sowohl privat als auch",
                       "kommerziell nutzen, verbreiten oder verändern. Bei der Verwendung des Datensatzes müssen",
                       "Sie die  Quelle angeben und alle änderungen, die Sie vornehmen, unter derselben Lizenz",
                       "veröffentlichen. Wenn der Datensatz auf empirischer Forschung basiert, sind die Autoren",
                       "im Abschnitt 'Referenzen' angegeben. Zitieren Sie in diesem Fall die Studie, wenn Sie",
                       "den Datensatz verwenden."))
    expect_equal(defLic("RP_GPL2", crrLng = "EN", licHld = "RPKG"),
                 paste("This data set was provided as part of the R-package 'RPKG', which is published under the",
                       "terms of the GNU General Public License 2.x. You may use the data both privately and",
                       "commercially, distribute or modify them. When using the data set, you need to disclose the",
                       "source, and publish any modifications you may make under the same license. If the data set",
                       "is based on empirical data, the authors are given in the 'References'-section. In such case,",
                       "please  cite them when using the dataset."))
    expect_error(defLic("RP_GPL2", crrLng = "JP", licHld = "RPKG"),
                 "No translation available \\(yet\\) for JP\\.")
    expect_equal(defLic("RP_GPL2", crrLng = "NB", licHld = "RPKG"),
                 paste("Dette datasettet er en del av R-pakken 'RPKG', som er publisert under vilkårene i GNU General",
                       "Public License 2.x. Du kan bruke dataene både privat og kommersielt, distribuere eller endre",
                       "dem. Når du bruker datasettet, må du oppgi kilden og publisere eventuelle endringer du gjør",
                       "under samme lisens. Hvis datasettet er basert på empiriske data, er forfatterne oppgitt under",
                       "'Referanser'. Vennligst oppgi referansene i slike tilfeller når du bruker datasettet."))
    expect_error(defLic("RP_GPL2", crrLng = "XX", licHld = "RPKG"),
                 "No translation available \\(yet\\) for XX\\.")
    expect_error(defLic("RP_GPL3"),
                 "When using this license, the R-package where the data are originating from needs to be defined\\.")
    expect_equal(defLic("RP_GPL3", licHld = "RPKG"),
                 paste("This data set was provided as part of the R-package 'RPKG', which is published under the",
                       "terms of the GNU General Public License 3.0. You may use the data both privately and",
                       "commercially, distribute or modify them. When using the data set, you need to disclose the",
                       "source, and publish any modifications you may make under the same license. If the data set",
                       "is based on empirical data, the authors are given in the 'References'-section. In such case,",
                       "please  cite them when using the dataset."))
    expect_error(defLic("RP_GPL3", crrLng = "XX", licHld = "RPKG"),
                 "No translation available \\(yet\\) for XX\\.")
    expect_error(defLic("RP_AGPL3"),
                 "When using this license, the R-package where the data are originating from needs to be defined\\.")
    expect_equal(defLic("RP_AGPL3", licHld = "RPKG"),
                 paste("This data set was provided as part of the R-package 'RPKG', which is published under the",
                       "terms of the GNU Affero General Public License 3.0. You may use the data both privately and",
                       "commercially, distribute or modify them. When using the data set, you need to disclose the",
                       "source, and publish any modifications you may make under the same license. If the data set",
                       "is based on empirical data, the authors are given in the 'References'-section. In such case,",
                       "please  cite them when using the dataset."))
    expect_error(defLic("RP_AGPL3", crrLng = "XX", licHld = "RPKG"),
                 "No translation available \\(yet\\) for XX\\.")
    expect_error(defLic("RP_LGPL3"),
                 "When using this license, the R-package where the data are originating from needs to be defined\\.")
    expect_equal(defLic("RP_LGPL3", licHld = "RPKG"),
                 paste("This data set was provided as part of the R-package 'RPKG', which is published under the",
                       "terms of the GNU Lesser General Public License 3.0. You may use the data both privately and",
                       "commercially, distribute or modify them. When using the data set, you need to disclose the",
                       "source, and publish any modifications you may make under the same license. If the data set",
                       "is based on empirical data, the authors are given in the 'References'-section. In such case,",
                       "please  cite them when using the dataset."))
    expect_error(defLic("RP_LGPL3", crrLng = "XX", licHld = "RPKG"),
                 "No translation available \\(yet\\) for XX\\.")
    expect_null(defLic())
    expect_null(defLic(licNme = ""))
    expect_null(defLic(licNme = "", crrLng = "DE", licHld = "Max Mustermann"))
    expect_null(defLic(licNme = "", crrLng = "EN", licHld = "John Doe"))
    expect_null(defLic(crrLng = "DE"))
    expect_null(defLic(crrLng = "DE", licHld = "Max Mustermann"))
    expect_null(defLic(licHld = "John Doe"))
    if (packageVersion("jmvcore") >= "2.4.3") {
        expect_warning(df4Chk <- describe_omv(dtaInp = file.path("..", "ToothGrowth.omv"), dtaTtl = "ToothGrowth"),
          regexp = "^The data set contains analyses\\. Those will be overwritten\\.")
        expect_s3_class(df4Chk, "data.frame")
        expect_equal(dim(df4Chk), c(60, 16))
        expect_equal(names(attributes(df4Chk)), c("names", "row.names", "class", "fltLst", "removedRows", "addedRows", "transforms", "HTML", "protobuf"))
        expect_length(attr(df4Chk, "protobuf")[[1]]$options$options, 1)
        expect_equal(attr(df4Chk, "protobuf")[[1]]$options$names, c("results//heading"))
        expect_equal(attr(df4Chk, "protobuf")[[1]]$options$options[[1]]$s, "ToothGrowth")
        expect_warning(df4Chk <- describe_omv(dtaInp = file.path("..", "ToothGrowth.omv"), fleOut = file.path("..", "ToothGrowth.omv"), dtaTtl = "ToothGrowth"),
          regexp = "^The data set contains analyses\\. Those will be overwritten\\.")
        expect_null(df4Chk)
        expect_error(splHTM("<p Trial with incomplete HTML tags.</p>"), regexp = paste("^Error when decoding HTML\\. Please check whether the HTML-string the you used for the description",
          "is valid \\(i\\.e\\., no typos, etc\\.\\)\\. If it is valid, send it to sebastian\\.jentschke@uib\\.no"))
        expect_equal(splHTM("Trial without HTML tags."), "Trial without HTML tags.")
        expect_warning(getAtt(list("<p>", "Trial", "</trial>"), 3, list()), regexp = "^The HTML tag \".*?\" isn't implemented, and hence the jamovi formatting attributes remain unchanged.")
        expect_warning(addHTM(gsub("</p>", "</p mismatch>", htmTxt()), "Title", "Description"), regexp = "^Unexpected HTML to be replaced with description:")
        suppressWarnings(expect_equal(getAtt(list("<p>", "Trial", "</trial>"), 3, list(bold = TRUE)), list(bold = TRUE)))
        suppressWarnings(expect_equal(getAtt(list("<p>", "Trial", "</trial>"), 3, list()), list()))
        rm(df4Chk)
    }

    rm(nmeOut, lstDsc, chrDsc)
})

#' Re-arrange columns / variables in .omv-files for the statistical spreadsheet 'jamovi' (<https://www.jamovi.org>)
#'
#' @param dtaInp Either a data frame or the name of a data file to be read (including the path, if required; "FILENAME.ext"; default: NULL); files can be of
#'               any supported file type, see Details below
#' @param fleOut Name of the data file to be written (including the path, if required; "FILE_OUT.omv"; default: ""); if empty, the resulting data frame is
#'               returned instead
#' @param varOrd Character vector with the desired order of variable(s) in the data frame (see Details; default: c())
#' @param varMve Named list defining to how much a particular variable (name of a list entry) should be moved up (neg. value of a list entry) or down (pos.
#'               value) in the data frame (see Details; default: c())
#' @param psvAnl Whether analyses that are contained in the input file shall be transferred to the output file (default: FALSE)
#' @param usePkg Name of the package: "foreign" or "haven" that shall be used to read SPSS, Stata and SAS files; "foreign" is the default (it comes with
#'               base R), but "haven" is newer and more comprehensive
#' @param selSet Name of the data set that is to be selected from the workspace (only applies when reading .RData-files)
#' @param ...    Additional arguments passed on to methods; see Details below
#'
#' @return a data frame (only returned if `fleOut` is empty) where the order of variables / columns of the input data set is re-arranged
#'
#' @details
#' * `varOrd` is a character vector. If not all variables of the original data set are contained in `varOrd`, a warning is issued but otherwise the list of
#'   variables defined in `varOrd` is used (removing variables not contained in `varOrd`).
#' * `varMve` is a named list. For example would `list(VARNAME = -3)` move the variable `VARNAME` three positions up in the list of variables (towards the
#'   first column), and `list(VARNAME = 3)` would move it three positions down (towards the last column). If the number of steps the variable is to be moved
#'   leads to the position being either lower than the first or higher than the total number of variables in the data set, an error message is issued. Please
#'   note that the list entries are processed one after another, that is, for a second list entry, you have to consider how the first list entry may have
#'   changed to order of variables.
#' * Using `varOrd` makes more sense for changing the position of several variables, whereas using `varMve` makes more sense for one variable. If
#'   both parameters are given, a warning is issued and `varOrd` takes precedence.
#' * The ellipsis-parameter (`...`) can be used to submit arguments / parameters to the functions that are used for reading and writing the data. By clicking
#'   on the respective function under “See also”, you can get a more detailed overview over which parameters each of those functions take. The functions are:
#'   `read_omv` and `write_omv` (for jamovi-files), `read.table` (for CSV / TSV files; using similar defaults as `read.csv` for CSV and `read.delim` for TSV
#'   which both are based upon `read.table`), `load` (for .RData-files), `readRDS` (for .rds-files), `read_sav` (needs the R-package `haven`) or `read.spss`
#'   (needs the R-package `foreign`) for SPSS-files, `read_dta` (`haven`) / `read.dta` (`foreign`) for Stata-files, `read_sas` (`haven`) for SAS-data-files,
#'   and `read_xpt` (`haven`) / `read.xport` (`foreign`) for SAS-transport-files. If you would like to use `haven`, you may need to install it using
#'   `install.packages("haven", dep = TRUE)`.
#'
#' @seealso `arrange_cols_omv` internally uses the following functions for reading and writing data files in different formats: [jmvReadWrite::read_omv()] and
#'   [jmvReadWrite::write_omv()] for jamovi-files, [utils::read.table()] for CSV / TSV files, [load()] for reading .RData-files, [readRDS()] for .rds-files,
#'   [haven::read_sav()] or [foreign::read.spss()] for SPSS-files, [haven::read_dta()] or [foreign::read.dta()] for Stata-files, [haven::read_sas()] for
#'   SAS-data-files, and [haven::read_xpt()] or [foreign::read.xport()] for SAS-transport-files.
#'
#' @examples
#' nmeInp <- system.file("extdata", "AlbumSales.omv", package = "jmvReadWrite")
#' nmeOut <- tempfile(fileext = ".omv")
#' # the original file has the variables in the order: "Adverts", "Airplay", "Image", "Sales"
#' names(read_omv(nmeInp))
#' # first, we move the variable "Sales" to the first place using the varOrd-parameter
#' jmvReadWrite::arrange_cols_omv(dtaInp = nmeInp, fleOut = nmeOut,
#'   varOrd = c("Sales", "Adverts", "Airplay", "Image"))
#' names(jmvReadWrite::read_omv(nmeOut))
#' unlink(nmeOut)
#' # now, we move the variable "Sales" to the first place using the varMve-parameter
#' jmvReadWrite::arrange_cols_omv(dtaInp = nmeInp, fleOut = nmeOut, varMve = list(Sales = -3))
#' names(jmvReadWrite::read_omv(nmeOut))
#' unlink(nmeOut)
#'
#'
#' @export arrange_cols_omv
#'
arrange_cols_omv <- function(dtaInp = NULL, fleOut = "", varOrd = c(), varMve = list(), psvAnl = FALSE, usePkg = c("foreign", "haven"), selSet = "", ...) {
    # check input parameters
    chkOnM(varOrd, varMve)

    # check and import input data set (either as data frame or from a file)
    if (!is.null(list(...)[["fleInp"]])) stop("Please use the argument dtaInp instead of fleInp.")
    dtaFrm <- inp2DF(dtaInp = dtaInp, usePkg = usePkg, selSet = selSet, ...)

    # check and update variable order
    varOrd <- updOrd(dtaFrm, varOrd, varMve)

    # re-arrange to order of variables, while storing and restoring the attributes attached to the whole data frame (column attributes are not affected)
    attLst <- bckAtt(dtaFrm)
    dtaFrm <- dtaFrm[, varOrd]
    dtaFrm <- rstAtt(dtaFrm, attLst)

    # rtnDta in globals.R (unified function to either write the data frame, open it in a new jamovi session or return it)
    rtnDta(dtaFrm = dtaFrm, fleOut = fleOut, dtaTtl = jmvTtl("_arr_cols"), psvAnl = psvAnl, dtaInp = dtaInp, ...)
}

chkOnM <- function(varOrd = c(), varMve = list()) {
    # check the input parameters: either varOrd or varMve need to be given; if varOrd is given, the given character vectore and all of its elements need to be not empty;
    # if varMve is given, it needs to be a list, the names can't be empty, and the values need to be integers and can't be zero
    if ((length(varOrd) < 1 || !is.character(varOrd) || !all(nzchar(varOrd))) &&
        (length(varMve) < 1 || !is.list(varMve) || !is.character(names(varMve)) || !all(vapply(unlist(varMve), function(e) is.numeric(e) && e != 0 && e %% 1 == 0, logical(1))))) {
        stop("Calling arrange_cols_omv requires either the parameter varOrd (a character vector) or the parameter varMve (a named list), using the correct format (see Details in help).")
    }
}

updOrd <- function(dtaFrm = NULL, varOrd = c(), varMve = list()) {
    # re-arrange the order of variables in the data set (varOrd)
    if (length(varOrd) > 0) {
        # [1] check whether all variables in varOrd are not empty and exist in the data set
        chkVar(dtaFrm, varOrd)
        # [2] check whether any variable of the original data set is not contained in varOrd, if so issue a warning but proceed
        if (!all(names(dtaFrm) %in% varOrd)) {
            warning(sprintf("The following variable(s) from the original data set are not contained in varOrd: %s", paste(setdiff(names(dtaFrm), varOrd), collapse = ", ")))
        }
    }
    if (length(varMve) > 0 && length(varOrd) == 0) {
        # [1] assign the original order of variables to varOrd
        varOrd <- names(dtaFrm)
        # [2] check whether all variables in varMve are not empty and exist in the data set
        chkVar(dtaFrm, names(varMve))
        for (crrVar in names(varMve)) {
            crrPos <- which(varOrd == crrVar)
            if (crrPos + varMve[[crrVar]] < 1 || crrPos + varMve[[crrVar]] > dim(dtaFrm)[2]) {
                stop("The value given in varMve must be chosen so that the element isn't moved before the first or after the last column.")
            }
            allPos <- seq(dim(dtaFrm)[2])
            if (varMve[[crrVar]] < 0) {
                rplPos <- seq(crrPos + varMve[[crrVar]], crrPos)
                allPos[allPos %in% rplPos] <- c(crrPos, setdiff(rplPos, crrPos))
            } else {
                rplPos <- seq(crrPos, crrPos + varMve[[crrVar]])
                allPos[allPos %in% rplPos] <- c(setdiff(rplPos, crrPos), crrPos)
            }
            varOrd <- varOrd[allPos]
        }
    } else if (length(varMve) > 0 && length(varOrd) > 0) {
        warning("Both, varOrd and varMve given as input parameters. varOrd takes precedence.")
    }

    varOrd
}

#' Search values in .omv-files for the statistical spreadsheet 'jamovi' (<https://www.jamovi.org>)
#'
#' @param dtaInp Either a data frame or the name of a jamovi data file to be read (including the path, if required; "FILENAME.omv"; default: NULL)
#' @param srcTrm (Character or numeric) Vector (with length = 1) with a search term to be found in the data frame (default: c())
#' @param whlTrm Whether the exact search term shall be found (TRUE) or whether a partial match is sufficient (FALSE; default: FALSE)
#' @param ignCse Whether to ignore the case of the search term (default: FALSE)
#' @param incNum Whether to include continuous variables in the search (default: TRUE)
#' @param incOrd Whether to include ordinal variables in the search (default: TRUE)
#' @param incNom Whether to include nominal variables in the search (default: TRUE)
#' @param incID  Whether to include ID variables in the search (default: TRUE)
#' @param incCmp Whether to include Computed variables in the search (default: TRUE)
#' @param incRcd Whether to include Recoded variables in the search (default: TRUE)
#' @param ...    Additional arguments passed on to methods; see Details below
#'
#' @return a named list with the places where the search term was found: names in the list are the variables / columns, the entries the respective row names
#'         within that variable / column (row names are used for being tolerant to filtered-out cases in jamovi, if a filter is used, row numbers would be
#'         incorrect)
#'
#' @details
#' * The ellipsis-parameter (`...`) can be used to submit arguments / parameters to the function that is used for reading and writing the data. Clicking on the
#'   respective function under “See also”, you can get a more detailed overview over which parameters each of those functions take. The functions are:
#'   `read_omv` and `write_omv` (for jamovi-files).
#'
#' @seealso `replace_omv` uses [jmvReadWrite::read_omv()] and [jmvReadWrite::write_omv()] for reading and writing jamovi-files.
#'
#' @examples
#' # the exact value 24 appears 13 times in age
#' bfi_sample <- jmvReadWrite::bfi_sample
#' jmvReadWrite::search_omv(bfi_sample, 24, whlTrm = TRUE)
#' # taking the fifth entry from the search results
#' bfi_sample["61", "age"]
#' # with the following search, both Males and Females are found
#' # (the M of Males, wouldn't be matched if ignCse were FALSE and males is
#' #  only a partial match within Females, thus whlTrm must be set to FALSE)
#' jmvReadWrite::search_omv(bfi_sample, "males", whlTrm = FALSE, ignCse = TRUE)
#' # the first entry is a female, the first entry is a male
#' bfi_sample["1", "gender"] # Females
#' bfi_sample["6", "gender"] # Males
#' # using the search results assigned to a variable
#' srcRes <- jmvReadWrite::search_omv(bfi_sample, "males", whlTrm = FALSE, ignCse = TRUE)
#' bfi_sample[srcRes[[1]][1], names(srcRes[1])] # Females
#' bfi_sample[srcRes[[1]][6], names(srcRes[1])] # Males
#'
#'
#' @export search_omv
#'
search_omv <- function(dtaInp = NULL, srcTrm = c(), whlTrm = FALSE, ignCse = FALSE, incNum = TRUE, incOrd = TRUE, incNom = TRUE, incID = TRUE, incCmp = TRUE, incRcd = TRUE, ...) {

    # check the input parameter: the search term needs to be a non-empty character vector with length 1
    if (!is.character(srcTrm) && is.vector(srcTrm)) srcTrm <- as.character(srcTrm)
    if (length(srcTrm) != 1 || !nzchar(srcTrm)) {
        stop("Calling search_omv requires the parameter srcTrm (a character vector with length 1).")
    }
    srcTrm <- gsub("\\(", "\\\\(", gsub("\\)", "\\\\)", gsub("\\.", "\\\\.", srcTrm)))

    # check and import input data set (either as data frame or from a file)
    if (!is.null(list(...)[["fleInp"]])) stop("Please use the argument dtaInp instead of fleInp.")
    dtaFrm <- inp2DF(dtaInp = dtaInp, ...)
    dtaFrm <- jmvAtt(dtaFrm)

    incClT <- c("Data", rep("Computed", incCmp), rep("Recoded", incRcd))
    incMsT <- c(rep("ID", incID), rep("Nominal", incNom), rep("Ordinal", incOrd), rep("Continuous", incNum))
    srcNme <- names(dtaFrm)
    srcNme <- srcNme[vapply(dtaFrm[srcNme], function(x) is.null(attr(x,  "columnType")) || any(attr(x,  "columnType") == incClT), logical(1))]
    srcNme <- srcNme[vapply(dtaFrm[srcNme], function(x) is.null(attr(x, "measureType")) || any(attr(x, "measureType") == incMsT), logical(1))]
    srcRes <- stats::setNames(rep(list(NULL), length(srcNme)), srcNme)
    nmeRow <- row.names(dtaFrm)
    for (i in seq_along(srcRes)) {
        srcRes[[i]] <- nmeRow[srcClm(dtaFrm[[srcNme[i]]], srcTrm, whlTrm, ignCse)]
    }
    return(srcRes[vapply(srcRes, length, integer(1)) > 0])
}

srcClm <- function(inpClm = NULL, srcTrm = "", whlTrm = FALSE, ignCse = FALSE) {
    if (!is.na(srcTrm)) {
        return(grepl(paste0(rep("^", whlTrm), srcTrm, rep("$", whlTrm)), as.character(inpClm), ignore.case = ignCse))
    } else if (is.na(srcTrm)) {
        return(is.na(inpClm))
    }
}

test_that("search_omv works", {
    tmpDF <- jmvReadWrite::bfi_sample
    expect_equal(search_omv(dtaInp = tmpDF, srcTrm = "A64365", whlTrm = FALSE), list(ID = "185", ID2 = "168"))
    expect_equal(search_omv(dtaInp = tmpDF, srcTrm = "a64365", whlTrm = FALSE), setNames(list(), character(0)))
    expect_equal(search_omv(dtaInp = tmpDF, srcTrm = "a64365", whlTrm = FALSE, ignCse = TRUE), list(ID = "185", ID2 = "168"))
    expect_equal(search_omv(dtaInp = tmpDF, srcTrm = 24, whlTrm = TRUE), list(age = c("2", "6", "11", "12", "61", "75", "77", "99", "109", "169", "197", "200")))
    expect_equal(vapply(search_omv(dtaInp = jmvReadWrite::bfi_sample, srcTrm = 24, whlTrm = FALSE), length, integer(1)), setNames(c(13, 12, 60, 8), c("ID", "age", "AD", "ID2")))
    expect_equal(search_omv(dtaInp = tmpDF, srcTrm = NA,  whlTrm = TRUE),  list(A1 = c("16", "17", "253", "254")))
    expect_equal(search_omv(dtaInp = tmpDF, srcTrm = NA,  whlTrm = FALSE), list(A1 = c("16", "17", "253", "254")))
    set.seed(1)
    tmpDF[sample(254, 10), "AD"] <- Inf
    tmpDF[sample(254, 10), "AD"] <- -Inf
    # whlTrm = TRUE only finds Inf, FALSE finds both Inf and -Inf
    expect_equal(search_omv(dtaInp = tmpDF, srcTrm = Inf, whlTrm = TRUE),  list(AD = c("14", "43", "68", "129", "162", "167", "187", "210", "215", "249")))
    expect_equal(search_omv(dtaInp = tmpDF, srcTrm = Inf, whlTrm = FALSE), list(AD = c("7", "14", "21", "43", "51", "68", "73", "74", "85", "106", "129",
      "162", "167", "182", "187", "210", "215", "225", "238", "249")))
    # NA's are in another column and shouldn't be affected
    expect_equal(search_omv(dtaInp = tmpDF, srcTrm = NA,  whlTrm = TRUE),  list(A1 = c("16", "17", "253", "254")))
    expect_equal(search_omv(dtaInp = tmpDF, srcTrm = NA,  whlTrm = FALSE), list(A1 = c("16", "17", "253", "254")))
    for (i in sample(33, 10)) tmpDF[sample(254, 2), i] <-  NA
    expect_equal(search_omv(dtaInp = tmpDF, srcTrm = NA,  whlTrm = TRUE),  list(A1 = c("16", "17", "44", "121", "253", "254"), A4 = c("33", "84"), C3 = c("70", "163"), C4 = c("70", "87"),
      E3 = c("111", "166"), E4 = c("89", "126"), N5 = c("172", "207"), O4 = c("42", "74"), gender = c("20", "250"), AG = c("148", "156")))

    expect_error(search_omv(fleInp = tmpDF),                       regexp = "^Calling search_omv requires the parameter srcTrm \\(a character vector with length 1\\)\\.")
    expect_error(search_omv(fleInp = tmpDF, srcTrm = c("1", "2")), regexp = "^Calling search_omv requires the parameter srcTrm \\(a character vector with length 1\\)\\.")
    expect_error(search_omv(fleInp = tmpDF, srcTrm = "A64365"),    regexp = "^Please use the argument dtaInp instead of fleInp\\.")
})

#' Label columns / variables in .omv-files for the statistical spreadsheet 'jamovi' (<https://www.jamovi.org>)
#'
#' @param dtaInp Either a data frame or the name of a data file to be read (including the path, if required; "FILENAME.ext"; default: NULL); files can be of
#'               any supported file type, see Details below
#' @param fleOut Name of the data file to be written (including the path, if required; "FILE_OUT.omv"; default: ""); if empty, the resulting data frame is
#'               returned instead
#' @param varLbl Variable (default: NULL) containing either a character (a file name; the file must contain two columns one with variable names, the other with
#'               the labels), a data frame (one column the variable names, the other the labels), or a character vector (with the same length as the data set,
#'               containing the variable labels). See Details for more information.
#' @param psvAnl Whether analyses that are contained in the input file shall be transferred to the output file (default: FALSE)
#' @param usePkg Name of the package: "foreign" or "haven" that shall be used to read SPSS, Stata and SAS files; "foreign" is the default (it comes with
#'               base R), but "haven" is newer and more comprehensive
#' @param selSet Name of the data set that is to be selected from the workspace (only applies when reading .RData-files)
#' @param ...    Additional arguments passed on to methods; see Details below
#'
#' @return a data frame (only returned if `fleOut` is empty) where the order of variables / columns of the input data set is re-arranged
#'
#' @details
#' * `varLbl` can be either (1) a character with a file name to read (the file must contain to columns, one with the variable names, the other with the
#'   variable labels); (2) a data frame with two columns (one with the variable names, the other with the variable labels), or (3) a character vector
#'   containing the variable labels (with a length equal to the number of variables in the input data set).
#' * The ellipsis-parameter (`...`) can be used to submit arguments / parameters to the functions that are used for reading and writing the data. By clicking
#'   on the respective function under “See also”, you can get a more detailed overview over which parameters each of those functions take. The functions are:
#'   `read_omv` and `write_omv` (for jamovi-files), `read.table` (for CSV / TSV files; using similar defaults as `read.csv` for CSV and `read.delim` for TSV
#'   which both are based upon `read.table`), `load` (for .RData-files), `readRDS` (for .rds-files), `read_sav` (needs the R-package `haven`) or `read.spss`
#'   (needs the R-package `foreign`) for SPSS-files, `read_dta` (`haven`) / `read.dta` (`foreign`) for Stata-files, `read_sas` (`haven`) for SAS-data-files,
#'   and `read_xpt` (`haven`) / `read.xport` (`foreign`) for SAS-transport-files. If you would like to use `haven`, you may need to install it using
#'   `install.packages("haven", dep = TRUE)`.
#'
#' @seealso `label_vars_omv` internally uses the following functions for reading and writing data files in different formats: [jmvReadWrite::read_omv()] and
#'   [jmvReadWrite::write_omv()] for jamovi-files, [utils::read.table()] for CSV / TSV files, [load()] for reading .RData-files, [readRDS()] for .rds-files,
#'   [haven::read_sav()] or [foreign::read.spss()] for SPSS-files, [haven::read_dta()] or [foreign::read.dta()] for Stata-files, [haven::read_sas()] for
#'   SAS-data-files, and [haven::read_xpt()] or [foreign::read.xport()] for SAS-transport-files.
#'
#' @examples
#' # use one of the data files included in the package, but only the first 28 columns
#' # (the latter columns contain data for testing calculations, etc.)
#' nmeInp <- system.file("extdata", "bfi_sample.omv", package = "jmvReadWrite")
#' dtaInp <- jmvReadWrite::read_omv(nmeInp)[1:28]
#' nmeOut <- tempfile(fileext = ".omv")
#' # in the original file, the variable labels – attr(*, "jmv-desc") - are empty
#' lapply(dtaInp, attr, "jmv-desc")
#' # the definition of the variable labels can be read from a file with two columns,
#' # the first containing the variable name, the second the variable labels
#' # you can easily create such a file in Excel and save it as CSV
#' # if your CSV contains column names (e.g., varNme and varLbl) in the first row are they ignored
#' lblFle <- system.file("extdata", "label_example.csv", package = "jmvReadWrite")
#' lblDtF <- utils::read.csv(lblFle, header = FALSE)
#' str(lblDtF)
#'
#' # there are three options to give the varLbl parameter:
#' # (1) as file name, ...
#' jmvReadWrite::label_vars_omv(dtaInp = dtaInp, fleOut = nmeOut, varLbl = lblFle)
#' lapply(jmvReadWrite::read_omv(nmeOut), attr, "jmv-desc")
#' unlink(nmeOut)
#'
#' # (2) as data frame (using lblDtF from above), or ...
#' jmvReadWrite::label_vars_omv(dtaInp = dtaInp, fleOut = nmeOut, varLbl = lblDtF)
#' lapply(jmvReadWrite::read_omv(nmeOut), attr, "jmv-desc")
#' unlink(nmeOut)
#'
#' # (3) as character vector (with the same length as there are columns in the input data set)
#' lblChr <- lblDtF[[2]]
#' head(lblChr)
#' jmvReadWrite::label_vars_omv(dtaInp = dtaInp, fleOut = nmeOut, varLbl = lblChr)
#' lapply(jmvReadWrite::read_omv(nmeOut), attr, "jmv-desc")
#' unlink(nmeOut)
#'
#'
#' @export label_vars_omv
#'
label_vars_omv <- function(dtaInp = NULL, fleOut = "", varLbl = NULL, psvAnl = FALSE, usePkg = c("foreign", "haven"),
                           selSet = "", ...) {
    # check the input parameter: varLbl needs to be given
    if (length(varLbl) < 1) {
        stop("Calling label_vars_omv requires the parameter varLbl, using the correct format (see Details in help).")
    }

    # check and import input data set (either as data frame or from a file)
    if (!is.null(list(...)[["fleInp"]])) stop("Please use the argument dtaInp instead of fleInp.")
    dtaFrm <- inp2DF(dtaInp = dtaInp, usePkg = usePkg, selSet = selSet, ...)

    # if varLbl is a character (a file name) or a character vector (with labels), transform it into a data frame with
    # two columns (variable names and labels)
    if (is.character(varLbl)) {
        if (length(varLbl) == 1 && file.exists(varLbl)) {
            varLbl <- utils::read.csv(varLbl, header = FALSE)
        } else if (length(varLbl) == length(dtaFrm)) {
            varLbl <- data.frame(varNme = names(dtaFrm), varLbl = varLbl)
        } else {
            stop(sprintf(paste0("If the parameter varLbl is a character, it eiter needs to be the name of an ",
                                "(exisiting) file or a vector with labels that has the same length as the number of ",
                                "variables in the input data set (%d)."), length(dtaFrm)))
        }
    }

    # assign the labels to the data frame
    if (is.data.frame(varLbl) && length(varLbl) == 2) {
        # determine which column of the data frame contains the variable names
        nmeClm <- which(vapply(varLbl, function(c) all(c %in% names(dtaFrm)), logical(1), USE.NAMES = FALSE))
        # ensure that there is exactly one such column
        if (length(nmeClm) == 1) {
            # iterate through the variable of the input data set
            nmeDtF <- names(dtaFrm)
            for (n in seq_along(nmeDtF)) {
                # if the variable name is found, set the variable label
                selRow <- varLbl[[nmeClm]] %in% nmeDtF[n]
                if (any(selRow)) attr(dtaFrm[, n], "jmv-desc") <- varLbl[selRow, -nmeClm]
            }
        } else {
            stop(sprintf(paste0("There must be exactly one column with the variable names (currently: %d).\n",
                                "All variable names in the label definition must be contained in the input data ",
                                "set.\n%s\n%s"),
                         length(nmeClm), paste0(varLbl[[1]], collapse = ", "), paste0(names(dtaFrm), collapse = ", ")))
        }
    } else {
        stop(sprintf("varLbl was either not a data frame or could not be converted into one.\n%s",
                     utils::capture.output(utils::str(varLbl))))
    }

    # rtnDta in globals.R (unified function to either write a data frame, open it in a new jamovi session or return it)
    rtnDta(dtaFrm = dtaFrm, fleOut = fleOut, dtaTtl = jmvTtl("_label_vars"), psvAnl = psvAnl, dtaInp = dtaInp, ...)
}

#' Convert data files (CSV, R, other statistics packages) into .omv-files for the statistical spreadsheet 'jamovi' (<https://www.jamovi.org>)
#'
#' @param fleInp Name (including the path, if required) of the data file to be read ("FILENAME.ext"; default: ""); supports CSV and R-files natively, or other
#'               file types if "foreign" or "haven" are installed, see Details below
#' @param fleOut Name (including the path, if required) of the data file to be written ("FILENAME.omv"; default: ""); if empty, the extension of fleInp is
#'               replaced with ".omv"
#' @param varSrt Variable(s) that are used to sort the data frame (see Details; if empty, the row order of the input file is kept; default: c())
#' @param usePkg Name of the package: "foreign" or "haven" that shall be used to read SPSS, Stata and SAS files; "foreign" is the default (it comes with base
#'               R), but "haven" is newer and more comprehensive
#' @param selSet Name of the data set that is to be selected from the workspace (only applies when reading .RData-files)
#' @param ...    Additional arguments passed on to methods; see Details below
#'
#' @return the function doesn't have a return value (it returns NULL)
#'
#' @details
#' * In difference to the remaining helper functions, `convert_to_omv` doesn't accept a data frame as input and it neither does return a data frame if `fleOut`
#'   is left empty: If you want to write a data frame, use `write_omv`. If you want to have a data frame returned use `read_omv` for jamovi-files or any of the
#'   functions listed in the bullet point below for any other file type.
#' * `varSrt` can be either a character or a character vector (with one or more variables respectively). The sorting order for a particular variable can be
#'   inverted with preceding the variable name with "-". Please note that this doesn't make sense and hence throws a warning for certain variable types (e.g.,
#'   factors).
#' * The ellipsis-parameter (`...`) can be used to submit arguments / parameters to the functions that are used for reading and writing the data. By clicking
#'   on the respective function under “See also”, you can get a more detailed overview over which parameters each of those functions take. The functions are:
#'   `read_omv` and `write_omv` (for jamovi-files), `read.table` (for CSV / TSV files; using similar defaults as `read.csv` for CSV and `read.delim` for TSV
#'   which both are based upon `read.table`), `load` (for .RData-files), `readRDS` (for .rds-files), `read_sav` (needs the R-package `haven`) or `read.spss`
#'   (needs the R-package `foreign`) for SPSS-files, `read_dta` (`haven`) / `read.dta` (`foreign`) for Stata-files, `read_sas` (`haven`) for SAS-data-files,
#'   and `read_xpt` (`haven`) / `read.xport` (`foreign`) for SAS-transport-files. If you would like to use `haven`, you may need to install it using
#'   `install.packages("haven", dep = TRUE)`.
#'
#' @seealso `convert_to_omv` internally uses the following functions for reading and writing data files in different formats: [jmvReadWrite::read_omv()] and
#'   [jmvReadWrite::write_omv()] for jamovi-files, [utils::read.table()] for CSV / TSV files, [load()] for reading .RData-files, [readRDS()] for .rds-files,
#'   [haven::read_sav()] or [foreign::read.spss()] for SPSS-files, [haven::read_dta()] or [foreign::read.dta()] for Stata-files, [haven::read_sas()] for
#'   SAS-data-files, and [haven::read_xpt()] or [foreign::read.xport()] for SAS-transport-files.
#'
#' @examples
#' # Example 1: Convert from RDS
#' # (use ToothGrowth as example, save it as RDS)
#' nmeInp <- tempfile(fileext = ".rds")
#' nmeOut <- tempfile(fileext = ".omv")
#' saveRDS(jmvReadWrite::ToothGrowth, nmeInp)
#' jmvReadWrite::convert_to_omv(fleInp = nmeInp, fleOut = nmeOut)
#' cat(list.files(dirname(nmeOut), basename(nmeOut)))
#' # -> "file[...].omv" ([...] contains a random combination of numbers / characters
#' cat(file.info(nmeOut)$size)
#' # -> 2448 (size may differ on different OSes)
#' cat(str(jmvReadWrite::read_omv(nmeOut, sveAtt = FALSE)))
#' # gives a overview of the dataframe (all columns and some attributes,
#' # sveAtt is intentionally set to FALSE to make the output not too overwhelming)
#' unlink(nmeInp)
#' unlink(nmeOut)
#'
#' # Example 2: Convert from CSV
#' # (use ToothGrowth again as example, this time save it as CSV)
#' nmeInp <- tempfile(fileext = ".csv")
#' nmeOut <- tempfile(fileext = ".omv")
#' write.csv(jmvReadWrite::ToothGrowth, nmeInp)
#' jmvReadWrite::convert_to_omv(fleInp = nmeInp, fleOut = nmeOut)
#' cat(list.files(dirname(nmeOut), basename(nmeOut)))
#' cat(file.info(nmeOut)$size)
#' # -> 2104 (size may differ acc. to OS; the size is smaller than for the RDS-file
#' # because CSV can store fewer attributes, e.g., labels)
#' cat(str(jmvReadWrite::read_omv(nmeOut, sveAtt = FALSE)))
#' # gives a overview of the dataframe (all columns and some attributes)
#' unlink(nmeInp)
#' unlink(nmeOut)
#'
#'
#' @export convert_to_omv
#'
convert_to_omv <- function(fleInp = "", fleOut = "", varSrt = c(), usePkg = c("foreign", "haven"), selSet = "", ...) {

    # check and format input and output files and handle / check further input arguments
    fleInp <- fmtFlI(fleInp, maxLng = 1)
    fleOut <- ifelse(nzchar(fleOut), fmtFlO(fleOut), fmtFlO(sub(paste0("\\.", tools::file_ext(fleInp)), ".omv", fleInp)))

    # read file and sort it (if varSrt is not empty)
    dtaFrm <- read_all(fleInp = fleInp, usePkg = usePkg, selSet = selSet, ...)
    dtaFrm <- srtFrm(dtaFrm, varSrt)

    # write file
    write_omv(dtaFrm = dtaFrm, fleOut = fleOut, ...)

    return(invisible(NULL))
}

test_that("convert_to_omv works", {
    # check whether writing the data is working (file existence, size, contents [.omv-files are ZIP archives and must contain files that include meta, metadata.json, data.bin])
    nmeInp <- tempfile(fileext = ".rds")
    nmeOut <- tempfile(fileext = ".omv")
    saveRDS(datasets::ToothGrowth, nmeInp)
    convert_to_omv(fleInp = nmeInp, fleOut = nmeOut)
    expect_true(file.exists(nmeOut))
    expect_gt(file.info(nmeOut)$size, 1)
    expect_true(chkFle(nmeOut, isZIP = TRUE))
    expect_true(chkFle(nmeOut, fleCnt = "meta"))
    expect_true(chkFle(nmeOut, fleCnt = "metadata.json"))
    expect_true(chkFle(nmeOut, fleCnt = "data.bin"))
    unlink(nmeInp)
    unlink(nmeOut)
})

#' Transform skewed variables (aiming at they conform to a normal distribution) in .omv-files for the statistical spreadsheet 'jamovi' (<https://www.jamovi.org>)
#'
#' @param dtaInp Either a data frame or the name of a data file to be read (including the path, if required; "FILENAME.ext"; default: NULL); files can be of
#'               any supported file type, see Details below
#' @param fleOut Name of the data file to be written (including the path, if required; "FILE_OUT.omv"; default: ""); if empty, the resulting data frame is
#'               returned instead
#' @param varXfm Named list variable where the name indicates which transformation is to be carried out and where each list entry points to one or more
#'               variables to be transformed using this transformation. See Details for more information.
#' @param psvAnl Whether analyses that are contained in the input file shall be transferred to the output file (default: FALSE)
#' @param usePkg Name of the package: "foreign" or "haven" that shall be used to read SPSS, Stata and SAS files; "foreign" is the default (it comes with
#'               base R), but "haven" is newer and more comprehensive
#' @param selSet Name of the data set that is to be selected from the workspace (only applies when reading .RData-files)
#' @param ...    Additional arguments passed on to methods; see Details below
#'
#' @return a data frame (only returned if `fleOut` is empty) where the order of variables / columns of the input data set is re-arranged
#'
#' @details
#' * `varXfm` has to be a named list variable where the names can either indicate the type of transformation or the kind and degree of skewness that shall be
#'   corrected. For the type of transformation, the following names are valid: `posSqr`, `negSqr`, `posLog`, `negLog`, `posInv`, `negInv`; where the second
#'   part of the name indicates the transformation to be carried out: `...Sqr` - square root, `...Log` - logarithm to the basis 10, `...Inv` - inversion, i.e.,
#'   1 / original value), and where the first part of the name indicates whether the original value is used (`pos...`) or whether the original value is
#'   subtracted from the maximum value of that variable (`neg...`; a constant of 1 is added to the maximum value for `...Log` and `...Inv` transformations).
#'   For the degree and kind of skewness, the following names are valid: `mdrPos`, `strPos`, `svrPos`, `mdrNeg`, `strNeg`, `svrNeg` (degree: moderate, strong,
#'   severe; kind: positive or negative).
#' * The ellipsis-parameter (`...`) can be used to submit arguments / parameters to the functions that are used for reading and writing the data. By clicking
#'   on the respective function under “See also”, you can get a more detailed overview over which parameters each of those functions take. The functions are:
#'   `read_omv` and `write_omv` (for jamovi-files), `read.table` (for CSV / TSV files; using similar defaults as `read.csv` for CSV and `read.delim` for TSV
#'   which both are based upon `read.table`), `load` (for .RData-files), `readRDS` (for .rds-files), `read_sav` (needs the R-package `haven`) or `read.spss`
#'   (needs the R-package `foreign`) for SPSS-files, `read_dta` (`haven`) / `read.dta` (`foreign`) for Stata-files, `read_sas` (`haven`) for SAS-data-files,
#'   and `read_xpt` (`haven`) / `read.xport` (`foreign`) for SAS-transport-files. If you would like to use `haven`, you may need to install it using
#'   `install.packages("haven", dep = TRUE)`.
#'
#' @seealso `transform_vars_omv` internally uses the following functions for reading and writing data files in different formats: [jmvReadWrite::read_omv()] and
#'   [jmvReadWrite::write_omv()] for jamovi-files, [utils::read.table()] for CSV / TSV files, [load()] for reading .RData-files, [readRDS()] for .rds-files,
#'   [haven::read_sav()] or [foreign::read.spss()] for SPSS-files, [haven::read_dta()] or [foreign::read.dta()] for Stata-files, [haven::read_sas()] for
#'   SAS-data-files, and [haven::read_xpt()] or [foreign::read.xport()] for SAS-transport-files.
#'
#' @examples
#' # generate skewed variables
#' set.seed(335)
#' dtaInp <- data.frame(MP = rnorm(1000) * 1e-1 + rexp(1000, 2) * (1 - 1e-1),
#'                      MN = rnorm(1000) * 1e-1 - rexp(1000, 2) * (1 - 1e-1),
#'                      SP = rnorm(1000) * 1e-2 + rexp(1000, 2) * (1 - 1e-2),
#'                      SN = rnorm(1000) * 1e-2 - rexp(1000, 2) * (1 - 1e-2),
#'                      EP = rnorm(1000) * 1e-4 + rexp(1000, 2) * (1 - 1e-4),
#'                      EN = rnorm(1000) * 1e-4 - rexp(1000, 2) * (1 - 1e-4))
#' jmv::descriptives(data = dtaInp, skew = TRUE, sw = TRUE)
#'
#' crrXfm <- list(posSqr = c("MP"), negSqr = c("MN"), posLog = c("MP", "SP"), negLog = c("SN"),
#'                posInv = c("MP", "SP", "EP"), negInv = c("EN"))
#' dtaOut <- jmvReadWrite::transform_vars_omv(dtaInp = dtaInp, varXfm = crrXfm)
#' jmv::descriptives(data = dtaOut, skew = TRUE, sw = TRUE)
#'
#' crrXfm <- list(mdrPos = c("MP"), mdrNeg = c("MN"), strPos = c("SP"), strNeg = c("SN"),
#'                svrPos = c("EP"), svrNeg = c("EN"))
#' dtaOut <- jmvReadWrite::transform_vars_omv(dtaInp = dtaInp, varXfm = crrXfm)
#' jmv::descriptives(data = dtaOut, skew = TRUE, sw = TRUE)
#'
#'
#' @export transform_vars_omv
#'
transform_vars_omv <- function(dtaInp = NULL, fleOut = "", varXfm = NULL, psvAnl = FALSE, usePkg = c("foreign", "haven"), selSet = "", ...) {
    # check the input parameter: varXfm needs to be given
    if (length(varXfm) < 1 || !is.list(varXfm)) {
        stop("Calling transform_vars_omv requires the parameter varXfm, using the correct format (see Details in help).")
    }
    names(varXfm) <- gsub("mdrPos", "posSqr", gsub("strPos", "posLog", gsub("svrPos", "posInv",
                     gsub("mdrNeg", "negSqr", gsub("strNeg", "negLog", gsub("svrNeg", "negInv", names(varXfm)))))))
    if (!all(names(varXfm) %in% c("posSqr", "negSqr", "posLog", "negLog", "posInv", "negInv"))) {
        stop("The parameter varXfm has an invalid entry (wrong name), please use the correct format (see Details in help).")
    }

    # check and import input data set (either as data frame or from a file)
    if (!is.null(list(...)[["fleInp"]])) stop("Please use the argument dtaInp instead of fleInp.")
    dtaFrm <- inp2DF(dtaInp = dtaInp, usePkg = usePkg, selSet = selSet, ...)

    # check whether all variables in varXfm are also contained in the input data frame
    if (!all(unique(unlist(varXfm, use.names = FALSE)) %in% names(dtaFrm))) {
        stop(sprintf("All columns / variables given in the parameter varXfm need to be contained in the input data frame (dtaInp), but variable(s) %s are missing.",
          paste(setdiff(unique(unlist(varXfm, use.names = FALSE)), names(dtaInp)), collapse = ", ")))
    }

    # TO-DO: replace the functionality underneath with compute_omv once implemented
    for (crrNme in names(varXfm)) {
        crrSfx <- toupper(substr(crrNme, 4, 6))
        cmdJmv <- ifelse(crrSfx == "INV", "1 / (RPL_VAR)", sprintf("%s(RPL_VAR)", gsub("SQR", "SQRT", gsub("LOG", "LOG10", crrSfx))))
        for (crrVar in varXfm[[crrNme]]) {
            if        (substr(crrNme, 1, 3) == "pos") {
                rplVar <- paste0("RPL_VAR", ifelse(crrSfx != "SQR" && min(dtaFrm[, crrVar], na.rm = TRUE) < 1, " - VMIN(RPL_VAR) + 1",
                                            ifelse(min(dtaFrm[, crrVar], na.rm = TRUE) < 0, " - VMIN(RPL_VAR)", "")))
            } else if (substr(crrNme, 1, 3) == "neg") {
                rplVar <- paste0("VMAX(RPL_VAR)", ifelse(substr(crrNme, 4, 6) != "Sqr", " + 1", ""), " - RPL_VAR")
            }
            tgtVar <- paste0(crrVar, "_", crrSfx)
            tgtAtt <- list(columnType = "Computed", formula = gsub("RPL_VAR", gsub("RPL_VAR", crrVar, rplVar), cmdJmv))
            dtaFrm[, tgtVar] <- eval(parse(text = gsub("SQRT", "sqrt", gsub("LOG10", "log10", gsub("RPL_VAR",
              gsub("RPL_VAR", paste0("dtaFrm[, \"", crrVar, "\"]"), gsub("VMIN", "min", gsub("VMAX", "max", rplVar))), cmdJmv)))))
            dtaFrm[tgtVar]   <- setAtt(names(tgtAtt), inpObj = tgtAtt, outObj = dtaFrm[tgtVar])
        }
    }

    # rtnDta in globals.R (unified function to either write the data frame, open it in a new jamovi session or return it)
    rtnDta(dtaFrm = dtaFrm, fleOut = fleOut, dtaTtl = jmvTtl("_transform_vars"), psvAnl = psvAnl, dtaInp = dtaInp, ...)
}

#' Imagine that you worked for a record company and that your boss was interested in predicting album sales from
#' advertising.
#'
#' The data set is fictional and was constructed by Andy Field who therefore owns the copyright. The data set is also publicly
#' available on the website that accompanies Andy Field's book, https://edge.sagepub.com/field5e.
#' Without Andy Field's explicit consent, this data set may not be distributed for commercial purposes, this data set may not be
#' edited, and this data set may not be presented without acknowledging its source (i.e., the terms of a CC BY-NC-ND license).
#'
#' Reference: Field, A. P. (2017). Discovering Statistics Using IBM SPSS Statistics (5th ed.). Sage.
#'
#'@format A data frame with 60 rows, each one representing a different album, and 5 variables:
#'\describe{
#'   \item{\code{selSbj}}{Select the data in this row (1) or not (0)}
#'   \item{\code{Adverts}}{Amount (in thousands of pounds) spent promoting the album before release}
#'   \item{\code{Airplay}}{How many times songs from the album were played on a prominent national radio station in the week before release}
#'   \item{\code{Image}}{How attractive people found the band's image (out of 10)}
#'   \item{\code{Sales}}{Sales (in thousands) of each album in the week after release}
#'}
"AlbumSales"

#' The Effect of Vitamin C on Tooth Growth in Guinea Pigs
#'
#'@format A data frame with 60 rows and 6 variables:
#'\describe{
#'   \item{\code{ID}}{ID of the guinea pig}
#'   \item{\code{supp}}{Supplement type (VC: Vitamin C or OJ: Orange juice)}
#'   \item{\code{supp2}}{Transformation of the supplement type (factor to numerical: VC = 1; OJ = 2)}
#'   \item{\code{dose}}{Dose in grams / day}
#'   \item{\code{dose2}}{Dose in grams / day}
#'   \item{\code{len}}{Tooth length}
#'   \item{\code{logLen}}{Natural logarithm of the tooth length (len)}
#'}
"ToothGrowth"

#' Twenty-five personality self-report items taken from the International Personality Item Pool
#'
#' The data set contains responses from 250 participants filling in twenty-five personality self-report items
#' taken from the International Personality Item Pool (https://ipip.ori.org) as part of the Synthetic Aperture
#' Personality Assessment (SAPA) web-based personality assessment (https://sapa-project.org) project.
#' The 25 items are organized by five putative factors: Agreeableness (A1 to A5), Conscientiousness (C1 to C5),
#' Extraversion (E1 to E5), Neuroticism (N1 to N5), and Openness (N1 to N5). The items were short phrases that
#' the respondent should answer by indicating how accurately the statement describes their typical behaviour or
#' attitude. Responses were collected using a 6-point scale: 1 - Very inaccurate, 2 - Moderately inaccurate,
#' 3 - Slightly inaccurate, 4 - Slightly accurate, 5 - Moderately accurate, 6 - Very accurate.
#'
#'@format A data frame with 254 rows (250 original respondents, 4 generated for testing) and 33 variables:
#'\describe{
#'   \item{\code{ID}}{Respondent ID}
#'   \item{\code{A1}}{Am indifferent to the feelings of others. (reversed)}
#'   \item{\code{A2}}{Inquire about others' well-being.}
#'   \item{\code{A3}}{Know how to comfort others.}
#'   \item{\code{A4}}{Love children.}
#'   \item{\code{A5}}{Make people feel at ease.}
#'   \item{\code{C1}}{Am exacting in my work.}
#'   \item{\code{C2}}{Continue until everything is perfect.}
#'   \item{\code{C3}}{Do things according to a plan.}
#'   \item{\code{C4}}{Do things in a half-way manner. (reversed)}
#'   \item{\code{C5}}{Waste my time. (reversed)}
#'   \item{\code{E1}}{Don't talk a lot. (reversed)}
#'   \item{\code{E2}}{Find it difficult to approach others. (reversed)}
#'   \item{\code{E3}}{Know how to captivate people.}
#'   \item{\code{E4}}{Make friends easily.}
#'   \item{\code{E5}}{Take charge.}
#'   \item{\code{N1}}{Get angry easily.}
#'   \item{\code{N2}}{Get irritated easily.}
#'   \item{\code{N3}}{Have frequent mood swings.}
#'   \item{\code{N4}}{Often feel blue.}
#'   \item{\code{N5}}{Panic easily.}
#'   \item{\code{O1}}{Am full of ideas.}
#'   \item{\code{O2}}{Avoid difficult reading material. (reversed)}
#'   \item{\code{O3}}{Carry the conversation to a higher level.}
#'   \item{\code{O4}}{Spend time reflecting on things.}
#'   \item{\code{O5}}{Will not probe deeply into a subject. (reversed)}
#'   \item{\code{gender}}{Gender of the respondent (female, male)}
#'   \item{\code{age}}{Age of the respondent (years)}
#'   \item{\code{AD}}{Exponent of age (computed: EXP(age))}
#'   \item{\code{AF}}{Random data (for testing)}
#'   \item{\code{AG}}{Random data (for testing)}
#'   \item{\code{age_tr}}{Age of the respondent (transformed, as decades: 1 - 10-19, 2 - 20-29, 3 - 30-39, 4 - 40-49, 5 - 50-59, 6 - 60 and over)}
#'   \item{\code{ID2}}{Respondent ID (for testing; "A" + random-generated 5-digit-code)}
#'}
"bfi_sample"

#' Twenty-five personality self-report items taken from the International Personality Item Pool
#' (includes jamovi-attributes; should result in a file identical to bfi_sample2.omv under "extdata" when
#' written with write_omv)
#'
#' The data set contains responses from 250 participants filling in twenty-five personality self-report items
#' taken from the International Personality Item Pool (https://ipip.ori.org) as part of the Synthetic Aperture
#' Personality Assessment (SAPA) web-based personality assessment (https://sapa-project.org) project.
#' The 25 items are organized by five putative factors: Agreeableness (A1 to A5), Conscientiousness (C1 to C5),
#' Extraversion (E1 to E5), Neuroticism (N1 to N5), and Openness (N1 to N5). The items were short phrases that
#' the respondent should answer by indicating how accurately the statement describes their typical behaviour or
#' attitude. Responses were collected using a 6-point scale: 1 - Very inaccurate, 2 - Moderately inaccurate,
#' 3 - Slightly inaccurate, 4 - Slightly accurate, 5 - Moderately accurate, 6 - Very accurate.
#' The data set includes the jamovi-attributes. It is supposed to result in an identical file compared to the
#' bfi_sample2.omv-file contained in the extdata-directory of the package when written with write_omv.
#'
#'@format A data.frame with 250 rows and 29 variables
#'\describe{
#'   \item{\code{ID}}{Respondent ID}
#'   \item{\code{A1}}{Am indifferent to the feelings of others. (reversed)}
#'   \item{\code{A2}}{Inquire about others' well-being.}
#'   \item{\code{A3}}{Know how to comfort others.}
#'   \item{\code{A4}}{Love children.}
#'   \item{\code{A5}}{Make people feel at ease.}
#'   \item{\code{C1}}{Am exacting in my work.}
#'   \item{\code{C2}}{Continue until everything is perfect.}
#'   \item{\code{C3}}{Do things according to a plan.}
#'   \item{\code{C4}}{Do things in a half-way manner. (reversed)}
#'   \item{\code{C5}}{Waste my time. (reversed)}
#'   \item{\code{E1}}{Don't talk a lot. (reversed)}
#'   \item{\code{E2}}{Find it difficult to approach others. (reversed)}
#'   \item{\code{E3}}{Know how to captivate people.}
#'   \item{\code{E4}}{Make friends easily.}
#'   \item{\code{E5}}{Take charge.}
#'   \item{\code{N1}}{Get angry easily.}
#'   \item{\code{N2}}{Get irritated easily.}
#'   \item{\code{N3}}{Have frequent mood swings.}
#'   \item{\code{N4}}{Often feel blue.}
#'   \item{\code{N5}}{Panic easily.}
#'   \item{\code{O1}}{Am full of ideas.}
#'   \item{\code{O2}}{Avoid difficult reading material. (reversed)}
#'   \item{\code{O3}}{Carry the conversation to a higher level.}
#'   \item{\code{O4}}{Spend time reflecting on things.}
#'   \item{\code{O5}}{Will not probe deeply into a subject. (reversed)}
#'   \item{\code{gender}}{Gender of the respondent (female, male)}
#'   \item{\code{age}}{Age of the respondent (years)}
#'   \item{\code{ID2}}{Respondent ID (for testing; "A" + random-generated 4-digit-code)}
#'}
"bfi_sample2"

#' Twenty-five personality self-report items taken from the International Personality Item Pool
#' (testing file for ordered factors / "Ordinal"-variables in jamovi)
#'
#' The data set contains responses from 250 participants filling in twenty-five personality self-report items
#' taken from the International Personality Item Pool (https://ipip.ori.org) as part of the Synthetic Aperture
#' Personality Assessment (SAPA) web-based personality assessment (https://sapa-project.org) project.
#' The 25 items are organized by five putative factors: Agreeableness (A1 to A5), Conscientiousness (C1 to C5),
#' Extraversion (E1 to E5), Neuroticism (N1 to N5), and Openness (N1 to N5). The items were short phrases that
#' the respondent should answer by indicating how accurately the statement describes their typical behaviour or
#' attitude. Responses were collected using a 6-point scale: 1 - Very inaccurate, 2 - Moderately inaccurate,
#' 3 - Slightly inaccurate, 4 - Slightly accurate, 5 - Moderately accurate, 6 - Very accurate.
#' The data set includes the jamovi-attributes. It is supposed to result in an identical file compared to the
#' bfi_sample2.omv-file contained in the extdata-directory of the package when written with write_omv.
#'
#'@format A data.frame with 250 rows and 28 variables
#'\describe{
#'   \item{\code{ID}}{Respondent ID}
#'   \item{\code{A1}}{Am indifferent to the feelings of others. (reversed)}
#'   \item{\code{A2}}{Inquire about others' well-being.}
#'   \item{\code{A3}}{Know how to comfort others.}
#'   \item{\code{A4}}{Love children.}
#'   \item{\code{A5}}{Make people feel at ease.}
#'   \item{\code{C1}}{Am exacting in my work.}
#'   \item{\code{C2}}{Continue until everything is perfect.}
#'   \item{\code{C3}}{Do things according to a plan.}
#'   \item{\code{C4}}{Do things in a half-way manner. (reversed)}
#'   \item{\code{C5}}{Waste my time. (reversed)}
#'   \item{\code{E1}}{Don't talk a lot. (reversed)}
#'   \item{\code{E2}}{Find it difficult to approach others. (reversed)}
#'   \item{\code{E3}}{Know how to captivate people.}
#'   \item{\code{E4}}{Make friends easily.}
#'   \item{\code{E5}}{Take charge.}
#'   \item{\code{N1}}{Get angry easily.}
#'   \item{\code{N2}}{Get irritated easily.}
#'   \item{\code{N3}}{Have frequent mood swings.}
#'   \item{\code{N4}}{Often feel blue.}
#'   \item{\code{N5}}{Panic easily.}
#'   \item{\code{O1}}{Am full of ideas.}
#'   \item{\code{O2}}{Avoid difficult reading material. (reversed)}
#'   \item{\code{O3}}{Carry the conversation to a higher level.}
#'   \item{\code{O4}}{Spend time reflecting on things.}
#'   \item{\code{O5}}{Will not probe deeply into a subject. (reversed)}
#'   \item{\code{gender}}{Gender of the respondent (Females, Males)}
#'   \item{\code{age}}{Age of the respondent (years)}
#'}
"bfi_sample3"

# binds the protcol buffer variables jamovi.coms.AnalysisResponse, jamovi.coms.AnalysisOptions,
# jamovi.coms.AnalysisOption, jamovi.coms.ResultsElement, and jamovi.coms.ResultsGroup locally
# to the function, otherwise devtools::check() - required before submitting to CRAN - throws an
# error
if (getRversion() >= "2.15.1") {
    utils::globalVariables(c("jamovi.coms.AnalysisResponse", "jamovi.coms.AnalysisOptions",
                             "jamovi.coms.AnalysisOption", "jamovi.coms.ResultsElement",
                             "jamovi.coms.ResultsGroup"))
}

# =================================================================================================
# the next lines store the currently supported versions (stored in meta / MANIFEST.MF)
# and the string that precedes the version number
lstMnf <- list(mnfVer = c("Manifest-Version",        "1.0"),
               datVer = c("Data-Archive-Version",    "1.0.2"),
               jmvVer = c("jamovi-Archive-Version",  "11.0"),
               crtStr = c("Created-By"))

# the next lines are dealing with storing the global and the data column attributes (that go into
# metadata.json inside the .omv-file; the currently defined defaults are in accordance with
# jamovi-Archive-Version: 11.0 (from jamovi 1.8)
mtaGlb <- list(rowCount = NA, columnCount = NA, removedRows = list(), addedRows = list(), fields = list(), transforms = list(), weights = NULL)
mtaFld <- list(name = "", id = NA, columnType = "Data", dataType = "Integer", measureType = "Nominal", formula = "", formulaMessage = "",
               parentId = 0, width = 100, type = "number", outputAnalysisId = NA, outputOptionName = "", outputName = "",
               outputDesiredColumnName = "", outputAssignedColumnName = "", importName = "", description = "", transform = 0,
               edits = list(), missingValues = list(), trimLevels = TRUE, filterNo = NA, active = FALSE)
grpMta <- paste0("^", paste(c(names(mtaGlb), names(mtaFld)), collapse = "$|^"), "$")

# =================================================================================================
# functions for checking parameters (file and directory existence, correct file extension, correct
# dimensions and existence of data frames) and normalizing the file name

#            jamovi        CSV    TSV    Rdata           RDS    SPSS           Stata  SAS
vldExt <- c("omv", "omt", "csv", "tsv", "rdata", "rda", "rds", "sav", "zsav", "dta", "sas7bdat", "sd2", "sd7", "xpt", "stx", "stc")

# REMEMBER: requires the full file name, NOT the directory
chkDir <- function(fleNme = "", wrtPrm = TRUE) {
    if (! utils::file_test("-d", dirname(fleNme))) {
        stop(sprintf("Directory (%s) doesn\'t exist.", dirname(fleNme)))
    }
    if (file.access(dirname(fleNme), mode = 2) != 0) {
        stop(sprintf("The directory (%s) exists, but you don\'t have writing permissions in that directory.", dirname(fleNme)))
    }
    TRUE
}

chkDtF <- function(dtaFrm = NULL, minSze = c(0, 1)) {
    if (length(minSze) != 2) minSze <- rep(minSze[1], 2)
    if (is.null(dtaFrm) || !is.data.frame(dtaFrm) || length(dim(dtaFrm)) != 2) {
        stop("Input data are either not a data frame or have incorrect (only one or more than two) dimensions.")
    } else if (any(dim(dtaFrm) < minSze)) {
        stop(sprintf("The %s dimension of the input data frame has not the required size (%d < %d).",
                     ifelse(which(dim(dtaFrm) < minSze)[1] == 1, "first", "second"), dim(dtaFrm)[dim(dtaFrm) < minSze][1], minSze[dim(dtaFrm) < minSze][1]))
    }
    TRUE
}

chkExt <- function(fleNme = "", extNme = c("")) {
    if (any(nzchar(tools::file_ext(fleNme))) && !hasExt(fleNme, extNme)) {
        stop(sprintf("File name (%s) contains an unsupported file extension (%s).",
                     basename(fleNme),
                     paste(paste0(".", tools::file_ext(fleNme)[tools::file_ext(fleNme) != extNme]), collapse = ", ")))
    }
    TRUE
}

chkFle <- function(fleNme = "", isZIP = FALSE, fleCnt = "") {
    if (!is.character(fleNme) || !is.logical(isZIP) || !is.character(fleCnt)) {
        stop("chkFle: Unsupported input parameter type.")
    }
    if (!utils::file_test("-f", fleNme)) {
        stop(sprintf("File \"%s\" not found.", fleNme))
    }
    if (isZIP) {
        hdrStr <- readBin(tmpHdl <- file(fleNme, "rb"), "character")
        close(tmpHdl)
        # only "PK\003\004" is considered, not "PK\005\006" (empty ZIP) or "PK\007\008" (spanned [over several files])
        if (hdrStr != "PK\003\004\024" && hdrStr != "PK\003\004") {
            stop(sprintf("chkFle: File \"%s\" has not the correct file format (is not a ZIP archive).", basename(fleNme)))
        }
    }
    if (nchar(fleCnt) > 0) {
        if (!any(grepl(fleCnt, zip::zip_list(fleNme)$filename))) {
            stop(sprintf("chkFle: File \"%s\" doesn\'t contain the file \"%s\".", basename(fleNme), fleCnt))
        }
    }
    TRUE
}

chkVar <- function(dtaFrm = NULL, varNme = c()) {
    if (is.null(varNme) || length(varNme) == 0 || !all(nzchar(varNme))) return(FALSE)
    if (!all(varNme %in% names(dtaFrm))) {
        stop(sprintf("The variable(s) %s are not contained in the current data set.", paste(varNme[! (varNme %in% names(dtaFrm))], collapse = ", ")))
    }
    TRUE
}

hasExt <- function(fleNme = "", extNme = c("")) {
    any(tolower(tools::file_ext(fleNme)) == tolower(extNme))
}

hasPkg <- function(usePkg = c()) {
    all(vapply(usePkg, function(X) nzchar(system.file(package = X)), logical(1)))
}

nrmFle <- function(fleNme = "") {
    file.path(normalizePath(dirname(fleNme)), basename(fleNme))
}

fmtFlI <- function(fleInp = c(), minLng = 1, maxLng = Inf, excExt = "") {
    # normalize the path of the input file and then check whether the file exists and whether it is of a supported file type
    if (length(fleInp) < minLng || length(fleInp) > maxLng) {
        clsRmv()
        stop(sprintf("The fleInp-argument is supposed to be a character vector with a minimal length of %.0f and a maximal length of %.0f (current length is %.0f).%s",
                     minLng, maxLng, length(fleInp), ifelse(length(fleInp) > maxLng, "\n  If you would like to process several files, call the function individually for each.", "")))
    }
    fleInp <- unname(vapply(fleInp, nrmFle, character(1)))
    all(vapply(fleInp, chkFle, logical(1)))
    all(vapply(fleInp, chkExt, logical(1), setdiff(vldExt, excExt)))
    fleInp
}

fmtFlO <- function(fleOut = "") {
    if (!nzchar(fleOut) || !hasExt(fleOut, c("omv", "omt"))) {
        clsRmv()
        stop("fleOut needs to be a valid non-empty file name (character), and the file extension for output file needs to be .omv or .omt.")
    }
    nrmFle(fleOut)
}

# close and remove files, and remove the file handles
clsRmv <- function() {
    for (i in getAllConnections()) {
        if (i < 3) next # on all OSes: stdin [0], stdout [1], stderr [2]
        fleHdl <- getConnection(i)
        crrFle <- summary(fleHdl)[["description"]]
        close(fleHdl)
        unlink(crrFle)
        rm(crrFle)
    }

    TRUE
}

# =================================================================================================
# convert columns to another class (e.g., integer to factor) preserving attributes, check whether
# a column contains / can be converted into integers, and convert undefined characters into UTF-8

cnvCol <- function(crrCol = NULL, tgtTyp = "character") {
    if (methods::is(crrCol, tgtTyp)) return(crrCol)

    # store attributes
    crrAtt <- attributes(crrCol)
    dffAtt <- setdiff(names(crrAtt), c("levels", "class"))
    # pre-processing (convert date, trim spaces and round where necessary)
    if (methods::is(crrCol, "POSIXct")) crrCol <- as.Date(crrCol)
    if (is.character(crrCol)) crrCol <- trimws(crrCol)
    if (is.numeric(crrCol) && tgtTyp ==  "integer") crrCol <- round(crrCol)
    # actual conversion; jamovi stores factors differently depending on whether they have the dataType Integer or Text
    if (is.factor(crrCol) && tgtTyp %in% c("integer", "numeric")) {
        crrCol <- if (intFnC(crrCol)) methods::as(as.character(crrCol), tgtTyp) else as.integer(crrCol) - 1L
    } else if (tgtTyp == "factor") {
        # tibble: conversion if the source is a column with the type dbl+lbl
        if ("labels" %in% names(crrAtt)) {
            crrCol <- factor(crrCol, levels = unname(crrAtt$labels), labels = names(crrAtt$labels))
            dffAtt <- setdiff(dffAtt, "labels")
        # foreign: conversion if the source is a column that has the attribute "value.labels"
        } else if ("value.labels" %in% names(crrAtt)) {
            crrCol <- factor(crrCol, levels = unname(crrAtt[["value.labels"]]), labels = cnvUTF(names(crrAtt[["value.labels"]])))
            dffAtt <- setdiff(dffAtt, "value.labels")
        # “usual” columns (without specified attributes)
        } else {
            crrCol <- as.factor(crrCol)
        }
    } else {
        crrCol <- methods::as(crrCol, tgtTyp)
    }
    if (length(dffAtt) > 0) crrCol <- setAtt(attLst = dffAtt, inpObj = crrAtt, outObj = as.data.frame(crrCol))[[1]]

    crrCol
}

intFnC <- function(crrCol = NULL) {
    facLvl <- if (is.factor(crrCol)) levels(crrCol) else unique(trimws(crrCol))

    all(!is.na(suppressWarnings(as.integer(facLvl)))) && all(as.character(as.integer(facLvl)) == facLvl)
}

cnvUTF <- function(inpStr = c()) {
    # assign "latin1" to those entries that have special characters (e.g., ä, æ, ß, etc.)
    Encoding(inpStr) == "latin1"
    # return a trimmed version of the input vector that is converted into UTF-8
    trimws(enc2utf8(inpStr))
}


# =================================================================================================
# initializing and handling ProtoBuffers

jmvPtB <- function() {
    # exit with TRUE if the ProtoBuffers are already initialized
    if (exists("jamovi.coms.Status")) return(TRUE)
    # check whether all required packages are present
    synPkg <- c("RProtoBuf", "jmvcore")
    if (!hasPkg(synPkg)) {
        warning(sprintf("For using protocol buffers, the package(s) \"%s\" need(s) to be installed.\n\n",
          paste0(synPkg[!vapply(synPkg, hasPkg, logical(1))], collapse = "\", \"")))
        return(FALSE)
    }
    # check the two possible places for the jamovi.proto file
    flePtB <- system.file("jamovi.proto", package = "jmvcore")
    if (!nzchar(flePtB)) flePtB <- system.file("inst", "jamovi.proto", package = "jmvcore")
    if (!nzchar(flePtB)) {
        warning("For using protocol buffers, the protocol file \"jamovi.proto\" (from the jmvcore-package) is required.\n\n")
        return(FALSE)
    }
    # read protocol file and initialize the protobuffers with it
    if (requireNamespace("RProtoBuf", quietly = TRUE)) {
        # try reading the protobuffer-file (if it can be read / parsed, tryCatch returns TRUE and the syntax can be extracted)
        # the is.null() is a way to enforce one-liners: either command readProtoFiles and message returns NULL and hence, either
        # TRUE (first line - is.null = TRUE) or FALSE (second line - !is.null = FALSE) are returned
        tryCatch(expr  = is.null(RProtoBuf::readProtoFiles(flePtB)),
                 error = function(e) !is.null(message("Error when loading protocol definition, syntax can\'t be extracted:\n", e)))
    } else {
        warning("The package RProtoBuf can not be initialized, try re-installing it.\n\n")
        FALSE
    }
}

var2PB <- function(inpVar = NULL) {
    # ensure that the jamovi protocol buffers are initiailized
    jmvPtB()
    # the protocol buffers in jamovi actually only support lists as data
    # structures, hence the as.list() conversions for converting vectors

    # NULL (o) ================================================================
    if        (is.null(inpVar)) {
        tmpPB   <- RProtoBuf::new(jamovi.coms.AnalysisOption)
        tmpPB$o <- 2
        return(tmpPB)
    # BOOLEAN (o) =============================================================
    } else if (is.logical(inpVar)) {
        if (length(inpVar) == 1) {
            tmpPB   <- RProtoBuf::new(jamovi.coms.AnalysisOption)
            tmpPB$o <- as.integer(inpVar)
            return(tmpPB)
        } else {
            var2PB(as.list(inpVar))
        }
    # INTEGER (i) =============================================================
    } else if (is.numeric(inpVar) &&  all(inpVar - floor(inpVar) == 0)) {
        if (length(inpVar) == 1) {
            return(RProtoBuf::new(jamovi.coms.AnalysisOption, i = inpVar))
        } else {
            var2PB(as.list(inpVar))
        }
    # DECIMAL (d) =============================================================
    } else if (is.numeric(inpVar)) {
        if (length(inpVar) == 1) {
            tmpPB   <- RProtoBuf::new(jamovi.coms.AnalysisOption)
            tmpPB$d <- inpVar
            return(tmpPB)
        } else {
            var2PB(as.list(inpVar))
        }
    # STRING (s) ==============================================================
    } else if (is.character(inpVar)) {
        if (length(inpVar) == 1) {
            return(RProtoBuf::new(jamovi.coms.AnalysisOption, s = inpVar))
        } else {
            var2PB(as.list(inpVar))
        }
    # CONTAINER (c) ===========================================================
    } else if (is.list(inpVar)) {
        resLst <- list()
        for (i in seq_along(inpVar)) {
            resLst[[i]] <- var2PB(inpVar[[i]])
        }
        tmpPB <- RProtoBuf::new(jamovi.coms.AnalysisOptions, options = resLst)
        if (!is.null(names(inpVar))) {
            tmpPB$hasNames <- TRUE
            tmpPB$names    <- names(inpVar)
        }
        return(RProtoBuf::new(jamovi.coms.AnalysisOption, c = tmpPB))
    # otherwise, throw error ==================================================
    } else {
        clsRmv()
        stop("Element not implemented for conversion to protocol buffer.")
    }
}


# =================================================================================================
# get function arguments and adjust them / select those valid for the current function call

adjArg <- function(fcnNme = c(), dflArg = list(), varArg = list(), fxdArg = c()) {
    chgArg <- setdiff(intersect(fcnArg(fcnNme), names(varArg)), fxdArg)
    c(dflArg[setdiff(names(dflArg), chgArg)], varArg[chgArg])
}

fcnArg <- function(fcnNme = c()) {
    if        (is.character(fcnNme) && length(fcnNme) == 1) {
        eval(parse(text = paste0("formalArgs(", fcnNme, ")")))
    } else if (is.character(fcnNme) && length(fcnNme) == 2) {
        eval(parse(text = paste0("formalArgs(getS3method(\"", fcnNme[1], "\", \"", fcnNme[2], "\"))")))
    } else {
        clsRmv()
        stop("The argument to fcnArg must be a character (vector) with 1 or 2 elements.")
    }
}


# =================================================================================================
# functions for handling setting and storing metadata-information / attributes

bckAtt <- function(dtaFrm = NULL, bckCol = FALSE) {
    attLst <- list(dtaFrm = list(), dtaCol = list())
    if (is.null(dtaFrm)) return(attLst)
    if (is.data.frame(dtaFrm)) dtaFrm <- list(dtaFrm)
    for (i in seq_along(dtaFrm)) {
        attLst[["dtaCol"]] <- c(attLst[["dtaCol"]], lapply(dtaFrm[[i]][, setdiff(names(dtaFrm[[i]]), names(attLst[["dtaCol"]]))], attributes))
    }
    attLst[["dtaFrm"]] <- attributes(dtaFrm[[1]])

    attLst
}

setAtt <- function(attLst = c(), inpObj = NULL, outObj = NULL) {
    if (!is.character(attLst)) stop("setAtt: The parameter attLst is supposed to be a character vector.")
    if (!is.list(inpObj))      stop("setAtt: The parameter inpObj is supposed to be either a list or a data frame.")
    if (!is.list(outObj))      stop("setAtt: The parameter outObj is supposed to be either a list or a data frame.")

    for (attNme in attLst) {
        # ensure that we have one data frame and one list; the problem is that data frames
        # frames are both lists and data frames, and therefore an error is thrown if BOTH
        # input and output objects are lists but not data frames
        if (identical(sort(c(class(inpObj), class(outObj))), c("data.frame", "list"))) {
            # if the output object is the mtaDta-variable, the input object must be the data frame
            # which contains the attribute in attNme (chkAtt), that are then stored in the mtaDta-
            # variable; the attribute might be empty (chkAtt == FALSE), and then the default is kept
            if        (is.data.frame(inpObj)) {
                if        (dim(inpObj)[2] >  1 &&  chkAtt(inpObj,      attNme)) {
                    outObj[[attNme]] <- attr(inpObj,      attNme)
                } else if (dim(inpObj)[2] == 1 &&  chkAtt(inpObj[[1]], attNme)) {
                    outObj[[attNme]] <- attr(inpObj[[1]], attNme)
                }
                eval(parse(text = paste0("")))
            # if the input object is the mtaDta-variable (which is a list), then the attribute is set
            # in the output object unless the attribute already exists in the ouput object (!chkAtt -
            # it shouldn't be overwritten)
            } else if (is.data.frame(outObj)) {
                if        (dim(outObj)[2] >  1 && !chkAtt(outObj,      attNme)) {
                    attr(outObj,      attNme) <- inpObj[[attNme]]
                } else if (dim(outObj)[2] == 1 && !chkAtt(outObj[[1]], attNme)) {
                    attr(outObj[[1]], attNme) <- inpObj[[attNme]]
                }
#               eval(parse(text = paste0("attr(outObj", ifelse(dim(outObj)[2] == 1, "[[1]]", ""), ", attNme) <- inpObj[[attNme]]")))
            }
        # the case which is critical is if both input and output objects are lists (then the first
        # part of the if-conditions above - is.list - wouldn't work)
        } else {
            errDsc <- paste0("\nOne input object (inpObj or outObj) must be a list, the other must be a data frame.\n\n",
                             "attNme: ", attNme, "\n",
                             "attLst: ", paste0(attLst, collapse = ", "), "\n\n",
                             "inpObj:\n", utils::capture.output(utils::str(inpObj)), "\n\n",
                             "outObj:\n", utils::capture.output(utils::str(outObj)), "\n\n")
            stop(sprintf("Error when storing or accessing meta-data information. Please send the file causing the error to sebastian.jentschke@uib.no\n%s", errDsc))
        }
    }

    outObj
}

rmvMsV <- function(dtaFrm = NULL) {
    for (N in names(dtaFrm))
        attr(dtaFrm[, N], "missingValues") <- NULL

    dtaFrm
}

rmvAtt <- function(attObj = NULL, att2Rm = NULL) {
    if (is.null(att2Rm))
        att2Rm <- setdiff(names(attributes(attObj)), c("class", "comment", "dim", "jmv-id", "jmv-desc", "levels", "names", "row.names", "values"))
    for (crrAtt in att2Rm) {
        attr(attObj, crrAtt) <- NULL
    }

    attObj
}

nllAtt <- function(attObj = NULL, att2Rs = c()) {
    for (crrAtt in att2Rs) {
        if (crrAtt %in% names(attributes(attObj))) attr(attObj, crrAtt) <- methods::as(c(), class(attr(attObj, crrAtt)))
    }

    attObj
}

rstAtt <- function(dtaFrm = NULL, attLst = NULL) {
    for (crrAtt in setdiff(names(attLst[["dtaFrm"]]), c("names", "row.names", "class", "fltLst"))) attr(dtaFrm, crrAtt) <- attLst[["dtaFrm"]][[crrAtt]]
    for (crrNme in names(dtaFrm)) {
        if (!is.null(attLst[["dtaCol"]][[crrNme]])) {
            dtaFrm[crrNme] <- setAtt(setdiff(names(attLst[["dtaCol"]][[crrNme]]), names(attributes(dtaFrm[crrNme]))), attLst[["dtaCol"]][[crrNme]], dtaFrm[crrNme])
        }
    }

    dtaFrm
}

chkAtt <- function(attObj = NULL, attNme = "", attVal = NULL) {
   ((attNme %in% names(attributes(attObj))) && length(attr(attObj, attNme)) > 0 &&
     ifelse(!is.null(attVal), grepl(attVal, attr(attObj, attNme)), TRUE))
}

chkFld <- function(fldObj = NULL, fldNme = "", fldVal = NULL) {
   ((fldNme %in% names(fldObj))    && length(fldObj[[fldNme]])     > 0 && ifelse(!is.null(fldVal), grepl(fldVal, fldObj[[fldNme]]),     TRUE))
}

# =================================================================================================
# function handling to have either a data frame or a character (pointing to a file) as input
inp2DF <- function(dtaInp = NULL, minDF = 1, maxDF = 1, rmvEmp = FALSE, usePkg = c("foreign", "haven"), selSet = "", ...) {
    usePkg <- match.arg(usePkg)
    # check and format input and output files, handle / check further input arguments:
    # if the input is a data frame, it is “embedded” in a list (in order to permit to read
    # and to concatenate this data frame with further data frames given as fleInp-attribute
    # and read via the lapply function)
    if (is.data.frame(dtaInp) && chkDtF(dtaInp)) {
        lstDF <- list(dtaInp)
        if (!is.null(attr(dtaInp, "fleInp"))) {
            lstDF <- c(lstDF, lapply(fmtFlI(attr(dtaInp, "fleInp"), minLng = minDF - 1, maxLng = maxDF - 1), function(x) read_all(fleInp = x, usePkg = usePkg, selSet = selSet, ...)))
        }
    # if the input is a character vector (with file names), all file names are read into
    # data frames (using the lapply function)
    } else if (is.character(dtaInp)) {
        lstDF <-              lapply(fmtFlI(dtaInp,                 minLng = minDF - 0, maxLng = maxDF - 0), function(x) read_all(fleInp = x, usePkg = usePkg, selSet = selSet, ...))
    } else {
        clsRmv()
        stop("dtaInp must either be a data frame or a character (pointing to a location where the input file can be found).")
    }
    # if rmvEmp is set, check for rows that are completely empty and remove them
    if (rmvEmp) {
        for (i in seq_along(lstDF)) {
            blnEmp <- apply(lstDF[[i]], 1, function(x) all(is.na(x)))
            if (blnEmp[1] && sum(diff(blnEmp) == -1) == 1) {
                lstDF[[i]] <- lstDF[[i]][-seq(1, which(diff(blnEmp) == -1)), ]
                blnEmp <- apply(lstDF[[i]], 1, function(x) all(is.na(x)))
            }
            if (blnEmp[length(blnEmp)] && sum(diff(blnEmp) == 1) == 1) {
                lstDF[[i]] <- lstDF[[i]][seq(which(diff(blnEmp) == 1)), ]
                blnEmp <- apply(lstDF[[i]], 1, function(x) all(is.na(x)))
            }
            if (any(blnEmp)) {
                stop("Empty rows are not permitted execpt from the begin or the end of an input data frame (in such case, they are automatically removed).")
            }
        }
    }
    # most functions expect only one data frame to be returned, thus, the list
    # used for reading processing those data frames is unpacked if there is
    # only one data frame to return
    if (maxDF == 1) lstDF[[1]] else lstDF
}

# =================================================================================================
# Unified function to handle data frames at the end of the helper functions
# * if the output file name is not empty, the data frame is written to the output file
# * if no output file name was given:
#   - open the data frame in a new session (only in jamovi, and if fleOut is an empty character vector)
#   - return the data frame (in R in any case, or in jamovi if fleOut is NULL)
#   NB: this makes opening the data frame in a new session the default, if in jamovi
rtnDta <- function(dtaFrm = NULL, fleOut = "", dtaTtl = "", wrtPtB = FALSE, psvAnl = FALSE, dtaInp = NULL, ...) {
    if (!is.null(fleOut) && nzchar(fleOut[1])) {
        fleOut <- fmtFlO(fleOut[1])
        write_omv(dtaFrm = dtaFrm, fleOut = fleOut, wrtPtB = wrtPtB, ...)
        # transfer analyses from input to output file
        if (psvAnl) {
            if (is.character(dtaInp)) {
                xfrAnl(dtaInp[1], fleOut)
            } else {
                warning("psvAnl is only possible if dtaInp is a file name (analyses are not stored in data frames, only in the jamovi files).")
            }
        }
        invisible(NULL)
    } else if (isJmv() && is.character(fleOut)) {
        if (psvAnl) warning("psvAnl is only possible if fleOut is a file name (analyses are not stored in data frames, only in the jamovi files).")
        jmvOpn(dtaFrm, dtaTtl = dtaTtl)
        invisible(NULL)
    } else {
        if (psvAnl) warning("psvAnl is only possible if fleOut is a file name (analyses are not stored in data frames, only in the jamovi files).")
        dtaFrm
    }
}

# =================================================================================================
# convert matrix from full to sparse - used for proximities_omv and distances_omv
mtxF2S <- function(dtaFrm = NULL, rmvTrU = FALSE, rmvDgn = FALSE, mtxXps = FALSE, mtxSps = FALSE) {
    if (diff(dim(dtaFrm)) == 0) rownames(dtaFrm) <- names(dtaFrm)
    if (!isSymmetric(as.matrix(dtaFrm))) stop("Input matrix needs to be symmetric.")

    C <- ncol(dtaFrm)
    if (rmvTrU || mtxSps) dtaFrm[upper.tri(dtaFrm)] <- NA
    if (rmvDgn || mtxSps) diag(dtaFrm) <- NA
    if (mtxXps) dtaFrm <- as.data.frame(t(dtaFrm))
    if (mtxSps) dtaFrm <- cbind(data.frame(Variable = names(dtaFrm)[seq(1, C)[ifelse(mtxXps, -C, -1)]]),
                                dtaFrm[seq(1, C)[ifelse(mtxXps, -C, -1)], seq(1, C)[ifelse(mtxXps, -1, -C)]])
    for (crrClm in names(dtaFrm))
        attr(dtaFrm[, crrClm], "measureType") <- ifelse(crrClm == "Variable", "Nominal", "Continuous")

    dtaFrm
}

# =================================================================================================
# function for copying analyses from one data file to another

xfrAnl <- function(fleOrg = "", fleTgt = "") {
    # check whether input and output files are valid and format input and output file names
    chkExt(fleOrg, c("omv", "omt")) && chkFle(fleOrg, isZIP = TRUE) && chkFle(fleOrg, fleCnt = "meta|MANIFEST.MF")
    fleOrg <- fmtFlI(fleOrg, maxLng = 1)
    fleTgt <- fmtFlI(fleTgt, maxLng = 1)

    # extract the list of files contained in the input file, assign tempdir()
    lstOrg <- zip::zip_list(fleOrg)$filename
    lst2Cp <- lstOrg[grepl("index.html|[0-9].*\\s[a-z].*?/", lstOrg)]
    lstCmb <- union(zip::zip_list(fleTgt)$filename, lst2Cp)
    xfrDir <- file.path(tempdir(), "xfrAnl")

    # create a list of files to be copied, extract them from the input file and
    # append them to the output file
    zip::unzip(fleTgt,                 exdir = xfrDir)
    zip::unzip(fleOrg, files = lst2Cp, exdir = xfrDir, overwrite = TRUE)
    zip::zip(fleTgt,   files = lstCmb, root  = xfrDir)

    # remove the files and directories from the list of files to be copied
    unlink(xfrDir, recursive = TRUE)

    TRUE
}

# =================================================================================================
# function for checking which OS is running, whether we are running inside jamovi, for adding
# attributes used by jamovi to data frames (e.g., those opened in Rj or via jTransform), and for
# opening a data set in jamovi (if fleOut is left blank, typically a data frame is returned; if we
# are running inside jamovi, a new data set is opened)

getOS <- function() {
    sysInf <- Sys.info()
    if (!is.null(sysInf)) {
        tolower(gsub("Darwin", "macos", sysInf[["sysname"]]))
    } else {
        ifelse(grepl("^darwin",   R.version$os), "macos",
          ifelse(grepl("linux-gnu", R.version$os), "linux", tolower(.Platform$OS.type)))
    }
}

isJmv <- function() {
    nzchar(Sys.getenv("JAMOVI_R_VERSION"))
}

jmvOpn <- function(dtaFrm = NULL, dtaTtl = "", rtnOut = TRUE) {
    # on both Windows and Linux, jamovi is in the path, and, hence,
    # Sys.which should give the full location
    jmvEXE <- Sys.which("jamovi")
    # if not, we have to determine the position of jamovi under the
    # current OS
    if (!nzchar(jmvEXE)) {
        crrOS <- getOS()
        if        (crrOS == "windows") {
            jmvHme <- jmvPth(R.home(), "Frameworks", TRUE)
            if (!is.null(jmvHme)) jmvEXE <- normalizePath(file.path(jmvHme, "bin", "jamovi.exe"))
        } else if (crrOS == "macos")   {
            jmvHme <- jmvPth(R.home(), "Contents", FALSE)
            if (!is.null(jmvHme)) jmvEXE <- file.path(jmvHme,  "MacOS", "jamovi")
        } else if (crrOS == "linux")   {
            jmvHme <- jmvPth(R.home(), "lib", TRUE)
            if (!is.null(jmvHme)) jmvEXE <- file.path(jmvHme, "bin", "jamovi")
        } else {
            stop(sprintf("Your OS (%s) is currently not implemented. Please report more details to sebastian.jentschke@uib.no to fix that.", crrOS))
        }
    }
    if (nzchar(jmvEXE) && file.exists(jmvEXE)) {
        tmpOut <- tempfile(fileext = ".omv")
        jmvReadWrite::write_omv(dtaFrm, fleOut = tmpOut)
        system2(jmvEXE, args = paste0(" --temp --title=\"", dtaTtl, "\" ", tmpOut), stderr = rtnOut, stdout = rtnOut)
    } else {
        stop(sprintf("The position of the jamovi executable could not be determined or it was not found at the determined position. Determined position: %s", jmvEXE))
    }
}

jmvPth <- function(inpPth = "", strTgt = "", bfrTgt = TRUE) {
    mtcTgt <- gregexpr(strTgt, inpPth)[[1]][1]
    if (mtcTgt > 0) {
        return(substr(inpPth, 1, mtcTgt + ifelse(bfrTgt, -2, nchar(strTgt) - 1)))
    } else {
        return()
    }
}

jmvTtl <- function(sfxTtl = "") {
    # return empty string when not inside jamove (then the title is irrelevant)
    if (!isJmv()) return("")
# TO-DO: replace Dataset with the name of the current data set (once this is implemented)
    return(paste0("Dataset", sfxTtl))
}

test_that("combine_cols_omv works", {
    dtaInp <- jmvReadWrite::bfi_sample2
    nmeOut <- tempfile(fileext = ".omv")
    set.seed(1)
    selRow <- rnorm(nrow(dtaInp)) < 0
    dtaInp[selRow, "A1_1"] <- dtaInp[selRow, "A1"]
    dtaInp[selRow, "A1"]   <- NA

    expect_null(combine_cols_omv(dtaInp = dtaInp, fleOut = nmeOut, varPrs = list(c("A1", "A1_1"))))
    expect_true(file.exists(nmeOut))
    expect_gt(file.info(nmeOut)$size, 1)
    expect_true(chkFle(nmeOut, isZIP = TRUE))
    expect_true(chkFle(nmeOut, fleCnt = "meta"))
    expect_true(chkFle(nmeOut, fleCnt = "metadata.json"))
    expect_true(chkFle(nmeOut, fleCnt = "data.bin"))
    unlink(nmeOut)

    dtaFrm <- combine_cols_omv(dtaInp, varPrs = list(list("A1", "A1_1")))
    expect_s3_class(dtaFrm, "data.frame")
    expect_equal(dim(dtaFrm), c(250, 29))
    expect_equal(vapply(dtaFrm, typeof, character(1), USE.NAMES = FALSE), c("character", rep("integer", 27), "character"))
    expect_equal(names(dtaFrm), names(jmvReadWrite::bfi_sample2))
    expect_equal(dtaFrm[!selRow, "A1"], dtaInp[!selRow, "A1"])
    expect_equal(dtaFrm[, "A1"], jmvReadWrite::bfi_sample2[, "A1"])
    expect_false(any(is.na(dtaFrm[, "A1"])))

    dtaInp <- jmvReadWrite::bfi_sample2
    dtaInp[selRow, "A1_1"] <- dtaInp[selRow,  "A1"] + 1
    class(dtaInp[, "A1_1"]) <- class(dtaInp[, "A1"])
    expect_false(any(is.na(dtaInp[selRow, "A1_1"])))
    expect_true(all(is.na(dtaInp[!selRow, "A1_1"])))

    dtaFrm <- combine_cols_omv(dtaInp, varPrs = list(c("A1", "A1_1")), mdeCmb = "first")
    expect_s3_class(dtaFrm, "data.frame")
    expect_equal(dim(dtaFrm), c(250, 29))
    expect_equal(vapply(dtaFrm, typeof, character(1), USE.NAMES = FALSE), c("character", rep("integer", 27), "character"))
    expect_equal(names(dtaFrm), names(jmvReadWrite::bfi_sample2))
    expect_equal(dtaFrm[, "A1"], dtaInp[, "A1"])
    expect_equal(dtaFrm[, "A1"], jmvReadWrite::bfi_sample2[, "A1"])
    expect_false(any(is.na(dtaFrm[, "A1"])))

    dtaFrm <- combine_cols_omv(dtaInp, varPrs = list(c("A1", "A1_1")), mdeCmb = "second")
    expect_s3_class(dtaFrm, "data.frame")
    expect_equal(dim(dtaFrm), c(250, 29))
    expect_equal(vapply(dtaFrm, typeof, character(1), USE.NAMES = FALSE), c("character", rep("integer", 27), "character"))
    expect_equal(names(dtaFrm), names(jmvReadWrite::bfi_sample2))
    expect_equal(dtaFrm[!selRow, "A1"], dtaInp[!selRow, "A1"])
    expect_equal(dtaFrm[selRow,  "A1"], dtaInp[selRow, "A1_1"])
    expect_equal(dtaFrm[!selRow, "A1"], jmvReadWrite::bfi_sample2[!selRow, "A1"])
    expect_equal(dtaFrm[selRow,  "A1"], jmvReadWrite::bfi_sample2[selRow,  "A1"] + 1)
    expect_false(any(is.na(dtaFrm[, "A1"])))

    # test cases for code coverage ============================================================================================================================
    expect_error(combine_cols_omv(fleInp = tempfile(fileext = ".omv"), varPrs = list(c("A1", "A1_1")), mdeCmb = "first"),
      regexp = "Please use the argument dtaInp instead of fleInp\\.")
    expect_error(combine_cols_omv(dtaInp = dtaInp, varPrs = list(c(1, 2))),
      regexp = "The parameter varPrs needs to be a list with at least one \\(valid\\) variable pair to combine\\.")
    expect_error(combine_cols_omv(dtaInp = dtaInp, varPrs = list(c("A1", "A1_2"))),
      regexp = "The parameter varPrs needs to be a list with at least one \\(valid\\) variable pair to combine\\.")
    expect_error(combine_cols_omv(dtaInp = dtaInp, varPrs = list("A1", "A1_2")),
      regexp = "The parameter varPrs needs to be a list with at least one \\(valid\\) variable pair to combine\\.")
    expect_error(combine_cols_omv(dtaInp = dtaInp, varPrs = list(c("A1", "A1_1", "A2"))),
      regexp = "The parameter varPrs needs to be a list with at least one \\(valid\\) variable pair to combine\\.")
    expect_error(combine_cols_omv(dtaInp = dtaInp, varPrs = list(c("A1", "A1_1"))),
      regexp = "Mismatching values in the variable pair A1 - A1_1\\.")
    expect_error(combine_cols_omv(dtaInp = dtaInp, varPrs = list(c("A1", "A1_1")), mdeCmb = "none"),
      regexp = "Mismatching values in the variable pair A1 - A1_1\\.")

})

test_that("globals work", {
    nmeOMV <- tempfile(fileext = ".omv")
    write_omv(stats::setNames(as.data.frame(matrix(runif(200, -10, 10), ncol = 8)), paste0("X_", sprintf("%02d", 1:8))), nmeOMV)
    expect_equal(adjArg("read_omv", list(fleInp = "Trial_dflArg.omv", getSyn = TRUE, getHTM = TRUE),
                                    list(fleInp = "Trial_varArg.omv", sveAtt = FALSE, getSyn = FALSE, getHTM = FALSE),
                                    fxdArg = c("fleInp", "getSyn")),
                 list(fleInp = "Trial_dflArg.omv", getSyn = TRUE, sveAtt = FALSE, getHTM = FALSE))
    expect_equal(fcnArg(c("merge", "data.frame")), c("x", "y", "by", "by.x", "by.y", "all", "all.x", "all.y", "sort", "suffixes", "no.dups", "incomparables", "..."))
    expect_true(chkDir(nmeOMV))
    if (.Platform$OS.type == "unix" && Sys.getenv("NO_CHKDIR") != "true") {
        # permissions on *nix-systems
        dir.create(file.path(tempdir(), "chkPrm"), mode = "0111")
        expect_error(chkDir(file.path(tempdir(), "chkPrm", "Trial.omv")),
          regexp = "The directory \\(.*\\) exists, but you don't have writing permissions in that directory\\.")
        unlink(file.path(tempdir(), "chkPrm"), recursive = TRUE)
    }
    expect_true(chkDtF(jmvReadWrite::ToothGrowth, minSze = c(60, 7)))
    expect_true(chkExt(nmeOMV, vldExt))
    expect_true(chkExt(gsub("omv$", "rds", nmeOMV), vldExt))
    expect_true(chkFle(nmeOMV, isZIP = TRUE))
    expect_true(chkFle(nmeOMV, fleCnt = "meta"))
    expect_true(chkVar(jmvReadWrite::ToothGrowth, c("len", "dose", "supp")))
    expect_error(chkDir(file.path(tempdir(), "not", "file")), regexp = "^Directory \\(.*?\\) doesn't exist\\.")
    expect_error(chkFle(NA),            regexp = "^chkFle: Unsupported input parameter type\\.")
    expect_error(chkFle("", isZIP = 1), regexp = "^chkFle: Unsupported input parameter type\\.")
    expect_error(chkFle(tempfile()), regexp = "^File \".*?\" not found\\.")
    expect_error(chkDtF(data.frame(A = runif(n = 100)), minSze = 2), regexp = "^The \\w+ dimension of the input data frame has not the required size")
    expect_true(chkExt(tempfile(), ""))
    expect_true(chkExt(tempfile(), "omv"))
    expect_true(chkExt(tempfile(fileext = ".omv"), "omv"))
    expect_error(chkExt(tempfile(fileext = ".omv"), "chk"), regexp = "^File name \\(.*?\\) contains an unsupported file extension \\(.*?\\)\\.")
    expect_error(chkFle("no_file"), regexp = "^File \".*?\" not found.")
    expect_error(chkFle(nmeOMV, fleCnt = "no_file"), regexp = "^chkFle: File \".*?\" doesn't contain the file \".*?\"\\.")
    expect_error(chkVar(dtaFrm = data.frame(A = runif(100)), varNme = c("A", "B")), regexp = "^The variable\\(s\\) \\w+ are not contained in the current data set\\.")

    set.seed(1)
    tmpDF <- setAtt("value.labels", list(value.labels = setNames(seq(4), sprintf("Level %d", seq(4)))), data.frame(tmpCol = sample(seq(4), 200, replace = TRUE)))
    tmpDF[, 1] <- cnvCol(tmpDF[, 1], "factor")
    expect_s3_class(tmpDF[, 1], "factor")
    expect_equal(levels(tmpDF[, 1]), sprintf("Level %d", seq(4)))
    expect_equal(as.integer(table(tmpDF[, 1])), c(51, 54, 51, 44))

    expect_error(fmtFlI(fleInp = tempfile(), minLng = 2), regexp = "^The fleInp-argument is supposed to be a character vector with a minimal length of \\d+ and a maximal length of")
    expect_error(fmtFlI(fleInp = c(tempfile(), tempfile()), maxLng = 1),
      regexp = "^The fleInp-argument is supposed to be a character vector with a minimal length of \\d+ and a maximal length of")
    expect_error(fmtFlO(fleOut = ""),
      regexp = "^fleOut needs to be a valid non-empty file name \\(character\\), and the file extension for output file needs to be \\.omv\\ or \\.omt\\.")
    expect_error(fmtFlO(fleOut = "Trial"),
      regexp = "^fleOut needs to be a valid non-empty file name \\(character\\), and the file extension for output file needs to be \\.omv\\ or \\.omt\\.")
    expect_error(fmtFlO(fleOut = "Trial.rds"),
      regexp = "^fleOut needs to be a valid non-empty file name \\(character\\), and the file extension for output file needs to be \\.omv\\ or \\.omt\\.")
    expect_error(fcnArg(c("stats::sd", "stats::mean", "C")), regexp = "^The argument to fcnArg must be a character \\(vector\\) with 1 or 2 elements.")
    expect_true(jmvPtB())
    tmpPB <- var2PB(inpVar = list(list(A = NULL, B = TRUE, C = 1, D = 0.01, E = "Trial", F = c(TRUE, FALSE, TRUE), G = c(1, 2, 3), H = c(0.01, 0.02, 0.03), I = c("A", "B"))))
    expect_s4_class(tmpPB, "Message")
    expect_equal(length(tmpPB), 1)
    expect_equal(length(tmpPB$c$options[[1]]$c$options), 9)
    expect_equal(tmpPB$as.character(), paste0("c {\n  options {\n    c {\n      options {\n        o: NONE\n      }\n      options {\n        o: TRUE\n      }\n      options {\n",
                                              "        i: 1\n      }\n      options {\n        d: 0.01\n      }\n      options {\n        s: \"Trial\"\n      }\n      options {\n",
                                              "        c {\n          options {\n            o: TRUE\n          }\n          options {\n            o: FALSE\n          }\n",
                                              "          options {\n            o: TRUE\n          }\n        }\n      }\n      options {\n        c {\n          options {\n",
                                              "            i: 1\n          }\n          options {\n            i: 2\n          }\n          options {\n            i: 3\n",
                                              "          }\n        }\n      }\n      options {\n        c {\n          options {\n            d: 0.01\n          }\n",
                                              "          options {\n            d: 0.02\n          }\n          options {\n            d: 0.03\n          }\n        }\n",
                                              "      }\n      options {\n        c {\n          options {\n            s: \"A\"\n          }\n          options {\n",
                                              "            s: \"B\"\n          }\n        }\n      }\n      hasNames: true\n      names: \"A\"\n      names: \"B\"\n",
                                              "      names: \"C\"\n      names: \"D\"\n      names: \"E\"\n      names: \"F\"\n      names: \"G\"\n      names: \"H\"\n",
                                              "      names: \"I\"\n    }\n  }\n}\n"))
    expect_error(var2PB(inpVar = as.complex(pi)), regexp = "^Element not implemented for conversion to protocol buffer.")
    expect_error(setAtt(attLst = "Trial", inpObj = list(),       outObj = list()),
      regexp = "^Error when storing or accessing meta-data information\\. Please send the file")
    expect_error(setAtt(attLst = "Trial", inpObj = data.frame(), outObj = data.frame()),
      regexp = "^Error when storing or accessing meta-data information\\. Please send the file")
    expect_error(setAtt(attLst = NULL,    inpObj = data.frame(), outObj = data.frame()), regexp = "^setAtt: The parameter attLst is supposed to be a character vector\\.")
    expect_error(setAtt(attLst = "Trial", inpObj = NULL,         outObj = data.frame()), regexp = "^setAtt: The parameter inpObj is supposed to be either a list or a data frame\\.")
    expect_error(setAtt(attLst = "Trial", inpObj = data.frame(), outObj = NULL),         regexp = "^setAtt: The parameter outObj is supposed to be either a list or a data frame\\.")
    expect_equal(names(attributes(rmvAtt(jmvReadWrite::AlbumSales))), c("names", "class", "row.names"))
    expect_equal(setdiff(names(attributes(jmvReadWrite::AlbumSales)), names(attributes(rmvAtt(jmvReadWrite::AlbumSales)))), c("datalabel", "var.labels"))
    unlink(nmeOMV)

    expect_equal(bckAtt(jmvReadWrite::AlbumSales),
      list(dtaFrm = attributes(jmvReadWrite::AlbumSales), dtaCol = setNames(rep(list(NULL), ncol(jmvReadWrite::AlbumSales)), names(jmvReadWrite::AlbumSales))))
    expect_equal(bckAtt(NULL), list(dtaFrm = list(), dtaCol = list()))

    tmpDF <- nllAtt(jmvReadWrite::AlbumSales, c("datalabel", "var.labels"))
    expect_equal(names(attributes(tmpDF)), names(attributes(jmvReadWrite::AlbumSales)))
    expect_equal(attr(tmpDF, "datalabel"),  character(0))
    expect_equal(attr(tmpDF, "var.labels"), character(0))
    expect_equal(length(attr(tmpDF, "datalabel")),  0)
    expect_equal(length(attr(tmpDF, "var.labels")), 0)

    expect_error(inp2DF(dtaInp = 1), regexp = "^dtaInp must either be a data frame or a character \\(pointing to a location where the input file can be found\\)\\.")
    inpDF <- jmvReadWrite::AlbumSales
    inpNme <- file.path("..", "ToothGrowth.omv")
    expect_s3_class(inp2DF(dtaInp = inpNme), "data.frame")
    expect_equal(dim(inp2DF(dtaInp = inpNme)), c(60, 16))

    inpDF <- jmvReadWrite::AlbumSales
    attr(inpDF, "fleInp") <- file.path("..", "ToothGrowth.omv")
    expect_error(inp2DF(dtaInp = inpDF), regexp = "")
    df4Chk <- inp2DF(dtaInp = inpDF, minDF = 2, maxDF = 2)
    expect_type(df4Chk, "list")
    expect_length(df4Chk, 2)
    expect_s3_class(df4Chk[[1]], "data.frame")
    expect_s3_class(df4Chk[[2]], "data.frame")
    expect_equal(dim(df4Chk[[1]]), c(200, 5))
    expect_equal(dim(df4Chk[[2]]), c(60, 16))
    expect_equal(names(df4Chk[[1]]), c("selSbj", "Adverts", "Airplay", "Image", "Sales"))
    expect_equal(names(df4Chk[[2]]),
      c("Filter 1", "ID", "logLen", "supp - Transform 1", "len", "supp", "dose", "dose2", "dose3", "Trial", "Residuals", "J", "K", "L", "M", "weights"))
    expect_equal(names(attributes(df4Chk[[1]])), c("datalabel", "names", "var.labels", "class", "row.names", "fleInp"))
    expect_equal(attr(df4Chk[[1]], "fleInp"), file.path("..", "ToothGrowth.omv"))

    inpDF <- rbind(jmvReadWrite::AlbumSales[seq(sample(10)), ] * NA, jmvReadWrite::AlbumSales, jmvReadWrite::AlbumSales[seq(sample(10)), ] * NA)
    df4Chk <- inp2DF(dtaInp = inpDF, rmvEmp = TRUE)
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(200, 5))
    expect_equal(names(df4Chk), c("selSbj", "Adverts", "Airplay", "Image", "Sales"))
    expect_equal(names(attributes(df4Chk)), c("names", "row.names", "class"))
    expect_true(all(df4Chk == jmvReadWrite::AlbumSales))

    inpDF <- rbind(jmvReadWrite::AlbumSales[seq(sample(10)), ] * NA, jmvReadWrite::AlbumSales[seq(1, 100), ], jmvReadWrite::AlbumSales[1, ] * NA,
                   jmvReadWrite::AlbumSales[seq(101, 200), ], jmvReadWrite::AlbumSales[seq(sample(10)), ] * NA)
    expect_error(df4Chk <- inp2DF(dtaInp = inpDF, rmvEmp = TRUE),
      regexp = "Empty rows are not permitted execpt from the begin or the end of an input data frame \\(in such case, they are automatically removed\\)\\.")

    Sys.setenv(JAMOVI_R_VERSION = paste0(R.version$major, ".", R.version$minor))
    expect_warning(expect_error(rtnDta(fleOut = "", psvAnl = TRUE),
      regexp = "The position of the jamovi executable could not be determined or it was not found at the determined position\\. Determined position:"),
      regexp = "psvAnl is only possible if fleOut is a file name \\(analyses are not stored in data frames, only in the jamovi files\\)\\.")
    Sys.unsetenv("JAMOVI_R_VERSION")
    expect_null(jmvPth(inpPth = R.home(), strTgt = "not_in_path", bfrTgt = TRUE))

    Sys.setenv(JAMOVI_R_VERSION = paste0(R.version$major, ".", R.version$minor))
    expect_equal(jmvTtl("_arr_col"), "Dataset_arr_col")
    Sys.unsetenv("JAMOVI_R_VERSION")
    expect_equal(jmvTtl("_arr_col"), "")

    set.seed(1)
    tmpDF <- as.data.frame(cor(matrix(rnorm(1000), nrow = 100)))
    expect_error(mtxF2S(tmpDF[-1, ]), regexp = "Input matrix needs to be symmetric.")
    df4Chk <- mtxF2S(tmpDF)
    expect_equal(dim(df4Chk), c(10, 10))
    expect_equal(names(df4Chk), row.names(df4Chk))
    expect_equal(names(df4Chk), sprintf("V%d", seq(10)))
    expect_false(any(is.na(df4Chk)))
    expect_equal(unname(colMeans(df4Chk)),
      c(0.128183893, 0.096625131, 0.070568445, 0.115919003, 0.139886614, 0.058323693, 0.092441925, 0.106711615, 0.168887776, 0.10403618))

    df4Chk <- mtxF2S(tmpDF, rmvTrU = TRUE)
    expect_equal(dim(df4Chk), c(10, 10))
    expect_equal(names(df4Chk), row.names(df4Chk))
    expect_equal(names(df4Chk), sprintf("V%d", seq(10)))
    expect_identical(as.integer(colSums(is.na(df4Chk))), seq(0, 9))
    expect_equal(unname(colMeans(df4Chk, na.rm = TRUE)),
      c(0.12818389, 0.10747174, 0.09210481, 0.16532417, 0.17717275, 0.18858648, 0.25467175, 0.34689183, 0.59969889, 1))

    df4Chk <- mtxF2S(tmpDF, rmvDgn = TRUE)
    expect_equal(dim(df4Chk), c(10, 10))
    expect_equal(names(df4Chk), row.names(df4Chk))
    expect_equal(names(df4Chk), sprintf("V%d", seq(10)))
    expect_identical(unname(colSums(is.na(df4Chk))), rep(1, 10))
    expect_equal(unname(colMeans(df4Chk, na.rm = TRUE)),
      c(0.031315436, -0.003749854, -0.032701727, 0.017687781, 0.044318459, -0.046307008, -0.008397862, 0.007457349, 0.076541973, 0.004484645))

    df4Chk <- mtxF2S(tmpDF, rmvTrU = TRUE, mtxXps = TRUE)
    expect_equal(dim(df4Chk), c(10, 10))
    expect_equal(names(df4Chk), row.names(df4Chk))
    expect_equal(names(df4Chk), sprintf("V%d", seq(10)))
    expect_identical(as.integer(colSums(is.na(df4Chk))), seq(9, 0))
    expect_equal(unname(colMeans(df4Chk, na.rm = TRUE)), c(1, 0.49950284, 0.322948658, 0.250480203, 0.26716593, 0.106717426, 0.129390323, 0.128305082, 0.165497776, 0.10403618))

    df4Chk <- mtxF2S(tmpDF, rmvTrU = TRUE, rmvDgn = TRUE, mtxXps = TRUE)
    expect_equal(dim(df4Chk), c(10, 10))
    expect_equal(names(df4Chk), row.names(df4Chk))
    expect_equal(names(df4Chk), sprintf("V%d", seq(10)))
    expect_identical(as.integer(colSums(is.na(df4Chk))), seq(10, 1))
    expect_equal(unname(colMeans(df4Chk, na.rm = TRUE)),
      c(NA, -0.0009943199, -0.0155770134, 0.0006402703, 0.0839574127, -0.0719390890, -0.0157112896, 0.0037772361, 0.0611849980, 0.0044846446))

    df4Chk <- mtxF2S(tmpDF, mtxSps = TRUE)
    expect_equal(dim(df4Chk), c(9, 10))
    expect_equal(names(df4Chk), c("Variable", sprintf("V%d", seq(1, 9))))
    expect_equal(row.names(df4Chk), sprintf("V%d", seq(2, 10)))
    expect_identical(unname(colSums(is.na(df4Chk))), c(0, seq(0, 8)))
    expect_equal(unname(colMeans(df4Chk[, -1], na.rm = TRUE)),
      c(0.031315436, -0.004094296, -0.037594503,  0.026211536, 0.012607297, -0.014266906, 0.006228995, 0.020337746, 0.199397772))
})

#' Combines pairs of columns from a raw data matrix in .omv-files for the statistical spreadsheet
#' 'jamovi' (<https://www.jamovi.org>)
#'
#' @param dtaInp Either a data frame or the name of a data file to be read (including the path,
#'               if required; "FILENAME.ext"; default: NULL); files can be of any supported file
#'               type, see Details below.
#' @param fleOut Name of the data file to be written (including the path, if required;
#'               "FILE_OUT.omv"; default: ""); if empty, the resulting data frame is returned
#'               instead.
#' @param varPrs Definition of variable pairs; a list containing either list(s) or character
#'               vector(s) with the names of pairs of variables to be combined (default: list()).
#' @param mdeCmb Mode of combining the variables when conflicting values occur, either "none",
#'               "first", or "second" (default: "none"), see Details below.
#' @param psvAnl Whether analyses that are contained in the input file shall be transferred to the
#'               output file (TRUE / FALSE; default: FALSE)
#' @param usePkg Name of the package: "foreign" or "haven" that shall be used to read SPSS, Stata,
#'               and SAS files; "foreign" is the default (it comes with base R), but "haven" is
#'               newer and more comprehensive.
#' @param selSet Name of the data set that is to be selected from the workspace (only applies when
#'               reading .RData-files)
#' @param ...    Additional arguments passed on to methods; see Details below.
#'
#' @return a data frame containing the column pairs given in `varPrs` combined and the original
#'         columns removed
#'
#' @details
#' * The need to combine two columns into one is quite common after merging columns or rows (using,
#'   e.g., merge_cols_omv or merge_rows_omv). `varPrs` defines the variable pairs to be combined.
#'   It is a list containing the pairs of variables to be combined, either as list or as character
#'   vector; e.g., list(c("A", "B"), c("C", "D")) or list(list("A", "B"), list("C", "D")). `mdeCmb`
#'   defines what to to if values in the first and the second variable of a variable pair contain
#'   conflicting / different values: "none" does not merge the variables (and instead throws an
#'   error), "first" makes that the values from the first variable of each pair are taken if the
#'   values are conflicting, and "second" use the values from the second variable of each pair in
#'   case of conflicts.
#' * The ellipsis-parameter (`...`) can be used to submit arguments / parameters to the functions
#'   that are used for reading and writing the data. By clicking on the respective function under
#'   “See also”, you can get a more detailed overview over which parameters each of those functions
#'   take. The functions are: `read_omv` and `write_omv` (for jamovi-files), `read.table` (for CSV
#'   / TSV files; using similar defaults as `read.csv` for CSV and `read.delim` for TSV which both
#'   are based upon `read.table`), `load` (for .RData-files), `readRDS` (for .rds-files),
#'   `read_sav` (needs the R-package `haven`) or `read.spss` (needs the R-package `foreign`) for
#'   SPSS-files, `read_dta` (`haven`) / `read.dta` (`foreign`) for Stata-files, `read_sas`
#'   (`haven`) for SAS-data-files, and `read_xpt` (`haven`) / `read.xport` (`foreign`) for
#'   SAS-transport-files. If you would like to use `haven`, you may need to install it using
#'   `install.packages("haven", dep = TRUE)`.
#'
#' @seealso `combine_cols_omv` uses the following functions for reading and writing data files in
#'   different formats: [jmvReadWrite::read_omv()] and [jmvReadWrite::write_omv()] for
#'   jamovi-files, [utils::read.table()] for CSV / TSV files, [load()] for reading .RData-files,
#'   [readRDS()] for .rds-files, [haven::read_sav()] or [foreign::read.spss()] for SPSS-files,
#'   [haven::read_dta()] or [foreign::read.dta()] for Stata-files, [haven::read_sas()] for
#'   SAS-data-files, and [haven::read_xpt()] or [foreign::read.xport()] for SAS-transport-files.
#'
#' @examples
#' dtaInp <- jmvReadWrite::bfi_sample2
#' # create a new column (A1_1) containing a subset of the values in the original variable
#' # whereas those lines are replaced with NAs
#' set.seed(1)
#' selRow <- rnorm(nrow(dtaInp)) < 0
#' dtaInp[selRow,  "A1_1"] <- dtaInp[selRow, "A1"]
#' dtaInp[selRow,  "A1"]   <- NA
#' head(dtaInp[, c("A1", "A1_1")])
#' dtaOut <- combine_cols_omv(dtaInp, varPrs = list(c("A1", "A1_1")))
#' # show the differences before and after combining the values in the columns and ensure
#' # that all values are the same as in the original data set
#' dtaInp[, "A1"]
#' dtaOut[, "A1"]
#' all(dtaOut[, "A1"] == jmvReadWrite::bfi_sample2[, "A1"])
#'
#' # create a new column, containing values that are different from the original variable
#' dtaInp <- jmvReadWrite::bfi_sample2
#' dtaInp[selRow,  "A1_1"] <- dtaInp[selRow,  "A1"] + 1
#' # [1] if mdeCmb is "none" (or if mdeCmb is not given - "none" is the default) an error would be
#' # thrown (therefore the next line is commented out)
#' # dtaOut <- combine_cols_omv(dtaInp, varPrs = list(c("A1", "A1_1")), mdeCmb = "none")
#' # [2] if mdeCmb is "first", missing values are replaced and values from the first column ("A1")
#' # take precedence if the values are unequal
#' dtaOut <- combine_cols_omv(dtaInp, varPrs = list(c("A1", "A1_1")), mdeCmb = "first")
#' head(cbind(dtaOut[, "A1"], dtaInp[, c("A1", "A1_1")]))
#' # [3] if mdeCmb is "second", missing values are replaced and values from the second column
#' # ("A1_1") take precedence if the values are unequal
#' dtaOut <- combine_cols_omv(dtaInp, varPrs = list(c("A1", "A1_1")), mdeCmb = "second")
#' head(cbind(dtaOut[, "A1"], dtaInp[, c("A1", "A1_1")]))
#'
#'
#' @export combine_cols_omv
#'
combine_cols_omv <- function(dtaInp = NULL, fleOut = "", varPrs = list(), mdeCmb = c("none", "first", "second"),
                             psvAnl = FALSE, usePkg = c("foreign", "haven"), selSet = "", ...) {

    # check and import input data set (either as data frame or from a file)
    if (!is.null(list(...)[["fleInp"]])) stop("Please use the argument dtaInp instead of fleInp.")
    mdeCmb <- match.arg(mdeCmb)
    dtaFrm <- inp2DF(dtaInp = dtaInp, usePkg = usePkg, selSet = selSet, ...)

    if (!all(is.character(unlist(varPrs))) || !all(nzchar(unlist(varPrs))) ||
        !all(unlist(varPrs) %in% names(dtaInp)) || !all(vapply(varPrs, length, numeric(1)) == 2) ||
        !all(vapply(varPrs, function(v) isa(v, "character") || isa(v, "list"), logical(1)))) {
        stop("The parameter varPrs needs to be a list with at least one (valid) variable pair to combine.")
    }

    for (crrPrs in varPrs) {
        if (is.list(crrPrs)) crrPrs <- as.character(crrPrs)
        # all values are equal (or NA) -> replace NAs in variable 1 with the respective rows from variable 2
        if (all(dtaFrm[, crrPrs[1]] == dtaFrm[, crrPrs[2]], na.rm = TRUE) || mdeCmb == "first") {
            dtaFrm[is.na(dtaFrm[, crrPrs[1]]), crrPrs[1]] <- dtaFrm[is.na(dtaFrm[, crrPrs[1]]), crrPrs[2]]
        } else if (mdeCmb == "second") {
            dtaFrm[!is.na(dtaFrm[, crrPrs[2]]), crrPrs[1]] <- dtaFrm[!is.na(dtaFrm[, crrPrs[2]]), crrPrs[2]]
        } else if (mdeCmb == "none") {
            stop(sprintf("Mismatching values in the variable pair %s.",  paste(crrPrs, collapse = " - ")))
        }
        dtaFrm <- dtaFrm[, setdiff(names(dtaFrm), crrPrs[2])]
    }

    # rtnDta in globals.R (unified function to either write the data frame, open it in a new jamovi session or return it)
    rtnDta(dtaFrm = dtaFrm, fleOut = fleOut, dtaTtl = jmvTtl("_cmb_cols"), psvAnl = psvAnl, dtaInp = dtaInp, ...)
}

#' Merges two .omv-files for the statistical spreadsheet 'jamovi' (<https://www.jamovi.org>) by adding the content of the second, etc.  file(s) as rows to the
#' first file
#'
#' @param dtaInp Either a data frame (with the attribute "fleInp" containing the files to merge) or vector with the names of the input files (including the
#'               path, if required; "FILENAME.ext"; default: NULL); files can be of any supported file type, see Details below
#' @param fleOut Name of the data file to be written (including the path, if required; "FILE_OUT.omv"; default: ""); if empty, the resulting data frame is
#'               returned instead
#' @param typMrg Type of merging operation: "all" (default) or  "common"; see also Details
#' @param colInd Add a column with an indicator (the basename of the file minus the extension) marking from which input data set the respective rows are coming
#'               (default: FALSE)
#' @param rstRwN Reset row names (i.e., do not keep the row names of the original input data sets but number them consecutively - one to the row number of all
#'               input data sets added up; default: TRUE)
#' @param rmvDpl Remove duplicated rows (i.e., rows with the same content as a previous row in all columns; default: FALSE)
#' @param varSrt Variable(s) that are used to sort the data frame (see Details; if empty, the order after merging is kept; default: c())
#' @param usePkg Name of the package: "foreign" or "haven" that shall be used to read SPSS, Stata and SAS files; "foreign" is the default (it comes with
#'               base R), but "haven" is newer and more comprehensive
#' @param selSet Name of the data set that is to be selected from the workspace (only applies when reading .RData-files)
#' @param ...    Additional arguments passed on to methods; see Details below
#'
#' @return a data frame (only returned if `fleOut` is empty) where the rows of all input data sets (given in the `dtaInp`-argument) are concatenated
#'
#' @details
#' * Using data frames with the input parameter `dtaInp` is primarily thought to be used when calling `merge_cols_omv` from the jamovi-modules `jTransform` and
#'   `Rj`. For the use in R, it is strongly recommended to use a character vector with the file names instead.
#' * There are four different types of merging operations (defined via `typMrg`): "all" keeps all existing variables / columns that are contained in any of the
#'   input data sets and fills them up with NA where the variable / column doesn't exist in an input data set. "common" only keeps the variables / columns that
#'   are common to all input data sets (i.e., that are contained in all data sets).
#' * `varSrt` can be either a character or a character vector (with one or more variables respectively). The sorting order for a particular variable can be
#'   inverted with preceding the variable name with "-". Please note that this doesn't make sense and hence throws a warning for certain variable types (e.g.,
#'   factors).
#' * The ellipsis-parameter (`...`) can be used to submit arguments / parameters to the functions that are used for merging or reading the data. By clicking on
#'   the respective function under “See also”, you can get a more detailed overview over which parameters each of those functions take.
#' * Adding columns uses `rbind` (with some further operation, adding missing columns (filled with NAs), if `typMrg` is "all").
#' * The functions for reading and writing the data are: `read_omv` and `write_omv` (for jamovi-files), `read.table` (for CSV / TSV files; using similar
#'   defaults as `read.csv` for CSV and `read.delim` for TSV which both are based upon `read.table`), `load` (for .RData-files), `readRDS` (for .rds-files),
#'   `read_sav` (needs R-package `haven`) or `read.spss` (needs R-package `foreign`) for SPSS-files, `read_dta` (`haven`) / `read.dta` (`foreign`) for
#'   Stata-files, `read_sas` (`haven`) for SAS-data-files, and `read_xpt` (`haven`) / `read.xport` (`foreign`) for SAS-transport-files. If you would like to
#'   use `haven`, you may need to install it using `install.packages("haven", dep = TRUE)`.
#'
#' @seealso `merge_rows_omv` internally uses the following functions: Adding columns uses [rbind()]. For reading and writing data files in different formats:
#'   [jmvReadWrite::read_omv()] and [jmvReadWrite::write_omv()] for jamovi-files, [utils::read.table()] for CSV / TSV files, [load()] for reading .RData-files,
#'   [readRDS()] for .rds-files, [haven::read_sav()] or [foreign::read.spss()] for SPSS-files, [haven::read_dta()] or [foreign::read.dta()] for Stata-files,
#'   [haven::read_sas()] for SAS-data-files, and [haven::read_xpt()] or [foreign::read.xport()] for SAS-transport-files.
#'
#' @examples
#' dtaInp <- jmvReadWrite::bfi_sample2
#' nmeInp <- paste0(tempfile(), "_", 1:3, ".rds")
#' nmeOut <- tempfile(fileext = ".omv")
#' for (i in seq_along(nmeInp)) saveRDS(dtaInp[-i - 1], nmeInp[i])
#' # save dtaInp three times (i.e., the length of nmeInp), removing one data columns in
#' # each data set (for demonstration purposes, A1 in the first, A2 in the second, ...)
#' jmvReadWrite::merge_rows_omv(dtaInp = nmeInp, fleOut = nmeOut, colInd = TRUE)
#' cat(file.info(nmeOut)$size)
#' # -> 10767 (size may differ on different OSes)
#' dtaOut <- jmvReadWrite::read_omv(nmeOut, sveAtt = FALSE)
#' unlink(nmeOut)
#' # read the data set where the three original datasets were added as rows and show
#' # the variable names
#' cat(names(dtaInp))
#' cat(names(dtaOut))
#' # compared to the input data set, we have the same variable names; fleInd (switched
#' # on by colInd = TRUE and showing from which data set the rows are coming from) is
#' # new and A1 is moved to the end of the list (the "original" order of variables may
#' # not always be preserved and columns missing from at least one of the input data
#' # sets may be added at the end)
#' cat(dim(dtaInp), dim(dtaOut))
#' # the first dimension of the data sets (rows) is now three times of that of the input
#' # data set (250 -> 750), the second dimension (columns / variables) is increased by 1
#' # (for "fleInd")
#'
#' jmvReadWrite::merge_rows_omv(dtaInp = nmeInp, fleOut = nmeOut, typMrg = "common")
#' # the argument typMrg = "common" removes the columns that are not present in all of
#' # the input data sets (i.e., A1, A2, A3)
#' dtaOut <- jmvReadWrite::read_omv(nmeOut, sveAtt = FALSE)
#' unlink(nmeOut)
#' # read the data set where the three original datasets were added as rows and show
#' # the variable names
#' cat(names(dtaInp))
#' cat(names(dtaOut))
#' # compared to the input data set, the variables that were missing in at least one
#' # data set (i.e., "A1", "A2" and "A3") are removed
#' cat(dim(dtaInp), dim(dtaOut))
#' # the first dimension of the data sets (rows) is now three times of that of the
#' # input data set (250 -> 750), the second dimension (columns / variables) is
#' # reduced by 3 (i.e., "A1", "A2", "A3")
#'
#' unlink(nmeInp)
#'
#'
#' @export merge_rows_omv
#'
merge_rows_omv <- function(dtaInp = NULL, fleOut = "", typMrg = c("all", "common"), colInd = FALSE, rstRwN = TRUE, rmvDpl = FALSE, varSrt = c(), usePkg = c("foreign", "haven"), selSet = "", ...) {

    # check and import input data set (either as data frame or from a file)
    if (!is.null(list(...)[["fleInp"]])) stop("Please use the argument dtaInp instead of fleInp.")
    dtaFrm <- inp2DF(dtaInp = dtaInp, minDF = 2, maxDF = Inf, usePkg = usePkg, selSet = selSet, ...)
    if (is.character(dtaInp)) fleInp <- dtaInp else fleInp <- c("input data frame", attr(dtaInp, "fleInp"))
    if (colInd) dtaFrm <- lapply(seq_along(dtaFrm), function(i) addIdx(dtaFrm[[i]], fleInp[i]))

    # merge files - the additional arguments are the same in either case
    typMrg <- match.arg(typMrg)
    crrArg <- adjArg(c("rbind", "data.frame"), list(), list(...), c())
    # keeping all existing variables, filling the void columns with NA
    if      (typMrg == "all") {
        # determine the variable names and types in all input data sets and remove
        # duplicates(order ensures that the data set that contains most variables
        # is prioritized, but yet, it is still impossible to preserve the whole
        # order - i.e., variables missing in one data set may end up at the end)
        varNme <- unlist(lapply(dtaFrm[order(-vapply(dtaFrm, function(x) dim(x)[2], integer(1)))], names))
        varTyp <- unlist(lapply(dtaFrm[order(-vapply(dtaFrm, function(x) dim(x)[2], integer(1)))], function(D) vapply(D, function(C) class(C)[1], character(1))))
        varNme <- varNme[!duplicated(varNme)]
        for (crrNme in varNme) {
            if (sum(names(varTyp) == crrNme) <= 1) next
            if (all(duplicated(varTyp[names(varTyp) == crrNme])[-1])) {
                varTyp <- varTyp[-which(names(varTyp) == crrNme)[-1]]
            } else {
                stop(sprintf("Variable %s has different types:\n%s\n", crrNme, paste(paste0(basename(fleInp),
                  rep(": ", sum(names(varTyp) == crrNme)), varTyp[names(varTyp) == crrNme]), collapse = "\n")))
            }
        }
        if (length(varNme) != length(varTyp)) stop("Something went wrong when comparing the variable types of the input data files. Please send the data files to sebastian.jentschke@uib.no for debugging.")
        tmpMrg <- addCol(dtaFrm[[1]], varNme, varTyp)
        for (i in setdiff(seq_along(dtaFrm), 1)) {
            crrInp <- addCol(dtaFrm[[i]], varNme, varTyp)
            tmpMrg <- do.call(rbind, c(list(tmpMrg, crrInp), crrArg))
        }
        dtaFrm <- tmpMrg
    # keeping only variables that are common to all input data sets
    } else if (typMrg == "common") {
        varNme <- Reduce(intersect, lapply(dtaFrm, names))
        if (identical(varNme, character(0))) {
            stop(paste("The data sets in the files that were given as dtaInp-argument do not contain variables that are overlapping (i.e., contained in all data sets).",
                       "You can either reduce the number of data sets given to dtaInp or use \"outer\" as argument for \"typMrg\" (see Details in the help for this function)."))
        }
        tmpMrg <- dtaFrm[[1]][, varNme]
        for (i in setdiff(seq_along(dtaFrm), 1)) {
            tmpMrg <- do.call(rbind, c(list(tmpMrg, dtaFrm[[i]][, varNme]), crrArg))
        }
        dtaFrm <- tmpMrg
    }

    # remove row names
    if (rstRwN == TRUE) {
        rownames(dtaFrm) <- NULL
    }

    # remove duplicate rows
    if (rmvDpl == TRUE) {
        dtaFrm <- dtaFrm[!duplicated(dtaFrm[, setdiff(varNme, "fleInd")]), ]
    }

    # sort data frame (if varSrt not empty)
    dtaFrm <- srtFrm(dtaFrm, varSrt)

    # rtnDta in globals.R (unified function to either write the data frame, open it in a new jamovi session or return it)
    rtnDta(dtaFrm = dtaFrm, fleOut = fleOut, dtaTtl = jmvTtl("_mrg_rows"), ...)
}

addCol <- function(dtaFrm = NULL, varNme = c(), varTyp = c()) {
    varDff <- setdiff(varNme, names(dtaFrm))
    for (i in seq_along(varDff)) {
        eval(parse(text = paste0("dtaFrm[varDff[i]] <- as.", varTyp[[varDff[i]]], "(NA)")))
    }
    dtaFrm
}

addIdx <- function(dtaFrm = NULL, fleNme = "") {
    if (nzchar(fleNme)) {
        cbind(list(fleInd = rep(gsub(paste0("\\.", tools::file_ext(fleNme)), "", basename(fleNme)), dim(dtaFrm)[1])), dtaFrm)
    } else {
        dtaFrm
    }
}

test_that("merge_rows_omv works", {
    tmpInp <- list(jmvReadWrite::bfi_sample, jmvReadWrite::bfi_sample2, jmvReadWrite::bfi_sample)
    nmeInp <- paste0(tempfile(), "_", 1:3, ".rds")
    for (i in seq_along(nmeInp)) saveRDS(tmpInp[[i]], nmeInp[i])

    # check merging rows with writing an output file and afterwards checking it (existence, size, whether it is a ZIP-file and content)
    nmeOut <- tempfile(fileext = ".omv")
    expect_null(merge_rows_omv(dtaInp = nmeInp, fleOut = nmeOut))
    expect_true(file.exists(nmeOut))
    expect_gt(file.info(nmeOut)$size, 1)
    expect_true(chkFle(nmeOut, isZIP = TRUE))
    expect_true(chkFle(nmeOut, fleCnt = "meta"))
    expect_true(chkFle(nmeOut, fleCnt = "metadata.json"))
    expect_true(chkFle(nmeOut, fleCnt = "data.bin"))
    unlink(nmeOut)

    # checking merging rows with returning the merged dataset as a variable
    # try out different setting for the arguments and their consequence for size, etc.
    dtaFrm <- merge_rows_omv(dtaInp = nmeInp)
    expect_s3_class(dtaFrm, "data.frame")
    expect_equal(dim(dtaFrm), c(758, 33))
    expect_equal(names(attributes(dtaFrm)), c("names", "row.names", "class"))
    expect_equal(names(attributes(dtaFrm[[2]])), "jmv-desc")
    expect_equal(names(attributes(dtaFrm[[30]])), c("levels", "class"))

    dtaFrm <- merge_rows_omv(dtaInp = nmeInp, typMrg = "common")
    expect_s3_class(dtaFrm, "data.frame")
    expect_equal(dim(dtaFrm), c(758, 29))

    dtaFrm <- merge_rows_omv(dtaInp = nmeInp, colInd = TRUE)
    expect_s3_class(dtaFrm, "data.frame")
    expect_equal(dim(dtaFrm), c(758, 34))
    expect_true(all(unname(table(dtaFrm$fleInd)) == vapply(tmpInp, dim, integer(2))[1, ]))

    dtaFrm <- merge_rows_omv(dtaInp = nmeInp, rstRwN = TRUE)
    expect_s3_class(dtaFrm, "data.frame")
    expect_equal(dim(dtaFrm), c(758, 33))
    expect_true(all(rownames(dtaFrm) == as.character(seq_len(dim(dtaFrm)[1]))))

    dtaFrm <- merge_rows_omv(dtaInp = nmeInp, colInd = TRUE, rmvDpl = TRUE)
    expect_s3_class(dtaFrm, "data.frame")
    expect_equal(dim(dtaFrm), c(504, 34))
    expect_true(all(table(dtaFrm$fleInd) == c(bfi_sample = 254, bfi_sample2 = 250)))

    tmpDF <- tmpInp[[1]]
    attr(tmpDF, "fleInp") <- nmeInp[2]
    dtaFrm <- merge_rows_omv(dtaInp = tmpDF, colInd = TRUE)
    expect_s3_class(dtaFrm, "data.frame")
    expect_equal(dim(dtaFrm), c(504, 34))
    expect_true(all(table(dtaFrm$fleInd) == setNames(254, "input data frame") | table(dtaFrm$fleInd) == setNames(250, nmeInp[2])))

    tmpDF$age <- as.numeric(tmpDF$age)
    expect_error(merge_rows_omv(dtaInp = tmpDF, colInd = TRUE), regexp = "^Variable age has different types:\\s+input data frame: numeric")
    unlink(nmeInp)

    # test cases for code coverage ============================================================================================================================
    nmeInp <- paste0(tempfile(), "_", 1:4, ".rds")
    for (i in seq_along(nmeInp)) saveRDS(stats::setNames(data.frame(runif(n = 100)), LETTERS[i]), nmeInp[i])
    expect_error(dtaFrm <- merge_rows_omv(fleInp = nmeInp, typMrg = "common"), regexp = "^Please use the argument dtaInp instead of fleInp\\.")
    expect_error(dtaFrm <- merge_rows_omv(dtaInp = nmeInp, typMrg = "common"), regexp = "^The data sets in the files that were given as dtaInp-argument do not contain variables that are overlapping")
    unlink(nmeInp)

    expect_equal(addIdx(jmvReadWrite::bfi_sample, "bfi_sample.omv")[[1]], rep("bfi_sample", 254))
    expect_equal(addIdx(jmvReadWrite::bfi_sample, ""), jmvReadWrite::bfi_sample)
    expect_equal(addIdx(jmvReadWrite::bfi_sample), jmvReadWrite::bfi_sample)
})

#' Calculates distances (returning a symmetric matrix) from a raw data matrix in .omv-files for the
#' statistical spreadsheet 'jamovi' (<https://www.jamovi.org>)
#'
#' @param dtaInp Either a data frame or the name of a data file to be read (including the path,
#'               if required; "FILENAME.ext"; default: NULL); files can be of any supported file
#'               type, see Details below.
#' @param fleOut Name of the data file to be written (including the path, if required;
#'               "FILE_OUT.omv"; default: ""); if empty, the resulting data frame is returned
#'               instead.
#' @param varDst Variable (default: c()) containing a character vector with the names of the
#'               variables for which distances are to be calculated. See Details for more
#'               information.
#' @param clmDst Whether the distances shall be calculated between columns (TRUE) or rows
#'               (FALSE; default: TRUE). See Details for more information.
#' @param stdDst Character string indicating whether the variables in varDst are to be standardized
#'               and how (default: "none"). See Details for more information.
#' @param nmeDst Character string indicating which distance measure is to be calculated calculated
#'               (default: "euclidean"). See Details for more information.
#' @param mtxSps Whether the symmetric matrix to be returned should be sparse (default: FALSE)
#' @param mtxTrL Whether the symmetric matrix to be returned should only contain the lower
#'               triangular (default: FALSE)
#' @param mtxDgn Whether the symmetric matrix to be returned should retain the values in the main
#'               diagonal (default: TRUE)
#' @param usePkg Name of the package: "foreign" or "haven" that shall be used to read SPSS, Stata,
#'               and SAS files; "foreign" is the default (it comes with base R), but "haven" is
#'               newer and more comprehensive.
#' @param selSet Name of the data set that is to be selected from the workspace (only applies when
#'               reading .RData-files)
#' @param ...    Additional arguments passed on to methods; see Details below.
#'
#' @return a data frame containing a symmetric matrix (only returned if `fleOut` is empty)
#'         containing the distances between the variables / columns (clmDst == TRUE) or rows
#'         (clmDst == FALSE)
#'
#' @details
#' * `varDst` must a character vector containing the variables to calculated distances over. If
#'   `clmDst` is set to TRUE, distances are calculated between all possible variable pairs and over
#'   subjects / rows in the original data frame. If `clmDst` is set to FALSE, distances are
#'   calculated between participants and over all variables given in `varDst`. If `clmDst` is set
#'   to `TRUE`, the symmetric matrix that is returned has the size V x V (V being the number of
#'   variables in varDst; if `mtxSps` is set to `TRUE`, the size is V - 1 x V - 1, see below); if
#'   `clmDst` is set to `FALSE`, the symmetric matrix that is returned has the size R x R (R being
#'   the number of rows in the original dataset; it is if `mtxSps` is set to `TRUE`, the size is
#'   R - 1 x R - 1, see below).
#' * `stdDst` can be one of the following calculations to standardize the selected variables before
#'   calculating the distances: `none` (do not standardize; default), `z` (z scores), `sd` (divide
#'   by the std. dev.), `range` (divide by the range), `max` (divide by the absolute maximum),
#'   `mean` (divide by the mean), `rescale` (subtract the mean and divide by the range).
#' * `nmeDst` can be one of the following distance measures.
#'   (1) For interval data: `euclid` (Euclidean), `seuclid` (squared Euclidean), `block` (city
#'   block / Manhattan), `canberra` (Canberra). `chebychev` (maximum distance / supremum norm /
#'   Chebychev), `minkowski_p` (Minkowski with power p; NB: needs p), `power_p_r` (Minkowski with
#'   power p, and the r-th root; NB: needs p and r), `cosine` (cosine between the two vectors),
#'   `correlation` (correlation between the two vectors).
#'   (2) For frequency count data: `chisq` (chi-square dissimilarity between two sets of
#'   frequencies), `ph2` (chi-square dissimilarity normalized by the square root of the number
#'   of values used in the calculation).
#'   (3) For binary data, all measure have to optional parts `p` and `np` which indicate presence
#'   (`p`; defaults to 1 if not given) or absence (`np`; defaults to zero if not given).
#'   (a) matching coefficients: `rr_p_np` (Russell and Rao), `sm_p_np` (simple matching),
#'   `jaccard_p_np` / `jaccards_p_np` (Jaccard similarity; as in SPSS), `jaccardd_p_np` (Jaccard
#'   dissimiliarity; as in `dist(..., "binary")` in R), `dice_p_np` (Dice or Czekanowski or
#'   Sorenson similarity), `ss1_p_np` (Sokal and Sneath measure 1), `rt_p_np` (Rogers and
#'   Tanimoto), `ss2_p_np` (Sokal and Sneath measure 2), `k1_p_np` (Kulczynski measure 1),
#'   `ss3_p_np` (Sokal and Sneath measure 3).
#'   (b) conditional probabilities: `k2_p_np` (Kulczynski measure 2), `ss4_p_np` (Sokal and Sneath
#'   measure 4), `hamann_p_np` (Hamann).
#'   (c) predictability measures: `lambda_p_np` (Goodman and Kruskal Lambda), `d_p_np` (Anderberg’s
#'   D), `y_p_np` (Yule’s Y coefficient of colligation), `q_p_np` (Yule’s Q).
#'   (d) other measures: `ochiai_p_np` (Ochiai), `ss5_p_np` (Sokal and Sneath measure 5),
#'   `phi_p_np` (fourfold point correlation), `beuclid_p_np` (binary Euclidean distance),
#'   `bseuclid_p_np` (binary squared Euclidean distance), `size_p_np` (size difference),
#'   `pattern_p_np` (pattern difference), `bshape_p_np` (binary Shape difference), `disper_p_np`
#'   (dispersion similarity), `variance_p_np` (variance dissimilarity), `blwmn_p_np` (binary Lance
#'   and Williams non-metric dissimilarity).
#'   (4) `none` (only carry out standardization, if stdDst is different from `none`).
#' * If `mtxSps` is set, a sparse matrix is returned. Those matrices are similar to the format one
#'   often finds for correlation matrices. The values are only retained in the lower triangular,
#'   the columns range from the first to the variable that is second to the last in `varDst` (or
#'   respectively, the columns contain the first to the second to the last row of the original
#'   dataset when `clmDst` is set to `FALSE`), and the rows contain the second to the last variable
#'   in `varDst` (or respectively, the rows contain the second to the last row of the original
#'   dataset when `clmDst` is set to `FALSE`).
#' * By default, a full symmetric matrix is returned (i.e., a matrix that has no NAs in any cell).
#'   This behaviour can be changed with setting `mtxTrL` and `mtxDgn`: If `mtxTrL` is set to
#'   `TRUE`, the values from the upper triangular matrix are removed / replaced with NAs; if
#'   `mtxDgn` is set to `FALSE`, the values from the main diagonal are removed / replaced with NAs.
#' * The ellipsis-parameter (`...`) can be used to submit arguments / parameters to the functions
#'   that are used for reading and writing the data. By clicking on the respective function under
#'   “See also”, you can get a more detailed overview over which parameters each of those functions
#'   take. The functions are: `read_omv` and `write_omv` (for jamovi-files), `read.table` (for CSV
#'   / TSV files; using similar defaults as `read.csv` for CSV and `read.delim` for TSV which both
#'   are based upon `read.table`), `load` (for .RData-files), `readRDS` (for .rds-files),
#'   `read_sav` (needs the R-package `haven`) or `read.spss` (needs the R-package `foreign`) for
#'   SPSS-files, `read_dta` (`haven`) / `read.dta` (`foreign`) for Stata-files, `read_sas`
#'   (`haven`) for SAS-data-files, and `read_xpt` (`haven`) / `read.xport` (`foreign`) for
#'   SAS-transport-files. If you would like to use `haven`, you may need to install it using
#'   `install.packages("haven", dep = TRUE)`.
#'
#' @seealso `distances_omv` internally uses the following function for calculating the distances
#'   for interval data [stats::dist()]. It furthermore uses the following functions for reading
#'   and writing data files in different formats: [jmvReadWrite::read_omv()] and
#'   [jmvReadWrite::write_omv()] for jamovi-files, [utils::read.table()] for CSV / TSV files,
#'   [load()] for reading .RData-files, [readRDS()] for .rds-files, [haven::read_sav()] or
#'   [foreign::read.spss()] for SPSS-files, [haven::read_dta()] or [foreign::read.dta()] for
#'   Stata-files, [haven::read_sas()] for SAS-data-files, and [haven::read_xpt()] or
#'   [foreign::read.xport()] for SAS-transport-files.
#'
#' @examples
#' # create matrices for the different types of distance measures: continuous
#' # (cntFrm), frequency counts (frqFrm) or binary (binFrm); all 20 R x 5 C
#' set.seed(1)
#' cntFrm <- stats::setNames(as.data.frame(matrix(rnorm(100, sd = 10),
#'             ncol = 5)), sprintf("C_%02d", seq(5)))
#' frqFrm <- stats::setNames(as.data.frame(matrix(sample(seq(10), 100,
#'             replace = TRUE), ncol = 5)), sprintf("F_%02d", seq(5)))
#' binFrm <- stats::setNames(as.data.frame(matrix(sample(c(TRUE, FALSE), 100,
#'             replace = TRUE), ncol = 5)), sprintf("B_%02d", seq(5)))
#' nmeOut <- tempfile(fileext = ".omv")
#'
#' # calculates the distances between columns, nmeDst is not required: "euclid"
#' # is the default
#' jmvReadWrite::distances_omv(dtaInp = cntFrm, fleOut = nmeOut, varDst =
#'   names(cntFrm), nmeDst = "euclid")
#' dtaFrm <- jmvReadWrite::read_omv(nmeOut)
#' unlink(nmeOut)
#' # the resulting matrix (10 x 10) with the Euclidian distances
#' print(dtaFrm)
#'
#' # calculates the (Euclidean) distances between rows (clmDst = FALSE)
#' jmvReadWrite::distances_omv(dtaInp = cntFrm, fleOut = nmeOut, varDst =
#'   names(cntFrm), clmDst = FALSE, nmeDst = "euclid")
#' dtaFrm <- jmvReadWrite::read_omv(nmeOut)
#' unlink(nmeOut)
#' # the resulting matrix (20 x 20) with the Euclidian distances
#' print(dtaFrm)
#'
#' # calculates the (Euclidean) distances between columns; the original data
#' # are z-standardized before calculating the distances (stdDst = "z")
#' jmvReadWrite::distances_omv(dtaInp = cntFrm, fleOut = nmeOut, varDst =
#'   names(cntFrm), stdDst = "z", nmeDst = "euclid")
#' dtaFrm <- jmvReadWrite::read_omv(nmeOut)
#' unlink(nmeOut)
#' # the resulting matrix (10 x 10) with the Euclidian distances using the
#' # z-standardized data
#' print(dtaFrm)
#'
#' # calculates the correlations between columns
#' jmvReadWrite::distances_omv(dtaInp = cntFrm, fleOut = nmeOut, varDst =
#'   names(cntFrm), nmeDst = "correlation")
#' dtaFrm <- jmvReadWrite::read_omv(nmeOut)
#' unlink(nmeOut)
#' # the resulting matrix (10 x 10) with the correlations
#' print(dtaFrm)
#'
#' # calculates the chi-square dissimilarity (nmeDst = "chisq") between columns
#' jmvReadWrite::distances_omv(dtaInp = frqFrm, fleOut = nmeOut, varDst =
#'   names(frqFrm), nmeDst = "chisq")
#' dtaFrm <- jmvReadWrite::read_omv(nmeOut)
#' unlink(nmeOut)
#' # the resulting matrix (10 x 10) with the chi-square dissimilarities
#' print(dtaFrm)
#'
#' # calculates the Jaccard similarity (nmeDst = "jaccard") between columns
#' jmvReadWrite::distances_omv(dtaInp = binFrm, fleOut = nmeOut, varDst =
#'   names(binFrm), nmeDst = "jaccard")
#' dtaFrm <- jmvReadWrite::read_omv(nmeOut)
#' unlink(nmeOut)
#' # the resulting matrix (10 x 10) with the Jaccard similarities
#' print(dtaFrm)
#'
#'
#' @export distances_omv
#'
distances_omv <- function(dtaInp = NULL, fleOut = "", varDst = c(), clmDst = TRUE, stdDst = "none",
                          nmeDst = "euclid", mtxSps = FALSE, mtxTrL = FALSE, mtxDgn = TRUE,
                          usePkg = c("foreign", "haven"), selSet = "", ...) {

    # check and import input data set (either as data frame or from a file)
    if (!is.null(list(...)[["fleInp"]])) stop("Please use the argument dtaInp instead of fleInp.")
    dtaFrm <- inp2DF(dtaInp = dtaInp, usePkg = usePkg, selSet = selSet, ...)

    if (!all(nzchar(varDst)) || length(intersect(varDst, names(dtaFrm))) < 2) {
        stop("Calling distances_omv requires giving at least two (valid) variables to calculate distances between.")
    }

    # convert to matrix, and transpose if necessary ===============================================
    if (clmDst) dtaMtx <- as.matrix(dtaFrm[, varDst]) else dtaMtx <- t(as.matrix(dtaFrm[, varDst]))

    # standardize the data ========================================================================
    if        (grepl("^none$",             stdDst)) {
    } else if (grepl("^z$",                stdDst)) {
        dtaMtx <- scale(dtaMtx)
    } else if (grepl("^sd$",               stdDst)) {
        dtaMtx <- scale(dtaMtx, center = FALSE)
    } else if (grepl("^range$",            stdDst)) {
        dtaMtx <- scale(dtaMtx, center = FALSE, scale = (apply(dtaMtx, 2, max) - apply(dtaMtx, 2, min)))
    } else if (grepl("^max$",              stdDst)) {
        dtaMtx <- scale(dtaMtx, center = FALSE, scale =  apply(abs(dtaMtx), 2, max))
    } else if (grepl("^mean$",             stdDst)) {
        dtaMtx <- scale(dtaMtx, center = FALSE, scale =  apply(dtaMtx, 2, mean))
    } else if (grepl("^rescale$",          stdDst)) {
        dtaMtx <- scale(dtaMtx, center = apply(dtaMtx, 2, min), scale = (apply(dtaMtx, 2, max) - apply(dtaMtx, 2, min)))
    } else {
        stop(sprintf("Invalid standardization: %s. See Details in the help for further information.", stdDst))
    }

    # calculate distances (or proximities) ========================================================
    # (1) interval data ---------------------------------------------------------------------------
    if        (grepl("^euclid",             nmeDst)) {
        dstMtx <- as.matrix(stats::dist(t(dtaMtx), "euclidean", TRUE, TRUE))
    } else if (grepl("^seuclid",            nmeDst)) {
        dstMtx <- as.matrix(stats::dist(t(dtaMtx), "euclidean", TRUE, TRUE) ^ 2)
    } else if (grepl("^block$|^manhattan$", nmeDst)) {
        dstMtx <- as.matrix(stats::dist(t(dtaMtx), "manhattan", TRUE, TRUE))
    } else if (grepl("^canberra$",          nmeDst)) {
        dstMtx <- as.matrix(stats::dist(t(dtaMtx), "canberra",  TRUE, TRUE))
    } else if (grepl("^chebychev$",         nmeDst)) {
        dstMtx <- as.matrix(stats::dist(t(dtaMtx), "maximum",   TRUE, TRUE))
    } else if (grepl("^minkowski_\\d+$",    nmeDst)) {
        dstMtx <- as.matrix(stats::dist(t(dtaMtx), "minkowski", TRUE, TRUE, p = getPw(nmeDst)))
    } else if (grepl("^power_\\d+_\\d+$",   nmeDst)) {
        dstMtx <- as.matrix(stats::dist(t(dtaMtx), "minkowski", TRUE, TRUE, p = getPw(nmeDst)) ^ (getPw(nmeDst) / getRt(nmeDst)))
    } else if (grepl("^cosine$",            nmeDst)) {
        dstMtx <- clcCos(dtaMtx)
    } else if (grepl("^correlation$",       nmeDst)) {
        dstMtx <- stats::cor(dtaMtx, use = "pairwise")
    # (2) frequency count data --------------------------------------------------------------------
    } else if (grepl("^chisq$|^ph2$",       nmeDst)) {
        dstMtx <- clcFrq(dtaMtx, nmeDst)
    # (3) binary data -----------------------------------------------------------------------------
    } else if (binDst(nmeDst)) {
        dstMtx <- clcBin(dtaMtx, nmeDst)
    # (4) none ------------------------------------------------------------------------------------
    } else if (grepl("^none$",              nmeDst)) {
        dstMtx <- dtaMtx
    # ---------------------------------------------------------------------------------------------
    } else {
        stop(sprintf("Invalid distance measure: %s. See Details in the help for further information.",     nmeDst))
    }

    # converting distance matrix back to data frame, and apply mtxF2S
    dstFrm <- as.data.frame(dstMtx)
    if (mtxSps || mtxTrL) dstFrm <- mtxF2S(dstFrm, rmvTrU = mtxTrL, rmvDgn = !mtxDgn, mtxSps = mtxSps)

    # rtnDta in globals.R (unified function to either write the data frame, open it in a new jamovi session or return it)
    rtnDta(dtaFrm = dstFrm, fleOut = fleOut, dtaTtl = jmvTtl("_dist"), ...)
}

# helper and calculation functions ================================================================
# binary measures: check whether name is valid
binDst <- function(nmeDst = "") {
    grepl(paste0("^beuclid$|^blwmn$|^bseuclid$|^bshape$|^d$|^dice$|^disper$|^hamann$|^jaccard$|^jaccard[s,d]$|^k[1-2]$|^lambda$|",
                 "^ochiai$|^pattern$|^phi$|^q$|^rr$|^rt$|^size$|^sm$|^ss[1-5]$|^variance$|^y$"), gsub("_\\w+_\\w+$", "", nmeDst))
}

# binary measures: calculation, calls mtcBin for each cell (variable pair comparison / matches)
clcBin <- function(m = NULL, t = "jaccard") {
    # transform data matrix into a logical matrix
    m <- mkeBin(m, getP(t), getNP(t))
    # extract transformation name
    t <- strsplit(t, "_")[[1]][1]
    n <- ncol(m)
    # create a result matrix
    r <- matrix(NA, n, n, dimnames = rep(dimnames(m)[2], 2))

    for (i in seq(1, n)) {
        for (j in seq(1, i)) {
            # skip calculation of distance in the main diagonal for k1 and ss3
            # (keep NA; if calculated, the limitation to 9999.999 [l. 366 / 392] would hit)
            if (i == j && grepl("^k1$|^ss3$", t)) next
            r[i, j] <- r[j, i] <- mtcBin(m[, i], m[, j], t)
        }
    }

    r
}

# measure for continuos variables: cosine (between variable pairs)
clcCos <- function(m = NULL) {
    n <- ncol(m)
    # create a result matrix
    r <- matrix(1, n, n, dimnames = rep(dimnames(m)[2], 2))

    for (i in seq(2, n)) {
        for (j in seq(1, i - 1)) {
            r[i, j] <- r[j, i] <- crossprod(m[, i], m[, j]) / sqrt(crossprod(m[, i]) * crossprod(m[, j]))
        }
    }

    r
}

clcFrq <- function(m = NULL, t = "chisq") {
    n <- ncol(m)
    # create a result matrix
    r <- matrix(0, n, n, dimnames = rep(dimnames(m)[2], 2))

    for (i in seq(2, n)) {
        for (j in seq(1, i - 1)) {
            if        (t == "chisq") {
                r[i, j] <- r[j, i] <- suppressWarnings(stats::chisq.test(rbind(table(m[, i]), table(m[, j])))[["statistic"]])
            } else if (t == "ph2") {
                r[i, j] <- r[j, i] <- suppressWarnings(stats::chisq.test(rbind(table(m[, i]), table(m[, j])))[["statistic"]]) / sqrt(nrow(m) * 2)
            } else {
                stop(sprintf("clcFrq: Method %s is not implemented.", t))
            }
        }
    }

    r
}

# helper functions: get P (power / present), NP (not present), and R (root)
getP  <- function(s) stats::na.omit(c(strsplit(s, "_")[[1]][2], "1"))[1]
getNP <- function(s) stats::na.omit(c(strsplit(s, "_")[[1]][3], "0"))[1]
getPw <- function(s) as.numeric(strsplit(s, "_")[[1]][2])
getRt <- function(s) as.numeric(strsplit(s, "_")[[1]][3])

# binary measures: transform the data matrix from numeric into logical
mkeBin <- function(m = NULL, p = 1, np = 0) {
    if (all(apply(m, 2, is.logical))) return(m)

    if        (all(apply(m, 2, function(c) is.numeric(c)   &&   all(as.numeric(c(p, np)) %in% unique(c))))) {
        r <- matrix(as.logical(NA), nrow = nrow(m), ncol = ncol(m), dimnames = dimnames(m))
        r[m == as.numeric(p)]  <- TRUE
        r[m == as.numeric(np)] <- FALSE
    } else if (all(apply(m, 2, function(c) is.character(c) && all(as.character(c(p, np)) %in% unique(c))))) {
        r <- matrix(as.logical(NA), nrow = nrow(m), ncol = ncol(m), dimnames = dimnames(m))
        r[m == as.character(p)]  <- TRUE
        r[m == as.character(np)] <- FALSE
    } else {
        stop(paste("The input matrix for binary data either needs to be logical (then it will be kept as it is),",
                   "numeric or character (for the latter two, p and np are used to derive TRUE and FALSE)."))
    }

    r
}

# binary measures: calculation of variable pair matches
# implemented based upon: www.sussex.ac.uk/its/pdfs/SPSS_Statistics_Algorithms_22.pdf
# see www.itl.nist.gov/div898/software/dataplot/refman2/auxillar/binmatch.htm for further possible measures
mtcBin <- function(x, y, t = "") {
    if (!all(is.logical(c(x, y)))) stop("mtcBin: Input columns to the calculation of binary measures must be logical.")
    o <- c(sum(x & y, na.rm = TRUE), sum(x & !y, na.rm = TRUE), sum(!x & y, na.rm = TRUE), sum(!x & !y, na.rm = TRUE))

    # binary - dissimilarity - Euclidian distance: binEuc - BEUCLID
    if      (t == "beuclid")  return(sqrt(sum(o[2:3])))
    # binary - dissimilarity - Lance and Williams: BLWMN
    else if (t == "blwmn")    return(sum(o[2], o[3]) / sum(o[-4], o[1]))
    # binary - dissimilarity - Squared Euclidian distance: BSEUCLID
    else if (t == "bseuclid") return(sum(o[2], o[3]))
    # binary - dissimilarity - Shape: BSHAPE
    else if (t == "bshape")   return((sum(o) * sum(o[2], o[3]) - sum(o[2], -o[3]) ^ 2) / sum(o) ^ 2)
    # binary - similarity - Anderberg's D: D
    else if (t == "d")        return((t1_Bin(o) - t2_Bin(o)) / (2 * sum(o)))
    # binary - similarity - Dice: DICE
    else if (t == "dice")     return((o[1] * 2) / sum(o[1], o[-4]))
    # binary - similarity - Dispersion: DISPER
    else if (t == "disper")   return((o[1] * o[4] - o[2] * o[3]) / (sum(o) ^ 2))
    # binary - similarity - Hamann: HAMANN
    else if (t == "hamann")   return(sum(o[1], -o[2], -o[3], o[4]) / sum(o))
    # binary - similarity - Jaccard: JACCARD
    else if (t == "jaccard")  return(o[1] / sum(o[-4]))
    # binary - similarity - Jaccard
    else if (t == "jaccards") return(o[1] / sum(o[-4]))
    # binary - distance - Jaccard (as in R - stats::dist - "binary")
    else if (t == "jaccardd") return(sum(o[2], o[3]) / sum(o[-4]))
    # binary - similarity - Kulczynski 1: K1
    else if (t == "k1")       return(min(o[1] / sum(o[2], o[3]), 9999.999))
    # binary - similarity - Kulczynski 2: K2
    else if (t == "k2")       return(sum(o[1] / sum(o[1], o[2]), o[1] / sum(o[1], o[3])) / 2)
    # binary - similarity - Lambda: LAMBDA
    else if (t == "lambda")   return((t1_Bin(o) - t2_Bin(o)) / (2 * sum(o) - t2_Bin(o)))
    # binary - similarity - Ochiai: OCHIAI
    else if (t == "ochiai")   return(sqrt((o[1] / sum(o[1], o[2])) * (o[1] / sum(o[1], o[3]))))
    # binary - similarity - Phi 4-point correlation: PHI
    else if (t == "phi")      return(((o[1] * o[4]) - (o[2] * o[3])) / sqrt(sum(o[1], o[2]) * sum(o[1], o[3]) * sum(o[2], o[4]) * sum(o[3], o[4])))
    # binary - dissimilarity - Pattern difference: PATTERN
    else if (t == "pattern")  return((o[2] * o[3]) / (sum(o) ^ 2))
    # binary - similarity - Yule's Q: Q
    else if (t == "q")        return((o[1] * o[4] - o[2] * o[3]) / sum(o[1] * o[4], o[2] * o[3]))
    # binary - similarity - Russel and Rao: RR
    else if (t == "rr")       return(o[1] / sum(o))
    # binary - similarity - Rogers and Tanimoto: RT
    else if (t == "rt")       return(sum(o[1], o[4]) / sum(o, o[2], o[3]))
    # binary - dissimilarity - Size difference: SIZE
    else if (t == "size")     return(((o[2] - o[3]) ^ 2) / (sum(o) ^ 2))
    # binary - similarity - Simple matching: SM
    else if (t == "sm")       return(sum(o[1], o[4]) / sum(o))
    # binary - similarity - Sokal and Sneath 1: SS1
    else if (t == "ss1")      return(2 * sum(o[1], o[4]) / sum(o, o[1], o[4]))
    # binary - similarity - Sokal and Sneath 2: SS2
    else if (t == "ss2")      return(o[1] / sum(o[-4], o[2], o[3]))
    # binary - similarity - Sokal and Sneath 3: SS3
    else if (t == "ss3")      return(min(sum(o[1], o[4]) / sum(o[2], o[3]), 9999.999))
    # binary - similarity - Sokal and Sneath 4: SS4
    else if (t == "ss4")      return(sum(o[1] / sum(o[1], o[2]), o[1] / sum(o[1], o[3]), o[4] / sum(o[2], o[4]), o[4] / sum(o[3], o[4])) / 4)
    # binary - similarity - Sokal and Sneath 5: SS5
    else if (t == "ss5")      return((o[1] * o[4]) / sqrt(sum(o[1], o[2]) * sum(o[1], o[3]) * sum(o[2], o[4]) * sum(o[3], o[4])))
    # binary - dissimilarity - Variance: VARIANCE
    else if (t == "variance") return(sum(o[2], o[3]) / (4 * sum(o)))
    # binary - similarity - Yule's Y: binYlY - Y
    else if (t == "y")        return((sqrt(o[1] * o[4]) - sqrt(o[2] * o[3])) / sum(sqrt(o[1] * o[4]), sqrt(o[2] * o[3])))
#   PLACEHOLDER FOR FUTURE IMPLEMENTATIONS
#   else if (t == "")         return()
    stop(sprintf("mtcBin: Method %s is not implemented.", t))
}

t1_Bin <- function(o = c()) {
    sum(max(o[1], o[2], na.rm = TRUE), max(o[3], o[4], na.rm = TRUE), max(o[1], o[3], na.rm = TRUE), max(o[2], o[4], na.rm = TRUE))
}

t2_Bin <- function(o = c()) {
    sum(max(sum(o[1], o[3], na.rm = TRUE), sum(o[2], o[4], na.rm = TRUE), na.rm = TRUE),
        max(sum(o[1], o[2], na.rm = TRUE), sum(o[3], o[4], na.rm = TRUE), na.rm = TRUE))
}

test_that("distances_omv works", {
    set.seed(1)
    cntFrm <- stats::setNames(as.data.frame(matrix(rnorm(1000, sd = 10),                         nrow = 50)), sprintf("C_%02d", seq(20)))
    frqFrm <- stats::setNames(as.data.frame(matrix(sample(seq(10),        1000, replace = TRUE), nrow = 50)), sprintf("C_%02d", seq(20)))
    binFrm <- stats::setNames(as.data.frame(matrix(sample(c(TRUE, FALSE), 1000, replace = TRUE), nrow = 50)), sprintf("C_%02d", seq(20)))
    binFm2 <- as.data.frame(lapply(binFrm, function(c) as.integer(c)  + 0))
    binFm3 <- as.data.frame(lapply(binFrm, function(c) as.integer(c)  + 1))
    binFm4 <- as.data.frame(lapply(binFrm, function(c) as.integer(!c) + 1))
    binFm5 <- stats::setNames(as.data.frame(matrix(sample(c(rep(1, 5), rep(2, 5), 4), 1000, replace = TRUE), nrow = 50)), sprintf("C_%02d", seq(20)))
    nmeInp <- tempfile(fileext = ".rds")
    nmeOut <- tempfile(fileext = "_D.omv")
    saveRDS(cntFrm, nmeInp)

    # Euclidean distances (column-wise), no standardization
    expect_null(distances_omv(dtaInp = cntFrm, fleOut = nmeOut, varDst = names(cntFrm)))
    expect_true(chkFle(nmeOut))
    expect_gt(file.info(nmeOut)$size, 1)
    expect_true(chkFle(nmeOut, isZIP = TRUE))
    expect_true(chkFle(nmeOut, fleCnt = "meta"))
    expect_true(chkFle(nmeOut, fleCnt = "metadata.json"))
    expect_true(chkFle(nmeOut, fleCnt = "data.bin"))
    df4Chk <- read_omv(nmeOut, sveAtt = FALSE)
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(20, 20))
    expect_equal(names(df4Chk), names(cntFrm))
    expect_equal(vapply(df4Chk, typeof, character(1), USE.NAMES = FALSE), rep("double", 20))
    expect_true(isSymmetric(unname(as.matrix(df4Chk))))
    expect_equal(unname(diag(as.matrix(df4Chk))), rep(0, 20))
    expect_equal(unname(apply(df4Chk, 2, mean)), c(90.18679, 93.75375, 93.49854,  95.84402, 100.25257, 96.79214,  95.25020, 96.61567, 97.88351, 108.10851,
                                                   91.54438, 96.83193, 99.66558, 104.26397, 102.46154, 99.06569, 104.29886, 89.90709, 97.13592, 101.83002),
                                                 tolerance = 1e-6)
    expect_equal(unname(apply(df4Chk, 2, sd)),   c(23.05755, 23.22398, 23.30027,  23.63582,  24.78627, 24.50848,  24.47311, 24.23558, 24.02617,  27.30439,
                                                   23.24474, 24.65757, 25.18948,  25.89397,  25.80464, 24.56810,  27.01149, 22.32947, 24.55965,  25.65252),
                                                 tolerance = 1e-6)
    unlink(nmeOut)

    # Euclidean distances (row-wise), no standardization
    expect_null(distances_omv(dtaInp = cntFrm, fleOut = nmeOut, varDst = names(cntFrm), clmDst = FALSE))
    expect_true(chkFle(nmeOut))
    expect_gt(file.info(nmeOut)$size, 1)
    expect_true(chkFle(nmeOut, isZIP = TRUE))
    expect_true(chkFle(nmeOut, fleCnt = "meta"))
    expect_true(chkFle(nmeOut, fleCnt = "metadata.json"))
    expect_true(chkFle(nmeOut, fleCnt = "data.bin"))
    df4Chk <- read_omv(nmeOut, sveAtt = FALSE)
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(50, 50))
    expect_equal(names(df4Chk), as.character(seq(50)))
    expect_equal(vapply(df4Chk, typeof, character(1), USE.NAMES = FALSE), rep("double", 50))
    expect_true(isSymmetric(unname(as.matrix(df4Chk))))
    expect_equal(unname(diag(as.matrix(df4Chk))), rep(0, 50))
    expect_equal(unname(apply(df4Chk, 2, mean)), c(59.23618, 64.02938, 62.95174, 60.92444, 66.50634, 76.44711, 68.87178, 57.46921, 64.34043, 64.93839,
                                                   68.70524, 58.17400, 63.75770, 67.27705, 64.39134, 67.93244, 56.30517, 68.48069, 56.41789, 67.58459,
                                                   63.14606, 65.99261, 57.17930, 69.23091, 62.35964, 62.27133, 61.43129, 65.34220, 61.63603, 56.70968,
                                                   61.44346, 66.03810, 57.51696, 64.25885, 62.64196, 64.52206, 61.71255, 57.74651, 63.29933, 54.38568,
                                                   65.92930, 66.02469, 67.45406, 54.62625, 76.43019, 66.94794, 67.35090, 60.45972, 70.91475, 63.27496),
                                                 tolerance = 1e-6)
    expect_equal(unname(apply(df4Chk, 2, sd)),   c(11.95075, 13.17128, 11.88307, 13.26369, 13.00918, 13.04877, 13.84253, 12.70425, 13.44566, 12.97648,
                                                   13.15931, 11.53839, 12.25830, 13.28330, 12.82124, 13.21762, 12.26558, 12.71885, 12.02038, 13.12634,
                                                   12.40275, 14.07321, 12.40036, 13.39745, 11.58568, 11.92935, 12.37711, 13.38654, 13.28438, 11.73747,
                                                   13.02186, 14.00567, 11.24524, 13.15485, 12.62797, 12.11199, 12.48867, 11.61827, 12.75744, 11.22085,
                                                   13.40663, 13.04554, 13.89652, 11.80896, 14.62947, 13.57283, 13.25454, 12.48747, 13.82016, 12.33878),
                                                 tolerance = 1e-6)
    unlink(nmeOut)

    # Euclidean distances (column-wise), no standardization, sparse matrix
    expect_null(distances_omv(dtaInp = cntFrm, fleOut = nmeOut, varDst = names(cntFrm), mtxSps = TRUE))
    expect_true(chkFle(nmeOut))
    expect_gt(file.info(nmeOut)$size, 1)
    expect_true(chkFle(nmeOut, isZIP = TRUE))
    expect_true(chkFle(nmeOut, fleCnt = "meta"))
    expect_true(chkFle(nmeOut, fleCnt = "metadata.json"))
    expect_true(chkFle(nmeOut, fleCnt = "data.bin"))
    df4Chk <- read_omv(nmeOut, sveAtt = FALSE)
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(19, 20))
    expect_equal(names(df4Chk), c("Variable", names(cntFrm)[-ncol(cntFrm)]))
    expect_equal(as.vector(df4Chk[, "Variable"]), names(cntFrm)[-1])
    expect_equal(vapply(df4Chk, typeof, character(1), USE.NAMES = FALSE), c("integer", rep("double", 19)))
    expect_true(all(is.na(df4Chk[, -1][upper.tri(matrix(NA, nrow = 19, ncol = 19))])))
    expect_equal(unname(apply(df4Chk[, -1], 2, mean, na.rm = TRUE)),
      c(94.93346,  99.11064,  99.36124, 101.83715, 107.08739, 103.51847, 103.06181, 102.75189, 105.06215, 116.83354,
        95.91321, 103.83493, 107.79650, 111.57576, 112.16093, 104.30065, 104.62081,  89.44986, 104.53152),
      tolerance = 1e-6)
    expect_equal(unname(apply(df4Chk[, -1], 2, sd,   na.rm = TRUE)),
       c(9.24849,   7.41334,   7.66104,   6.74420,   7.47457,   9.51210,  10.06599,   9.95656,   7.32652,
        10.28519,   9.26913,   9.05088,  12.73900,   8.10547,   7.71139,   3.56249,  13.96156,   1.45965,  NA),
      tolerance = 1e-6)
    unlink(nmeOut)

    # Euclidean distances (column-wise), no standardization, lower triangular matrix
    expect_null(distances_omv(dtaInp = cntFrm, fleOut = nmeOut, varDst = names(cntFrm), mtxTrL = TRUE))
    expect_true(chkFle(nmeOut))
    expect_gt(file.info(nmeOut)$size, 1)
    expect_true(chkFle(nmeOut, isZIP = TRUE))
    expect_true(chkFle(nmeOut, fleCnt = "meta"))
    expect_true(chkFle(nmeOut, fleCnt = "metadata.json"))
    expect_true(chkFle(nmeOut, fleCnt = "data.bin"))
    df4Chk <- read_omv(nmeOut, sveAtt = FALSE)
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(20, 20))
    expect_equal(names(df4Chk), names(cntFrm))
    expect_equal(vapply(df4Chk, typeof, character(1), USE.NAMES = FALSE), c(rep("double", 19), "integer"))
    expect_true(all(is.na(df4Chk[upper.tri(matrix(NA, nrow = 20, ncol = 20))])))
    expect_equal(unname(apply(df4Chk, 2, mean, na.rm = TRUE)),
      c(90.18679, 93.89429, 93.84117, 95.84673, 100.39442, 96.61724, 95.70025, 94.84790, 96.30697, 106.21231,
        86.32188, 92.29771, 94.32194, 95.63636,  93.46744, 83.44052, 78.46561, 59.63324, 52.26576,   0.00000),
      tolerance = 1e-6)
    expect_equal(unname(apply(df4Chk, 2, sd,   na.rm = TRUE)),
      c(23.05755, 23.85163, 24.57072, 25.54777,  27.72862, 28.25636, 29.19290, 30.05032, 31.12292,  36.55301,
        31.56429, 35.63207, 39.89499, 42.81587,  46.30606, 46.74659, 53.53810, 51.65422, 73.91495,  NA),
    tolerance = 1e-6)
    unlink(nmeOut)

    # Euclidean distances (column-wise), no standardization, lower triangular matrix, lower triangular matrix without elements in the main diagonal
    expect_null(distances_omv(dtaInp = cntFrm, fleOut = nmeOut, varDst = names(cntFrm), mtxTrL = TRUE, mtxDgn = FALSE))
    expect_true(chkFle(nmeOut))
    expect_gt(file.info(nmeOut)$size, 1)
    expect_true(chkFle(nmeOut, isZIP = TRUE))
    expect_true(chkFle(nmeOut, fleCnt = "meta"))
    expect_true(chkFle(nmeOut, fleCnt = "metadata.json"))
    expect_true(chkFle(nmeOut, fleCnt = "data.bin"))
    df4Chk <- read_omv(nmeOut, sveAtt = FALSE)
    expect_s3_class(df4Chk, "data.frame")
    expect_equal(dim(df4Chk), c(20, 20))
    expect_equal(names(df4Chk), names(cntFrm))
    expect_equal(vapply(df4Chk, typeof, character(1), USE.NAMES = FALSE), rep("double", 20))
    expect_true(all(is.na(df4Chk[upper.tri(matrix(NA, nrow = 20, ncol = 20), TRUE)])))
    expect_equal(unname(apply(df4Chk[, -ncol(cntFrm)], 2, mean, na.rm = TRUE)),
      c(94.93346,  99.11064,  99.36124, 101.83715, 107.08739, 103.51847, 103.06181, 102.75189, 105.06215, 116.83354,
        95.91321, 103.83493, 107.79650, 111.57576, 112.16093, 104.30065, 104.62081,  89.44986, 104.53152),
      tolerance = 1e-6)
    expect_equal(unname(apply(df4Chk[, -ncol(cntFrm)], 2, sd,   na.rm = TRUE)),
       c(9.24849,   7.41334,   7.66104,   6.74420,   7.47457,   9.51210,  10.06599,   9.95656,   7.32652,  10.28519,
         9.26913,   9.05088,  12.73900,   8.10547,   7.71139,   3.56249,  13.96156,   1.45965,  NA),
      tolerance = 1e-6)
    unlink(nmeOut)

    # =================================================================================================================
    # standardization methods: `none`, `z`, `sd`, `range`, `max`, `mean`, `rescale`; no calculation of distances
    # standardization: none, distance measure: none
    # with no standardization and no distance calculation, the output data set should be equal to the input data set
    expect_equal(cntFrm, rmvMsV(distances_omv(dtaInp = cntFrm, varDst = names(cntFrm), stdDst = "none", nmeDst = "none")))

    # standardization: z, distance measure: none
    df4Chk <- distances_omv(dtaInp = cntFrm, varDst = names(cntFrm), stdDst = "z", nmeDst = "none")
    expect_equal(dim(df4Chk), dim(cntFrm))
    expect_equal(unname(apply(df4Chk, 2, mean)), rep(0, 20))
    expect_equal(unname(apply(df4Chk, 2, sd)),   rep(1, 20))
    expect_equal(as.vector(as.matrix(rmvMsV(df4Chk) / apply(cntFrm, 2, scale))), rep(1, 1000))
    unlink(nmeOut)

    # standardization: sd, distance measure: none
    df4Chk <- distances_omv(dtaInp = cntFrm, varDst = names(cntFrm), stdDst = "sd", nmeDst = "none")
    expect_equal(dim(df4Chk), dim(cntFrm))
    expect_equal(unname((apply(cntFrm, 2, mean) / apply(df4Chk, 2, mean)) / apply(cntFrm, 2, sd)), rep(1, 20), tolerance = 1e-2)
    expect_equal(unname(apply(df4Chk, 2, sd)),   rep(1, 20), tolerance = 1e-2)
    expect_equal(as.vector(as.matrix(rmvMsV(df4Chk) / apply(cntFrm, 2, scale, center = FALSE))), rep(1, 1000))
    unlink(nmeOut)

    # standardization: range, distance measure: none
    df4Chk <- distances_omv(dtaInp = cntFrm, varDst = names(cntFrm), stdDst = "range", nmeDst = "none")
    expect_equal(dim(df4Chk), dim(cntFrm))
    expect_equal(unname(apply(df4Chk, 2, mean)), as.vector(apply(cntFrm, 2, mean) / diff(apply(cntFrm, 2, range))))
    expect_equal(unname(apply(df4Chk, 2, sd)),   as.vector(apply(cntFrm, 2, sd)   / diff(apply(cntFrm, 2, range))))
    expect_equal(unname(apply(df4Chk, 2, max) - apply(df4Chk, 2, min)), rep(1, 20))
    unlink(nmeOut)

    # standardization: max, distance measure: none
    df4Chk <- distances_omv(dtaInp = cntFrm, varDst = names(cntFrm), stdDst = "max", nmeDst = "none")
    expect_equal(dim(df4Chk), dim(cntFrm))
    expect_equal(unname(apply(df4Chk, 2, mean)), unname(apply(cntFrm, 2, mean) / apply(abs(cntFrm), 2, max)))
    expect_equal(unname(apply(df4Chk, 2, sd)),   unname(apply(cntFrm, 2, sd)   / apply(abs(cntFrm), 2, max)))
    expect_equal(colSums(abs(rbind(unname(apply(df4Chk, 2, min)), unname(apply(df4Chk, 2, max)))) == 1), rep(1, 20))
    unlink(nmeOut)

    # standardization: mean, distance measure: none
    df4Chk <- distances_omv(dtaInp = cntFrm, varDst = names(cntFrm), stdDst = "mean", nmeDst = "none")
    expect_equal(dim(df4Chk), dim(cntFrm))
    expect_equal(unname(apply(df4Chk, 2, mean)), rep(1, 20))
    expect_equal(unname(apply(df4Chk, 2, sd)),   unname(apply(cntFrm, 2, sd) / abs(apply(cntFrm, 2, mean))))
    unlink(nmeOut)

    # standardization: rescale, distance measure: none
    df4Chk <- distances_omv(dtaInp = cntFrm, varDst = names(cntFrm), stdDst = "rescale", nmeDst = "none")
    expect_equal(dim(df4Chk), dim(cntFrm))
    expect_equal(unname(apply(df4Chk, 2, min)),  rep(0, 20))
    expect_equal(unname(apply(df4Chk, 2, max)),  rep(1, 20))
    expect_equal(unname(apply(df4Chk, 2, sd)), as.vector(apply(cntFrm, 2, sd) / diff(apply(cntFrm, 2, range))))
    unlink(nmeOut)

    # =================================================================================================================
    # distance meausres for continuos data: euclid, seuclid, block, canberra, chebychev, minkowski_p, power_p_r, cosine,
    # correlation; no standardization
    # distance measure: euclid, standardization: none
    df4Chk <- distances_omv(dtaInp = cntFrm, varDst = names(cntFrm), stdDst = "none", nmeDst = "euclid")
    expect_equal(dim(df4Chk), rep(dim(cntFrm)[2], 2))
    expect_equal(names(df4Chk), names(cntFrm))
    expect_true(isSymmetric(unname(as.matrix(df4Chk))))
    expect_equal(unname(diag(as.matrix(df4Chk))), rep(0, 20))
    expect_equal(unname(apply(df4Chk, 2, mean)), c(90.18679, 93.75375, 93.49854,  95.84402, 100.25257, 96.79214,  95.25020, 96.61567, 97.88351, 108.10851,
                                                   91.54438, 96.83193, 99.66558, 104.26397, 102.46154, 99.06569, 104.29886, 89.90709, 97.13592, 101.83002),
                                                 tolerance = 1e-6)
    expect_equal(unname(apply(df4Chk, 2, sd)),   c(23.05755, 23.22398, 23.30027,  23.63582,  24.78627, 24.50848,  24.47311, 24.23558, 24.02617,  27.30439,
                                                   23.24474, 24.65757, 25.18948,  25.89397,  25.80464, 24.56810,  27.01149, 22.32947, 24.55965,  25.65252),
                                                 tolerance = 1e-6)
    # euclidean == euclid
    expect_equal(distances_omv(dtaInp = cntFrm, varDst = names(cntFrm), stdDst = "none", nmeDst = "euclidean"), df4Chk)

    # distance measure: seuclid, standardization: none
    df4Chk <- distances_omv(dtaInp = cntFrm, varDst = names(cntFrm), stdDst = "none", nmeDst = "seuclid")
    expect_equal(dim(df4Chk), rep(dim(cntFrm)[2], 2))
    expect_equal(names(df4Chk), names(cntFrm))
    expect_true(isSymmetric(unname(as.matrix(df4Chk))))
    expect_equal(unname(diag(as.matrix(df4Chk))), rep(0, 20))
    expect_equal(unname(apply(df4Chk, 2, mean)), c(8638.72502, 9302.15162,  9257.73442,  9716.79546, 10634.21924,  9939.35126,  9641.58669, 9892.58203, 10129.57618, 12395.70229,
                                                   8893.67530, 9954.01874, 10536.01146, 11507.94836, 11130.95307, 10387.42201, 11571.39124, 8556.95973, 10008.40523, 10994.50290),
                                                 tolerance = 1e-6)
    expect_equal(unname(apply(df4Chk, 2, sd)),   c(2706.95087, 2612.18298,  2657.96515,  2682.29126,  2974.69185,  2976.31068,  3042.57714, 2883.08202,  2770.73040,  3710.39530,
                                                   2683.03967, 3049.76569,  3146.80743,  3249.34860,  3258.55781,  2914.40577,  3724.65631, 2435.83129,  2965.04294,  3220.06283),
                                                 tolerance = 1e-6)
    # seuclidean == seuclid
    expect_equal(distances_omv(dtaInp = cntFrm, varDst = names(cntFrm), stdDst = "none", nmeDst = "seuclidean"), df4Chk)

    # distance measure: block, standardization: none
    df4Chk <- distances_omv(dtaInp = cntFrm, varDst = names(cntFrm), stdDst = "none", nmeDst = "block")
    expect_equal(dim(df4Chk), rep(dim(cntFrm)[2], 2))
    expect_equal(names(df4Chk), names(cntFrm))
    expect_true(isSymmetric(unname(as.matrix(df4Chk))))
    expect_equal(unname(diag(as.matrix(df4Chk))), rep(0, 20))
    expect_equal(unname(apply(df4Chk, 2, mean)), c(510.11848, 531.72723, 527.03184, 541.47101, 563.65123, 552.49981, 539.29376, 547.19063, 560.14147, 619.83929,
                                                   514.77227, 548.20019, 566.08846, 594.66377, 580.03450, 560.21567, 599.48532, 514.68472, 552.15867, 578.28306),
                                                 tolerance = 1e-6)
    expect_equal(unname(apply(df4Chk, 2, sd)),   c(130.57798, 134.10532, 128.83121, 135.26859, 141.28170, 145.57155, 137.79444, 136.52438, 138.00417, 156.99662,
                                                   133.49714, 142.72953, 144.74001, 151.31945, 149.44571, 141.14608, 154.58736, 128.54925, 140.88010, 144.38364),
                                                 tolerance = 1e-6)
    # manhattan == block
    expect_equal(distances_omv(dtaInp = cntFrm, varDst = names(cntFrm), stdDst = "none", nmeDst = "manhattan"), df4Chk)

    # distance measure: canberra, standardization: none
    df4Chk <- distances_omv(dtaInp = cntFrm, varDst = names(cntFrm), stdDst = "none", nmeDst = "canberra")
    expect_equal(dim(df4Chk), rep(dim(cntFrm)[2], 2))
    expect_equal(names(df4Chk), names(cntFrm))
    expect_true(isSymmetric(unname(as.matrix(df4Chk))))
    expect_equal(unname(diag(as.matrix(df4Chk))), rep(0, 20))
    expect_equal(unname(apply(df4Chk, 2, mean)), c(34.49320, 34.03633, 34.97755, 33.71225, 34.58030, 34.32198, 33.97928, 34.50087, 33.79525, 35.26895,
                                                   34.28703, 32.95946, 34.23159, 34.90192, 33.03279, 34.09206, 34.74617, 33.79520, 34.19561, 35.30816),
                                                 tolerance = 1e-6)
    expect_equal(unname(apply(df4Chk, 2, sd)),    c(8.47184,  8.47098,  8.32530,  8.30726,  8.40608,  8.44222,  8.41203,  8.31785,  8.30473,  8.58373,
                                                    8.39323,  8.19873,  8.37566,  8.47766,  8.23869,  8.30194,  8.54616,  8.17902,  8.31659,  8.51508),
                                                 tolerance = 1e-6)

    # distance measure: chebychev, standardization: none
    df4Chk <- distances_omv(dtaInp = cntFrm, varDst = names(cntFrm), stdDst = "none", nmeDst = "chebychev")
    expect_equal(dim(df4Chk), rep(dim(cntFrm)[2], 2))
    expect_equal(names(df4Chk), names(cntFrm))
    expect_true(isSymmetric(unname(as.matrix(df4Chk))))
    expect_equal(unname(diag(as.matrix(df4Chk))), rep(0, 20))
    expect_equal(unname(apply(df4Chk, 2, mean)), c(33.35363, 34.63043, 33.21321, 33.21751, 36.14409, 34.50494, 35.56581, 34.76410, 33.96757, 39.74970,
                                                   33.26305, 34.51619, 34.50881, 38.05685, 35.67290, 35.33634, 36.56601, 30.28183, 33.55651, 35.04377),
                                                 tolerance = 1e-6)
    expect_equal(unname(apply(df4Chk, 2, sd)),   c(10.22961, 10.39461, 10.23548,  9.94799, 10.46579,  9.75279, 10.94785,  9.79454,  8.88057, 12.89452,
                                                    9.69607, 10.54373,  9.16828, 11.26589, 10.33764,  9.49605, 11.07437,  7.94100,  9.46650,  9.83562),
                                                 tolerance = 1e-6)

    # distance measure: minkowski_p, standardization: none
    df4Chk <- distances_omv(dtaInp = cntFrm, varDst = names(cntFrm), stdDst = "none", nmeDst = "minkowski_3")
    expect_equal(dim(df4Chk), rep(dim(cntFrm)[2], 2))
    expect_equal(names(df4Chk), names(cntFrm))
    expect_true(isSymmetric(unname(as.matrix(df4Chk))))
    expect_equal(unname(diag(as.matrix(df4Chk))), rep(0, 20))
    expect_equal(unname(apply(df4Chk, 2, mean)), c(54.67712, 56.99251, 56.82751, 57.96354, 61.19111, 58.47326, 57.96689, 58.45795, 58.71571, 65.33263,
                                                   55.87355, 58.55490, 59.94234, 63.21778, 61.92123, 60.09938, 62.57737, 53.89934, 58.56803, 61.46659),
                                                 tolerance = 1e-6)
    expect_equal(unname(apply(df4Chk, 2, sd)),   c(14.18010, 14.04591, 14.62228, 14.40142, 15.23499, 14.66176, 15.14758, 14.87106, 14.46439, 16.73602,
                                                   14.03609, 15.05682, 15.14051, 15.63788, 15.78960, 14.90891, 16.55806, 13.43597, 14.88961, 15.68011),
                                                 tolerance = 1e-6)
    # minkowski_2 == euclid
    expect_equal(distances_omv(dtaInp = cntFrm, varDst = names(cntFrm), stdDst = "none", nmeDst = "minkowski_2"),
                 distances_omv(dtaInp = cntFrm, varDst = names(cntFrm), stdDst = "none", nmeDst = "euclid"))
    # minkowski_1 == block
    expect_equal(distances_omv(dtaInp = cntFrm, varDst = names(cntFrm), stdDst = "none", nmeDst = "minkowski_1"),
                 distances_omv(dtaInp = cntFrm, varDst = names(cntFrm), stdDst = "none", nmeDst = "block"))

    # distance measure: power_p_r, standardization: none
    df4Chk <- distances_omv(dtaInp = cntFrm, varDst = names(cntFrm), stdDst = "none", nmeDst = "power_3_2")
    expect_equal(dim(df4Chk), rep(dim(cntFrm)[2], 2))
    expect_equal(names(df4Chk), names(cntFrm))
    expect_true(isSymmetric(unname(as.matrix(df4Chk))))
    expect_equal(unname(diag(as.matrix(df4Chk))), rep(0, 20))
    expect_equal(unname(apply(df4Chk, 2, mean)), c(416.45985, 442.22845, 441.11214, 453.74088, 492.19694, 459.91073, 454.75426, 460.00974, 462.43128, 543.66607,
                                                   429.61070, 461.35882, 477.49091, 516.71302, 501.55846, 479.01644, 510.38992, 406.90613, 461.30873, 496.06172),
                                                 tolerance = 1e-6)
    expect_equal(unname(apply(df4Chk, 2, sd)),   c(119.68786, 114.87371, 124.92054, 119.39272, 130.33552, 123.43098, 132.83875, 127.15831, 119.87500, 152.16527,
                                                   115.95281, 130.67957, 130.34089, 135.02989, 138.66486, 125.73364, 151.64728, 108.35405, 126.96224, 136.96509),
                                                 tolerance = 1e-6)
    # power_2_2 == minkowski_2
    expect_equal(distances_omv(dtaInp = cntFrm, varDst = names(cntFrm), stdDst = "none", nmeDst = "power_2_2"),
                 distances_omv(dtaInp = cntFrm, varDst = names(cntFrm), stdDst = "none", nmeDst = "minkowski_2"))
    # power_4_4 == minkowski_4
    expect_equal(distances_omv(dtaInp = cntFrm, varDst = names(cntFrm), stdDst = "none", nmeDst = "power_4_4"),
                 distances_omv(dtaInp = cntFrm, varDst = names(cntFrm), stdDst = "none", nmeDst = "minkowski_4"))

    # distance measure: cosine, standardization: none
    df4Chk <- distances_omv(dtaInp = cntFrm, varDst = names(cntFrm), stdDst = "none", nmeDst = "cosine")
    expect_equal(dim(df4Chk), rep(dim(cntFrm)[2], 2))
    expect_equal(names(df4Chk), names(cntFrm))
    expect_true(isSymmetric(unname(as.matrix(df4Chk))))
    expect_equal(unname(diag(as.matrix(df4Chk))), rep(1, 20))
    expect_equal(unname(apply(df4Chk, 2, mean)), c(0.0261721, 0.0727720, 0.0255531,  0.0594165, 0.0500769, 0.0181540, 0.0526752, 0.0438387, 0.1013408, 0.0327727,
                                                   0.0789248, 0.0405412, 0.0554114, -0.0059643, 0.0553022, 0.0349486, 0.0960476, 0.1117722, 0.0674172, 0.0028155),
                                                 tolerance = 1e-6)
    expect_equal(unname(apply(df4Chk, 2, sd)),   c(0.2857365, 0.2518821, 0.2552830,  0.2632538, 0.2516072, 0.2659546, 0.2764758, 0.2578368, 0.2352152, 0.2818756,
                                                   0.2667546, 0.2699523, 0.2699902,  0.2642751, 0.2621859, 0.2623423, 0.2600199, 0.2525837, 0.2512283, 0.2679373),
                                                 tolerance = 1e-6)

    # distance measure: correlation, standardization: none
    df4Chk <- distances_omv(dtaInp = cntFrm, varDst = names(cntFrm), stdDst = "none", nmeDst = "correlation")
    expect_equal(dim(df4Chk), rep(dim(cntFrm)[2], 2))
    expect_equal(names(df4Chk), names(cntFrm))
    expect_true(isSymmetric(unname(as.matrix(df4Chk))))
    expect_equal(unname(diag(as.matrix(df4Chk))), rep(1, 20))
    expect_equal(unname(apply(df4Chk, 2, mean)), c(0.0282471, 0.0750318, 0.0246367,  0.0603677, 0.0499838, 0.0191165, 0.0545289, 0.0434007, 0.1012484, 0.0330809,
                                                   0.0784566, 0.0402646, 0.0545752, -0.0073858, 0.0559124, 0.0343765, 0.0965748, 0.1115375, 0.0698606, 0.0015639),
                                                 tolerance = 1e-6)
    expect_equal(unname(apply(df4Chk, 2, sd)),   c(0.2866331, 0.2517791, 0.2594964,  0.2642852, 0.2517053, 0.2675609, 0.2756468, 0.2583735, 0.2358153, 0.2829945,
                                                   0.2672026, 0.2703650, 0.2736224,  0.2718538, 0.2623670, 0.2627199, 0.2594415, 0.2535597, 0.2497000, 0.2686121),
                                                 tolerance = 1e-6)

    # =================================================================================================================
    # distance measures for frequency data: chisq, ph2; no standardization
    # distance measure: chisq, standardization: none
    df4Chk <- distances_omv(dtaInp = frqFrm, varDst = names(frqFrm), stdDst = "none", nmeDst = "chisq")
    expect_equal(dim(df4Chk), rep(dim(frqFrm)[2], 2))
    expect_equal(names(df4Chk), names(frqFrm))
    expect_true(isSymmetric(unname(as.matrix(df4Chk))))
    expect_equal(unname(diag(as.matrix(df4Chk))), rep(0, 20))
    expect_equal(unname(apply(df4Chk, 2, mean)) - c(7.747222, 10.663834, 10.396158, 7.620559,  7.808307, 6.532916, 10.848895, 6.095187, 7.102744,  8.519774,
                                                    9.919973,  8.877894,  9.220077, 6.708244, 10.539968, 9.265459,  7.193613, 6.704473, 7.738743, 11.206640),
                                                  rep(0, dim(frqFrm)[2]), tolerance = 1e-6)
    expect_equal(unname(apply(df4Chk, 2, sd))   - c(3.220719,  4.143478,  4.850509, 3.472757,  3.441681, 3.346350,  4.644328, 2.437293, 3.515119,  4.077974,
                                                    3.678504,  4.087233,  4.132481, 2.957086,  4.274442, 3.846881,  3.410927, 2.797160, 4.067294,  4.667570),
                                                  rep(0, dim(frqFrm)[2]), tolerance = 1e-6)

    # distance measure: ph2, standardization: none
    df4Chk <- distances_omv(dtaInp = frqFrm, varDst = names(frqFrm), stdDst = "none", nmeDst = "ph2")
    expect_equal(dim(df4Chk), rep(dim(frqFrm)[2], 2))
    expect_equal(names(df4Chk), names(frqFrm))
    expect_true(isSymmetric(unname(as.matrix(df4Chk))))
    expect_equal(unname(diag(as.matrix(df4Chk))), rep(0, 20))
    expect_equal(unname(apply(df4Chk, 2, mean)) - c(0.774722, 1.066383, 1.039616, 0.762056, 0.780831, 0.653292, 1.084890, 0.609519, 0.710274, 0.851977,
                                                    0.991997, 0.887789, 0.922008, 0.670824, 1.053997, 0.926546, 0.719361, 0.670447, 0.773874, 1.120664),
                                                  rep(0, dim(frqFrm)[2]), tolerance = 1e-6)
    expect_equal(unname(apply(df4Chk, 2, sd))   - c(0.322072, 0.414348, 0.485051, 0.347276, 0.344168, 0.334635, 0.464433, 0.243729, 0.351512, 0.407797,
                                                    0.367850, 0.408723, 0.413248, 0.295709, 0.427444, 0.384688, 0.341093, 0.279716, 0.406729, 0.466757),
                                                  rep(0, dim(frqFrm)[2]), tolerance = 1e-6)

    # =================================================================================================================
    # distance measures for binary data: beuclid, blwmn, bseuclid, bshape, d, dice, disper, hamann, jaccard, k1, k2,
    # lambda, ochiai, pattern, phi, q, rr, rt, size, sm, ss1, ss2, ss3, ss4, ss5, variance, y; no standardization
    # distance measure: beuclid, standardization: none
    df4Chk <- distances_omv(dtaInp = binFrm, varDst = names(binFrm), stdDst = "none", nmeDst = "beuclid")
    expect_equal(dim(df4Chk), rep(dim(binFrm)[2], 2))
    expect_equal(names(df4Chk), names(binFrm))
    expect_true(isSymmetric(unname(as.matrix(df4Chk))))
    expect_equal(unname(diag(as.matrix(df4Chk))), rep(0, 20))
    expect_equal(unname(apply(df4Chk, 2, mean)) - c(4.689500, 4.858037, 4.623153, 4.728271, 4.765985, 4.797172, 4.704819, 4.646293, 4.836648, 4.722894,
                                                    4.716488, 4.704718, 4.728416, 4.614760, 4.743886, 4.658794, 4.741526, 4.654649, 4.583997, 4.792678),
                                                  rep(0, dim(frqFrm)[2]), tolerance = 1e-6)
    expect_equal(unname(apply(df4Chk, 2, sd))   - c(1.151013, 1.191848, 1.112825, 1.166847, 1.185611, 1.186388, 1.176382, 1.152555, 1.195094, 1.189547,
                                                    1.171927, 1.176806, 1.166225, 1.148909, 1.190112, 1.145075, 1.155278, 1.162683, 1.141086, 1.160872),
                                                  rep(0, dim(frqFrm)[2]), tolerance = 1e-6)
    expect_equal(df4Chk, distances_omv(dtaInp = binFm2, varDst = names(binFm2), stdDst = "none", nmeDst = "beuclid"))
    expect_equal(df4Chk, distances_omv(dtaInp = binFm3, varDst = names(binFm3), stdDst = "none", nmeDst = "beuclid_2_1"))
    expect_equal(df4Chk, distances_omv(dtaInp = binFm4, varDst = names(binFm4), stdDst = "none", nmeDst = "beuclid_1_2"))

    # distance measure: blwmn, standardization: none
    df4Chk <- distances_omv(dtaInp = binFrm, varDst = names(binFrm), stdDst = "none", nmeDst = "blwmn")
    expect_equal(dim(df4Chk), rep(dim(binFrm)[2], 2))
    expect_equal(names(df4Chk), names(binFrm))
    expect_true(isSymmetric(unname(as.matrix(df4Chk))))
    expect_equal(unname(diag(as.matrix(df4Chk))), rep(0, 20))
    expect_equal(unname(apply(df4Chk, 2, mean)) - c(0.471614, 0.462666, 0.424808, 0.454607, 0.497064, 0.523678, 0.484746, 0.522466, 0.520449, 0.480162,
                                                    0.443808, 0.503348, 0.528201, 0.457711, 0.450691, 0.493759, 0.531400, 0.440961, 0.468459, 0.500525),
                                                  rep(0, dim(frqFrm)[2]), tolerance = 1e-6)
    expect_equal(unname(apply(df4Chk, 2, sd))   - c(0.133840, 0.134304, 0.113729, 0.135335, 0.143792, 0.150475, 0.143404, 0.146677, 0.143397, 0.147464,
                                                    0.131636, 0.144830, 0.142485, 0.134467, 0.140197, 0.139029, 0.141272, 0.134183, 0.133866, 0.131386),
                                                  rep(0, dim(frqFrm)[2]), tolerance = 1e-6)
    expect_equal(df4Chk, distances_omv(dtaInp = binFm2, varDst = names(binFm2), stdDst = "none", nmeDst = "blwmn"))
    expect_equal(df4Chk, distances_omv(dtaInp = binFm3, varDst = names(binFm3), stdDst = "none", nmeDst = "blwmn_2_1"))
    expect_equal(df4Chk, distances_omv(dtaInp = binFm4, varDst = names(binFm4), stdDst = "none", nmeDst = "blwmn_1_2"))

    # distance measure: bseuclid, standardization: none
    df4Chk <- distances_omv(dtaInp = binFrm, varDst = names(binFrm), stdDst = "none", nmeDst = "bseuclid")
    expect_equal(dim(df4Chk), rep(dim(binFrm)[2], 2))
    expect_equal(names(df4Chk), names(binFrm))
    expect_true(isSymmetric(unname(as.matrix(df4Chk))))
    expect_equal(unname(diag(as.matrix(df4Chk))), rep(0, 20))
    expect_equal(unname(apply(df4Chk, 2, mean)) - c(23.250000, 24.950000, 22.550000, 23.650000, 24.050000, 24.350000, 23.450000, 22.850000, 24.750000, 23.650000,
                                                    23.550000, 23.450000, 23.650000, 22.550000, 23.850000, 22.950000, 23.750000, 22.950000, 22.250000, 24.250000),
                                                  rep(0, dim(frqFrm)[2]), tolerance = 1e-6)
    expect_equal(unname(apply(df4Chk, 2, sd))   -  c(6.373258,  6.816968,  5.789964,  6.579674,  6.870800,  6.753362,  6.770485,  6.474850,  6.912041,  7.005825,
                                                     6.692297,  6.739319,  6.531503,  6.411133,  6.937882,  6.303508,  6.323515,  6.613185,  6.364995,  6.323515),
                                                  rep(0, dim(frqFrm)[2]), tolerance = 1e-6)
    expect_equal(df4Chk, distances_omv(dtaInp = binFm2, varDst = names(binFm2), stdDst = "none", nmeDst = "bseuclid"))
    expect_equal(df4Chk, distances_omv(dtaInp = binFm3, varDst = names(binFm3), stdDst = "none", nmeDst = "bseuclid_2_1"))
    expect_equal(df4Chk, distances_omv(dtaInp = binFm4, varDst = names(binFm4), stdDst = "none", nmeDst = "bseuclid_1_2"))

    # distance measure: bshape, standardization: none
    df4Chk <- distances_omv(dtaInp = binFrm, varDst = names(binFrm), stdDst = "none", nmeDst = "bshape")
    expect_equal(dim(df4Chk), rep(dim(binFrm)[2], 2))
    expect_equal(names(df4Chk), names(binFrm))
    expect_true(isSymmetric(unname(as.matrix(df4Chk))))
    expect_equal(unname(diag(as.matrix(df4Chk))), rep(0, 20))
    expect_equal(unname(apply(df4Chk, 2, mean)) - c(0.460700, 0.482900, 0.438860, 0.464020, 0.476660, 0.480180, 0.464660, 0.440460, 0.489820, 0.468700,
                                                    0.458860, 0.462180, 0.460500, 0.446700, 0.464860, 0.452180, 0.462500, 0.450020, 0.439820, 0.480660),
                                                  rep(0, dim(frqFrm)[2]), tolerance = 1e-6)
    expect_equal(unname(apply(df4Chk, 2, sd))   - c(0.126045, 0.131225, 0.113546, 0.128046, 0.137094, 0.133547, 0.134332, 0.125128, 0.135745, 0.138601,
                                                    0.131514, 0.132424, 0.126058, 0.127841, 0.134177, 0.124929, 0.122934, 0.130107, 0.126138, 0.124965),
                                                  rep(0, dim(frqFrm)[2]), tolerance = 1e-6)
    expect_equal(df4Chk, distances_omv(dtaInp = binFm2, varDst = names(binFm2), stdDst = "none", nmeDst = "bshape"))
    expect_equal(df4Chk, distances_omv(dtaInp = binFm3, varDst = names(binFm3), stdDst = "none", nmeDst = "bshape_2_1"))
    expect_equal(df4Chk, distances_omv(dtaInp = binFm4, varDst = names(binFm4), stdDst = "none", nmeDst = "bshape_1_2"))

    # distance measure: d, standardization: none
    df4Chk <- distances_omv(dtaInp = binFrm, varDst = names(binFrm), stdDst = "none", nmeDst = "d")
    expect_equal(dim(df4Chk), rep(dim(binFrm)[2], 2))
    expect_equal(names(df4Chk), names(binFrm))
    expect_true(isSymmetric(unname(as.matrix(df4Chk))))
    expect_equal(unname(diag(as.matrix(df4Chk))), c(0.500000, 0.400000, 0.420000, 0.440000, 0.480000, 0.440000, 0.480000, 0.380000, 0.460000, 0.500000,
                                                    0.420000, 0.440000, 0.400000, 0.500000, 0.420000, 0.440000, 0.400000, 0.440000, 0.460000, 0.480000))
    expect_equal(unname(apply(df4Chk, 2, mean)) - c(0.063000, 0.036000, 0.028000, 0.039000, 0.059500, 0.041000, 0.057500, 0.031500, 0.049500, 0.073000,
                                                    0.039000, 0.044000, 0.031000, 0.068500, 0.045500, 0.039500, 0.029500, 0.043500, 0.052000, 0.045500),
                                                  rep(0, dim(frqFrm)[2]), tolerance = 1e-6)
    expect_equal(unname(apply(df4Chk, 2, sd))   - c(0.107415, 0.090403, 0.092884, 0.098723, 0.106152, 0.096349, 0.106715, 0.082925, 0.102211, 0.109405,
                                                    0.094195, 0.099441, 0.088133, 0.107863, 0.097142, 0.096979, 0.089058, 0.098209, 0.101650, 0.105555),
                                                  rep(0, dim(frqFrm)[2]), tolerance = 1e-6)
    expect_equal(df4Chk, distances_omv(dtaInp = binFm2, varDst = names(binFm2), stdDst = "none", nmeDst = "d"))
    expect_equal(df4Chk, distances_omv(dtaInp = binFm3, varDst = names(binFm3), stdDst = "none", nmeDst = "d_2_1"))
    expect_equal(df4Chk, distances_omv(dtaInp = binFm4, varDst = names(binFm4), stdDst = "none", nmeDst = "d_1_2"))

    # distance measure: dice, standardization: none
    df4Chk <- distances_omv(dtaInp = binFrm, varDst = names(binFrm), stdDst = "none", nmeDst = "dice")
    expect_equal(dim(df4Chk), rep(dim(binFrm)[2], 2))
    expect_equal(names(df4Chk), names(binFrm))
    expect_true(isSymmetric(unname(as.matrix(df4Chk))))
    expect_equal(unname(diag(as.matrix(df4Chk))), rep(1, 20))
    expect_equal(unname(apply(df4Chk, 2, mean)) - c(0.528386, 0.537334, 0.575192, 0.545393, 0.502936, 0.476322, 0.515254, 0.477534, 0.479551, 0.519838,
                                                    0.556192, 0.496652, 0.471799, 0.542289, 0.549309, 0.506241, 0.468600, 0.559039, 0.531541, 0.499475),
                                                  rep(0, dim(frqFrm)[2]), tolerance = 1e-6)
    expect_equal(unname(apply(df4Chk, 2, sd))   - c(0.133840, 0.134304, 0.113729, 0.135335, 0.143792, 0.150475, 0.143404, 0.146677, 0.143397, 0.147464,
                                                    0.131636, 0.144830, 0.142485, 0.134467, 0.140197, 0.139029, 0.141272, 0.134183, 0.133866, 0.131386),
                                                  rep(0, dim(frqFrm)[2]), tolerance = 1e-6)
    expect_equal(df4Chk, distances_omv(dtaInp = binFm2, varDst = names(binFm2), stdDst = "none", nmeDst = "dice"))
    expect_equal(df4Chk, distances_omv(dtaInp = binFm3, varDst = names(binFm3), stdDst = "none", nmeDst = "dice_2_1"))
    expect_equal(df4Chk, distances_omv(dtaInp = binFm4, varDst = names(binFm4), stdDst = "none", nmeDst = "dice_1_2"))

    # distance measure: disper, standardization: none
    df4Chk <- distances_omv(dtaInp = binFrm, varDst = names(binFrm), stdDst = "none", nmeDst = "disper")
    expect_equal(dim(df4Chk), rep(dim(binFrm)[2], 2))
    expect_equal(names(df4Chk), names(binFrm))
    expect_true(isSymmetric(unname(as.matrix(df4Chk))))
    expect_equal(unname(diag(as.matrix(df4Chk))), c(0.250000, 0.240000, 0.243600, 0.246400, 0.249600, 0.246400, 0.249600, 0.235600, 0.248400, 0.250000,
                                                    0.243600, 0.246400, 0.240000, 0.250000, 0.243600, 0.246400, 0.240000, 0.246400, 0.248400, 0.249600))
    expect_equal(unname(apply(df4Chk, 2, mean)) - c(0.017500, 0.001400, 0.025220, 0.014040, 0.009320, 0.005960, 0.015320, 0.020420, 0.002140, 0.013500,
                                                    0.015220, 0.014960, 0.012600, 0.024500, 0.012220, 0.019960, 0.011600, 0.021040, 0.027140, 0.007320),
                                                  rep(0, dim(frqFrm)[2]), tolerance = 1e-6)
    expect_equal(unname(apply(df4Chk, 2, sd))   - c(0.063733, 0.064905, 0.056230, 0.064434, 0.068640, 0.066833, 0.067634, 0.061603, 0.068427, 0.070058,
                                                    0.065290, 0.066312, 0.062216, 0.064111, 0.066939, 0.062432, 0.060675, 0.064851, 0.063193, 0.063035),
                                                  rep(0, dim(frqFrm)[2]), tolerance = 1e-6)
    expect_equal(df4Chk, distances_omv(dtaInp = binFm2, varDst = names(binFm2), stdDst = "none", nmeDst = "disper"))
    expect_equal(df4Chk, distances_omv(dtaInp = binFm3, varDst = names(binFm3), stdDst = "none", nmeDst = "disper_2_1"))
    expect_equal(df4Chk, distances_omv(dtaInp = binFm4, varDst = names(binFm4), stdDst = "none", nmeDst = "disper_1_2"))

    # distance measure: hamann, standardization: none
    df4Chk <- distances_omv(dtaInp = binFrm, varDst = names(binFrm), stdDst = "none", nmeDst = "hamann")
    expect_equal(dim(df4Chk), rep(dim(binFrm)[2], 2))
    expect_equal(names(df4Chk), names(binFrm))
    expect_true(isSymmetric(unname(as.matrix(df4Chk))))
    expect_equal(unname(diag(as.matrix(df4Chk))), rep(1, 20))
    expect_equal(unname(apply(df4Chk, 2, mean)) - c(0.070000, 0.002000, 0.098000, 0.054000, 0.038000, 0.026000, 0.062000, 0.086000, 0.010000, 0.054000,
                                                    0.058000, 0.062000, 0.054000, 0.098000, 0.046000, 0.082000, 0.050000, 0.082000, 0.110000, 0.030000),
                                                  rep(0, dim(frqFrm)[2]), tolerance = 1e-6)
    expect_equal(unname(apply(df4Chk, 2, sd))   - c(0.254930, 0.272679, 0.231599, 0.263187, 0.274832, 0.270134, 0.270819, 0.258994, 0.276482, 0.280233,
                                                    0.267692, 0.269573, 0.261260, 0.256445, 0.277515, 0.252140, 0.252941, 0.264527, 0.254600, 0.252941),
                                                  rep(0, dim(frqFrm)[2]), tolerance = 1e-6)
    expect_equal(df4Chk, distances_omv(dtaInp = binFm2, varDst = names(binFm2), stdDst = "none", nmeDst = "hamann"))
    expect_equal(df4Chk, distances_omv(dtaInp = binFm3, varDst = names(binFm3), stdDst = "none", nmeDst = "hamann_2_1"))
    expect_equal(df4Chk, distances_omv(dtaInp = binFm4, varDst = names(binFm4), stdDst = "none", nmeDst = "hamann_1_2"))

    # distance measure: jaccard, standardization: none
    df4Chk <- distances_omv(dtaInp = binFrm, varDst = names(binFrm), stdDst = "none", nmeDst = "jaccard")
    expect_equal(dim(df4Chk), rep(dim(binFrm)[2], 2))
    expect_equal(names(df4Chk), names(binFrm))
    expect_true(isSymmetric(unname(as.matrix(df4Chk))))
    expect_equal(unname(diag(as.matrix(df4Chk))), rep(1, 20))
    expect_equal(unname(apply(df4Chk, 2, mean)) - c(0.372793, 0.381299, 0.414720, 0.389012, 0.351160, 0.328574, 0.362228, 0.329147, 0.330369, 0.367062,
                                                    0.398811, 0.345661, 0.323499, 0.386091, 0.393669, 0.353358, 0.320593, 0.402027, 0.375827, 0.346207),
                                                  rep(0, dim(frqFrm)[2]), tolerance = 1e-6)
    expect_equal(unname(apply(df4Chk, 2, sd))   - c(0.161525, 0.162120, 0.146869, 0.161925, 0.169331, 0.173258, 0.168425, 0.171797, 0.169527, 0.170670,
                                                    0.159361, 0.169905, 0.169132, 0.162884, 0.166213, 0.165965, 0.168725, 0.161480, 0.162228, 0.161989),
                                                  rep(0, dim(frqFrm)[2]), tolerance = 1e-6)
    expect_equal(df4Chk, distances_omv(dtaInp = binFm2, varDst = names(binFm2), stdDst = "none", nmeDst = "jaccard"))
    expect_equal(df4Chk, distances_omv(dtaInp = binFm3, varDst = names(binFm3), stdDst = "none", nmeDst = "jaccard_2_1"))
    expect_equal(df4Chk, distances_omv(dtaInp = binFm4, varDst = names(binFm4), stdDst = "none", nmeDst = "jaccard_1_2"))
    expect_equal(df4Chk, distances_omv(dtaInp = binFrm, varDst = names(binFrm), stdDst = "none", nmeDst = "jaccards"))
    # distance measure: jaccards, standardization: none
    expect_equal(df4Chk, distances_omv(dtaInp = binFrm, varDst = names(binFrm), stdDst = "none", nmeDst = "jaccards"))
    # distance measure: jaccard, standardization: none
    expect_true(all(df4Chk + distances_omv(dtaInp = binFrm, varDst = names(binFrm), stdDst = "none", nmeDst = "jaccardd") == 1))
    expect_equal(as.matrix(distances_omv(dtaInp = binFm2, varDst = names(binFm2), stdDst = "none", nmeDst = "jaccardd")),
                 as.matrix(stats::dist(t(as.matrix(binFrm)), "binary", upper = TRUE, diag = TRUE)))

    # distance measure: k1, standardization: none
    df4Chk <- distances_omv(dtaInp = binFrm, varDst = names(binFrm), stdDst = "none", nmeDst = "k1")
    expect_equal(dim(df4Chk), rep(dim(binFrm)[2], 2))
    expect_equal(names(df4Chk), names(binFrm))
    expect_true(isSymmetric(unname(as.matrix(df4Chk))))
    expect_true(all(is.na(diag(as.matrix(df4Chk)))))
    expect_equal(unname(apply(df4Chk, 2, mean, na.rm = TRUE)),
                 c(0.529244, 0.554404, 0.634104, 0.575568, 0.481972, 0.429257, 0.509615, 0.429813, 0.430074, 0.525043,
                   0.601247, 0.468599, 0.413306, 0.570882, 0.596938, 0.483544, 0.406245, 0.614369, 0.539410, 0.461229),
                 tolerance = 1e-6)
    expect_equal(unname(apply(df4Chk, 2, sd,   na.rm = TRUE)),
                 c(0.151432, 0.178762, 0.136927, 0.185283, 0.172207, 0.147623, 0.182650, 0.147587, 0.132521, 0.199638,
                   0.187701, 0.167360, 0.116168, 0.210129, 0.235243, 0.153660, 0.109127, 0.210847, 0.171373, 0.114502),
                 tolerance = 1e-5)
    expect_equal(df4Chk, distances_omv(dtaInp = binFm2, varDst = names(binFm2), stdDst = "none", nmeDst = "k1"))
    expect_equal(df4Chk, distances_omv(dtaInp = binFm3, varDst = names(binFm3), stdDst = "none", nmeDst = "k1_2_1"))
    expect_equal(df4Chk, distances_omv(dtaInp = binFm4, varDst = names(binFm4), stdDst = "none", nmeDst = "k1_1_2"))

    # distance measure: k2, standardization: none
    df4Chk <- distances_omv(dtaInp = binFrm, varDst = names(binFrm), stdDst = "none", nmeDst = "k2")
    expect_equal(dim(df4Chk), rep(dim(binFrm)[2], 2))
    expect_equal(names(df4Chk), names(binFrm))
    expect_true(isSymmetric(unname(as.matrix(df4Chk))))
    expect_equal(unname(diag(as.matrix(df4Chk))), rep(1, 20))
    expect_equal(unname(apply(df4Chk, 2, mean)) - c(0.530571, 0.544611, 0.581644, 0.549715, 0.505157, 0.479658, 0.517415, 0.486805, 0.481851, 0.521943,
                                                    0.562362, 0.500027, 0.478165, 0.544657, 0.555061, 0.509834, 0.475082, 0.563743, 0.534299, 0.501551),
                                                  rep(0, dim(frqFrm)[2]), tolerance = 1e-3)
    expect_equal(unname(apply(df4Chk, 2, sd))   - c(0.133118, 0.131063, 0.111346, 0.133039, 0.143951, 0.150667, 0.143232, 0.147246, 0.142928, 0.146791,
                                                    0.129895, 0.144653, 0.141873, 0.134176, 0.137326, 0.139532, 0.141270, 0.132616, 0.134015, 0.130903),
                                                  rep(0, dim(frqFrm)[2]), tolerance = 1e-4)
    expect_equal(df4Chk, distances_omv(dtaInp = binFm2, varDst = names(binFm2), stdDst = "none", nmeDst = "k2"))
    expect_equal(df4Chk, distances_omv(dtaInp = binFm3, varDst = names(binFm3), stdDst = "none", nmeDst = "k2_2_1"))
    expect_equal(df4Chk, distances_omv(dtaInp = binFm4, varDst = names(binFm4), stdDst = "none", nmeDst = "k2_1_2"))

    # distance measure: lambda, standardization: none
    df4Chk <- distances_omv(dtaInp = binFrm, varDst = names(binFrm), stdDst = "none", nmeDst = "lambda")
    expect_equal(dim(df4Chk), rep(dim(binFrm)[2], 2))
    expect_equal(names(df4Chk), names(binFrm))
    expect_true(isSymmetric(unname(as.matrix(df4Chk))))
    expect_equal(unname(diag(as.matrix(df4Chk))), rep(1, 20))
    expect_equal(unname(apply(df4Chk, 2, mean)) - c(0.129500, 0.086240, 0.065687, 0.087403, 0.125925, 0.092580, 0.122388, 0.079794, 0.108648, 0.151168,
                                                    0.090533, 0.098449, 0.075527, 0.141509, 0.104574, 0.088939, 0.071768, 0.097606, 0.113185, 0.095457),
                                                  rep(0, dim(frqFrm)[2]), tolerance = 1e-6)
    expect_equal(unname(apply(df4Chk, 2, sd))   - c(0.214238, 0.224516, 0.221229, 0.223766, 0.220777, 0.218954, 0.222921, 0.218527, 0.222756, 0.219136,
                                                    0.223520, 0.225204, 0.220354, 0.216099, 0.229415, 0.220264, 0.222340, 0.222695, 0.220523, 0.219662),
                                                  rep(0, dim(frqFrm)[2]), tolerance = 1e-6)
    expect_equal(df4Chk, distances_omv(dtaInp = binFm2, varDst = names(binFm2), stdDst = "none", nmeDst = "lambda"))
    expect_equal(df4Chk, distances_omv(dtaInp = binFm3, varDst = names(binFm3), stdDst = "none", nmeDst = "lambda_2_1"))
    expect_equal(df4Chk, distances_omv(dtaInp = binFm4, varDst = names(binFm4), stdDst = "none", nmeDst = "lambda_1_2"))

    # distance measure: ochiai, standardization: none
    df4Chk <- distances_omv(dtaInp = binFrm, varDst = names(binFrm), stdDst = "none", nmeDst = "ochiai")
    expect_equal(dim(df4Chk), rep(dim(binFrm)[2], 2))
    expect_equal(names(df4Chk), names(binFrm))
    expect_true(isSymmetric(unname(as.matrix(df4Chk))))
    expect_equal(unname(diag(as.matrix(df4Chk))), rep(1, 20))
    expect_equal(unname(apply(df4Chk, 2, mean)) - c(0.529476, 0.540947, 0.578398, 0.547543, 0.504044, 0.477983, 0.516332, 0.482130, 0.480698, 0.520888,
                                                    0.559257, 0.498333, 0.474961, 0.543470, 0.552167, 0.508030, 0.471819, 0.561378, 0.532916, 0.500511),
                                                  rep(0, dim(frqFrm)[2]), tolerance = 1e-6)
    expect_equal(unname(apply(df4Chk, 2, sd))   - c(0.133476, 0.132665, 0.112496, 0.134183, 0.143866, 0.150557, 0.143314, 0.146892, 0.143157, 0.147125,
                                                    0.130728, 0.144729, 0.142143, 0.134318, 0.138748, 0.139263, 0.141228, 0.133375, 0.133931, 0.131141),
                                                  rep(0, dim(frqFrm)[2]), tolerance = 1e-6)
    expect_equal(df4Chk, distances_omv(dtaInp = binFm2, varDst = names(binFm2), stdDst = "none", nmeDst = "ochiai"))
    expect_equal(df4Chk, distances_omv(dtaInp = binFm3, varDst = names(binFm3), stdDst = "none", nmeDst = "ochiai_2_1"))
    expect_equal(df4Chk, distances_omv(dtaInp = binFm4, varDst = names(binFm4), stdDst = "none", nmeDst = "ochiai_1_2"))

    # distance measure: pattern, standardization: none
    df4Chk <- distances_omv(dtaInp = binFrm, varDst = names(binFrm), stdDst = "none", nmeDst = "pattern")
    expect_equal(dim(df4Chk), rep(dim(binFrm)[2], 2))
    expect_equal(names(df4Chk), names(binFrm))
    expect_true(isSymmetric(unname(as.matrix(df4Chk))))
    expect_equal(unname(diag(as.matrix(df4Chk))), rep(0, 20))
    expect_equal(unname(apply(df4Chk, 2, mean)) - c(0.056840, 0.062640, 0.051000, 0.057800, 0.061240, 0.061920, 0.058260, 0.052060, 0.064500, 0.059520,
                                                    0.056680, 0.057600, 0.056860, 0.053680, 0.058420, 0.054740, 0.057080, 0.054580, 0.052060, 0.061520),
                                                  rep(0, dim(frqFrm)[2]), tolerance = 1e-6)
    expect_equal(unname(apply(df4Chk, 2, sd))   - c(0.021072, 0.023033, 0.016784, 0.021758, 0.025124, 0.022724, 0.023894, 0.021234, 0.024409, 0.025806,
                                                    0.023233, 0.022714, 0.020641, 0.021306, 0.024060, 0.020606, 0.019603, 0.022429, 0.021230, 0.019601),
                                                  rep(0, dim(frqFrm)[2]), tolerance = 1e-6)
    expect_equal(df4Chk, distances_omv(dtaInp = binFm2, varDst = names(binFm2), stdDst = "none", nmeDst = "pattern"))
    expect_equal(df4Chk, distances_omv(dtaInp = binFm3, varDst = names(binFm3), stdDst = "none", nmeDst = "pattern_2_1"))
    expect_equal(df4Chk, distances_omv(dtaInp = binFm4, varDst = names(binFm4), stdDst = "none", nmeDst = "pattern_1_2"))

    # distance measure: phi, standardization: none
    df4Chk <- distances_omv(dtaInp = binFrm, varDst = names(binFrm), stdDst = "none", nmeDst = "phi")
    expect_equal(dim(df4Chk), rep(dim(binFrm)[2], 2))
    expect_equal(names(df4Chk), names(binFrm))
    expect_true(isSymmetric(unname(as.matrix(df4Chk))))
    expect_equal(unname(diag(as.matrix(df4Chk))), rep(1, 20))
    expect_equal(unname(apply(df4Chk, 2, mean)) - c(0.069881, 0.006466, 0.103550, 0.056606, 0.037509, 0.024129, 0.061340, 0.085827, 0.008034, 0.053614,
                                                    0.062668, 0.060670, 0.052676, 0.098636, 0.050075, 0.081125, 0.048516, 0.085693, 0.109710, 0.028984),
                                                  rep(0, dim(frqFrm)[2]), tolerance = 1e-6)
    expect_equal(unname(apply(df4Chk, 2, sd))   - c(0.255569, 0.269311, 0.230680, 0.261642, 0.275595, 0.271525, 0.271811, 0.259880, 0.276028, 0.281245,
                                                    0.267533, 0.269014, 0.258417, 0.256921, 0.273895, 0.253377, 0.252156, 0.263321, 0.254730, 0.252888),
                                                  rep(0, dim(frqFrm)[2]), tolerance = 1e-6)
    expect_equal(df4Chk, distances_omv(dtaInp = binFm2, varDst = names(binFm2), stdDst = "none", nmeDst = "phi"))
    expect_equal(df4Chk, distances_omv(dtaInp = binFm3, varDst = names(binFm3), stdDst = "none", nmeDst = "phi_2_1"))
    expect_equal(df4Chk, distances_omv(dtaInp = binFm4, varDst = names(binFm4), stdDst = "none", nmeDst = "phi_1_2"))

    # distance measure: q, standardization: none
    df4Chk <- distances_omv(dtaInp = binFrm, varDst = names(binFrm), stdDst = "none", nmeDst = "q")
    expect_equal(dim(df4Chk), rep(dim(binFrm)[2], 2))
    expect_equal(names(df4Chk), names(binFrm))
    expect_true(isSymmetric(unname(as.matrix(df4Chk))))
    expect_equal(unname(diag(as.matrix(df4Chk))), rep(1, 20))
    expect_equal(unname(apply(df4Chk, 2, mean)) - c(0.089590, -0.032455, 0.158563, 0.063247, 0.029430, -0.003108, 0.072154, 0.121641, -0.029004, 0.059150,
                                                    0.076419,  0.067936, 0.054189, 0.141785, 0.050632,  0.110989, 0.047092, 0.118760,  0.165151, 0.007468),
                                                  rep(0, dim(frqFrm)[2]), tolerance = 1e-6)
    expect_equal(unname(apply(df4Chk, 2, sd))   - c(0.335152,  0.355854, 0.271792, 0.347915, 0.377516,  0.368495, 0.370935, 0.356046,  0.370574, 0.392907,
                                                    0.364581,  0.366533, 0.341529, 0.341989, 0.370806,  0.331959, 0.320924, 0.360723,  0.342162, 0.316062),
                                                  rep(0, dim(frqFrm)[2]), tolerance = 1e-6)
    expect_equal(df4Chk, distances_omv(dtaInp = binFm2, varDst = names(binFm2), stdDst = "none", nmeDst = "q"))
    expect_equal(df4Chk, distances_omv(dtaInp = binFm3, varDst = names(binFm3), stdDst = "none", nmeDst = "q_2_1"))
    expect_equal(df4Chk, distances_omv(dtaInp = binFm4, varDst = names(binFm4), stdDst = "none", nmeDst = "q_1_2"))

    # distance measure: rr, standardization: none
    df4Chk <- distances_omv(dtaInp = binFrm, varDst = names(binFrm), stdDst = "none", nmeDst = "rr")
    expect_equal(dim(df4Chk), rep(dim(binFrm)[2], 2))
    expect_equal(names(df4Chk), names(binFrm))
    expect_equal(unname(diag(as.matrix(df4Chk))), c(0.500000, 0.600000, 0.580000, 0.560000, 0.480000, 0.440000, 0.480000, 0.380000, 0.460000, 0.500000,
                                                    0.580000, 0.440000, 0.400000, 0.500000, 0.580000, 0.440000, 0.400000, 0.560000, 0.460000, 0.480000))
    expect_equal(unname(apply(df4Chk, 2, mean)) - c(0.263000, 0.296000, 0.310000, 0.289000, 0.245000, 0.222000, 0.251000, 0.207000, 0.228000, 0.259000,
                                                    0.300000, 0.231000, 0.209000, 0.270000, 0.297000, 0.236000, 0.208000, 0.296000, 0.253000, 0.243000),
                                                  rep(0, dim(frqFrm)[2]), tolerance = 1e-6)
    expect_equal(unname(apply(df4Chk, 2, sd))   - c(0.072627, 0.090286, 0.076365, 0.083974, 0.074516, 0.068947, 0.073549, 0.058138, 0.067870, 0.079598,
                                                    0.084355, 0.065687, 0.056373, 0.074410, 0.090443, 0.065727, 0.057455, 0.083754, 0.065943, 0.066578),
                                                  rep(0, dim(frqFrm)[2]), tolerance = 1e-6)
    expect_equal(df4Chk, distances_omv(dtaInp = binFm2, varDst = names(binFm2), stdDst = "none", nmeDst = "rr"))
    expect_equal(df4Chk, distances_omv(dtaInp = binFm3, varDst = names(binFm3), stdDst = "none", nmeDst = "rr_2_1"))
    expect_equal(df4Chk, distances_omv(dtaInp = binFm4, varDst = names(binFm4), stdDst = "none", nmeDst = "rr_1_2"))

    # distance measure: rt, standardization: none
    df4Chk <- distances_omv(dtaInp = binFrm, varDst = names(binFrm), stdDst = "none", nmeDst = "rt")
    expect_equal(dim(df4Chk), rep(dim(binFrm)[2], 2))
    expect_equal(names(df4Chk), names(binFrm))
    expect_true(isSymmetric(unname(as.matrix(df4Chk))))
    expect_equal(unname(diag(as.matrix(df4Chk))), rep(1, 20))
    expect_equal(unname(apply(df4Chk, 2, mean)) - c(0.378029, 0.348230, 0.389600, 0.371211, 0.364678, 0.359001, 0.375532, 0.385861, 0.352103, 0.372481,
                                                    0.373400, 0.375518, 0.371140, 0.391471, 0.368649, 0.383495, 0.368652, 0.384443, 0.397012, 0.359507),
                                                  rep(0, dim(frqFrm)[2]), tolerance = 1e-6)
    expect_equal(unname(apply(df4Chk, 2, sd))   - c(0.157430, 0.164380, 0.149532, 0.160539, 0.164160, 0.164271, 0.163041, 0.158383, 0.165378, 0.165616,
                                                    0.161833, 0.163321, 0.160556, 0.158189, 0.165618, 0.156568, 0.157813, 0.160687, 0.156697, 0.158442),
                                                  rep(0, dim(frqFrm)[2]), tolerance = 1e-6)
    expect_equal(df4Chk, distances_omv(dtaInp = binFm2, varDst = names(binFm2), stdDst = "none", nmeDst = "rt"))
    expect_equal(df4Chk, distances_omv(dtaInp = binFm3, varDst = names(binFm3), stdDst = "none", nmeDst = "rt_2_1"))
    expect_equal(df4Chk, distances_omv(dtaInp = binFm4, varDst = names(binFm4), stdDst = "none", nmeDst = "rt_1_2"))

    # distance measure: size, standardization: none
    df4Chk <- distances_omv(dtaInp = binFrm, varDst = names(binFrm), stdDst = "none", nmeDst = "size")
    expect_equal(dim(df4Chk), rep(dim(binFrm)[2], 2))
    expect_equal(names(df4Chk), names(binFrm))
    expect_true(isSymmetric(unname(as.matrix(df4Chk))))
    expect_equal(unname(diag(as.matrix(df4Chk))), rep(0, 20))
    expect_equal(unname(apply(df4Chk, 2, mean)) - c(0.004300, 0.016100, 0.012140, 0.008980, 0.004340, 0.006820, 0.004340, 0.016540, 0.005180, 0.004300,
                                                    0.012140, 0.006820, 0.012500, 0.004300, 0.012140, 0.006820, 0.012500, 0.008980, 0.005180, 0.004340),
                                                  rep(0, dim(frqFrm)[2]), tolerance = 1e-6)
    expect_equal(unname(apply(df4Chk, 2, sd))   - c(0.004173, 0.014547, 0.012014, 0.009555, 0.004605, 0.008452, 0.004605, 0.015917, 0.006262, 0.004173,
                                                    0.012014, 0.008452, 0.013361, 0.004173, 0.012014, 0.008452, 0.013361, 0.009555, 0.006262, 0.004605),
                                                  rep(0, dim(frqFrm)[2]), tolerance = 1e-6)
    expect_equal(df4Chk, distances_omv(dtaInp = binFm2, varDst = names(binFm2), stdDst = "none", nmeDst = "size"))
    expect_equal(df4Chk, distances_omv(dtaInp = binFm3, varDst = names(binFm3), stdDst = "none", nmeDst = "size_2_1"))
    expect_equal(df4Chk, distances_omv(dtaInp = binFm4, varDst = names(binFm4), stdDst = "none", nmeDst = "size_1_2"))

    # distance measure: sm, standardization: none
    df4Chk <- distances_omv(dtaInp = binFrm, varDst = names(binFrm), stdDst = "none", nmeDst = "sm")
    expect_equal(dim(df4Chk), rep(dim(binFrm)[2], 2))
    expect_equal(names(df4Chk), names(binFrm))
    expect_true(isSymmetric(unname(as.matrix(df4Chk))))
    expect_equal(unname(diag(as.matrix(df4Chk))), rep(1, 20))
    expect_equal(unname(apply(df4Chk, 2, mean)) - c(0.535000, 0.501000, 0.549000, 0.527000, 0.519000, 0.513000, 0.531000, 0.543000, 0.505000, 0.527000,
                                                    0.529000, 0.531000, 0.527000, 0.549000, 0.523000, 0.541000, 0.525000, 0.541000, 0.555000, 0.515000),
                                                  rep(0, dim(frqFrm)[2]), tolerance = 1e-6)
    expect_equal(unname(apply(df4Chk, 2, sd))   - c(0.127465, 0.136339, 0.115799, 0.131593, 0.137416, 0.135067, 0.135410, 0.129497, 0.138241, 0.140116,
                                                    0.133846, 0.134786, 0.130630, 0.128223, 0.138758, 0.126070, 0.126470, 0.132264, 0.127300, 0.126470),
                                                  rep(0, dim(frqFrm)[2]), tolerance = 1e-6)
    expect_equal(df4Chk, distances_omv(dtaInp = binFm2, varDst = names(binFm2), stdDst = "none", nmeDst = "sm"))
    expect_equal(df4Chk, distances_omv(dtaInp = binFm3, varDst = names(binFm3), stdDst = "none", nmeDst = "sm_2_1"))
    expect_equal(df4Chk, distances_omv(dtaInp = binFm4, varDst = names(binFm4), stdDst = "none", nmeDst = "sm_1_2"))

    # distance measure: ss1, standardization: none
    df4Chk <- distances_omv(dtaInp = binFrm, varDst = names(binFrm), stdDst = "none", nmeDst = "ss1")
    expect_equal(dim(df4Chk), rep(dim(binFrm)[2], 2))
    expect_equal(names(df4Chk), names(binFrm))
    expect_true(isSymmetric(unname(as.matrix(df4Chk))))
    expect_equal(unname(diag(as.matrix(df4Chk))), rep(1, 20))
    expect_equal(unname(apply(df4Chk, 2, mean)) - c(0.689748, 0.658743, 0.703130, 0.682303, 0.674427, 0.669715, 0.685174, 0.696240, 0.661988, 0.680932,
                                                    0.683675, 0.685360, 0.682522, 0.701551, 0.677754, 0.695067, 0.681365, 0.694201, 0.706625, 0.672704),
                                                  rep(0, dim(frqFrm)[2]), tolerance = 1e-6)
    expect_equal(unname(apply(df4Chk, 2, sd))   - c(0.093731, 0.103262, 0.081069, 0.097985, 0.104996, 0.099912, 0.102170, 0.096130, 0.105630, 0.108121,
                                                    0.100670, 0.100162, 0.095781, 0.093540, 0.105752, 0.091856, 0.091631, 0.098304, 0.093489, 0.090954),
                                                  rep(0, dim(frqFrm)[2]), tolerance = 1e-6)
    expect_equal(df4Chk, distances_omv(dtaInp = binFm2, varDst = names(binFm2), stdDst = "none", nmeDst = "ss1"))
    expect_equal(df4Chk, distances_omv(dtaInp = binFm3, varDst = names(binFm3), stdDst = "none", nmeDst = "ss1_2_1"))
    expect_equal(df4Chk, distances_omv(dtaInp = binFm4, varDst = names(binFm4), stdDst = "none", nmeDst = "ss1_1_2"))

    # distance measure: ss2, standardization: none
    df4Chk <- distances_omv(dtaInp = binFrm, varDst = names(binFrm), stdDst = "none", nmeDst = "ss2")
    expect_equal(dim(df4Chk), rep(dim(binFrm)[2], 2))
    expect_equal(names(df4Chk), names(binFrm))
    expect_true(isSymmetric(unname(as.matrix(df4Chk))))
    expect_equal(unname(diag(as.matrix(df4Chk))), rep(1, 20))
    expect_equal(unname(apply(df4Chk, 2, mean)) - c(0.246189, 0.252788, 0.276838, 0.258670, 0.231113, 0.215139, 0.239207, 0.215393, 0.215954, 0.243143,
                                                    0.265948, 0.227137, 0.210972, 0.256644, 0.262973, 0.232265, 0.208855, 0.268798, 0.248634, 0.226481),
                                                  rep(0, dim(frqFrm)[2]), tolerance = 1e-6)
    expect_equal(unname(apply(df4Chk, 2, sd))   - c(0.183543, 0.183579, 0.174552, 0.182865, 0.188470, 0.191038, 0.187394, 0.190644, 0.189540, 0.188066,
                                                    0.181188, 0.188987, 0.189734, 0.184140, 0.185180, 0.186800, 0.189778, 0.182197, 0.183985, 0.185591),
                                                  rep(0, dim(frqFrm)[2]), tolerance = 1e-6)
    expect_equal(df4Chk, distances_omv(dtaInp = binFm2, varDst = names(binFm2), stdDst = "none", nmeDst = "ss2"))
    expect_equal(df4Chk, distances_omv(dtaInp = binFm3, varDst = names(binFm3), stdDst = "none", nmeDst = "ss2_2_1"))
    expect_equal(df4Chk, distances_omv(dtaInp = binFm4, varDst = names(binFm4), stdDst = "none", nmeDst = "ss2_1_2"))

    # distance measure: ss3, standardization: none
    df4Chk <- distances_omv(dtaInp = binFrm, varDst = names(binFrm), stdDst = "none", nmeDst = "ss3")
    expect_equal(dim(df4Chk), rep(dim(binFrm)[2], 2))
    expect_equal(names(df4Chk), names(binFrm))
    expect_true(isSymmetric(unname(as.matrix(df4Chk))))
    expect_true(all(is.na(diag(as.matrix(df4Chk)))))
    expect_equal(unname(apply(df4Chk, 2, mean, na.rm = TRUE)),
                 c(1.078226, 0.936743, 1.125170, 1.048680, 1.021168, 0.994574, 1.079788, 1.124509, 0.958833, 1.067129,
                   1.063652, 1.081734, 1.049790, 1.159014, 1.048984, 1.107437, 1.029689, 1.122979, 1.185479, 0.982876),
                 tolerance = 1e-6)
    expect_equal(unname(apply(df4Chk, 2, sd,   na.rm = TRUE)),
                 c(0.272941, 0.258805, 0.199197, 0.296731, 0.311652, 0.326131, 0.361464, 0.317131, 0.289661, 0.359135,
                   0.321630, 0.375726, 0.312558, 0.363458, 0.375300, 0.291035, 0.262476, 0.349591, 0.340473, 0.233838),
                 tolerance = 1e-6)
    expect_equal(df4Chk, distances_omv(dtaInp = binFm2, varDst = names(binFm2), stdDst = "none", nmeDst = "ss3"))
    expect_equal(df4Chk, distances_omv(dtaInp = binFm3, varDst = names(binFm3), stdDst = "none", nmeDst = "ss3_2_1"))
    expect_equal(df4Chk, distances_omv(dtaInp = binFm4, varDst = names(binFm4), stdDst = "none", nmeDst = "ss3_1_2"))

    # distance measure: ss4, standardization: none
    df4Chk <- distances_omv(dtaInp = binFrm, varDst = names(binFrm), stdDst = "none", nmeDst = "ss4")
    expect_equal(dim(df4Chk), rep(dim(binFrm)[2], 2))
    expect_equal(names(df4Chk), names(binFrm))
    expect_true(isSymmetric(unname(as.matrix(df4Chk))))
    expect_equal(unname(diag(as.matrix(df4Chk))), rep(1, 20))
    expect_equal(unname(apply(df4Chk, 2, mean)) - c(0.534940, 0.503229, 0.551776, 0.528301, 0.518756, 0.512065, 0.530669, 0.542918, 0.504015, 0.526804,
                                                    0.531335, 0.530336, 0.526337, 0.549320, 0.525038, 0.540563, 0.524257, 0.542848, 0.554857, 0.514490),
                                                  rep(0, dim(frqFrm)[2]), tolerance = 1e-6)
    expect_equal(unname(apply(df4Chk, 2, sd))   - c(0.127788, 0.134661, 0.115341, 0.130823, 0.137800, 0.135765, 0.135909, 0.129949, 0.138015, 0.140627,
                                                    0.133769, 0.134508, 0.129212, 0.128462, 0.136950, 0.126689, 0.126081, 0.131663, 0.127367, 0.126445),
                                                  rep(0, dim(frqFrm)[2]), tolerance = 1e-6)
    expect_equal(df4Chk, distances_omv(dtaInp = binFm2, varDst = names(binFm2), stdDst = "none", nmeDst = "ss4"))
    expect_equal(df4Chk, distances_omv(dtaInp = binFm3, varDst = names(binFm3), stdDst = "none", nmeDst = "ss4_2_1"))
    expect_equal(df4Chk, distances_omv(dtaInp = binFm4, varDst = names(binFm4), stdDst = "none", nmeDst = "ss4_1_2"))

    # distance measure: ss5, standardization: none
    df4Chk <- distances_omv(dtaInp = binFrm, varDst = names(binFrm), stdDst = "none", nmeDst = "ss5")
    expect_equal(dim(df4Chk), rep(dim(binFrm)[2], 2))
    expect_equal(names(df4Chk), names(binFrm))
    expect_true(isSymmetric(unname(as.matrix(df4Chk))))
    expect_equal(unname(diag(as.matrix(df4Chk))), rep(1, 20))
    expect_equal(unname(apply(df4Chk, 2, mean)) - c(0.299469, 0.264204, 0.311828, 0.291720, 0.284751, 0.275816, 0.296759, 0.301993, 0.269360, 0.294081,
                                                    0.294157, 0.294795, 0.286582, 0.315214, 0.288819, 0.303570, 0.283345, 0.307405, 0.320439, 0.277582),
                                                  rep(0, dim(frqFrm)[2]), tolerance = 1e-6)
    expect_equal(unname(apply(df4Chk, 2, sd))   - c(0.177577, 0.183828, 0.168467, 0.180540, 0.184313, 0.185869, 0.183833, 0.180206, 0.185693, 0.186395,
                                                    0.182245, 0.184307, 0.180703, 0.178606, 0.185307, 0.177226, 0.178403, 0.180671, 0.176808, 0.178720),
                                                  rep(0, dim(frqFrm)[2]), tolerance = 1e-6)
    expect_equal(df4Chk, distances_omv(dtaInp = binFm2, varDst = names(binFm2), stdDst = "none", nmeDst = "ss5"))
    expect_equal(df4Chk, distances_omv(dtaInp = binFm3, varDst = names(binFm3), stdDst = "none", nmeDst = "ss5_2_1"))
    expect_equal(df4Chk, distances_omv(dtaInp = binFm4, varDst = names(binFm4), stdDst = "none", nmeDst = "ss5_1_2"))

    # distance measure: variance, standardization: none
    df4Chk <- distances_omv(dtaInp = binFrm, varDst = names(binFrm), stdDst = "none", nmeDst = "variance")
    expect_equal(dim(df4Chk), rep(dim(binFrm)[2], 2))
    expect_equal(names(df4Chk), names(binFrm))
    expect_true(isSymmetric(unname(as.matrix(df4Chk))))
    expect_equal(unname(diag(as.matrix(df4Chk))), rep(0, 20))
    expect_equal(unname(apply(df4Chk, 2, mean)) - c(0.116250, 0.124750, 0.112750, 0.118250, 0.120250, 0.121750, 0.117250, 0.114250, 0.123750, 0.118250,
                                                    0.117750, 0.117250, 0.118250, 0.112750, 0.119250, 0.114750, 0.118750, 0.114750, 0.111250, 0.121250),
                                                  rep(0, dim(frqFrm)[2]), tolerance = 1e-6)
    expect_equal(unname(apply(df4Chk, 2, sd))   - c(0.031866, 0.034085, 0.028950, 0.032898, 0.034354, 0.033767, 0.033852, 0.032374, 0.034560, 0.035029,
                                                    0.033461, 0.033697, 0.032658, 0.032056, 0.034689, 0.031518, 0.031618, 0.033066, 0.031825, 0.031618),
                                                  rep(0, dim(frqFrm)[2]), tolerance = 1e-6)
    expect_equal(df4Chk, distances_omv(dtaInp = binFm2, varDst = names(binFm2), stdDst = "none", nmeDst = "variance"))
    expect_equal(df4Chk, distances_omv(dtaInp = binFm3, varDst = names(binFm3), stdDst = "none", nmeDst = "variance_2_1"))
    expect_equal(df4Chk, distances_omv(dtaInp = binFm4, varDst = names(binFm4), stdDst = "none", nmeDst = "variance_1_2"))

    # distance measure: y, standardization: none
    df4Chk <- distances_omv(dtaInp = binFrm, varDst = names(binFrm), stdDst = "none", nmeDst = "y")
    expect_equal(dim(df4Chk), rep(dim(binFrm)[2], 2))
    expect_equal(names(df4Chk), names(binFrm))
    expect_true(isSymmetric(unname(as.matrix(df4Chk))))
    expect_equal(unname(diag(as.matrix(df4Chk))), rep(1, 20))
    expect_equal(unname(apply(df4Chk, 2, mean)) - c(0.069727, 0.005456, 0.105310, 0.056502, 0.037736, 0.023116, 0.061273, 0.087993, 0.007216, 0.053119,
                                                    0.063911, 0.060523, 0.052541, 0.099385, 0.050221, 0.081807, 0.048464, 0.086953, 0.110664, 0.028590),
                                                  rep(0, dim(frqFrm)[2]), tolerance = 1e-6)
    expect_equal(unname(apply(df4Chk, 2, sd))   - c(0.256289, 0.271226, 0.231255, 0.262755, 0.276392, 0.273275, 0.272829, 0.263103, 0.276997, 0.282468,
                                                    0.269575, 0.270260, 0.260290, 0.257506, 0.275180, 0.254593, 0.253612, 0.264899, 0.255524, 0.253307),
                                                  rep(0, dim(frqFrm)[2]), tolerance = 1e-6)
    expect_equal(df4Chk, distances_omv(dtaInp = binFm2, varDst = names(binFm2), stdDst = "none", nmeDst = "y"))
    expect_equal(df4Chk, distances_omv(dtaInp = binFm3, varDst = names(binFm3), stdDst = "none", nmeDst = "y_2_1"))
    expect_equal(df4Chk, distances_omv(dtaInp = binFm4, varDst = names(binFm4), stdDst = "none", nmeDst = "y_1_2"))

    # check mkeBin
    mt4Chk <- mkeBin(as.matrix(binFm5), 1, 2)
    expect_equal(dim(mt4Chk), dim(binFm5))
    expect_equal(unname(apply(mt4Chk, 2, class)), rep("logical", dim(binFm5)[2]))
    expect_equal(unname(apply(mt4Chk, 2, table)), unname(apply(binFm5, 2, table)[c(2, 1), ]))
    expect_equal(mt4Chk, mkeBin(as.matrix(as.data.frame(lapply(binFm5, as.factor))), 1, 2))
    expect_equal(binFrm, as.data.frame(mkeBin(apply(apply(as.matrix(binFrm), 2, as.integer), 2, as.factor))))
    expect_equal(binFrm, as.data.frame(mkeBin(apply(apply(as.matrix(binFrm), 2, as.integer), 2, as.factor), 1, 0)))
    expect_equal(binFrm, as.data.frame(mkeBin(as.matrix(as.data.frame(lapply(binFrm, function(c) as.character(as.integer(!c) + 1)))), 1, 2)))

    # test cases for error messages ===================================================================================
    expect_error(distances_omv(fleInp = nmeInp, fleOut = nmeOut, varDst = names(cntFrm)),
      regexp = "^Please use the argument dtaInp instead of fleInp\\.")
    expect_error(distances_omv(dtaInp = cntFrm, fleOut = nmeOut, varDst = c("A1", "A2", "A3")),
      regexp = "^Calling distances_omv requires giving at least two \\(valid\\) variables to calculate distances between\\.")
    expect_error(distances_omv(dtaInp = cntFrm, fleOut = nmeOut, varDst = names(cntFrm)[1]),
      regexp = "^Calling distances_omv requires giving at least two \\(valid\\) variables to calculate distances between\\.")
    expect_error(distances_omv(dtaInp = cntFrm, fleOut = nmeOut, varDst = names(cntFrm), stdDst = "wrong"),
      regexp = "^Invalid standardization: \\w+\\. See Details in the help for further information\\.")
    expect_error(distances_omv(dtaInp = cntFrm, fleOut = nmeOut, varDst = names(cntFrm), nmeDst = "wrong"),
      regexp = "^Invalid distance measure: \\w+\\. See Details in the help for further information\\.")
    expect_error(clcFrq(as.matrix(frqFrm), "wrong"), regexp = "clcFrq: Method wrong is not implemented\\.")
    expect_error(mkeBin(apply(as.matrix(binFrm), 2, as.complex), "TRUE", "FALSE"),
      regexp = paste("The input matrix for binary data either needs to be logical \\(then it will be kept as it is\\),",
                     "numeric or character \\(for the latter two, p and np are used to derive TRUE and FALSE\\)\\."))
    expect_error(mtcBin(binFm3[, 1], binFm3[, 1], "beuclid"),
      regexp = "mtcBin: Input columns to the calculation of binary measures must be logical\\.")
    expect_error(mtcBin(binFrm[, 1], binFrm[, 2], "wrong"), regexp = "mtcBin: Method wrong is not implemented\\.")
})

---
title: "R package `jmvReadWrite`"
author: "Sebastian Jentschke"
date: "`r Sys.Date()`"
vignette: >
  %\VignetteIndexEntry{R package `jmvReadWrite`}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
output:
  rmarkdown::html_vignette:
    keep_md: TRUE  
---

```{r, include = FALSE}
knitr::opts_chunk$set(echo = TRUE,
                      comment = "#>",
                      collapse = TRUE,
                      message = FALSE,
                      warning = FALSE,
                      fig.height = 5,
                      fig.width = 7,
                      fig.align = "center",
                      out.width = "100%")
```


The R-package `jmvReadWrite` reads and writes the .omv-files that are used by
the statistical spreadsheet `jamovi` (<https://www.jamovi.org>). It is supposed
to ease using jamovi and R together, provide helper functions for some often 
required data management tasks, and to adjust and use syntax for statistical
analyses that were created using the GUI in `jamovi` in R (in connection with
the R-library `jmv`). More recently, `jmvReadWrite` became easily available
from within `jamovi` by becoming part of the `Rj` module (where you can use it
by writing R commands, documented below), and via the `jTransform` module that
provides a graphical user interface for most helper functions.

## Installation

In R, you can either install a stable version of `jmvReadWrite` which is
available on [CRAN](https://cran.r-project.org/package=jmvReadWrite) using the
following command:

```{r, eval=FALSE}
install.packages("jmvReadWrite")
```

or you can install the development version of the `jmvReadWrite` package from
[GitHub](https://github.com/sjentsch/jmvReadWrite/):

```{r, eval=FALSE}
if (!require(devtools)) install.packages("devtools")
devtools::install_github("sjentsch/jmvReadWrite")
```

 
## How to use the package?

[**read_omv**](https://sjentsch.github.io/jmvReadWrite/reference/read_omv.html)

The following code uses the ToothGrowth-data set that is part of the data sets
included in R (the current file contains some modifications though for testing
the reading and writing routines: `read_omv` and `write_omv`). With this data
set, a syntax to conduct an ANOVA is run.

The results should be similar to those obtained when running the same analysis
in jamovi (using the GUI). To do so, open the file menu (☰) choose `Open`,
`Data Library` and `ToothGrowth`. Afterwards, click on the `ANOVA`-button in
the `Analyses`-tab and choose `ANOVA`. There, you assign the variable `len` to
`Dependent Variable` and `supp` and `dose` to `Fixed Factors`. Afterwards, you
choose / tick `Overall Model Test` and `ω²`. Open the drop-down menu
`Assumption Checks` and tick `Homogeneity test` and `Normality test`.
The results should be identical apart from that the table output looks nicer in
`jamovi` (not only text, as below), numbers are rounded and maybe one or two
other cosmetic differences.

If you want to copy the syntax generated in jamovi, you have to switch on the
[`Syntax Mode`](https://jamovi.readthedocs.io/en/latest/usermanual/um_6_jamovi_and_R/#syntax-mode).
Afterwards, the syntax is shown at the top of the analysis and can be copied
from there.

```{r, echo=TRUE}
fleOMV <- system.file("extdata", "ToothGrowth.omv", package = "jmvReadWrite")
data <- jmvReadWrite::read_omv(fleOMV)
# if the "jmv"-package is installed, we can run a test analysis with the data
if (nzchar(system.file(package = "jmv"))) {
    jmv::ANOVA(
        formula = len ~ supp + dose + supp:dose,
        data = data,
        effectSize = "omega",
        modelTest = TRUE,
        homo = TRUE,
        norm = TRUE)
    }
```

Since version 0.2.0,
[`read_omv`](https://sjentsch.github.io/jmvReadWrite/reference/read_omv.html)
also extracts the syntax from analyses that you may have conducted in the
jamovi-GUI and that are stored in the .omv-file. To extract them, you have to
set the parameter `getSyn = TRUE` when calling
[`read_omv`](https://sjentsch.github.io/jmvReadWrite/reference/read_omv.html)
(default is `FALSE`). When the parameter is set, the analyses are stored in the
attribute `syntax`. They can be used as shown in the following examples:

```{r, echo=TRUE}
fleOMV <- system.file("extdata", "ToothGrowth.omv", package = "jmvReadWrite")
data <- jmvReadWrite::read_omv(fleOMV, getSyn = TRUE)
# shows the syntax of the analyses from the .omv-file
# please note that syntax extraction may not work on all systems
# if the syntax couldn't be extracted, an empty list (length = 0) is returned,
# otherwise, the syntax of the analyses from the .omv-file is shown and
# the commands of the first and the second analysis are run, with the
# output of the second analysis assigned to the variable result
if (length(attr(data, "syntax")) >= 2) {
    attr(data, "syntax")
    # if the "jmv"-package is installed, we can run the analyses in "syntax"
    if (nzchar(system.file(package = "jmv"))) {
        eval(parse(text = attr(data, "syntax")[[1]]))
        eval(parse(text = paste0("result = ", attr(data, "syntax")[[2]])))
        names(result)
        # -> "main"      "assump"    "contrasts" "postHoc"   "emm"
        # (the names of the five output tables)
    }
}
```
<br/>

[**write_omv**](https://sjentsch.github.io/jmvReadWrite/reference/write_omv.html)

The `jmvReadWrite`-package also enables you to write `.omv`-files in order to
use them in `jamovi`. Let's assume that you have a large collection of
log-files (e.g., from an experiment) that you compile and process (summarize,
filter, etc.) in R in order to later analyse them in `jamovi`. You will have
those processed log-files stored in a data frame (called, e.g., `data`) which
you then write to a file that you can open in jamovi afterwards.

```{r, echo=TRUE}
# use the data set "ToothGrowth" and, if it exists, write it as jamovi-file
# using write_omv()
data("ToothGrowth", package = "jmvReadWrite")
# "retDbg" has to be set in order to return debug information to wrtDta
wrtDta <- jmvReadWrite::write_omv(ToothGrowth, "Trial.omv", retDbg = TRUE)
names(wrtDta)
# -> "mtaDta" "xtdDta" "dtaFrm"
# this debug information contains a list with the metadata ("mtaDta", e.g.,
# column and data type), the extended data ("xtdDta", e.g., variable lables),
# and the data frame (dtaFrm) for checking (understanding the file format) and
# debugging

# check whether the file was written to the disk, get the file information (size, etc.)
# and delete the file afterwards
list.files(".", "Trial.omv")
file.info("Trial.omv")[, c("size", "isdir", "mode")]
unlink("Trial.omv")
```

Although jamovi reads R-data files (.RData, .rda, .rds)
[`write_omv`](https://sjentsch.github.io/jmvReadWrite/reference/write_omv.html)
permits to store `jamovi`-specific attributes (such as variable labels) in
addition. Please note that if you are reading from an `.omv`-file in order to
write back to an `.omv`-file (perhaps after some modifications), it is
recommended to leave the `sveAtt`-attribute set to `TRUE` (which is the
default).

```{r, echo=TRUE}
# reading and writing a file with the "sveAtt"-parameter permits you to keep
# essential meta-data to ensure that the written file looks and works like the
# original file (plus you modifications)
fleOMV <- system.file("extdata", "ToothGrowth.omv", package = "jmvReadWrite")
data <- jmvReadWrite::read_omv(fleOMV, sveAtt = TRUE)
# shows the names of the attributes for the whole data set (e.g., number of
# rows and columns) and the names of the attributes of the first column
names(attributes(data))
names(attributes(data[[1]]))
#
# perhaps do some modifications to the file here and write it back afterwards
jmvReadWrite::write_omv(data, "Trial.omv")
unlink("Trial.omv")
```

If `Trial.omv` in the example above would have been kept, it should look like
the original file (plus your possible modifications). If you, e.g., added a new
column, you could adjust some attributes (e.g., to enforce a specific
`columnType` or `measurementType`): just look at how attributes are stored for
other columns.
<br /><br />

**Helper functions**

`jmvReadWrite` contains a number of helper functions that assist you with data
management tasks that are frequently required:

* [`aggregate_omv`](https://sjentsch.github.io/jmvReadWrite/reference/aggregate_omv.html):
  Aggregates data (calculating, e.g., mean, median, etc.) from data sets in long
  format.

* [`arrange_cols_omv`](https://sjentsch.github.io/jmvReadWrite/reference/arrange_cols_omv.html):
  Re-arranges the columns of your data file in a requested order.

* [`combine_cols_omv`](https://sjentsch.github.io/jmvReadWrite/reference/combine_cols_omv.html):
  Combines the data from two columns into one. The function assumes that the data contained in
  the two columns are the same. It checks for (and thereby ensures) equality and replaces missing
  values in one column by replacing those with values from the second column (if those are not
  missing as well).

* [`convert_to_omv`](https://sjentsch.github.io/jmvReadWrite/reference/convert_to_omv.html):
  Converts data sets from other file formats into jamovi-format. This function
  may be helpful if you have to convert a larger amount of files.

* [`describe_omv`](https://sjentsch.github.io/jmvReadWrite/reference/describe_omv.html):
  Adds a title and a description to a data set. This function may be helpful
  for documenting what is contained in a data set, e.g. for publishing them in
  a repository such as OSF, or for generated data sets, e.g. those used in
  teaching.

* [`distances_omv`](https://sjentsch.github.io/jmvReadWrite/reference/distances_omv.html):
  Calculates a wide range of distances measures (for continuous, frequency or binary data).
  If can be determined, whether the calculation of the distances should be carried out
  between columns / variables or between rows / units of observation. The original data can
  be standardized before the distances are calculated.

* [`label_vars_omv`](https://sjentsch.github.io/jmvReadWrite/reference/label_vars_omv.html):
  Labels columns / variables in jamovi data files, using either a CSV file, a data frame (each
  containing two columns, one with the variable names, the other with the variable labels), or
  a character vector with the variable labels.

* [`long2wide_omv`](https://sjentsch.github.io/jmvReadWrite/reference/long2wide_omv.html):
  Converts a data set from long to wide format: Time points for repeated
  measurements are arranged as rows in the original and converted into
  columns.

* [`wide2long_omv`](https://sjentsch.github.io/jmvReadWrite/reference/wide2long_omv.html):
  Converts a data set from wide to long format: Time points for repeated
  measurements are arranged as columns in the original and converted into
  rows.

* [`merge_cols_omv`](https://sjentsch.github.io/jmvReadWrite/reference/merge_cols_omv.html):
  Add variables from several data sets, i.e. the variables / columns in the
  second, etc. input data set are added as columns to the first data set.

* [`merge_rows_omv`](https://sjentsch.github.io/jmvReadWrite/reference/merge_rows_omv.html):
  Add cases from several data sets, i.e. the cases / rows in the second, etc.
  data set are added as rows to the first data set.

* [`sort_omv`](https://sjentsch.github.io/jmvReadWrite/reference/sort_omv.html):
  Sort a data set according to one or more variable(s).

* [`transform_vars_omv`](https://sjentsch.github.io/jmvReadWrite/reference/transform_vars_omv.html):
  Transform skewed variables (aiming at they better conform to a normal distribution).

* [`transpose_omv`](https://sjentsch.github.io/jmvReadWrite/reference/transpose_omv.html):
  Transpose a data set: Make rows into columns and vice versa.

---
title: "R package `jmvReadWrite`"
author: "Sebastian Jentschke"
date: "`r Sys.Date()`"
vignette: >
  %\VignetteIndexEntry{R package `jmvReadWrite`}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
output:
  rmarkdown::html_vignette:
    keep_md: TRUE  
---

```{r, include = FALSE}
knitr::opts_chunk$set(echo = TRUE,
                      comment = "#>",
                      collapse = TRUE,
                      message = FALSE,
                      warning = FALSE,
                      fig.height = 5,
                      fig.width = 7,
                      fig.align = "center",
                      out.width = "100%")
```


The R-package `jmvReadWrite` reads and writes the .omv-files that are used by
the statistical spreadsheet `jamovi` (<https://www.jamovi.org>). It is supposed
to ease using jamovi and R together, provide helper functions for some often 
required data management tasks, and to adjust and use syntax for statistical
analyses that were created using the GUI in `jamovi` in R (in connection with
the R-library `jmv`). More recently, `jmvReadWrite` became easily available
from within `jamovi` by becoming part of the `Rj` module (where you can use it
by writing R commands, documented below), and via the `jTransform` module that
provides a graphical user interface for most helper functions.

## Installation

In R, you can either install a stable version of `jmvReadWrite` which is
available on [CRAN](https://cran.r-project.org/package=jmvReadWrite) using the
following command:

```{r, eval=FALSE}
install.packages("jmvReadWrite")
```

or you can install the development version of the `jmvReadWrite` package from
[GitHub](https://github.com/sjentsch/jmvReadWrite/):

```{r, eval=FALSE}
if (!require(devtools)) install.packages("devtools")
devtools::install_github("sjentsch/jmvReadWrite")
```

 
## How to use the package?

[**read_omv**](https://sjentsch.github.io/jmvReadWrite/reference/read_omv.html)

The following code uses the ToothGrowth-data set that is part of the data sets
included in R (the current file contains some modifications though for testing
the reading and writing routines: `read_omv` and `write_omv`). With this data
set, a syntax to conduct an ANOVA is run.

The results should be similar to those obtained when running the same analysis
in jamovi (using the GUI). To do so, open the file menu (☰) choose `Open`,
`Data Library` and `ToothGrowth`. Afterwards, click on the `ANOVA`-button in
the `Analyses`-tab and choose `ANOVA`. There, you assign the variable `len` to
`Dependent Variable` and `supp` and `dose` to `Fixed Factors`. Afterwards, you
choose / tick `Overall Model Test` and `ω²`. Open the drop-down menu
`Assumption Checks` and tick `Homogeneity test` and `Normality test`.
The results should be identical apart from that the table output looks nicer in
`jamovi` (not only text, as below), numbers are rounded and maybe one or two
other cosmetic differences.

If you want to copy the syntax generated in jamovi, you have to switch on the
[`Syntax Mode`](https://jamovi.readthedocs.io/en/latest/usermanual/um_6_jamovi_and_R/#syntax-mode).
Afterwards, the syntax is shown at the top of the analysis and can be copied
from there.

```{r, echo=TRUE}
fleOMV <- system.file("extdata", "ToothGrowth.omv", package = "jmvReadWrite")
data <- jmvReadWrite::read_omv(fleOMV)
# if the "jmv"-package is installed, we can run a test analysis with the data
if ("jmv" %in% rownames(installed.packages())) {
    jmv::ANOVA(
        formula = len ~ supp + dose + supp:dose,
        data = data,
        effectSize = c("omega"),
        modelTest = TRUE,
        homo = TRUE,
        norm = TRUE)
    }
```

Since version 0.2.0,
[`read_omv`](https://sjentsch.github.io/jmvReadWrite/reference/read_omv.html)
also extracts the syntax from analyses that you may have conducted in the
jamovi-GUI and that are stored in the .omv-file. To extract them, you have to
set the parameter `getSyn = TRUE` when calling
[`read_omv`](https://sjentsch.github.io/jmvReadWrite/reference/read_omv.html)
(default is `FALSE`). When the parameter is set, the analyses are stored in the
attribute `syntax`. They can be used as shown in the following examples:

```{r, echo=TRUE}
fleOMV <- system.file("extdata", "ToothGrowth.omv", package = "jmvReadWrite")
data <- jmvReadWrite::read_omv(fleOMV, getSyn = TRUE)
# shows the syntax of the analyses from the .omv-file
# please note that syntax extraction may not work on all systems
# if the syntax couldn't be extracted, an empty list (length = 0) is returned,
# otherwise, the syntax of the analyses from the .omv-file is shown and
# the commands of the first and the second analysis are run, with the
# output of the second analysis assigned to the variable result
if (length(attr(data, "syntax")) >= 2) {
    attr(data, "syntax")
    # if the "jmv"-package is installed, we can run the analyses in "syntax"
    if ("jmv" %in% rownames(installed.packages())) {
        eval(parse(text = attr(data, "syntax")[[1]]))
        eval(parse(text = paste0("result = ", attr(data, "syntax")[[2]])))
        names(result)
        # -> "main"      "assump"    "contrasts" "postHoc"   "emm"
        # (the names of the five output tables)
    }
}
```
<br/>

[**write_omv**](https://sjentsch.github.io/jmvReadWrite/reference/write_omv.html)

The `jmvReadWrite`-package also enables you to write `.omv`-files in order to
use them in `jamovi`. Let's assume that you have a large collection of
log-files (e.g., from an experiment) that you compile and process (summarize,
filter, etc.) in R in order to later analyse them in `jamovi`. You will have
those processed log-files stored in a data frame (called, e.g., `data`) which
you then write to a file that you can open in jamovi afterwards.

```{r, echo=TRUE}
# use the data set "ToothGrowth" and, if it exists, write it as jamovi-file
# using write_omv()
data("ToothGrowth", package = "jmvReadWrite")
# "retDbg" has to be set in order to return debug information to wrtDta
wrtDta <- jmvReadWrite::write_omv(ToothGrowth, "Trial.omv", retDbg = TRUE)
names(wrtDta)
# -> "mtaDta" "xtdDta" "dtaFrm"
# this debug information contains a list with the metadata ("mtaDta", e.g.,
# column and data type), the extended data ("xtdDta", e.g., variable lables),
# and the data frame (dtaFrm) for checking (understanding the file format) and
# debugging

# check whether the file was written to the disk, get the file information (size, etc.)
# and delete the file afterwards
list.files(".", "Trial.omv")
file.info("Trial.omv")[, c("size", "isdir", "mode")]
unlink("Trial.omv")
```

Although jamovi reads R-data files (.RData, .rda, .rds)
[`write_omv`](https://sjentsch.github.io/jmvReadWrite/reference/write_omv.html)
permits to store `jamovi`-specific attributes (such as variable labels) in
addition. Please note that if you are reading from an `.omv`-file in order to
write back to an `.omv`-file (perhaps after some modifications), it is
recommended to leave the `sveAtt`-attribute set to `TRUE` (which is the
default).

```{r, echo=TRUE}
# reading and writing a file with the "sveAtt"-parameter permits you to keep
# essential meta-data to ensure that the written file looks and works like the
# original file (plus you modifications)
fleOMV <- system.file("extdata", "ToothGrowth.omv", package = "jmvReadWrite")
data <- jmvReadWrite::read_omv(fleOMV, sveAtt = TRUE)
# shows the names of the attributes for the whole data set (e.g., number of
# rows and columns) and the names of the attributes of the first column
names(attributes(data))
names(attributes(data[[1]]))
#
# perhaps do some modifications to the file here and write it back afterwards
jmvReadWrite::write_omv(data, "Trial.omv")
unlink("Trial.omv")
```

If `Trial.omv` in the example above would have been kept, it should look like
the original file (plus your possible modifications). If you, e.g., added a new
column, you could adjust some attributes (e.g., to enforce a specific
`columnType` or `measurementType`): just look at how attributes are stored for
other columns.
<br /><br />

**Helper functions**

`jmvReadWrite` contains a number of helper functions that assist you with data
management tasks that are frequently required:

* [`aggregate_omv`](https://sjentsch.github.io/jmvReadWrite/reference/aggregate_omv.html):
  Aggregates data (calculating, e.g., mean, median, etc.) from data sets in long
  format.

* [`arrange_cols_omv`](https://sjentsch.github.io/jmvReadWrite/reference/arrange_cols_omv.html):
  Re-arranges the columns of your data file in a requested order.

* [`combine_cols_omv`](https://sjentsch.github.io/jmvReadWrite/reference/combine_cols_omv.html):
  Combines the data from two columns into one. The function assumes that the data contained in
  the two columns are the same. It checks for (and thereby ensures) equality and replaces missing
  values in one column by replacing those with values from the second column (if those are not
  missing as well).

* [`convert_to_omv`](https://sjentsch.github.io/jmvReadWrite/reference/convert_to_omv.html):
  Converts data sets from other file formats into jamovi-format. This function
  may be helpful if you have to convert a larger amount of files.

* [`describe_omv`](https://sjentsch.github.io/jmvReadWrite/reference/describe_omv.html):
  Adds a title and a description to a data set. This function may be helpful
  for documenting what is contained in a data set, e.g. for publishing them in
  a repository such as OSF, or for generated data sets, e.g. those used in
  teaching.

* [`distances_omv`](https://sjentsch.github.io/jmvReadWrite/reference/distances_omv.html):
  Calculates a wide range of distances measures (for continuous, frequency or binary data).
  If can be determined, whether the calculation of the distances should be carried out
  between columns / variables or between rows / units of observation. The original data can
  be standardized before the distances are calculated.

* [`label_vars_omv`](https://sjentsch.github.io/jmvReadWrite/reference/label_vars_omv.html):
  Labels columns / variables in jamovi data files, using either a CSV file, a data frame (each
  containing two columns, one with the variable names, the other with the variable labels), or
  a character vector with the variable labels.

* [`long2wide_omv`](https://sjentsch.github.io/jmvReadWrite/reference/long2wide_omv.html):
  Converts a data set from long to wide format: Time points for repeated
  measurements are arranged as rows in the original and converted into
  columns.

* [`wide2long_omv`](https://sjentsch.github.io/jmvReadWrite/reference/wide2long_omv.html):
  Converts a data set from wide to long format: Time points for repeated
  measurements are arranged as columns in the original and converted into
  rows.

* [`merge_cols_omv`](https://sjentsch.github.io/jmvReadWrite/reference/merge_cols_omv.html):
  Add variables from several data sets, i.e. the variables / columns in the
  second, etc. input data set are added as columns to the first data set.

* [`merge_rows_omv`](https://sjentsch.github.io/jmvReadWrite/reference/merge_rows_omv.html):
  Add cases from several data sets, i.e. the cases / rows in the second, etc.
  data set are added as rows to the first data set.

* [`sort_omv`](https://sjentsch.github.io/jmvReadWrite/reference/sort_omv.html):
  Sort a data set according to one or more variable(s).

* [`transform_vars_omv`](https://sjentsch.github.io/jmvReadWrite/reference/transform_vars_omv.html):
  Transform skewed variables (aiming at they better conform to a normal distribution).

* [`transpose_omv`](https://sjentsch.github.io/jmvReadWrite/reference/transpose_omv.html):
  Transpose a data set: Make rows into columns and vice versa.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_omv.R
\name{read_omv}
\alias{read_omv}
\title{Read files created of the statistical spreadsheet 'jamovi' (\url{https://www.jamovi.org})}
\usage{
read_omv(
  fleInp = "",
  useFlt = FALSE,
  rmMsVl = FALSE,
  sveAtt = TRUE,
  getSyn = FALSE,
  getHTM = FALSE
)
}
\arguments{
\item{fleInp}{Name (including the path, if required) of the 'jamovi'-file to be read ("FILENAME.omv"; default: "")}

\item{useFlt}{Apply filters (remove the lines where the filter is set to 0; default: FALSE)?}

\item{rmMsVl}{Remove values defined as missing values (replace them with NA; default: FALSE)?}

\item{sveAtt}{Store attributes that are not required in the data set (if you want to write the same data set using write_omv; default: FALSE)?}

\item{getSyn}{Extract syntax from the analyses in the 'jamovi'-file and store it in the attribute "syntax" (default: FALSE)?}

\item{getHTM}{Store index.html in the attribute "HTML" (default: FALSE)?}
}
\value{
data frame (can be directly used with functions included in the R-package \code{jmv} and syntax from 'jamovi'; also compatible with the format of the
R-package \code{foreign})
}
\description{
Read files created of the statistical spreadsheet 'jamovi' (\url{https://www.jamovi.org})
}
\examples{
nmeInp <- system.file("extdata", "ToothGrowth.omv", package = "jmvReadWrite")
data <- jmvReadWrite::read_omv(fleInp = nmeInp)
str(data)
# shows the data frame including all jamovi attributes:
# 'data.frame':	60 obs. of  16 variables:
#  $ Filter 1          : logi  TRUE TRUE TRUE TRUE TRUE TRUE ...
#   ..- attr(*, "name")= chr "Filter 1"
#   ..- attr(*, "id")= int 18
#   ..- attr(*, "columnType")= chr "Filter"
#   ..- attr(*, "dataType")= chr "Integer"
#   ..- attr(*, "measureType")= chr "Nominal"
#   ..- attr(*, "formula")= chr "logLen < 1.5"
#   ..- attr(*, "formulaMessage")= chr ""
#   ..- attr(*, "parentId")= int 0
#   ..- attr(*, "width")= int 78
#   ..- attr(*, "type")= chr "integer"
#   ..- attr(*, "importName")= chr ""
#   ..- attr(*, "description")= chr ""
#   ..- attr(*, "transform")= int 0
#   ..- attr(*, "edits")= list()
#   ..- attr(*, "missingValues")= list()
#   ..- attr(*, "filterNo")= int 0
#   ..- attr(*, "active")= logi FALSE
# ... (continued)

# getSyn requires jmvcore and RProtoBuf, and is thus not run here
\dontrun{
data <- jmvReadWrite::read_omv(fleInp = nmeInp, getSyn = TRUE)
}
# if the syntax couldn't be extracted, an empty list - length = 0 - is returned,
# otherwise, the commands are shown and the first analysis is run, with the output
# from the second analysis being assigned to the variable result
if (length(attr(data, "syntax")) >= 1) {
    print(attr(data, "syntax"))
    if (nzchar(system.file(package = "jmv"))) {
        # the print-function is only used to force devtools::run_examples() to show output
        eval(parse(text = paste0("result = ", attr(data, "syntax")[1])))
        # without assigning the output to a variable, the command would be:
        # eval(parse(text = attr(data, "syntax")[1]))
        print(names(result))
        print(result$main)
        # -> "main"      "assump"    "contrasts" "postHoc"   "emm"       "residsOV"
        # (the names of the six output tables)
    }
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/search_omv.R
\name{search_omv}
\alias{search_omv}
\title{Search values in .omv-files for the statistical spreadsheet 'jamovi' (\url{https://www.jamovi.org})}
\usage{
search_omv(
  dtaInp = NULL,
  srcTrm = c(),
  whlTrm = FALSE,
  ignCse = FALSE,
  incNum = TRUE,
  incOrd = TRUE,
  incNom = TRUE,
  incID = TRUE,
  incCmp = TRUE,
  incRcd = TRUE,
  ...
)
}
\arguments{
\item{dtaInp}{Either a data frame or the name of a jamovi data file to be read (including the path, if required; "FILENAME.omv"; default: NULL)}

\item{srcTrm}{(Character or numeric) Vector (with length = 1) with a search term to be found in the data frame (default: c())}

\item{whlTrm}{Whether the exact search term shall be found (TRUE) or whether a partial match is sufficient (FALSE; default: FALSE)}

\item{ignCse}{Whether to ignore the case of the search term (default: FALSE)}

\item{incNum}{Whether to include continuous variables in the search (default: TRUE)}

\item{incOrd}{Whether to include ordinal variables in the search (default: TRUE)}

\item{incNom}{Whether to include nominal variables in the search (default: TRUE)}

\item{incID}{Whether to include ID variables in the search (default: TRUE)}

\item{incCmp}{Whether to include Computed variables in the search (default: TRUE)}

\item{incRcd}{Whether to include Recoded variables in the search (default: TRUE)}

\item{...}{Additional arguments passed on to methods; see Details below}
}
\value{
a named list with the places where the search term was found: names in the list are the variables / columns, the entries the respective row names
within that variable / column (row names are used for being tolerant to filtered-out cases in jamovi, if a filter is used, row numbers would be
incorrect)
}
\description{
Search values in .omv-files for the statistical spreadsheet 'jamovi' (\url{https://www.jamovi.org})
}
\details{
\itemize{
\item The ellipsis-parameter (\code{...}) can be used to submit arguments / parameters to the function that is used for reading and writing the data. Clicking on the
respective function under “See also”, you can get a more detailed overview over which parameters each of those functions take. The functions are:
\code{read_omv} and \code{write_omv} (for jamovi-files).
}
}
\examples{
# the exact value 24 appears 13 times in age
bfi_sample <- jmvReadWrite::bfi_sample
jmvReadWrite::search_omv(bfi_sample, 24, whlTrm = TRUE)
# taking the fifth entry from the search results
bfi_sample["61", "age"]
# with the following search, both Males and Females are found
# (the M of Males, wouldn't be matched if ignCse were FALSE and males is
#  only a partial match within Females, thus whlTrm must be set to FALSE)
jmvReadWrite::search_omv(bfi_sample, "males", whlTrm = FALSE, ignCse = TRUE)
# the first entry is a female, the first entry is a male
bfi_sample["1", "gender"] # Females
bfi_sample["6", "gender"] # Males
# using the search results assigned to a variable
srcRes <- jmvReadWrite::search_omv(bfi_sample, "males", whlTrm = FALSE, ignCse = TRUE)
bfi_sample[srcRes[[1]][1], names(srcRes[1])] # Females
bfi_sample[srcRes[[1]][6], names(srcRes[1])] # Males


}
\seealso{
\code{replace_omv} uses \code{\link[=read_omv]{read_omv()}} and \code{\link[=write_omv]{write_omv()}} for reading and writing jamovi-files.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/replace_omv.R
\name{replace_omv}
\alias{replace_omv}
\title{Search values in .omv-files for the statistical spreadsheet 'jamovi' (\url{https://www.jamovi.org})}
\usage{
replace_omv(
  dtaInp = NULL,
  fleOut = "",
  rplLst = list(),
  whlTrm = TRUE,
  varInc = c(),
  varExc = c(),
  incNum = TRUE,
  incOrd = TRUE,
  incNom = TRUE,
  incID = TRUE,
  incCmp = TRUE,
  incRcd = TRUE,
  psvAnl = FALSE,
  ...
)
}
\arguments{
\item{dtaInp}{Either a data frame or the name of a jamovi data file to be read (including the path, if required; "FILENAME.omv"; default: NULL)}

\item{fleOut}{Name of the data file to be written (including the path, if required; "FILE_OUT.omv"; default: ""); if empty, the resulting data frame is
returned instead}

\item{rplLst}{A list where each entry is a vector (with length 2) containing the original value and the to-replace-value (default: list())}

\item{whlTrm}{Whether the search term (first entry in the vectors) must be found exactly (TRUE) or whether a partial match is sufficient (FALSE; default:
TRUE)}

\item{varInc}{Names of variables (character vector) to be included in the replacement (default: c())}

\item{varExc}{Names of variables (character vector) to be excluded from the replacement (default: c())}

\item{incNum}{Whether to include continuous variables in the replacement (default: TRUE)}

\item{incOrd}{Whether to include ordinal variables in the replacement (default: TRUE)}

\item{incNom}{Whether to include nominal variables in the replacement (default: TRUE)}

\item{incID}{Whether to include ID variables in the replacement (default: TRUE)}

\item{incCmp}{Whether to include Computed variables in the replacement (default: TRUE)}

\item{incRcd}{Whether to include Recoded variables in the replacement (default: TRUE)}

\item{psvAnl}{Whether analyses that are contained in the input file shall be transferred to the output file (default: FALSE)}

\item{...}{Additional arguments passed on to methods; see Details below}
}
\value{
a data frame (only returned if \code{fleOut} is empty) with the replaced values
}
\description{
Search values in .omv-files for the statistical spreadsheet 'jamovi' (\url{https://www.jamovi.org})
}
\details{
\itemize{
\item \code{rplLst} is a list. Each list entry contains a vector (with length 2), where the first entry is the original value, and the second entry is the value the
original value is to be replaced with.
\item \code{whlTrm} indicates whether partial matches of the original value(s) shall replaced (e.g., for original: 24 and replacement: 34, 241 will be changed into
341).
\item \code{varInc} and \code{varExc} determine which variables are included or excluded from the replacement. If both are given, a warning is issued and \code{varInc} takes
precedence. \code{varInc} makes that only in these variables, the replacement requested by \code{rplLst} is carried out, if \code{varExc} is given, for all variables of
the input data set, except those defined in \code{varExc}, the replacement is carried out.
\item The ellipsis-parameter (\code{...}) can be used to submit arguments / parameters to the function that is used for reading and writing the data. Clicking on the
respective function under “See also”, you can get a more detailed overview over which parameters each of those functions take. The functions are:
\code{read_omv} and \code{write_omv} (for jamovi-files).
}
}
\examples{
\dontrun{
bfi_sample <- jmvReadWrite::bfi_sample
# the gender in the original data file is plural...
table(bfi_sample$gender)
# and shall be converted to singular
rplDF <- jmvReadWrite::replace_omv(dtaInp = bfi_sample,
           rplLst = list(c("Females", "Female"), c("Males", "Male")))
table(rplDF$gender)
# with giving an output file name, the data set is written
nmeOut <- tempfile(fileext = ".omv")
jmvReadWrite::replace_omv(bfi_sample, fleOut = nmeOut,
  rplLst = list(c("Females", "Female"), c("Males", "Male")))
file.exists(nmeOut)
rplDF <- jmvReadWrite::read_omv(nmeOut)
table(rplDF$gender)
unlink(nmeOut)
# it is sensible to check / search for the original values before running replace_omv
jmvReadWrite::search_omv(bfi_sample, 24, whlTrm = TRUE)
rplDF <- jmvReadWrite::replace_omv(bfi_sample, rplLst = list(c(24, NA)))
table(rplDF$age)
}

}
\seealso{
\code{replace_omv} uses \code{\link[=read_omv]{read_omv()}} and \code{\link[=write_omv]{write_omv()}} for reading and writing jamovi-files.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/arrange_cols_omv.R
\name{arrange_cols_omv}
\alias{arrange_cols_omv}
\title{Re-arrange columns / variables in .omv-files for the statistical spreadsheet 'jamovi' (\url{https://www.jamovi.org})}
\usage{
arrange_cols_omv(
  dtaInp = NULL,
  fleOut = "",
  varOrd = c(),
  varMve = list(),
  psvAnl = FALSE,
  usePkg = c("foreign", "haven"),
  selSet = "",
  ...
)
}
\arguments{
\item{dtaInp}{Either a data frame or the name of a data file to be read (including the path, if required; "FILENAME.ext"; default: NULL); files can be of
any supported file type, see Details below}

\item{fleOut}{Name of the data file to be written (including the path, if required; "FILE_OUT.omv"; default: ""); if empty, the resulting data frame is
returned instead}

\item{varOrd}{Character vector with the desired order of variable(s) in the data frame (see Details; default: c())}

\item{varMve}{Named list defining to how much a particular variable (name of a list entry) should be moved up (neg. value of a list entry) or down (pos.
value) in the data frame (see Details; default: c())}

\item{psvAnl}{Whether analyses that are contained in the input file shall be transferred to the output file (default: FALSE)}

\item{usePkg}{Name of the package: "foreign" or "haven" that shall be used to read SPSS, Stata and SAS files; "foreign" is the default (it comes with
base R), but "haven" is newer and more comprehensive}

\item{selSet}{Name of the data set that is to be selected from the workspace (only applies when reading .RData-files)}

\item{...}{Additional arguments passed on to methods; see Details below}
}
\value{
a data frame (only returned if \code{fleOut} is empty) where the order of variables / columns of the input data set is re-arranged
}
\description{
Re-arrange columns / variables in .omv-files for the statistical spreadsheet 'jamovi' (\url{https://www.jamovi.org})
}
\details{
\itemize{
\item \code{varOrd} is a character vector. If not all variables of the original data set are contained in \code{varOrd}, a warning is issued but otherwise the list of
variables defined in \code{varOrd} is used (removing variables not contained in \code{varOrd}).
\item \code{varMve} is a named list. For example would \code{list(VARNAME = -3)} move the variable \code{VARNAME} three positions up in the list of variables (towards the
first column), and \code{list(VARNAME = 3)} would move it three positions down (towards the last column). If the number of steps the variable is to be moved
leads to the position being either lower than the first or higher than the total number of variables in the data set, an error message is issued. Please
note that the list entries are processed one after another, that is, for a second list entry, you have to consider how the first list entry may have
changed to order of variables.
\item Using \code{varOrd} makes more sense for changing the position of several variables, whereas using \code{varMve} makes more sense for one variable. If
both parameters are given, a warning is issued and \code{varOrd} takes precedence.
\item The ellipsis-parameter (\code{...}) can be used to submit arguments / parameters to the functions that are used for reading and writing the data. By clicking
on the respective function under “See also”, you can get a more detailed overview over which parameters each of those functions take. The functions are:
\code{read_omv} and \code{write_omv} (for jamovi-files), \code{read.table} (for CSV / TSV files; using similar defaults as \code{read.csv} for CSV and \code{read.delim} for TSV
which both are based upon \code{read.table}), \code{load} (for .RData-files), \code{readRDS} (for .rds-files), \code{read_sav} (needs the R-package \code{haven}) or \code{read.spss}
(needs the R-package \code{foreign}) for SPSS-files, \code{read_dta} (\code{haven}) / \code{read.dta} (\code{foreign}) for Stata-files, \code{read_sas} (\code{haven}) for SAS-data-files,
and \code{read_xpt} (\code{haven}) / \code{read.xport} (\code{foreign}) for SAS-transport-files. If you would like to use \code{haven}, you may need to install it using
\code{install.packages("haven", dep = TRUE)}.
}
}
\examples{
nmeInp <- system.file("extdata", "AlbumSales.omv", package = "jmvReadWrite")
nmeOut <- tempfile(fileext = ".omv")
# the original file has the variables in the order: "Adverts", "Airplay", "Image", "Sales"
names(read_omv(nmeInp))
# first, we move the variable "Sales" to the first place using the varOrd-parameter
jmvReadWrite::arrange_cols_omv(dtaInp = nmeInp, fleOut = nmeOut,
  varOrd = c("Sales", "Adverts", "Airplay", "Image"))
names(jmvReadWrite::read_omv(nmeOut))
unlink(nmeOut)
# now, we move the variable "Sales" to the first place using the varMve-parameter
jmvReadWrite::arrange_cols_omv(dtaInp = nmeInp, fleOut = nmeOut, varMve = list(Sales = -3))
names(jmvReadWrite::read_omv(nmeOut))
unlink(nmeOut)


}
\seealso{
\code{arrange_cols_omv} internally uses the following functions for reading and writing data files in different formats: \code{\link[=read_omv]{read_omv()}} and
\code{\link[=write_omv]{write_omv()}} for jamovi-files, \code{\link[utils:read.table]{utils::read.table()}} for CSV / TSV files, \code{\link[=load]{load()}} for reading .RData-files, \code{\link[=readRDS]{readRDS()}} for .rds-files,
\code{\link[haven:read_spss]{haven::read_sav()}} or \code{\link[foreign:read.spss]{foreign::read.spss()}} for SPSS-files, \code{\link[haven:read_dta]{haven::read_dta()}} or \code{\link[foreign:read.dta]{foreign::read.dta()}} for Stata-files, \code{\link[haven:read_sas]{haven::read_sas()}} for
SAS-data-files, and \code{\link[haven:read_xpt]{haven::read_xpt()}} or \code{\link[foreign:read.xport]{foreign::read.xport()}} for SAS-transport-files.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merge_cols_omv.R
\name{merge_cols_omv}
\alias{merge_cols_omv}
\title{Merges two or more data files by adding the content of other input files as columns to the first
input file and outputs them as files for the statistical spreadsheet 'jamovi'
(\url{https://www.jamovi.org})}
\usage{
merge_cols_omv(
  dtaInp = NULL,
  fleOut = "",
  typMrg = c("outer", "inner", "left", "right"),
  varBy = list(),
  varSrt = c(),
  psvAnl = FALSE,
  usePkg = c("foreign", "haven"),
  selSet = "",
  ...
)
}
\arguments{
\item{dtaInp}{Either a data frame (with the attribute "fleInp" containing the files to merge) or
vector with the names of the input files (including the path, if required;
"FILENAME.ext"; default: NULL); files can be of any supported file type, see
Details below}

\item{fleOut}{Name of the data file to be written (including the path, if required;
"FILE_OUT.omv"; default: ""); if empty, the resulting data frame is returned
instead}

\item{typMrg}{Type of merging operation: "outer" (default), "inner", "left" or "right"; see
Details below}

\item{varBy}{Name of the variable by which the data sets are matched, can either be a string,
a character or a list (see Details below; default: list())}

\item{varSrt}{Variable(s) that are used to sort the data frame (see Details; if empty, the order
after merging is kept; default: c())}

\item{psvAnl}{Whether analyses that are contained in the input file shall be transferred to the
output file (TRUE / FALSE; default: FALSE)}

\item{usePkg}{Name of the package: "foreign" or "haven" that shall be used to read SPSS, Stata
and SAS files; "foreign" is the default (it comes with base R), but "haven" is
newer and more comprehensive}

\item{selSet}{Name of the data set that is to be selected from the workspace (only applies when
reading .RData-files)}

\item{...}{Additional arguments passed on to methods; see Details below}
}
\value{
a data frame (only returned if \code{fleOut} is empty) where the columns of all input data
sets (given in the \code{dtaInp}-argument) are concatenated
}
\description{
Merges two or more data files by adding the content of other input files as columns to the first
input file and outputs them as files for the statistical spreadsheet 'jamovi'
(\url{https://www.jamovi.org})
}
\details{
\itemize{
\item Using data frames with the input parameter \code{dtaInp} is primarily thought to be used when
calling \code{merge_cols_omv} from the jamovi-modules \code{jTransform} and \code{Rj}. For the use in R, it
is strongly recommended to use a character vector with the file names instead.
\item There are four different types of merging operations (defined via \code{typMrg}): "outer" keeps all
cases (but columns in the resulting data set may contain empty cells / missing values if same
input data sets did not have a row containing the matching variable (defined in \code{varBy}).
"inner" keeps only those cases where all datasets contain the same value in the matching
variable, for "left" all cases from the first data set in \code{dtaInp} are kept (whereas cases
that are only contained in the second or any later input data set are dropped), for "right"
all cases from the second (or any higher) data set in \code{dtaInp} are kept. The behaviour of
"left" and "right" may be somewhat difficult to predict in case of merging several data sets,
therefore "outer" might be a safer choice if several data sets are merged.
\item The variable that is used for matching (\code{varBy}) can either be a string (if all datasets
contain a matching variable with the same name), a character vector (containing more than one
matching variables that are contained in / the same for all data sets) or a list with the same
length as dtaInp. In such list, each cell can again contain either a string (one matching
variable for each data set in dtaInp) or a character vector (several matching variables for
each data set in dtaInp; NB: all character vectors in the cells of the list must have the same
length as it is necessary to always use the same number of matching variables when merging).
\item \code{varSrt} can be either a character or a character vector (with one or more variables
respectively). The sorting order for a particular variable can be inverted with preceding the
variable name with "-". Please note that this doesn't make sense and hence throws a warning
for certain variable types (e.g., factors).
\item The ellipsis-parameter (\code{...}) can be used to submit arguments / parameters to the functions
that are used for transforming or reading the data. By clicking on the respective function
under “See also”, you can get a more detailed overview over which parameters each of those
functions take.
\item Adding columns uses \code{merge}. \code{typMrg} is implemented by setting \code{TRUE} or \code{FALSE} to \code{all.x}
and \code{all.y} in \code{merge}, \code{varBy} matches \code{by.x} and \code{by.y}. The help for \code{merge} can be
accessed by clicking on the link under “See also”.
\item The functions for reading and writing the data are: \code{read_omv} and \code{write_omv} (for
jamovi-files), \code{read.table} (for CSV / TSV files; using similar defaults as \code{read.csv} for CSV
and \code{read.delim} for TSV which both are based upon \code{read.table}), \code{load} (for .RData-files),
\code{readRDS} (for .rds-files), \code{read_sav} (needs R-package \code{haven}) or \code{read.spss} (needs
R-package \code{foreign}) for SPSS-files, \code{read_dta} (\code{haven}) / \code{read.dta} (\code{foreign}) for
Stata-files, \code{read_sas} (\code{haven}) for SAS-data-files, and \code{read_xpt} (\code{haven}) / \code{read.xport}
(\code{foreign}) for SAS-transport-files. If you would like to use \code{haven}, you may need to install
it using \code{install.packages("haven", dep = TRUE)}.
}
}
\examples{
dtaInp <- jmvReadWrite::bfi_sample2
nmeInp <- paste0(tempfile(), "_", 1:3, ".rds")
nmeOut <- tempfile(fileext = ".omv")
for (i in seq_along(nmeInp)) {
    saveRDS(stats::setNames(dtaInp, c("ID", paste0(names(dtaInp)[-1], "_", i))), nmeInp[i])
}
# save dtaInp three times (i.e., the length of nmeInp), adding "_" + 1 ... 3 as index
# to the data variables (A1 ... O5, gender, age → A1_1, ...)
jmvReadWrite::merge_cols_omv(dtaInp = nmeInp, fleOut = nmeOut, varBy = "ID")
cat(file.info(nmeOut)$size)
# -> 17731 (size may differ on different OSes)
dtaOut <- jmvReadWrite::read_omv(nmeOut, sveAtt = FALSE)
# read the data set where the three original datasets were added as columns and show
# the variable names
cat(names(dtaOut))
cat(names(dtaInp))
# compared to the input data set, we have the same names (expect for "ID" which was
# used for matching and that each variable had added an indicator from which data
# set they came)
cat(dim(dtaInp), dim(dtaOut))
# the first dimension of the data sets (rows) stayed the same (250), whereas the
# second dimension is now approx. three times as large (28 -> 82):
# 28 - 1 (for "ID") = 27 * 3 + 1 (for "ID") = 82
cat(colMeans(dtaInp[2:11]))
cat(colMeans(dtaOut[2:11]))
# it's therefore not much surprise that the values of the column means for the first
# 10 variables of dtaInp and dtaOut are the same too

unlink(nmeInp)
unlink(nmeOut)


}
\seealso{
\code{merge_cols_omv} internally uses the following functions: Adding columns uses
\code{\link[=merge]{merge()}}. For reading and writing data files in different formats: \code{\link[=read_omv]{read_omv()}}
and \code{\link[=write_omv]{write_omv()}} for jamovi-files, \code{\link[utils:read.table]{utils::read.table()}} for CSV / TSV files,
\code{\link[=load]{load()}} for reading .RData-files, \code{\link[=readRDS]{readRDS()}} for .rds-files, \code{\link[haven:read_spss]{haven::read_sav()}} or
\code{\link[foreign:read.spss]{foreign::read.spss()}} for SPSS-files, \code{\link[haven:read_dta]{haven::read_dta()}} or \code{\link[foreign:read.dta]{foreign::read.dta()}} for
Stata-files, \code{\link[haven:read_sas]{haven::read_sas()}} for SAS-data-files, and \code{\link[haven:read_xpt]{haven::read_xpt()}} or
\code{\link[foreign:read.xport]{foreign::read.xport()}} for SAS-transport-files.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/write_omv.R
\name{write_omv}
\alias{write_omv}
\title{Write files to be used with the statistical spreadsheet 'jamovi'
(\url{https://www.jamovi.org})}
\usage{
write_omv(
  dtaFrm = NULL,
  fleOut = "",
  wrtPtB = FALSE,
  frcWrt = FALSE,
  retDbg = FALSE,
  vldExt = TRUE
)
}
\arguments{
\item{dtaFrm}{Data frame to be exported (default: NULL)}

\item{fleOut}{Name / position of the output file to be generated
("FILENAME.omv"; default: "")}

\item{wrtPtB}{Whether to write protocol buffers (see Details; default: FALSE)}

\item{frcWrt}{Whether to overwrite existing files with the same name (see
Details; default: FALSE)}

\item{retDbg}{Whether to return a list with debugging information (see
Value; default: FALSE)}

\item{vldExt}{Whether to validate that the file name has a correct extension
(default: TRUE)}
}
\value{
a list (if retDbg == TRUE), containing the meta data (mtaDta,
metadata.json in the OMV-file), the extended data (xtdDta,
xdata.json in the OMV-file) and the original data frame (dtaFrm)
}
\description{
Write files to be used with the statistical spreadsheet 'jamovi'
(\url{https://www.jamovi.org})
}
\details{
\itemize{
\item jamovi has a specific measurement level / type "ID" (in addition to the
"standard" ones "Nominal", "Ordinal", and "Continuous"). "ID" is used for
columns that contain some form of ID (e.g., a participant code). In order
to set a variable of your data frame to "ID", you have to set the
attribute \code{jmv-id} (e.g., \code{attr(dtaFrm$column, "jmv-id") = TRUE}).
\item CAUTION: Setting wrtPtB to TRUE currently overwrites analyses that already
exist in a data file. It is meant to be used for \code{describe_omv} only. If
you set wrtPtB to TRUE, ensure to use an output file name that isn't would
not overwrite any existing file. Protocol buffers are used to exchange
data between the different parts of jamovi (the server and the client) and
also the format in which analyses are stored in the jamovi data files.
\item \code{write_omv} checks whether the output file already exists and throws an
error if this is the case. frcWrt permits you to overwrite the existing
file.
}
}
\examples{
# use the data set "ToothGrowth" and, if it exists, write it as
# jamovi-file using write_omv()
jmvReadWrite::ToothGrowth
nmeOut <- tempfile(fileext = ".omv")
# typically, one would use a "real" file name instead of tempfile(),
# e.g., "Data1.omv"
dtaDbg = jmvReadWrite::write_omv(dtaFrm = ToothGrowth, fleOut = nmeOut, retDbg = TRUE)
print(names(dtaDbg))
# the print-function is only used to force devtools::run_examples()
# to show output
# -> "mtaDta" "xtdDta" "dtaFrm"
# returns a list with the metadata (mtaDta, e.g., column and data type),
# the extended data (xtdDta, e.g., variable lables), and the data frame
# (dtaFrm) the purpose of these variables is merely for checking (under-
# standing the file format) and debugging

# check whether the file was written to the disk, get the file informa-
# tion (size, etc.) and delete the file afterwards
print(list.files(dirname(nmeOut), basename(nmeOut)))
# -> "file[...].omv" ([...] is a combination of random numbers / characters
print(file.info(nmeOut)$size)
# -> approx. 2600 (size may differ on different OSes)
unlink(nmeOut)


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{bfi_sample}
\alias{bfi_sample}
\title{Twenty-five personality self-report items taken from the International Personality Item Pool}
\format{
A data frame with 254 rows (250 original respondents, 4 generated for testing) and 33 variables:
\describe{
\item{\code{ID}}{Respondent ID}
\item{\code{A1}}{Am indifferent to the feelings of others. (reversed)}
\item{\code{A2}}{Inquire about others' well-being.}
\item{\code{A3}}{Know how to comfort others.}
\item{\code{A4}}{Love children.}
\item{\code{A5}}{Make people feel at ease.}
\item{\code{C1}}{Am exacting in my work.}
\item{\code{C2}}{Continue until everything is perfect.}
\item{\code{C3}}{Do things according to a plan.}
\item{\code{C4}}{Do things in a half-way manner. (reversed)}
\item{\code{C5}}{Waste my time. (reversed)}
\item{\code{E1}}{Don't talk a lot. (reversed)}
\item{\code{E2}}{Find it difficult to approach others. (reversed)}
\item{\code{E3}}{Know how to captivate people.}
\item{\code{E4}}{Make friends easily.}
\item{\code{E5}}{Take charge.}
\item{\code{N1}}{Get angry easily.}
\item{\code{N2}}{Get irritated easily.}
\item{\code{N3}}{Have frequent mood swings.}
\item{\code{N4}}{Often feel blue.}
\item{\code{N5}}{Panic easily.}
\item{\code{O1}}{Am full of ideas.}
\item{\code{O2}}{Avoid difficult reading material. (reversed)}
\item{\code{O3}}{Carry the conversation to a higher level.}
\item{\code{O4}}{Spend time reflecting on things.}
\item{\code{O5}}{Will not probe deeply into a subject. (reversed)}
\item{\code{gender}}{Gender of the respondent (female, male)}
\item{\code{age}}{Age of the respondent (years)}
\item{\code{AD}}{Exponent of age (computed: EXP(age))}
\item{\code{AF}}{Random data (for testing)}
\item{\code{AG}}{Random data (for testing)}
\item{\code{age_tr}}{Age of the respondent (transformed, as decades: 1 - 10-19, 2 - 20-29, 3 - 30-39, 4 - 40-49, 5 - 50-59, 6 - 60 and over)}
\item{\code{ID2}}{Respondent ID (for testing; "A" + random-generated 5-digit-code)}
}
}
\usage{
bfi_sample
}
\description{
The data set contains responses from 250 participants filling in twenty-five personality self-report items
taken from the International Personality Item Pool (https://ipip.ori.org) as part of the Synthetic Aperture
Personality Assessment (SAPA) web-based personality assessment (https://sapa-project.org) project.
The 25 items are organized by five putative factors: Agreeableness (A1 to A5), Conscientiousness (C1 to C5),
Extraversion (E1 to E5), Neuroticism (N1 to N5), and Openness (N1 to N5). The items were short phrases that
the respondent should answer by indicating how accurately the statement describes their typical behaviour or
attitude. Responses were collected using a 6-point scale: 1 - Very inaccurate, 2 - Moderately inaccurate,
3 - Slightly inaccurate, 4 - Slightly accurate, 5 - Moderately accurate, 6 - Very accurate.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/convert_to_omv.R
\name{convert_to_omv}
\alias{convert_to_omv}
\title{Convert data files (CSV, R, other statistics packages) into .omv-files for the statistical spreadsheet 'jamovi' (\url{https://www.jamovi.org})}
\usage{
convert_to_omv(
  fleInp = "",
  fleOut = "",
  varSrt = c(),
  usePkg = c("foreign", "haven"),
  selSet = "",
  ...
)
}
\arguments{
\item{fleInp}{Name (including the path, if required) of the data file to be read ("FILENAME.ext"; default: ""); supports CSV and R-files natively, or other
file types if "foreign" or "haven" are installed, see Details below}

\item{fleOut}{Name (including the path, if required) of the data file to be written ("FILENAME.omv"; default: ""); if empty, the extension of fleInp is
replaced with ".omv"}

\item{varSrt}{Variable(s) that are used to sort the data frame (see Details; if empty, the row order of the input file is kept; default: c())}

\item{usePkg}{Name of the package: "foreign" or "haven" that shall be used to read SPSS, Stata and SAS files; "foreign" is the default (it comes with base
R), but "haven" is newer and more comprehensive}

\item{selSet}{Name of the data set that is to be selected from the workspace (only applies when reading .RData-files)}

\item{...}{Additional arguments passed on to methods; see Details below}
}
\value{
the function doesn't have a return value (it returns NULL)
}
\description{
Convert data files (CSV, R, other statistics packages) into .omv-files for the statistical spreadsheet 'jamovi' (\url{https://www.jamovi.org})
}
\details{
\itemize{
\item In difference to the remaining helper functions, \code{convert_to_omv} doesn't accept a data frame as input and it neither does return a data frame if \code{fleOut}
is left empty: If you want to write a data frame, use \code{write_omv}. If you want to have a data frame returned use \code{read_omv} for jamovi-files or any of the
functions listed in the bullet point below for any other file type.
\item \code{varSrt} can be either a character or a character vector (with one or more variables respectively). The sorting order for a particular variable can be
inverted with preceding the variable name with "-". Please note that this doesn't make sense and hence throws a warning for certain variable types (e.g.,
factors).
\item The ellipsis-parameter (\code{...}) can be used to submit arguments / parameters to the functions that are used for reading and writing the data. By clicking
on the respective function under “See also”, you can get a more detailed overview over which parameters each of those functions take. The functions are:
\code{read_omv} and \code{write_omv} (for jamovi-files), \code{read.table} (for CSV / TSV files; using similar defaults as \code{read.csv} for CSV and \code{read.delim} for TSV
which both are based upon \code{read.table}), \code{load} (for .RData-files), \code{readRDS} (for .rds-files), \code{read_sav} (needs the R-package \code{haven}) or \code{read.spss}
(needs the R-package \code{foreign}) for SPSS-files, \code{read_dta} (\code{haven}) / \code{read.dta} (\code{foreign}) for Stata-files, \code{read_sas} (\code{haven}) for SAS-data-files,
and \code{read_xpt} (\code{haven}) / \code{read.xport} (\code{foreign}) for SAS-transport-files. If you would like to use \code{haven}, you may need to install it using
\code{install.packages("haven", dep = TRUE)}.
}
}
\examples{
# Example 1: Convert from RDS
# (use ToothGrowth as example, save it as RDS)
nmeInp <- tempfile(fileext = ".rds")
nmeOut <- tempfile(fileext = ".omv")
saveRDS(jmvReadWrite::ToothGrowth, nmeInp)
jmvReadWrite::convert_to_omv(fleInp = nmeInp, fleOut = nmeOut)
cat(list.files(dirname(nmeOut), basename(nmeOut)))
# -> "file[...].omv" ([...] contains a random combination of numbers / characters
cat(file.info(nmeOut)$size)
# -> 2448 (size may differ on different OSes)
cat(str(jmvReadWrite::read_omv(nmeOut, sveAtt = FALSE)))
# gives a overview of the dataframe (all columns and some attributes,
# sveAtt is intentionally set to FALSE to make the output not too overwhelming)
unlink(nmeInp)
unlink(nmeOut)

# Example 2: Convert from CSV
# (use ToothGrowth again as example, this time save it as CSV)
nmeInp <- tempfile(fileext = ".csv")
nmeOut <- tempfile(fileext = ".omv")
write.csv(jmvReadWrite::ToothGrowth, nmeInp)
jmvReadWrite::convert_to_omv(fleInp = nmeInp, fleOut = nmeOut)
cat(list.files(dirname(nmeOut), basename(nmeOut)))
cat(file.info(nmeOut)$size)
# -> 2104 (size may differ acc. to OS; the size is smaller than for the RDS-file
# because CSV can store fewer attributes, e.g., labels)
cat(str(jmvReadWrite::read_omv(nmeOut, sveAtt = FALSE)))
# gives a overview of the dataframe (all columns and some attributes)
unlink(nmeInp)
unlink(nmeOut)


}
\seealso{
\code{convert_to_omv} internally uses the following functions for reading and writing data files in different formats: \code{\link[=read_omv]{read_omv()}} and
\code{\link[=write_omv]{write_omv()}} for jamovi-files, \code{\link[utils:read.table]{utils::read.table()}} for CSV / TSV files, \code{\link[=load]{load()}} for reading .RData-files, \code{\link[=readRDS]{readRDS()}} for .rds-files,
\code{\link[haven:read_spss]{haven::read_sav()}} or \code{\link[foreign:read.spss]{foreign::read.spss()}} for SPSS-files, \code{\link[haven:read_dta]{haven::read_dta()}} or \code{\link[foreign:read.dta]{foreign::read.dta()}} for Stata-files, \code{\link[haven:read_sas]{haven::read_sas()}} for
SAS-data-files, and \code{\link[haven:read_xpt]{haven::read_xpt()}} or \code{\link[foreign:read.xport]{foreign::read.xport()}} for SAS-transport-files.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/distances_omv.R
\name{distances_omv}
\alias{distances_omv}
\title{Calculates distances (returning a symmetric matrix) from a raw data matrix in .omv-files for the
statistical spreadsheet 'jamovi' (\url{https://www.jamovi.org})}
\usage{
distances_omv(
  dtaInp = NULL,
  fleOut = "",
  varDst = c(),
  clmDst = TRUE,
  stdDst = "none",
  nmeDst = "euclid",
  mtxSps = FALSE,
  mtxTrL = FALSE,
  mtxDgn = TRUE,
  usePkg = c("foreign", "haven"),
  selSet = "",
  ...
)
}
\arguments{
\item{dtaInp}{Either a data frame or the name of a data file to be read (including the path,
if required; "FILENAME.ext"; default: NULL); files can be of any supported file
type, see Details below.}

\item{fleOut}{Name of the data file to be written (including the path, if required;
"FILE_OUT.omv"; default: ""); if empty, the resulting data frame is returned
instead.}

\item{varDst}{Variable (default: c()) containing a character vector with the names of the
variables for which distances are to be calculated. See Details for more
information.}

\item{clmDst}{Whether the distances shall be calculated between columns (TRUE) or rows
(FALSE; default: TRUE). See Details for more information.}

\item{stdDst}{Character string indicating whether the variables in varDst are to be standardized
and how (default: "none"). See Details for more information.}

\item{nmeDst}{Character string indicating which distance measure is to be calculated calculated
(default: "euclidean"). See Details for more information.}

\item{mtxSps}{Whether the symmetric matrix to be returned should be sparse (default: FALSE)}

\item{mtxTrL}{Whether the symmetric matrix to be returned should only contain the lower
triangular (default: FALSE)}

\item{mtxDgn}{Whether the symmetric matrix to be returned should retain the values in the main
diagonal (default: TRUE)}

\item{usePkg}{Name of the package: "foreign" or "haven" that shall be used to read SPSS, Stata,
and SAS files; "foreign" is the default (it comes with base R), but "haven" is
newer and more comprehensive.}

\item{selSet}{Name of the data set that is to be selected from the workspace (only applies when
reading .RData-files)}

\item{...}{Additional arguments passed on to methods; see Details below.}
}
\value{
a data frame containing a symmetric matrix (only returned if \code{fleOut} is empty)
containing the distances between the variables / columns (clmDst == TRUE) or rows
(clmDst == FALSE)
}
\description{
Calculates distances (returning a symmetric matrix) from a raw data matrix in .omv-files for the
statistical spreadsheet 'jamovi' (\url{https://www.jamovi.org})
}
\details{
\itemize{
\item \code{varDst} must a character vector containing the variables to calculated distances over. If
\code{clmDst} is set to TRUE, distances are calculated between all possible variable pairs and over
subjects / rows in the original data frame. If \code{clmDst} is set to FALSE, distances are
calculated between participants and over all variables given in \code{varDst}. If \code{clmDst} is set
to \code{TRUE}, the symmetric matrix that is returned has the size V x V (V being the number of
variables in varDst; if \code{mtxSps} is set to \code{TRUE}, the size is V - 1 x V - 1, see below); if
\code{clmDst} is set to \code{FALSE}, the symmetric matrix that is returned has the size R x R (R being
the number of rows in the original dataset; it is if \code{mtxSps} is set to \code{TRUE}, the size is
R - 1 x R - 1, see below).
\item \code{stdDst} can be one of the following calculations to standardize the selected variables before
calculating the distances: \code{none} (do not standardize; default), \code{z} (z scores), \code{sd} (divide
by the std. dev.), \code{range} (divide by the range), \code{max} (divide by the absolute maximum),
\code{mean} (divide by the mean), \code{rescale} (subtract the mean and divide by the range).
\item \code{nmeDst} can be one of the following distance measures.
(1) For interval data: \code{euclid} (Euclidean), \code{seuclid} (squared Euclidean), \code{block} (city
block / Manhattan), \code{canberra} (Canberra). \code{chebychev} (maximum distance / supremum norm /
Chebychev), \code{minkowski_p} (Minkowski with power p; NB: needs p), \code{power_p_r} (Minkowski with
power p, and the r-th root; NB: needs p and r), \code{cosine} (cosine between the two vectors),
\code{correlation} (correlation between the two vectors).
(2) For frequency count data: \code{chisq} (chi-square dissimilarity between two sets of
frequencies), \code{ph2} (chi-square dissimilarity normalized by the square root of the number
of values used in the calculation).
(3) For binary data, all measure have to optional parts \code{p} and \code{np} which indicate presence
(\code{p}; defaults to 1 if not given) or absence (\code{np}; defaults to zero if not given).
(a) matching coefficients: \code{rr_p_np} (Russell and Rao), \code{sm_p_np} (simple matching),
\code{jaccard_p_np} / \code{jaccards_p_np} (Jaccard similarity; as in SPSS), \code{jaccardd_p_np} (Jaccard
dissimiliarity; as in \code{dist(..., "binary")} in R), \code{dice_p_np} (Dice or Czekanowski or
Sorenson similarity), \code{ss1_p_np} (Sokal and Sneath measure 1), \code{rt_p_np} (Rogers and
Tanimoto), \code{ss2_p_np} (Sokal and Sneath measure 2), \code{k1_p_np} (Kulczynski measure 1),
\code{ss3_p_np} (Sokal and Sneath measure 3).
(b) conditional probabilities: \code{k2_p_np} (Kulczynski measure 2), \code{ss4_p_np} (Sokal and Sneath
measure 4), \code{hamann_p_np} (Hamann).
(c) predictability measures: \code{lambda_p_np} (Goodman and Kruskal Lambda), \code{d_p_np} (Anderberg’s
D), \code{y_p_np} (Yule’s Y coefficient of colligation), \code{q_p_np} (Yule’s Q).
(d) other measures: \code{ochiai_p_np} (Ochiai), \code{ss5_p_np} (Sokal and Sneath measure 5),
\code{phi_p_np} (fourfold point correlation), \code{beuclid_p_np} (binary Euclidean distance),
\code{bseuclid_p_np} (binary squared Euclidean distance), \code{size_p_np} (size difference),
\code{pattern_p_np} (pattern difference), \code{bshape_p_np} (binary Shape difference), \code{disper_p_np}
(dispersion similarity), \code{variance_p_np} (variance dissimilarity), \code{blwmn_p_np} (binary Lance
and Williams non-metric dissimilarity).
(4) \code{none} (only carry out standardization, if stdDst is different from \code{none}).
\item If \code{mtxSps} is set, a sparse matrix is returned. Those matrices are similar to the format one
often finds for correlation matrices. The values are only retained in the lower triangular,
the columns range from the first to the variable that is second to the last in \code{varDst} (or
respectively, the columns contain the first to the second to the last row of the original
dataset when \code{clmDst} is set to \code{FALSE}), and the rows contain the second to the last variable
in \code{varDst} (or respectively, the rows contain the second to the last row of the original
dataset when \code{clmDst} is set to \code{FALSE}).
\item By default, a full symmetric matrix is returned (i.e., a matrix that has no NAs in any cell).
This behaviour can be changed with setting \code{mtxTrL} and \code{mtxDgn}: If \code{mtxTrL} is set to
\code{TRUE}, the values from the upper triangular matrix are removed / replaced with NAs; if
\code{mtxDgn} is set to \code{FALSE}, the values from the main diagonal are removed / replaced with NAs.
\item The ellipsis-parameter (\code{...}) can be used to submit arguments / parameters to the functions
that are used for reading and writing the data. By clicking on the respective function under
“See also”, you can get a more detailed overview over which parameters each of those functions
take. The functions are: \code{read_omv} and \code{write_omv} (for jamovi-files), \code{read.table} (for CSV
/ TSV files; using similar defaults as \code{read.csv} for CSV and \code{read.delim} for TSV which both
are based upon \code{read.table}), \code{load} (for .RData-files), \code{readRDS} (for .rds-files),
\code{read_sav} (needs the R-package \code{haven}) or \code{read.spss} (needs the R-package \code{foreign}) for
SPSS-files, \code{read_dta} (\code{haven}) / \code{read.dta} (\code{foreign}) for Stata-files, \code{read_sas}
(\code{haven}) for SAS-data-files, and \code{read_xpt} (\code{haven}) / \code{read.xport} (\code{foreign}) for
SAS-transport-files. If you would like to use \code{haven}, you may need to install it using
\code{install.packages("haven", dep = TRUE)}.
}
}
\examples{
# create matrices for the different types of distance measures: continuous
# (cntFrm), frequency counts (frqFrm) or binary (binFrm); all 20 R x 5 C
set.seed(1)
cntFrm <- stats::setNames(as.data.frame(matrix(rnorm(100, sd = 10),
            ncol = 5)), sprintf("C_\%02d", seq(5)))
frqFrm <- stats::setNames(as.data.frame(matrix(sample(seq(10), 100,
            replace = TRUE), ncol = 5)), sprintf("F_\%02d", seq(5)))
binFrm <- stats::setNames(as.data.frame(matrix(sample(c(TRUE, FALSE), 100,
            replace = TRUE), ncol = 5)), sprintf("B_\%02d", seq(5)))
nmeOut <- tempfile(fileext = ".omv")

# calculates the distances between columns, nmeDst is not required: "euclid"
# is the default
jmvReadWrite::distances_omv(dtaInp = cntFrm, fleOut = nmeOut, varDst =
  names(cntFrm), nmeDst = "euclid")
dtaFrm <- jmvReadWrite::read_omv(nmeOut)
unlink(nmeOut)
# the resulting matrix (10 x 10) with the Euclidian distances
print(dtaFrm)

# calculates the (Euclidean) distances between rows (clmDst = FALSE)
jmvReadWrite::distances_omv(dtaInp = cntFrm, fleOut = nmeOut, varDst =
  names(cntFrm), clmDst = FALSE, nmeDst = "euclid")
dtaFrm <- jmvReadWrite::read_omv(nmeOut)
unlink(nmeOut)
# the resulting matrix (20 x 20) with the Euclidian distances
print(dtaFrm)

# calculates the (Euclidean) distances between columns; the original data
# are z-standardized before calculating the distances (stdDst = "z")
jmvReadWrite::distances_omv(dtaInp = cntFrm, fleOut = nmeOut, varDst =
  names(cntFrm), stdDst = "z", nmeDst = "euclid")
dtaFrm <- jmvReadWrite::read_omv(nmeOut)
unlink(nmeOut)
# the resulting matrix (10 x 10) with the Euclidian distances using the
# z-standardized data
print(dtaFrm)

# calculates the correlations between columns
jmvReadWrite::distances_omv(dtaInp = cntFrm, fleOut = nmeOut, varDst =
  names(cntFrm), nmeDst = "correlation")
dtaFrm <- jmvReadWrite::read_omv(nmeOut)
unlink(nmeOut)
# the resulting matrix (10 x 10) with the correlations
print(dtaFrm)

# calculates the chi-square dissimilarity (nmeDst = "chisq") between columns
jmvReadWrite::distances_omv(dtaInp = frqFrm, fleOut = nmeOut, varDst =
  names(frqFrm), nmeDst = "chisq")
dtaFrm <- jmvReadWrite::read_omv(nmeOut)
unlink(nmeOut)
# the resulting matrix (10 x 10) with the chi-square dissimilarities
print(dtaFrm)

# calculates the Jaccard similarity (nmeDst = "jaccard") between columns
jmvReadWrite::distances_omv(dtaInp = binFrm, fleOut = nmeOut, varDst =
  names(binFrm), nmeDst = "jaccard")
dtaFrm <- jmvReadWrite::read_omv(nmeOut)
unlink(nmeOut)
# the resulting matrix (10 x 10) with the Jaccard similarities
print(dtaFrm)


}
\seealso{
\code{distances_omv} internally uses the following function for calculating the distances
for interval data \code{\link[stats:dist]{stats::dist()}}. It furthermore uses the following functions for reading
and writing data files in different formats: \code{\link[=read_omv]{read_omv()}} and
\code{\link[=write_omv]{write_omv()}} for jamovi-files, \code{\link[utils:read.table]{utils::read.table()}} for CSV / TSV files,
\code{\link[=load]{load()}} for reading .RData-files, \code{\link[=readRDS]{readRDS()}} for .rds-files, \code{\link[haven:read_spss]{haven::read_sav()}} or
\code{\link[foreign:read.spss]{foreign::read.spss()}} for SPSS-files, \code{\link[haven:read_dta]{haven::read_dta()}} or \code{\link[foreign:read.dta]{foreign::read.dta()}} for
Stata-files, \code{\link[haven:read_sas]{haven::read_sas()}} for SAS-data-files, and \code{\link[haven:read_xpt]{haven::read_xpt()}} or
\code{\link[foreign:read.xport]{foreign::read.xport()}} for SAS-transport-files.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/long2wide_omv.R
\name{long2wide_omv}
\alias{long2wide_omv}
\title{Converts .omv-files for the statistical spreadsheet 'jamovi' (\url{https://www.jamovi.org}) from long to wide format}
\usage{
long2wide_omv(
  dtaInp = NULL,
  fleOut = "",
  varTgt = c(),
  varExc = c(),
  varID = "ID",
  varTme = c(),
  varSep = "_",
  varOrd = c("times", "vars"),
  varAgg = c("mean", "first"),
  varSrt = c(),
  usePkg = c("foreign", "haven"),
  selSet = "",
  ...
)
}
\arguments{
\item{dtaInp}{Either a data frame or the name of a data file to be read (including the path, if required; "FILENAME.ext"; default: NULL); files can be of
any supported file type, see Details below}

\item{fleOut}{Name of the data file to be written (including the path, if required; "FILE_OUT.omv"; default: ""); if empty, the resulting data frame is
returned instead}

\item{varTgt}{Names of one or more variables to be transformed / reshaped (other variables are excluded, if empty(c()) all variables except \code{varTme},
\code{varID} and \code{varExc} are included; default: c())}

\item{varExc}{Name of the variable(s) should be excluded from the transformation, typically this will be between-subject-variable(s) (default: c())}

\item{varID}{Names of one or more variables that identify the same group / individual (default: c())}

\item{varTme}{Name of the variable(s) that differentiates multiple records from the same group / individual (default: c())}

\item{varSep}{Separator character when concatenating the fixed and time-varying part of the variable name ("VAR1_1", "VAR1_2"; default: "_")}

\item{varOrd}{How variables / columns are organized: for "times" (default) the steps of the time varying variable are adjacent, for "vars" the steps of
the original columns in the long dataset}

\item{varAgg}{How multiple occurrences of particular combinations of time varying variables are aggregated: either "mean" (calculate the mean over
occurrences), or "first" (take the first occurrence)}

\item{varSrt}{Variable(s) that are used to sort the data frame (see Details; if empty, the order returned from reshape is kept; default: c())}

\item{usePkg}{Name of the package: "foreign" or "haven" that shall be used to read SPSS, Stata and SAS files; "foreign" is the default (it comes with
base R), but "haven" is newer and more comprehensive}

\item{selSet}{Name of the data set that is to be selected from the workspace (only applies when reading .RData-files)}

\item{...}{Additional arguments passed on to methods; see Details below}
}
\value{
a data frame (only returned if \code{fleOut} is empty) where the input data set is converted from long to wide format
}
\description{
Converts .omv-files for the statistical spreadsheet 'jamovi' (\url{https://www.jamovi.org}) from long to wide format
}
\details{
\itemize{
\item If \code{varTgt} is empty, it is tried to generate it using all variables in the data frame except those defined by \code{varID}, \code{varTme} and \code{varExc}. The
variable(s) in \code{varID} need to be unique identifiers (in the original dataset), those in \code{varExc} don't have this requirement. It is generally recommended
that the variable names in \code{varExc} and \code{varID} should not contain the variable separator (defined in \code{varSep}; default: "_").
\item \code{varSrt} can be either a character or a character vector (with one or more variables respectively). The sorting order for a particular variable can be
inverted with preceding the variable name with "-". Please note that this doesn't make sense and hence throws a warning for certain variable types (e.g.,
factors).
\item The ellipsis-parameter (\code{...}) can be used to submit arguments / parameters to the functions that are used for transforming, reading or writing the data.
By clicking on the respective function under “See also”, you can get a more detailed overview over which parameters each of those functions take.
\item The transformation from long to wide uses \code{reshape}. \code{varTgt} matches (~) \code{v.names} in \code{reshape}, \code{varID} ~ \code{idvar}, \code{varTme} ~ \code{timevar}, and \code{varSep} ~
\code{sep}. The help for \code{reshape} is very explanatory, click on the link under “See also” to access it, particularly what is explained under “Details”.
\item The functions for reading and writing the data are: \code{read_omv} and \code{write_omv} (for jamovi-files), \code{read.table} (for CSV / TSV files; using similar
defaults as \code{read.csv} for CSV and \code{read.delim} for TSV which both are based upon \code{read.table}), \code{load} (for .RData-files), \code{readRDS} (for .rds-files),
\code{read_sav} (needs R-package \code{haven}) or \code{read.spss} (needs R-package \code{foreign}) for SPSS-files, \code{read_dta} (\code{haven}) / \code{read.dta} (\code{foreign}) for
Stata-files, \code{read_sas} (\code{haven}) for SAS-data-files, and \code{read_xpt} (\code{haven}) / \code{read.xport} (\code{foreign}) for SAS-transport-files. If you would like to
use \code{haven}, you may need to install it using \code{install.packages("haven", dep = TRUE)}.
}
}
\examples{
# generate a test dataframe with 100 (imaginary) participants / units of
#  observation (ID), 8 measurement (measure) of one variable (X)
dtaInp <- data.frame(ID = rep(as.character(seq(1, 100)), each = 8),
                     measure = rep(seq(1, 8), times = 100),
                     X = runif(800, -10, 10))
cat(str(dtaInp))
# the output should look like this
# 'data.frame': 800 obs. of  3 variables:
#  $ ID     : chr  "1" "1" "1" "1" ...
#  $ measure: int  1 2 3 4 5 6 7 8 1 2 ...
#  $ X      : num  ...
# this data set is stored as (temporary) RDS-file and later processed by long2wide
nmeInp <- tempfile(fileext = ".rds")
nmeOut <- tempfile(fileext = ".omv")
saveRDS(dtaInp, nmeInp)
jmvReadWrite::long2wide_omv(dtaInp = nmeInp, fleOut = nmeOut, varTgt = "X", varID = "ID",
  varTme = "measure")
# it is required to give at least the arguments dtaInp, varID and varTme
# check whether the file was created and its size
cat(list.files(dirname(nmeOut), basename(nmeOut)))
# -> "file[...].omv" ([...] contains a random combination of numbers / characters
cat(file.info(nmeOut)$size)
# -> 6851 (approximate size; size may differ in every run [in dependence of
#          how well the generated random data can be compressed])
cat(str(jmvReadWrite::read_omv(nmeOut, sveAtt = FALSE)))
# the data set is now transformed into wide (and each the measurements is now
# indicated as a suffix to X; X_1, X_2, ...)
# 'data.frame':	100 obs. of  9 variables:
#  $ ID : chr  "1" "2" "3" "4" "5" "6" "7" "8" "9" "10" ...
#   ..- attr(*, "jmv-id")= logi TRUE
#   ..- attr(*, "missingValues")= list()
#  $ X_1: num  ...
#   ..- attr(*, "missingValues")= list()
#  $ X_2: num  ...
#   ..- attr(*, "missingValues")= list()
#  $ X_3: num  ...
#   ..- attr(*, "missingValues")= list()
#  $ X_4: num  ...
#   ..- attr(*, "missingValues")= list()
#  $ X_5: num  ...
#   ..- attr(*, "missingValues")= list()
#  $ X_6: num  ...
#   ..- attr(*, "missingValues")= list()
#  $ X_7: num  ...
#   ..- attr(*, "missingValues")= list()
#  $ X_8: num  ...
#   ..- attr(*, "missingValues")= list()

unlink(nmeInp)
unlink(nmeOut)


}
\seealso{
\code{long2wide_omv} internally uses the following functions: The transformation from long to wide uses \code{\link[stats:reshape]{stats::reshape()}}. For reading and writing data
files in different formats: \code{\link[=read_omv]{read_omv()}} and \code{\link[=write_omv]{write_omv()}} for jamovi-files, \code{\link[utils:read.table]{utils::read.table()}} for CSV / TSV files,
\code{\link[=load]{load()}} for reading .RData-files, \code{\link[=readRDS]{readRDS()}} for .rds-files, \code{\link[haven:read_spss]{haven::read_sav()}} or \code{\link[foreign:read.spss]{foreign::read.spss()}} for SPSS-files, \code{\link[haven:read_dta]{haven::read_dta()}} or
\code{\link[foreign:read.dta]{foreign::read.dta()}} for Stata-files, \code{\link[haven:read_sas]{haven::read_sas()}} for SAS-data-files, and \code{\link[haven:read_xpt]{haven::read_xpt()}} or \code{\link[foreign:read.xport]{foreign::read.xport()}} for SAS-transport-files.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sort_omv.R
\name{sort_omv}
\alias{sort_omv}
\title{Sort data (using one or more variables) in .omv-files for the statistical spreadsheet 'jamovi' (\url{https://www.jamovi.org})}
\usage{
sort_omv(
  dtaInp = NULL,
  fleOut = "",
  varSrt = c(),
  psvAnl = FALSE,
  usePkg = c("foreign", "haven"),
  selSet = "",
  ...
)
}
\arguments{
\item{dtaInp}{Either a data frame or the name of a data file to be read (including the path, if required; "FILENAME.ext"; default: NULL); files can be of
any supported file type, see Details below}

\item{fleOut}{Name of the data file to be written (including the path, if required; "FILE_OUT.omv"; default: ""); if empty, the resulting data frame is
returned instead}

\item{varSrt}{Variable(s) that are used to sort the data frame (see Details; default: c())}

\item{psvAnl}{Whether analyses that are contained in the input file shall be transferred to the output file (TRUE / FALSE; default: FALSE)}

\item{usePkg}{Name of the package: "foreign" or "haven" that shall be used to read SPSS, Stata and SAS files; "foreign" is the default (it comes with
base R), but "haven" is newer and more comprehensive}

\item{selSet}{Name of the data set that is to be selected from the workspace (only applies when reading .RData-files)}

\item{...}{Additional arguments passed on to methods; see Details below}
}
\value{
a data frame (only returned if \code{fleOut} is empty) where the input data set is sorted (according to the variables in \code{varSrt})
}
\description{
Sort data (using one or more variables) in .omv-files for the statistical spreadsheet 'jamovi' (\url{https://www.jamovi.org})
}
\details{
\itemize{
\item \code{varSrt} can be either a character or a character vector (with one or more variables respectively). The sorting order for a particular variable can be
inverted with preceding the variable name with "-". Please note that this doesn't make sense and hence throws a warning for certain variable types (e.g.,
factors).
\item The ellipsis-parameter (\code{...}) can be used to submit arguments / parameters to the functions that are used for reading and writing the data. By clicking
on the respective function under “See also”, you can get a more detailed overview over which parameters each of those functions take. The functions are:
\code{read_omv} and \code{write_omv} (for jamovi-files), \code{read.table} (for CSV / TSV files; using similar defaults as \code{read.csv} for CSV and \code{read.delim} for TSV
which both are based upon \code{read.table}), \code{load} (for .RData-files), \code{readRDS} (for .rds-files), \code{read_sav} (needs the R-package \code{haven}) or \code{read.spss}
(needs the R-package \code{foreign}) for SPSS-files, \code{read_dta} (\code{haven}) / \code{read.dta} (\code{foreign}) for Stata-files, \code{read_sas} (\code{haven}) for SAS-data-files,
and \code{read_xpt} (\code{haven}) / \code{read.xport} (\code{foreign}) for SAS-transport-files. If you would like to use \code{haven}, you may need to install it using
\code{install.packages("haven", dep = TRUE)}.
}
}
\examples{
nmeInp <- system.file("extdata", "AlbumSales.omv", package = "jmvReadWrite")
nmeOut <- tempfile(fileext = ".omv")
jmvReadWrite::sort_omv(dtaInp = nmeInp, fleOut = nmeOut, varSrt = "Image")
dtaFrm <- jmvReadWrite::read_omv(nmeOut)
unlink(nmeOut)
cat(dtaFrm$Image)
# shows that the variable "Image" is sorted in ascending order
cat(is.unsorted(dtaFrm$Image))
# is.unsorted (which checks for whether the variable is NOT sorted) returns FALSE
jmvReadWrite::sort_omv(dtaInp = nmeInp, fleOut = nmeOut, varSrt = "-Image")
# variables can also be sorted in descending order by preceding them with "-"
dtaFrm <- jmvReadWrite::read_omv(nmeOut)
unlink(nmeOut)
cat(dtaFrm$Image)
# shows that the variable "Image" is now sorted in descending order
cat(is.unsorted(dtaFrm$Image))
# this first returns TRUE (the variable is not in ascending order, i.e., unsorted)
cat(is.unsorted(-dtaFrm$Image))
# if the sign of the variable is changed, it returns FALSE (i.e., the variable is
# NOT unsorted)


}
\seealso{
\code{sort_omv} internally uses the following functions for reading and writing data files in different formats: \code{\link[=read_omv]{read_omv()}} and
\code{\link[=write_omv]{write_omv()}} for jamovi-files, \code{\link[utils:read.table]{utils::read.table()}} for CSV / TSV files, \code{\link[=load]{load()}} for reading .RData-files, \code{\link[=readRDS]{readRDS()}} for .rds-files,
\code{\link[haven:read_spss]{haven::read_sav()}} or \code{\link[foreign:read.spss]{foreign::read.spss()}} for SPSS-files, \code{\link[haven:read_dta]{haven::read_dta()}} or \code{\link[foreign:read.dta]{foreign::read.dta()}} for Stata-files, \code{\link[haven:read_sas]{haven::read_sas()}} for
SAS-data-files, and \code{\link[haven:read_xpt]{haven::read_xpt()}} or \code{\link[foreign:read.xport]{foreign::read.xport()}} for SAS-transport-files.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{AlbumSales}
\alias{AlbumSales}
\title{Imagine that you worked for a record company and that your boss was interested in predicting album sales from
advertising.}
\format{
A data frame with 60 rows, each one representing a different album, and 5 variables:
\describe{
\item{\code{selSbj}}{Select the data in this row (1) or not (0)}
\item{\code{Adverts}}{Amount (in thousands of pounds) spent promoting the album before release}
\item{\code{Airplay}}{How many times songs from the album were played on a prominent national radio station in the week before release}
\item{\code{Image}}{How attractive people found the band's image (out of 10)}
\item{\code{Sales}}{Sales (in thousands) of each album in the week after release}
}
}
\usage{
AlbumSales
}
\description{
The data set is fictional and was constructed by Andy Field who therefore owns the copyright. The data set is also publicly
available on the website that accompanies Andy Field's book, https://edge.sagepub.com/field5e.
Without Andy Field's explicit consent, this data set may not be distributed for commercial purposes, this data set may not be
edited, and this data set may not be presented without acknowledging its source (i.e., the terms of a CC BY-NC-ND license).
}
\details{
Reference: Field, A. P. (2017). Discovering Statistics Using IBM SPSS Statistics (5th ed.). Sage.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transform_vars_omv.R
\name{transform_vars_omv}
\alias{transform_vars_omv}
\title{Transform skewed variables (aiming at they conform to a normal distribution) in .omv-files for the statistical spreadsheet 'jamovi' (\url{https://www.jamovi.org})}
\usage{
transform_vars_omv(
  dtaInp = NULL,
  fleOut = "",
  varXfm = NULL,
  psvAnl = FALSE,
  usePkg = c("foreign", "haven"),
  selSet = "",
  ...
)
}
\arguments{
\item{dtaInp}{Either a data frame or the name of a data file to be read (including the path, if required; "FILENAME.ext"; default: NULL); files can be of
any supported file type, see Details below}

\item{fleOut}{Name of the data file to be written (including the path, if required; "FILE_OUT.omv"; default: ""); if empty, the resulting data frame is
returned instead}

\item{varXfm}{Named list variable where the name indicates which transformation is to be carried out and where each list entry points to one or more
variables to be transformed using this transformation. See Details for more information.}

\item{psvAnl}{Whether analyses that are contained in the input file shall be transferred to the output file (default: FALSE)}

\item{usePkg}{Name of the package: "foreign" or "haven" that shall be used to read SPSS, Stata and SAS files; "foreign" is the default (it comes with
base R), but "haven" is newer and more comprehensive}

\item{selSet}{Name of the data set that is to be selected from the workspace (only applies when reading .RData-files)}

\item{...}{Additional arguments passed on to methods; see Details below}
}
\value{
a data frame (only returned if \code{fleOut} is empty) where the order of variables / columns of the input data set is re-arranged
}
\description{
Transform skewed variables (aiming at they conform to a normal distribution) in .omv-files for the statistical spreadsheet 'jamovi' (\url{https://www.jamovi.org})
}
\details{
\itemize{
\item \code{varXfm} has to be a named list variable where the names can either indicate the type of transformation or the kind and degree of skewness that shall be
corrected. For the type of transformation, the following names are valid: \code{posSqr}, \code{negSqr}, \code{posLog}, \code{negLog}, \code{posInv}, \code{negInv}; where the second
part of the name indicates the transformation to be carried out: \code{...Sqr} - square root, \code{...Log} - logarithm to the basis 10, \code{...Inv} - inversion, i.e.,
1 / original value), and where the first part of the name indicates whether the original value is used (\code{pos...}) or whether the original value is
subtracted from the maximum value of that variable (\code{neg...}; a constant of 1 is added to the maximum value for \code{...Log} and \code{...Inv} transformations).
For the degree and kind of skewness, the following names are valid: \code{mdrPos}, \code{strPos}, \code{svrPos}, \code{mdrNeg}, \code{strNeg}, \code{svrNeg} (degree: moderate, strong,
severe; kind: positive or negative).
\item The ellipsis-parameter (\code{...}) can be used to submit arguments / parameters to the functions that are used for reading and writing the data. By clicking
on the respective function under “See also”, you can get a more detailed overview over which parameters each of those functions take. The functions are:
\code{read_omv} and \code{write_omv} (for jamovi-files), \code{read.table} (for CSV / TSV files; using similar defaults as \code{read.csv} for CSV and \code{read.delim} for TSV
which both are based upon \code{read.table}), \code{load} (for .RData-files), \code{readRDS} (for .rds-files), \code{read_sav} (needs the R-package \code{haven}) or \code{read.spss}
(needs the R-package \code{foreign}) for SPSS-files, \code{read_dta} (\code{haven}) / \code{read.dta} (\code{foreign}) for Stata-files, \code{read_sas} (\code{haven}) for SAS-data-files,
and \code{read_xpt} (\code{haven}) / \code{read.xport} (\code{foreign}) for SAS-transport-files. If you would like to use \code{haven}, you may need to install it using
\code{install.packages("haven", dep = TRUE)}.
}
}
\examples{
# generate skewed variables
set.seed(335)
dtaInp <- data.frame(MP = rnorm(1000) * 1e-1 + rexp(1000, 2) * (1 - 1e-1),
                     MN = rnorm(1000) * 1e-1 - rexp(1000, 2) * (1 - 1e-1),
                     SP = rnorm(1000) * 1e-2 + rexp(1000, 2) * (1 - 1e-2),
                     SN = rnorm(1000) * 1e-2 - rexp(1000, 2) * (1 - 1e-2),
                     EP = rnorm(1000) * 1e-4 + rexp(1000, 2) * (1 - 1e-4),
                     EN = rnorm(1000) * 1e-4 - rexp(1000, 2) * (1 - 1e-4))
jmv::descriptives(data = dtaInp, skew = TRUE, sw = TRUE)

crrXfm <- list(posSqr = c("MP"), negSqr = c("MN"), posLog = c("MP", "SP"), negLog = c("SN"),
               posInv = c("MP", "SP", "EP"), negInv = c("EN"))
dtaOut <- jmvReadWrite::transform_vars_omv(dtaInp = dtaInp, varXfm = crrXfm)
jmv::descriptives(data = dtaOut, skew = TRUE, sw = TRUE)

crrXfm <- list(mdrPos = c("MP"), mdrNeg = c("MN"), strPos = c("SP"), strNeg = c("SN"),
               svrPos = c("EP"), svrNeg = c("EN"))
dtaOut <- jmvReadWrite::transform_vars_omv(dtaInp = dtaInp, varXfm = crrXfm)
jmv::descriptives(data = dtaOut, skew = TRUE, sw = TRUE)


}
\seealso{
\code{transform_vars_omv} internally uses the following functions for reading and writing data files in different formats: \code{\link[=read_omv]{read_omv()}} and
\code{\link[=write_omv]{write_omv()}} for jamovi-files, \code{\link[utils:read.table]{utils::read.table()}} for CSV / TSV files, \code{\link[=load]{load()}} for reading .RData-files, \code{\link[=readRDS]{readRDS()}} for .rds-files,
\code{\link[haven:read_spss]{haven::read_sav()}} or \code{\link[foreign:read.spss]{foreign::read.spss()}} for SPSS-files, \code{\link[haven:read_dta]{haven::read_dta()}} or \code{\link[foreign:read.dta]{foreign::read.dta()}} for Stata-files, \code{\link[haven:read_sas]{haven::read_sas()}} for
SAS-data-files, and \code{\link[haven:read_xpt]{haven::read_xpt()}} or \code{\link[foreign:read.xport]{foreign::read.xport()}} for SAS-transport-files.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merge_rows_omv.R
\name{merge_rows_omv}
\alias{merge_rows_omv}
\title{Merges two .omv-files for the statistical spreadsheet 'jamovi' (\url{https://www.jamovi.org}) by adding the content of the second, etc.  file(s) as rows to the
first file}
\usage{
merge_rows_omv(
  dtaInp = NULL,
  fleOut = "",
  typMrg = c("all", "common"),
  colInd = FALSE,
  rstRwN = TRUE,
  rmvDpl = FALSE,
  varSrt = c(),
  usePkg = c("foreign", "haven"),
  selSet = "",
  ...
)
}
\arguments{
\item{dtaInp}{Either a data frame (with the attribute "fleInp" containing the files to merge) or vector with the names of the input files (including the
path, if required; "FILENAME.ext"; default: NULL); files can be of any supported file type, see Details below}

\item{fleOut}{Name of the data file to be written (including the path, if required; "FILE_OUT.omv"; default: ""); if empty, the resulting data frame is
returned instead}

\item{typMrg}{Type of merging operation: "all" (default) or  "common"; see also Details}

\item{colInd}{Add a column with an indicator (the basename of the file minus the extension) marking from which input data set the respective rows are coming
(default: FALSE)}

\item{rstRwN}{Reset row names (i.e., do not keep the row names of the original input data sets but number them consecutively - one to the row number of all
input data sets added up; default: TRUE)}

\item{rmvDpl}{Remove duplicated rows (i.e., rows with the same content as a previous row in all columns; default: FALSE)}

\item{varSrt}{Variable(s) that are used to sort the data frame (see Details; if empty, the order after merging is kept; default: c())}

\item{usePkg}{Name of the package: "foreign" or "haven" that shall be used to read SPSS, Stata and SAS files; "foreign" is the default (it comes with
base R), but "haven" is newer and more comprehensive}

\item{selSet}{Name of the data set that is to be selected from the workspace (only applies when reading .RData-files)}

\item{...}{Additional arguments passed on to methods; see Details below}
}
\value{
a data frame (only returned if \code{fleOut} is empty) where the rows of all input data sets (given in the \code{dtaInp}-argument) are concatenated
}
\description{
Merges two .omv-files for the statistical spreadsheet 'jamovi' (\url{https://www.jamovi.org}) by adding the content of the second, etc.  file(s) as rows to the
first file
}
\details{
\itemize{
\item Using data frames with the input parameter \code{dtaInp} is primarily thought to be used when calling \code{merge_cols_omv} from the jamovi-modules \code{jTransform} and
\code{Rj}. For the use in R, it is strongly recommended to use a character vector with the file names instead.
\item There are four different types of merging operations (defined via \code{typMrg}): "all" keeps all existing variables / columns that are contained in any of the
input data sets and fills them up with NA where the variable / column doesn't exist in an input data set. "common" only keeps the variables / columns that
are common to all input data sets (i.e., that are contained in all data sets).
\item \code{varSrt} can be either a character or a character vector (with one or more variables respectively). The sorting order for a particular variable can be
inverted with preceding the variable name with "-". Please note that this doesn't make sense and hence throws a warning for certain variable types (e.g.,
factors).
\item The ellipsis-parameter (\code{...}) can be used to submit arguments / parameters to the functions that are used for merging or reading the data. By clicking on
the respective function under “See also”, you can get a more detailed overview over which parameters each of those functions take.
\item Adding columns uses \code{rbind} (with some further operation, adding missing columns (filled with NAs), if \code{typMrg} is "all").
\item The functions for reading and writing the data are: \code{read_omv} and \code{write_omv} (for jamovi-files), \code{read.table} (for CSV / TSV files; using similar
defaults as \code{read.csv} for CSV and \code{read.delim} for TSV which both are based upon \code{read.table}), \code{load} (for .RData-files), \code{readRDS} (for .rds-files),
\code{read_sav} (needs R-package \code{haven}) or \code{read.spss} (needs R-package \code{foreign}) for SPSS-files, \code{read_dta} (\code{haven}) / \code{read.dta} (\code{foreign}) for
Stata-files, \code{read_sas} (\code{haven}) for SAS-data-files, and \code{read_xpt} (\code{haven}) / \code{read.xport} (\code{foreign}) for SAS-transport-files. If you would like to
use \code{haven}, you may need to install it using \code{install.packages("haven", dep = TRUE)}.
}
}
\examples{
dtaInp <- jmvReadWrite::bfi_sample2
nmeInp <- paste0(tempfile(), "_", 1:3, ".rds")
nmeOut <- tempfile(fileext = ".omv")
for (i in seq_along(nmeInp)) saveRDS(dtaInp[-i - 1], nmeInp[i])
# save dtaInp three times (i.e., the length of nmeInp), removing one data columns in
# each data set (for demonstration purposes, A1 in the first, A2 in the second, ...)
jmvReadWrite::merge_rows_omv(dtaInp = nmeInp, fleOut = nmeOut, colInd = TRUE)
cat(file.info(nmeOut)$size)
# -> 10767 (size may differ on different OSes)
dtaOut <- jmvReadWrite::read_omv(nmeOut, sveAtt = FALSE)
unlink(nmeOut)
# read the data set where the three original datasets were added as rows and show
# the variable names
cat(names(dtaInp))
cat(names(dtaOut))
# compared to the input data set, we have the same variable names; fleInd (switched
# on by colInd = TRUE and showing from which data set the rows are coming from) is
# new and A1 is moved to the end of the list (the "original" order of variables may
# not always be preserved and columns missing from at least one of the input data
# sets may be added at the end)
cat(dim(dtaInp), dim(dtaOut))
# the first dimension of the data sets (rows) is now three times of that of the input
# data set (250 -> 750), the second dimension (columns / variables) is increased by 1
# (for "fleInd")

jmvReadWrite::merge_rows_omv(dtaInp = nmeInp, fleOut = nmeOut, typMrg = "common")
# the argument typMrg = "common" removes the columns that are not present in all of
# the input data sets (i.e., A1, A2, A3)
dtaOut <- jmvReadWrite::read_omv(nmeOut, sveAtt = FALSE)
unlink(nmeOut)
# read the data set where the three original datasets were added as rows and show
# the variable names
cat(names(dtaInp))
cat(names(dtaOut))
# compared to the input data set, the variables that were missing in at least one
# data set (i.e., "A1", "A2" and "A3") are removed
cat(dim(dtaInp), dim(dtaOut))
# the first dimension of the data sets (rows) is now three times of that of the
# input data set (250 -> 750), the second dimension (columns / variables) is
# reduced by 3 (i.e., "A1", "A2", "A3")

unlink(nmeInp)


}
\seealso{
\code{merge_rows_omv} internally uses the following functions: Adding columns uses \code{\link[=rbind]{rbind()}}. For reading and writing data files in different formats:
\code{\link[=read_omv]{read_omv()}} and \code{\link[=write_omv]{write_omv()}} for jamovi-files, \code{\link[utils:read.table]{utils::read.table()}} for CSV / TSV files, \code{\link[=load]{load()}} for reading .RData-files,
\code{\link[=readRDS]{readRDS()}} for .rds-files, \code{\link[haven:read_spss]{haven::read_sav()}} or \code{\link[foreign:read.spss]{foreign::read.spss()}} for SPSS-files, \code{\link[haven:read_dta]{haven::read_dta()}} or \code{\link[foreign:read.dta]{foreign::read.dta()}} for Stata-files,
\code{\link[haven:read_sas]{haven::read_sas()}} for SAS-data-files, and \code{\link[haven:read_xpt]{haven::read_xpt()}} or \code{\link[foreign:read.xport]{foreign::read.xport()}} for SAS-transport-files.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/describe_omv.R
\name{describe_omv}
\alias{describe_omv}
\title{Adds a title and a description for a data set stored as .omv-file for the statistical spreadsheet 'jamovi' (\url{https://www.jamovi.org})}
\usage{
describe_omv(
  dtaInp = NULL,
  fleOut = "",
  dtaTtl = c(),
  dtaDsc = c(),
  lngDsc = "EN",
  usePkg = c("foreign", "haven"),
  selSet = "",
  ...
)
}
\arguments{
\item{dtaInp}{Either a data frame or the name of a data file to be read (including the path, if required; "FILENAME.ext"; default: NULL); files can be of
any supported file type, see Details below}

\item{fleOut}{Name of the data file to be written (including the path, if required; "FILE_OUT.omv"; default: ""); if empty, the resulting data frame is
returned instead}

\item{dtaTtl}{Character vector with a title to be added to the data set (see Details; default: "")}

\item{dtaDsc}{Description of the data set, either as character vector (HTML-formatted) or as named list with the entries "description", "variables",
"references", and "license" (see Details; default: "")}

\item{lngDsc}{Language of the description (localizes the description components: "Description", "Variables", "References", and "License";
default: "EN")}

\item{usePkg}{Name of the package: "foreign" or "haven" that shall be used to read SPSS, Stata and SAS files; "foreign" is the default (it comes with
base R), but "haven" is newer and more comprehensive}

\item{selSet}{Name of the data set that is to be selected from the workspace (only applies when reading .RData-files)}

\item{...}{Additional arguments passed on to methods; see Details below}
}
\value{
a data frame (only returned if \code{fleOut} is empty) where a description and a title are added to the input data
}
\description{
Adds a title and a description for a data set stored as .omv-file for the statistical spreadsheet 'jamovi' (\url{https://www.jamovi.org})
}
\details{
\itemize{
\item The aim of this function is to add a title and a data set description to jamovi data files. Two typical use cases would be (1) to help creating data sets
to be used in teaching (i.e., either creating or using data sets in R, and afterwards adding a description to those), and (2) to provide "properly
described" data when publishing in a repository such as the OSF).
\item NB: The data set should not contain any existing analyses. These will be overwritten (a  warning is issued informing you about that).
\item \code{dtaTtl} is a title for the dataset (at the top of the results output, i.e., that title which initially is "Results" when you create a new data set in
jamovi).
\item \code{dtaDsc} can either be a character vector (with length = 1) containing HTML-formatted text that describes the data set (see \code{chrDsc} in the examples for
HTML tags that are currently implemented; putting "unformatted" text is not a problem, but then the result is just plain text without formatting).
Alternatively, \code{dtaDcs} can be a named list with the entries \code{description}, \code{variables}, \code{references}, \code{license}. All entries except from \code{variables}
contain character vectors (length = 1); \code{variables} shall be a named list with the variable name as name and a description what the variable contains as
entry. \code{description} and \code{variables} must be given, \code{references} and \code{license} can be left blank (""; but the names must be present in the list). An
example for both a named list with a description (\code{lstDsc}), as well as a character vector with all HTML tags that are implemented (\code{chrDsc}) can be found
in the examples below.
\item The ellipsis-parameter (\code{...}) can be used to submit arguments / parameters to the functions that are used for reading and writing the data. By clicking
on the respective function under "See also", you can get a more detailed overview over which parameters each of those functions take. The functions are:
\code{read_omv} and \code{write_omv} (for jamovi-files), \code{read.table} (for CSV / TSV files; using similar defaults as \code{read.csv} for CSV and \code{read.delim} for TSV
which both are based upon \code{read.table}), \code{load} (for .RData-files), \code{readRDS} (for .rds-files), \code{read_sav} (needs the R-package \code{haven}) or \code{read.spss}
(needs the R-package \code{foreign}) for SPSS-files, \code{read_dta} (\code{haven}) / \code{read.dta} (\code{foreign}) for Stata-files, \code{read_sas} (\code{haven}) for SAS-data-files,
and \code{read_xpt} (\code{haven}) / \code{read.xport} (\code{foreign}) for SAS-transport-files. If you would like to use \code{haven}, you may need to install it using
\code{install.packages("haven", dep = TRUE)}.
}
}
\examples{
dtaFrm <- jmvReadWrite::ToothGrowth[, c("len", "supp", "dose")]
nmeOut <- tempfile(fileext = ".omv")

# the paste's underneath are only for readability (without them, the vignettes are misformatted)
lstDsc <- list(description = paste("The response is the length of odontoblasts (cells responsible",
                                   "for tooth growth) in 60 guinea pigs. Each animal received one",
                                   "of three dose levels of vitamin C (0.5, 1, and 2 mg / day) by",
                                   "one of two delivery methods, orange juice or ascorbic acid (a",
                                   "form of vitamin C and coded as VC)."),
               variables = list(len  = "Tooth length",
                                supp = "Supplement type (VC or OJ)",
                                dose = "Dose (in milligrams / day"),
               references = paste("Crampton, E. W. (1947). The growth of the odontoblast of the",
                                  "incisor teeth as a criterion of vitamin C intake of the guinea",
                                  "pig. <em>The Journal of Nutrition, 33</em>(5), 491-504.",
                                  "https://doi.org/10.1093/jn/33.5.491"),
               license = "")
jmvReadWrite::describe_omv(dtaInp = dtaFrm, fleOut = nmeOut, dtaTtl = "ToothGrowth",
  dtaDsc = lstDsc)
# don't include the unlink, if you copy the code and want to look at the resulting output file
unlink(nmeOut)

# the code underneath should cover all formatting options jamovi is able to use (paste0 is only
# for readability)
chrDsc <- paste0("<p><strong>Trial - all formattings:</strong><br/><strong>bold</strong><br/>",
                 "<strong><em>bold, italics</em></strong><br/><em>italics</em><br/><u>underlined",
                 "</u><br/><s>strikethrough</s><br/>C<sub>2</sub>H<sub>5</sub>OH<br/>R<sup>2",
                 "</sup><br/><span style=\"background-color:#e60000\">background colour: red",
                 "</span><br/><span style=\"color:#e60000\">foreground color: red",
                 "</span></p><p class=\"ql-align-center\">centered</p><p class=\"ql-align-right\">",
                 "right</p><p class=\"ql-align-justify\">justify justify justify justify justify ",
                 "justify justify justify justify justify justify justify justify justify justify ",
                 "justify justify justify justify justify justify justify justify justify justify",
                 "</p><p><br/></p><ol><li>numbered list</li><li>numbered list</li></ol><p><br/>",
                 "</p><ul><li>bullet point</li><li>bullet point</li></ul><p class=\"ql-indent-1\">",
                 "indented once</p><p class=\"ql-indent-2\">indented twice</p><p ",
                 "class=\"ql-indent-1\">indented once</p><p>Formula: <span class=\"ql-formula\">",
                 "e=mc^2</span></p><pre>Preformatted</pre><p>normal again</p><h2>Heading</h2>")
jmvReadWrite::describe_omv(dtaInp = dtaFrm, fleOut = nmeOut, dtaTtl = "ToothGrowth",
  dtaDsc = chrDsc)
unlink(nmeOut)


}
\seealso{
\code{describe_omv} internally uses the following functions for reading and writing data files in different formats: \code{\link[=read_omv]{read_omv()}} and
\code{\link[=write_omv]{write_omv()}} for jamovi-files, \code{\link[utils:read.table]{utils::read.table()}} for CSV / TSV files, \code{\link[=load]{load()}} for reading .RData-files, \code{\link[=readRDS]{readRDS()}} for .rds-files,
\code{\link[haven:read_spss]{haven::read_sav()}} or \code{\link[foreign:read.spss]{foreign::read.spss()}} for SPSS-files, \code{\link[haven:read_dta]{haven::read_dta()}} or \code{\link[foreign:read.dta]{foreign::read.dta()}} for Stata-files, \code{\link[haven:read_sas]{haven::read_sas()}} for
SAS-data-files, and \code{\link[haven:read_xpt]{haven::read_xpt()}} or \code{\link[foreign:read.xport]{foreign::read.xport()}} for SAS-transport-files.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/label_vars_omv.R
\name{label_vars_omv}
\alias{label_vars_omv}
\title{Label columns / variables in .omv-files for the statistical spreadsheet 'jamovi' (\url{https://www.jamovi.org})}
\usage{
label_vars_omv(
  dtaInp = NULL,
  fleOut = "",
  varLbl = NULL,
  psvAnl = FALSE,
  usePkg = c("foreign", "haven"),
  selSet = "",
  ...
)
}
\arguments{
\item{dtaInp}{Either a data frame or the name of a data file to be read (including the path, if required; "FILENAME.ext"; default: NULL); files can be of
any supported file type, see Details below}

\item{fleOut}{Name of the data file to be written (including the path, if required; "FILE_OUT.omv"; default: ""); if empty, the resulting data frame is
returned instead}

\item{varLbl}{Variable (default: NULL) containing either a character (a file name; the file must contain two columns one with variable names, the other with
the labels), a data frame (one column the variable names, the other the labels), or a character vector (with the same length as the data set,
containing the variable labels). See Details for more information.}

\item{psvAnl}{Whether analyses that are contained in the input file shall be transferred to the output file (default: FALSE)}

\item{usePkg}{Name of the package: "foreign" or "haven" that shall be used to read SPSS, Stata and SAS files; "foreign" is the default (it comes with
base R), but "haven" is newer and more comprehensive}

\item{selSet}{Name of the data set that is to be selected from the workspace (only applies when reading .RData-files)}

\item{...}{Additional arguments passed on to methods; see Details below}
}
\value{
a data frame (only returned if \code{fleOut} is empty) where the order of variables / columns of the input data set is re-arranged
}
\description{
Label columns / variables in .omv-files for the statistical spreadsheet 'jamovi' (\url{https://www.jamovi.org})
}
\details{
\itemize{
\item \code{varLbl} can be either (1) a character with a file name to read (the file must contain to columns, one with the variable names, the other with the
variable labels); (2) a data frame with two columns (one with the variable names, the other with the variable labels), or (3) a character vector
containing the variable labels (with a length equal to the number of variables in the input data set).
\item The ellipsis-parameter (\code{...}) can be used to submit arguments / parameters to the functions that are used for reading and writing the data. By clicking
on the respective function under “See also”, you can get a more detailed overview over which parameters each of those functions take. The functions are:
\code{read_omv} and \code{write_omv} (for jamovi-files), \code{read.table} (for CSV / TSV files; using similar defaults as \code{read.csv} for CSV and \code{read.delim} for TSV
which both are based upon \code{read.table}), \code{load} (for .RData-files), \code{readRDS} (for .rds-files), \code{read_sav} (needs the R-package \code{haven}) or \code{read.spss}
(needs the R-package \code{foreign}) for SPSS-files, \code{read_dta} (\code{haven}) / \code{read.dta} (\code{foreign}) for Stata-files, \code{read_sas} (\code{haven}) for SAS-data-files,
and \code{read_xpt} (\code{haven}) / \code{read.xport} (\code{foreign}) for SAS-transport-files. If you would like to use \code{haven}, you may need to install it using
\code{install.packages("haven", dep = TRUE)}.
}
}
\examples{
# use one of the data files included in the package, but only the first 28 columns
# (the latter columns contain data for testing calculations, etc.)
nmeInp <- system.file("extdata", "bfi_sample.omv", package = "jmvReadWrite")
dtaInp <- jmvReadWrite::read_omv(nmeInp)[1:28]
nmeOut <- tempfile(fileext = ".omv")
# in the original file, the variable labels – attr(*, "jmv-desc") - are empty
lapply(dtaInp, attr, "jmv-desc")
# the definition of the variable labels can be read from a file with two columns,
# the first containing the variable name, the second the variable labels
# you can easily create such a file in Excel and save it as CSV
# if your CSV contains column names (e.g., varNme and varLbl) in the first row are they ignored
lblFle <- system.file("extdata", "label_example.csv", package = "jmvReadWrite")
lblDtF <- utils::read.csv(lblFle, header = FALSE)
str(lblDtF)

# there are three options to give the varLbl parameter:
# (1) as file name, ...
jmvReadWrite::label_vars_omv(dtaInp = dtaInp, fleOut = nmeOut, varLbl = lblFle)
lapply(jmvReadWrite::read_omv(nmeOut), attr, "jmv-desc")
unlink(nmeOut)

# (2) as data frame (using lblDtF from above), or ...
jmvReadWrite::label_vars_omv(dtaInp = dtaInp, fleOut = nmeOut, varLbl = lblDtF)
lapply(jmvReadWrite::read_omv(nmeOut), attr, "jmv-desc")
unlink(nmeOut)

# (3) as character vector (with the same length as there are columns in the input data set)
lblChr <- lblDtF[[2]]
head(lblChr)
jmvReadWrite::label_vars_omv(dtaInp = dtaInp, fleOut = nmeOut, varLbl = lblChr)
lapply(jmvReadWrite::read_omv(nmeOut), attr, "jmv-desc")
unlink(nmeOut)


}
\seealso{
\code{label_vars_omv} internally uses the following functions for reading and writing data files in different formats: \code{\link[=read_omv]{read_omv()}} and
\code{\link[=write_omv]{write_omv()}} for jamovi-files, \code{\link[utils:read.table]{utils::read.table()}} for CSV / TSV files, \code{\link[=load]{load()}} for reading .RData-files, \code{\link[=readRDS]{readRDS()}} for .rds-files,
\code{\link[haven:read_spss]{haven::read_sav()}} or \code{\link[foreign:read.spss]{foreign::read.spss()}} for SPSS-files, \code{\link[haven:read_dta]{haven::read_dta()}} or \code{\link[foreign:read.dta]{foreign::read.dta()}} for Stata-files, \code{\link[haven:read_sas]{haven::read_sas()}} for
SAS-data-files, and \code{\link[haven:read_xpt]{haven::read_xpt()}} or \code{\link[foreign:read.xport]{foreign::read.xport()}} for SAS-transport-files.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/combine_cols_omv.R
\name{combine_cols_omv}
\alias{combine_cols_omv}
\title{Combines pairs of columns from a raw data matrix in .omv-files for the statistical spreadsheet
'jamovi' (\url{https://www.jamovi.org})}
\usage{
combine_cols_omv(
  dtaInp = NULL,
  fleOut = "",
  varPrs = list(),
  mdeCmb = c("none", "first", "second"),
  psvAnl = FALSE,
  usePkg = c("foreign", "haven"),
  selSet = "",
  ...
)
}
\arguments{
\item{dtaInp}{Either a data frame or the name of a data file to be read (including the path,
if required; "FILENAME.ext"; default: NULL); files can be of any supported file
type, see Details below.}

\item{fleOut}{Name of the data file to be written (including the path, if required;
"FILE_OUT.omv"; default: ""); if empty, the resulting data frame is returned
instead.}

\item{varPrs}{Definition of variable pairs; a list containing either list(s) or character
vector(s) with the names of pairs of variables to be combined (default: list()).}

\item{mdeCmb}{Mode of combining the variables when conflicting values occur, either "none",
"first", or "second" (default: "none"), see Details below.}

\item{psvAnl}{Whether analyses that are contained in the input file shall be transferred to the
output file (TRUE / FALSE; default: FALSE)}

\item{usePkg}{Name of the package: "foreign" or "haven" that shall be used to read SPSS, Stata,
and SAS files; "foreign" is the default (it comes with base R), but "haven" is
newer and more comprehensive.}

\item{selSet}{Name of the data set that is to be selected from the workspace (only applies when
reading .RData-files)}

\item{...}{Additional arguments passed on to methods; see Details below.}
}
\value{
a data frame containing the column pairs given in \code{varPrs} combined and the original
columns removed
}
\description{
Combines pairs of columns from a raw data matrix in .omv-files for the statistical spreadsheet
'jamovi' (\url{https://www.jamovi.org})
}
\details{
\itemize{
\item The need to combine two columns into one is quite common after merging columns or rows (using,
e.g., merge_cols_omv or merge_rows_omv). \code{varPrs} defines the variable pairs to be combined.
It is a list containing the pairs of variables to be combined, either as list or as character
vector; e.g., list(c("A", "B"), c("C", "D")) or list(list("A", "B"), list("C", "D")). \code{mdeCmb}
defines what to to if values in the first and the second variable of a variable pair contain
conflicting / different values: "none" does not merge the variables (and instead throws an
error), "first" makes that the values from the first variable of each pair are taken if the
values are conflicting, and "second" use the values from the second variable of each pair in
case of conflicts.
\item The ellipsis-parameter (\code{...}) can be used to submit arguments / parameters to the functions
that are used for reading and writing the data. By clicking on the respective function under
“See also”, you can get a more detailed overview over which parameters each of those functions
take. The functions are: \code{read_omv} and \code{write_omv} (for jamovi-files), \code{read.table} (for CSV
/ TSV files; using similar defaults as \code{read.csv} for CSV and \code{read.delim} for TSV which both
are based upon \code{read.table}), \code{load} (for .RData-files), \code{readRDS} (for .rds-files),
\code{read_sav} (needs the R-package \code{haven}) or \code{read.spss} (needs the R-package \code{foreign}) for
SPSS-files, \code{read_dta} (\code{haven}) / \code{read.dta} (\code{foreign}) for Stata-files, \code{read_sas}
(\code{haven}) for SAS-data-files, and \code{read_xpt} (\code{haven}) / \code{read.xport} (\code{foreign}) for
SAS-transport-files. If you would like to use \code{haven}, you may need to install it using
\code{install.packages("haven", dep = TRUE)}.
}
}
\examples{
dtaInp <- jmvReadWrite::bfi_sample2
# create a new column (A1_1) containing a subset of the values in the original variable
# whereas those lines are replaced with NAs
set.seed(1)
selRow <- rnorm(nrow(dtaInp)) < 0
dtaInp[selRow,  "A1_1"] <- dtaInp[selRow, "A1"]
dtaInp[selRow,  "A1"]   <- NA
head(dtaInp[, c("A1", "A1_1")])
dtaOut <- combine_cols_omv(dtaInp, varPrs = list(c("A1", "A1_1")))
# show the differences before and after combining the values in the columns and ensure
# that all values are the same as in the original data set
dtaInp[, "A1"]
dtaOut[, "A1"]
all(dtaOut[, "A1"] == jmvReadWrite::bfi_sample2[, "A1"])

# create a new column, containing values that are different from the original variable
dtaInp <- jmvReadWrite::bfi_sample2
dtaInp[selRow,  "A1_1"] <- dtaInp[selRow,  "A1"] + 1
# [1] if mdeCmb is "none" (or if mdeCmb is not given - "none" is the default) an error would be
# thrown (therefore the next line is commented out)
# dtaOut <- combine_cols_omv(dtaInp, varPrs = list(c("A1", "A1_1")), mdeCmb = "none")
# [2] if mdeCmb is "first", missing values are replaced and values from the first column ("A1")
# take precedence if the values are unequal
dtaOut <- combine_cols_omv(dtaInp, varPrs = list(c("A1", "A1_1")), mdeCmb = "first")
head(cbind(dtaOut[, "A1"], dtaInp[, c("A1", "A1_1")]))
# [3] if mdeCmb is "second", missing values are replaced and values from the second column
# ("A1_1") take precedence if the values are unequal
dtaOut <- combine_cols_omv(dtaInp, varPrs = list(c("A1", "A1_1")), mdeCmb = "second")
head(cbind(dtaOut[, "A1"], dtaInp[, c("A1", "A1_1")]))


}
\seealso{
\code{combine_cols_omv} uses the following functions for reading and writing data files in
different formats: \code{\link[=read_omv]{read_omv()}} and \code{\link[=write_omv]{write_omv()}} for
jamovi-files, \code{\link[utils:read.table]{utils::read.table()}} for CSV / TSV files, \code{\link[=load]{load()}} for reading .RData-files,
\code{\link[=readRDS]{readRDS()}} for .rds-files, \code{\link[haven:read_spss]{haven::read_sav()}} or \code{\link[foreign:read.spss]{foreign::read.spss()}} for SPSS-files,
\code{\link[haven:read_dta]{haven::read_dta()}} or \code{\link[foreign:read.dta]{foreign::read.dta()}} for Stata-files, \code{\link[haven:read_sas]{haven::read_sas()}} for
SAS-data-files, and \code{\link[haven:read_xpt]{haven::read_xpt()}} or \code{\link[foreign:read.xport]{foreign::read.xport()}} for SAS-transport-files.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/wide2long_omv.R
\name{wide2long_omv}
\alias{wide2long_omv}
\title{Converts .omv-files for the statistical spreadsheet 'jamovi' (\url{https://www.jamovi.org}) from wide to long format}
\usage{
wide2long_omv(
  dtaInp = NULL,
  fleOut = "",
  varLst = c(),
  varExc = c(),
  varID = NULL,
  varTme = "cond",
  varSep = "_",
  varOrd = TRUE,
  varSrt = c(),
  excLvl = NULL,
  usePkg = c("foreign", "haven"),
  selSet = "",
  ...
)
}
\arguments{
\item{dtaInp}{Either a data frame or the name of a data file to be read (including the path, if required; "FILENAME.ext"; default: NULL); files can be of
any supported file type, see Details below}

\item{fleOut}{Name of the data file to be written (including the path, if required; "FILE_OUT.omv"; default: ""); if empty, the resulting data frame is
returned instead}

\item{varLst}{List / set of variables that are to be transformed into single (time-varying) variables in long format (default: c())}

\item{varExc}{Name of the variable(s) should be excluded from the transformation, typically this will be between-subject-variable(s) (default: c())}

\item{varID}{Name(s) of one or more variables that (is created to) identify the same group / individual (if empty, "ID" is added with row numbers
identifying cases; default: NULL)}

\item{varTme}{Name of the variable (or vector with variable names) that (is created to) differentiate multiple records from the same group / individual #'               (default: "cond"; if required, a counter is added for each time-varying part)}

\item{varSep}{Character that separates the variables in varLst into a time-varying part and a part that forms the variable name in long format ("\emph{" in
"VAR_1", "VAR_2", default: "}")}

\item{varOrd}{Whether to arrange the variables before the transformation, so that they are in accordance with the different split levels (default: TRUE)}

\item{varSrt}{Variable(s) that are used to sort the data frame (see Details; if empty, the order returned from reshape is kept; default: c())}

\item{excLvl}{Integer (or vector of integers) determining which parts of the variable names in varLst shall not be transformed (default: NULL), see Details
below}

\item{usePkg}{Name of the package: "foreign" or "haven" that shall be used to read SPSS, Stata and SAS files; "foreign" is the default (it comes with
base R), but "haven" is newer and more comprehensive}

\item{selSet}{Name of the data set that is to be selected from the workspace (only applies when reading .RData-files)}

\item{...}{Additional arguments passed on to methods; see Details below}
}
\value{
a data frame (only returned if \code{fleOut} is empty) where the input data set is converted from wide to long format
}
\description{
Converts .omv-files for the statistical spreadsheet 'jamovi' (\url{https://www.jamovi.org}) from wide to long format
}
\details{
\itemize{
\item If \code{varLst} is empty, it is tried to generate it using all variables in the data frame except those defined by \code{varExc} and \code{varID}. The variable(s) in
\code{varID} need to be unique identifiers (in the original dataset), those in \code{varExc} don't have this requirement. It is recommended that the variable names
in \code{varExc} and \code{varID} should not contain the variable separator (defined in \code{varSep}; default: "_").
\item \code{varOrd} determines whether the variables are rearranged to match the order of split levels. Consider the \code{varLst} X_1, Y_1, X_2, Y_2. If \code{varOrd} were
set to FALSE, the original order would be preserved and the second part of the variable name (1, 2, ...) would become condition 1, and the first part
condition 2. In most cases, leaving \code{varOrd} set to TRUE is recommended.
\item \code{varSrt} can be either a character or a character vector (with one or more variables respectively). The sorting order for a particular variable can be
inverted with preceding the variable name with "-". Please note that this doesn't make sense and hence throws a warning for certain variable types (e.g.,
factors).
\item \code{exclLvl} points to a part of the variable names in \code{varLst} to be excluded. For example, if the variable name is \code{PART1_PART2_PART3} (split at _), then
\code{excLvl} = 1 would exclude PART1 from the transformation. Quite often, one has more that one variable of a particular type (e.g., responses, reaction
times, etc.). Those would typically be the first part of each variable name in \code{varLst} (the conditions then being PART2, PART3, and so on). \code{excLvl} = 1
would exclude those variable types / categories from being transformed into long (i.e., they would be kept as separate columns).
\item The ellipsis-parameter (\code{...}) can be used to submit arguments / parameters to the functions that are used for transforming or reading the data. By
clicking on the respective function under “See also”, you can get a more detailed overview over which parameters each of those functions take.
\item The transformation from long to wide uses \code{reshape}: \code{varID} matches (~) \code{idvar} in \code{reshape}, \code{varTme} ~ \code{timevar}, \code{varLst} ~ \code{varying}, and \code{varSep} ~
\code{sep}. The help for \code{reshape} is very explanatory, click on the link under “See also” to access it, particularly what is explained under “Details”.
\item The functions for reading and writing the data are: \code{read_omv} and \code{write_omv} (for jamovi-files), \code{read.table} (for CSV / TSV files; using similar
defaults as \code{read.csv} for CSV and \code{read.delim} for TSV which both are based upon \code{read.table}), \code{load} (for .RData-files), \code{readRDS} (for .rds-files),
\code{read_sav} (needs R-package \code{haven}) or \code{read.spss} (needs R-package \code{foreign}) for SPSS-files, \code{read_dta} (\code{haven}) / \code{read.dta} (\code{foreign}) for
Stata-files, \code{read_sas} (\code{haven}) for SAS-data-files, and \code{read_xpt} (\code{haven}) / \code{read.xport} (\code{foreign}) for SAS-transport-files. If you would like to
use \code{haven}, you may need to install it using \code{install.packages("haven", dep = TRUE)}.
}
}
\examples{
\dontrun{
# generate a test dataframe with 100 (imaginary) participants / units of
# observation (ID), and 8 repeated measurements of variable (X_1, X_2, ...)
dtaInp <- cbind(data.frame(ID = as.character(seq(1:100))),
                stats::setNames(
                    as.data.frame(matrix(runif(800, -10, 10), nrow = 100)),
                    paste0("X_", 1:8)))
cat(str(dtaInp))
# 'data.frame':	100 obs. of  9 variables:
#  $ ID : chr  "1" "2" "3" "4" ...
#  $ X_1: num  ...
#  $ X_2: num  ...
#  $ X_3: num  ...
#  $ X_4: num  ...
#  $ X_5: num  ...
#  $ X_6: num  ...
#  $ X_7: num  ...
#  $ X_8: num  ...
# this data set is stored as (temporary) RDS-file and later processed by wide2long
nmeInp <- tempfile(fileext = ".rds")
nmeOut <- tempfile(fileext = ".omv")
saveRDS(dtaInp, nmeInp)
jmvReadWrite::wide2long_omv(dtaInp = nmeInp, fleOut = nmeOut, varID = "ID",
    varTme = "measure", varLst = setdiff(names(dtaInp), "ID"),
    varSrt = c("ID", "measure"))
# it is required to give at least the arguments dtaInp (if dtaInp is a data frame,
# fleOut needs to be provided too) and varID
# "reshape" then assigns all variables expect the variable defined by varID to
# varLst (but throws a warning)
# varSrt enforces sorting the data set after the transformation (sorted, the
# measurements within one person come after another; unsorted all measurements
# for one repetition would come after another)

# check whether the file was created and its size
cat(list.files(dirname(nmeOut), basename(nmeOut)))
# -> "file[...].omv" ([...] contains a random combination of numbers / characters
cat(file.info(nmeOut)$size)
# -> 6939 (approximate size; size may differ in every run [in dependence of how
#          well the generated random data can be compressed])
cat(str(jmvReadWrite::read_omv(nmeOut, sveAtt = FALSE)))
# the data set is now transformed into long (and each the measurements is now
# indicated by the "measure")
# 'data.frame':	800 obs. of  3 variables:
#  $ ID     : Factor w/ 100 levels "1","2","3","4",..: 1 1 1 1 1 1 1 1 2 2 ...
#   ..- attr(*, "missingValues")= list()
#  $ measure: Factor w/ 8 levels "1","2","3","4",..: 1 2 3 4 5 6 7 8 1 2 ...
#   ..- attr(*, "missingValues")= list()
#  $ X      : num  ...
#   ..- attr(*, "missingValues")= list()

unlink(nmeInp)
unlink(nmeOut)
}

}
\seealso{
\code{long2wide_omv} internally uses the following functions: The transformation from long to wide uses \code{\link[stats:reshape]{stats::reshape()}}. For reading and writing data
files in different formats: \code{\link[=read_omv]{read_omv()}} and \code{\link[=write_omv]{write_omv()}} for jamovi-files, \code{\link[utils:read.table]{utils::read.table()}} for CSV / TSV files,
\code{\link[=load]{load()}} for reading .RData-files, \code{\link[=readRDS]{readRDS()}} for .rds-files, \code{\link[haven:read_spss]{haven::read_sav()}} or \code{\link[foreign:read.spss]{foreign::read.spss()}} for SPSS-files, \code{\link[haven:read_dta]{haven::read_dta()}} or
\code{\link[foreign:read.dta]{foreign::read.dta()}} for Stata-files, \code{\link[haven:read_sas]{haven::read_sas()}} for SAS-data-files, and \code{\link[haven:read_xpt]{haven::read_xpt()}} or \code{\link[foreign:read.xport]{foreign::read.xport()}} for SAS-transport-files.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{ToothGrowth}
\alias{ToothGrowth}
\title{The Effect of Vitamin C on Tooth Growth in Guinea Pigs}
\format{
A data frame with 60 rows and 6 variables:
\describe{
\item{\code{ID}}{ID of the guinea pig}
\item{\code{supp}}{Supplement type (VC: Vitamin C or OJ: Orange juice)}
\item{\code{supp2}}{Transformation of the supplement type (factor to numerical: VC = 1; OJ = 2)}
\item{\code{dose}}{Dose in grams / day}
\item{\code{dose2}}{Dose in grams / day}
\item{\code{len}}{Tooth length}
\item{\code{logLen}}{Natural logarithm of the tooth length (len)}
}
}
\usage{
ToothGrowth
}
\description{
The Effect of Vitamin C on Tooth Growth in Guinea Pigs
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/aggregate_omv.R
\name{aggregate_omv}
\alias{aggregate_omv}
\title{Aggregates data from a data set / frame (in long format) and returns them as a .omv-file for the
statistical spreadsheet 'jamovi' (\url{https://www.jamovi.org})}
\usage{
aggregate_omv(
  dtaInp = NULL,
  fleOut = "",
  varAgg = c(),
  grpAgg = c(),
  clcN = FALSE,
  clcMss = FALSE,
  clcMn = FALSE,
  clcMdn = FALSE,
  clcMde = FALSE,
  clcSum = FALSE,
  clcSD = FALSE,
  clcVar = FALSE,
  clcRng = FALSE,
  clcMin = FALSE,
  clcMax = FALSE,
  clcIQR = FALSE,
  drpNA = TRUE,
  usePkg = c("foreign", "haven"),
  selSet = "",
  ...
)
}
\arguments{
\item{dtaInp}{Either a data frame or the name of a data file to be read (including the path,
if required; "FILENAME.ext"; default: NULL); files can be of any supported file
type, see Details below.}

\item{fleOut}{Name of the aggregated data set / file to be written (including the path, if
required; "FILE_OUT.omv"; default: ""); if empty, the resulting data frame is
returned instead.}

\item{varAgg}{A character vector (default: c()) with the names of the variables which shall be
aggregated.}

\item{grpAgg}{A character vector (default: c()) with the variables to group the aggregation
variable (\code{varAgg}) by. If no grouping variable is given, the aggregation happens
over the whole data set. If several grouping variables are given, the aggregation
happens for each step / possible combination of these variables. See Details for
more information.}

\item{clcN}{If TRUE, counts the number of valid values for each step / combination of values
in the grouping variable. The suffix appended `"_N"´ is appended to each variable
in the resulting data set.}

\item{clcMss}{If TRUE, counts the number of missing values for each step / combination of values
in the grouping variable. The suffix appended \code{"_Mss"} is appended to each
variable in the resulting data set.}

\item{clcMn}{If TRUE, calculates the mean for each step / combination of values in the grouping
variable. The suffix appended \code{"_Mn"} is appended to each variable in the
resulting data set.}

\item{clcMdn}{If TRUE, calculates the median for each step / combination of values in the
grouping variable. The suffix appended \code{"_Mdn"} is appended to each variable in
the resulting data set.}

\item{clcMde}{If TRUE, calculates the mode for each step / combination of values in the grouping
variable. The suffix appended \code{"_Mde"} is appended to each variable in the
resulting data set.}

\item{clcSum}{If TRUE, calculates the sum for each step / combination of values in the grouping
variable. The suffix appended \code{"_Sum"} is appended to each variable in the
resulting data set.}

\item{clcSD}{If TRUE, calculates the std. deviation for each step / combination of values in
the grouping variable. The suffix appended \code{"_SD"} is appended to each variable in
the resulting data set.}

\item{clcVar}{If TRUE, calculates the variance for each step / combination of values in the
grouping variable. The suffix appended \code{"_Var"} is appended to each variable in
the resulting data set.}

\item{clcRng}{If TRUE, calculates the range for each step / combination of values in the
grouping variable. The suffix appended \code{"_Rng"} is appended to each variable in
the resulting data set.}

\item{clcMin}{If TRUE, determines the minimum at each step / combination of values in the
grouping variable. The suffix appended \code{"_Min"} is appended to each variable in
the resulting data set.}

\item{clcMax}{If TRUE, determines the maximum at each step / combination of values in the
grouping variable. The suffix appended \code{"_Max"} is appended to each variable in
the resulting data set.}

\item{clcIQR}{If TRUE, calculates the IQR for each step / combination of values in the grouping
variable. The suffix appended \code{"_IQR"} is appended to each variable in the
resulting data set.}

\item{drpNA}{If TRUE (default: TRUE), NA values are removed before the aggregation, and the
aggregation is calculated using the valid values. If FALSE, the result would be NA
for any step / combination of values in the grouping variable that contains a NA
value.}

\item{usePkg}{Name of the package: "foreign" or "haven" that shall be used to read SPSS, Stata,
and SAS files; "foreign" is the default (it comes with base R), but "haven" is
newer and more comprehensive.}

\item{selSet}{Name of the data set that is to be selected from the workspace (only applies when
reading .RData-files)}

\item{...}{Additional arguments passed on to methods; see Details below.}
}
\value{
a data frame containing a symmetric matrix (only returned if \code{fleOut} is empty)
containing the distances between the variables / columns (clmDst == TRUE) or rows
(clmDst == FALSE)
}
\description{
Aggregates data from a data set / frame (in long format) and returns them as a .omv-file for the
statistical spreadsheet 'jamovi' (\url{https://www.jamovi.org})
}
\details{
\itemize{
\item \code{varAgg} must be a character vector containing the variables to be aggregated. From these
variables, a new data set is created where the values in the original data set are aggregated
for each possible combination of steps in the grouping variable(s) given in \code{grpAgg}. For
example, if one grouping variable were a participant ID, and another grouping variable were
measurement time points, the resulting new dataset would contain as many rows as the number of
participants multiplied by the number of measurement time points. For each variable in
\code{varAgg}, and each possible aggregation (\code{clcN}, \code{clcMn}, etc. if set to TRUE), a new column
would be generated in the resulting new data set.
\item \code{drpNA} determines whether NA should be dropped / removed before calculating an aggregation.
If set to FALSE, the result for a combination of steps in the grouping variable(s) where at
least one value is NA would be NA. If set to TRUE, only the values for any given combination
of steps in the grouping variable(s) that are not NA are considered for the calculation. For
\code{clcN} and \code{clcMss}, \code{drpNA} has no effect.
NB: If \code{drpNA} is set to TRUE, any row where any of the grouping variable(s) has the value NA
is excluded from the aggregation, if \code{drpNA} is set to FALSE and there is any row where any of
the group variables has the value NA, an error is thrown and no aggregation is carried out.
\item The ellipsis-parameter (\code{...}) can be used to submit arguments / parameters to the functions
that are used for reading and writing the data. By clicking on the respective function under
“See also”, you can get a more detailed overview over which parameters each of those functions
take. The functions are: \code{read_omv} and \code{write_omv} (for jamovi-files), \code{read.table} (for
CSV / TSV files; using similar defaults as \code{read.csv} for CSV and \code{read.delim} for TSV which
both are based upon \code{read.table}), \code{load} (for .RData-files), \code{readRDS} (for .rds-files),
\code{read_sav} (needs the R-package \code{haven}) or \code{read.spss} (needs the R-package \code{foreign}) for
SPSS-files, \code{read_dta} (\code{haven}) / \code{read.dta} (\code{foreign}) for Stata-files, \code{read_sas}
(\code{haven}) for SAS-data-files, and \code{read_xpt} (\code{haven}) / \code{read.xport} (\code{foreign}) for
SAS-transport-files. If you would like to use \code{haven}, you may need to install it using
\code{install.packages("haven", dep = TRUE)}.
}
}
\examples{
# generate a test dataframe with 100 (imaginary) participants / units of
#  observation (ID), 10 measurement (measure) of two variables (V1, V2)
dtaInp <- data.frame(ID = rep(as.character(seq(1, 100)), each = 10),
                     Measure = rep(seq(1, 10), times = 100),
                     V1 = runif(1000,   0, 100),
                     V2 = round(rnorm(1000, 100,  15)))
cat(str(dtaInp))
# the output should look like this
# 'data.frame': 1000 obs. of  4 variables:
#  $ ID     : chr  "1" "1" "1" "1" ...
#  $ Measure: int  1 2 3 4 5 6 7 8 9 10 ...
#  $ V1     : num  ...
#  $ V2     : num  ...
# this data set is stored as (temporary) RDS-file and later processed by long2wide
nmeInp <- tempfile(fileext = ".rds")
nmeOut <- tempfile(fileext = ".omv")
saveRDS(dtaInp, nmeInp)
jmvReadWrite::aggregate_omv(dtaInp = nmeInp, fleOut = nmeOut, varAgg = c("V1", "V2"),
                            grpAgg = "ID", clcN = TRUE, clcMn = TRUE, clcSD = TRUE)
# it is required to give at least the arguments dtaInp, varAgg and grpAgg, each of
# the different switches (clc...) requests a aggregation measure (e.g., mean, median,
# SD, IQR, etc.) to be calculated
# check whether the file was created and its size
cat(list.files(dirname(nmeOut), basename(nmeOut)))
# -> "file[...].omv" ([...] contains a random combination of numbers / characters
cat(file.info(nmeOut)$size)
# -> 4898 (approximate size; size may differ in every run [in dependence of
#          how well the generated random data can be compressed])
cat(str(jmvReadWrite::read_omv(nmeOut, sveAtt = FALSE)))
# the data set contains now the ID variable identifying the different steps of
# aggregation and one column for each combination of aggregation variable (V1 / V2)
# and which calculation was requested (N, mean and SD)
# 'data.frame':	100 obs. of  7 variables:
#  $ ID   : chr  "1" "10" "100" "11" ...
#   ..- attr(*, "jmv-id")= logi TRUE
#   ..- attr(*, "missingValues")= list()
#  $ V1_N : int  10 10 10 10 10 10 10 10 10 10 ...
#   ..- attr(*, "jmv-desc")= chr "V1 (N)"
#   ..- attr(*, "missingValues")= list()
#  $ V1_Mn: num  45.4 51.9 49.4 54.8 47.2 ...
#   ..- attr(*, "jmv-desc")= chr "V1 (Mean)"
#   ..- attr(*, "missingValues")= list()
#  $ V1_SD: num  31.7 31.4 26.5 20.2 29.1 ...
#   ..- attr(*, "jmv-desc")= chr "V1 (SD)"
#   ..- attr(*, "missingValues")= list()
#  $ V2_N : int  10 10 10 10 10 10 10 10 10 10 ...
#   ..- attr(*, "jmv-desc")= chr "V2 (N)"
#   ..- attr(*, "missingValues")= list()
#  $ V2_Mn: num  96.4 102.3 101.6 104.6 108.7 ...
#   ..- attr(*, "jmv-desc")= chr "V2 (Mean)"
#   ..- attr(*, "missingValues")= list()
#  $ V2_SD: num  14.8 18.4 11.2 10.1 14.3 ...
#   ..- attr(*, "jmv-desc")= chr "V2 (SD)"
#   ..- attr(*, "missingValues")= list()

unlink(nmeInp)
unlink(nmeOut)


}
\seealso{
\code{aggregate_omv} internally uses \code{\link[stats:aggregate]{stats::aggregate()}} as function for the aggregation,
and \code{\link[base:NA]{base::is.na()}}, \code{\link[base:mean]{base::mean()}}, \code{\link[stats:median]{stats::median()}}, \code{\link[base:sum]{base::sum()}}, \code{\link[stats:sd]{stats::sd()}},
\code{\link[stats:cor]{stats::var()}}, \code{\link[base:range]{base::range()}}, \code{\link[base:Extremes]{base::min()}}, \code{\link[base:Extremes]{base::max()}}, and
\code{\link[stats:quantile]{stats::quantile()}} for calculation. It furthermore uses the following functions for reading
and writing data files in different formats: \code{\link[=read_omv]{read_omv()}} and
\code{\link[=write_omv]{write_omv()}} for jamovi-files, \code{\link[utils:read.table]{utils::read.table()}} for CSV / TSV files,
\code{\link[=load]{load()}} for reading .RData-files, \code{\link[=readRDS]{readRDS()}} for .rds-files, \code{\link[haven:read_spss]{haven::read_sav()}} or
\code{\link[foreign:read.spss]{foreign::read.spss()}} for SPSS-files, \code{\link[haven:read_dta]{haven::read_dta()}} or \code{\link[foreign:read.dta]{foreign::read.dta()}} for
Stata-files, \code{\link[haven:read_sas]{haven::read_sas()}} for SAS-data-files, and \code{\link[haven:read_xpt]{haven::read_xpt()}} or
\code{\link[foreign:read.xport]{foreign::read.xport()}} for SAS-transport-files.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transpose_omv.R
\name{transpose_omv}
\alias{transpose_omv}
\title{Transpose .omv-files for the statistical spreadsheet 'jamovi' (\url{https://www.jamovi.org})}
\usage{
transpose_omv(
  dtaInp = NULL,
  fleOut = "",
  varNme = "",
  usePkg = c("foreign", "haven"),
  selSet = "",
  ...
)
}
\arguments{
\item{dtaInp}{Either a data frame or the name of a data file to be read (including the path, if required; "FILENAME.ext"; default: NULL); files can be of
any supported file type, see Details below}

\item{fleOut}{Name of the data file to be written (including the path, if required; "FILE_OUT.omv"; default: ""); if empty, the resulting data frame is
returned instead}

\item{varNme}{Name of the variables in the output data frame; see Details below}

\item{usePkg}{Name of the package: "foreign" or "haven" that shall be used to read SPSS, Stata and SAS files; "foreign" is the default (it comes with
base R), but "haven" is newer and more comprehensive}

\item{selSet}{Name of the data set that is to be selected from the workspace (only applies when reading .RData-files)}

\item{...}{Additional arguments passed on to methods; see Details below}
}
\value{
a data frame (only returned if \code{fleOut} is empty) where the input data set is transposed
}
\description{
Transpose .omv-files for the statistical spreadsheet 'jamovi' (\url{https://www.jamovi.org})
}
\details{
\itemize{
\item If \code{varNme} empty, the row names of the input data set are used (preceded by "V_" if all row names are numbers); if \code{varNme} has length 1, then it is
supposed to point to a variable in the input data frame; if \code{varNme} has the same length as the number of rows in the input data frame, then the values
in \code{varNme} are assigned as column names to the output data frame.
\item The ellipsis-parameter (\code{...}) can be used to submit arguments / parameters to the functions that are used for reading and writing the data. By clicking
on the respective function under “See also”, you can get a more detailed overview over which parameters each of those functions take. The functions are:
\code{read_omv} and \code{write_omv} (for jamovi-files), \code{read.table} (for CSV / TSV files; using similar defaults as \code{read.csv} for CSV and \code{read.delim} for TSV
which both are based upon \code{read.table}), \code{load} (for .RData-files), \code{readRDS} (for .rds-files), \code{read_sav} (needs the R-package \code{haven}) or \code{read.spss}
(needs the R-package \code{foreign}) for SPSS-files, \code{read_dta} (\code{haven}) / \code{read.dta} (\code{foreign}) for Stata-files, \code{read_sas} (\code{haven}) for SAS-data-files,
and \code{read_xpt} (\code{haven}) / \code{read.xport} (\code{foreign}) for SAS-transport-files. If you would like to use \code{haven}, you may need to install it using
\code{install.packages("haven", dep = TRUE)}.
}
}
\examples{
set.seed(1)
tmpDF <- stats::setNames(as.data.frame(matrix(sample(6, 1200, replace = TRUE), nrow = 16)),
                         sprintf("sbj_\%03d", seq(75)))
str(tmpDF)
# Data sets that were extracted, e.g., from PsychoPy, may look like this (trials as rows
# and participants as columns, one for each participant, manually assembled / copy-and-pasted).
# However, for analyses, one wants the data set transposed (units / participants as columns)...
nmeOut <- tempfile(fileext = ".omv")
jmvReadWrite::transpose_omv(dtaInp = tmpDF, fleOut = nmeOut)
dtaFrm <- jmvReadWrite::read_omv(nmeOut)
unlink(nmeOut)
str(dtaFrm)
# if no varNme-parameter is given, generic variable names are created (V_...)
jmvReadWrite::transpose_omv(dtaInp = tmpDF, fleOut = nmeOut, varNme = sprintf("Trl_\%02d", seq(16)))
dtaFrm <- jmvReadWrite::read_omv(nmeOut)
unlink(nmeOut)
str(dtaFrm)
# alternatively, the character vector with the desired variable names (of the same length as
# the number of rows in tmpDF) may be given, "Trl" can easily be exchanged by the name of your
# questionnaire, experimental conditions, etc.


}
\seealso{
\code{transpose_omv} internally uses the following functions for reading and writing data files in different formats: \code{\link[=read_omv]{read_omv()}} and
\code{\link[=write_omv]{write_omv()}} for jamovi-files, \code{\link[utils:read.table]{utils::read.table()}} for CSV / TSV files, \code{\link[=load]{load()}} for reading .RData-files, \code{\link[=readRDS]{readRDS()}} for .rds-files,
\code{\link[haven:read_spss]{haven::read_sav()}} or \code{\link[foreign:read.spss]{foreign::read.spss()}} for SPSS-files, \code{\link[haven:read_dta]{haven::read_dta()}} or \code{\link[foreign:read.dta]{foreign::read.dta()}} for Stata-files, \code{\link[haven:read_sas]{haven::read_sas()}} for
SAS-data-files, and \code{\link[haven:read_xpt]{haven::read_xpt()}} or \code{\link[foreign:read.xport]{foreign::read.xport()}} for SAS-transport-files.
}
